/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: SerializableCookie.java,v 1.7 2000/02/18 07:24:51 shawn Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.http;

import java.io.Serializable;
import javax.servlet.http.Cookie;


/**
 *   This class is a substitute for Cookies, but it is serializable.
 *   It is used to pass cookies back and forth when using RMI.
 *   Rather than pasing or returning a Cookie, which would cause a
 *   marshalling error, the Cookies are turned into SerializabeCookies,
 *   passed through RMI, then turned back into Cookies at the other end. <P>
 *
 *   There is no reason why Cookies could not be serializable. Sun may well
 *   fix this oversight in version 2 of Java, in which case this class will
 *   no longer be necissary to use RMI with the MultiServer.
 *
 * @see javax.servlet.http.Cookie
 * @see com.lutris.servlet.connectionMethod.RMIConnectionMethodImpl
 * @see com.lutris.servlet.rmi.http.RMIHttpServletRequestImpl
 * @see com.lutris.servlet.rmi.http.RMIHttpServletResponseImpl
 */
public class SerializableCookie implements Serializable {

    private String name;
    private String value;
    private String comment;
    private String domain; 
    private int maxAge; 
    private String path; 
    private boolean secure; 
    private int version; 

    /**
     * Constructor. Saves all the data about the Cookie, so it can be
     * reconstructed later.
     *
     * @param cookie The cookie to make a copy of. 
     */
    public SerializableCookie(Cookie cookie) {
        name = cookie.getName();
        value = cookie.getValue();
        comment = cookie.getComment();
        domain = cookie.getDomain();
        maxAge = cookie.getMaxAge();
        path = cookie.getPath();
        secure = cookie.getSecure();
        version = cookie.getVersion();
    }


    /**
     * Recreate the cookie that was passed into the constructor.
     *
     * @return A Cookie just like the one passed into the constructor.
     */
    public Cookie getCookie() {
        Cookie cookie = new Cookie(name, value);
        if (comment != null)
            cookie.setComment(comment);
        if (domain != null)
            cookie.setDomain(domain);
        cookie.setMaxAge(maxAge);
        if (path != null)
            cookie.setPath(path);
        cookie.setSecure(secure);
        cookie.setVersion(version);
        return cookie;
    }


    /**
     *  Returns a multi-line description of the data stored about
     *  the cookie. Usefull for debugging.
     *
     *  @return A string description of the cookie.
     */
    public String toString() {
        return "SerializableCookie:" +
               "\n    name = " + name +
               "\n    value = " + value +
               "\n    comment = " + comment +
               "\n    domain = " + domain +
               "\n    maxAge = " + maxAge +
               "\n    path = " + path +
               "\n    secure = " + secure +
               "\n    version = " + version;
    }

}


