/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: LogWriter.java,v 1.5 2000/02/12 18:01:58 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.logging;

import java.io.PrintWriter;
import java.io.Writer;
import java.io.IOException;

/**
 * Class use to write log output to a particular LogChannel and level.
 * This class is PrintWriter, with println() causing a write.
 * One should use println() with this rather than write, as with a
 * LogChannel, since write doesn't write a full line.
 */
public class LogWriter extends PrintWriter {
    /*
     * Log channel and associated level.
     */
    private LogChannel channel;
    private int level;

    /*
     * Are we enabled.
     */
    private boolean enabled;

    /**
     * Constructor.
     * @param logChannel The log channel to write to.
     * @param logLevel The level associated with this channel.
     */
    protected LogWriter(LogChannel logChannel,
                        String logLevel) {
        this(logChannel, logChannel.getLevel(logLevel));
    }

    /**
     * Constructor.
     * @param logChannel The log channel to write to.
     * @param logLevel The level associated with this channel.
     */
    protected LogWriter(LogChannel logChannel,
                        int logLevel) {
        // Flushing logic is in ChannelWriter, so don't buffer here
        super(new ChannelWriter(logChannel, logLevel), true);
        channel = logChannel;
        level = logLevel;
        enabled = logChannel.isEnabled(level);
    }

    /**
     * Get the associate channel.
     */
    public LogChannel getChannel() {
        return channel;
    }

    /**
     * Get the associate level.
     */
    public int getLevel() {
        return level;
    }

    /**
     * Determine if logging is enabled.  This is useful to prevent a series of
     * unnecessary logging calls, as often encountered with debug logging, or
     * a call where generating the message is expensive.
     *
     * @return <CODE>true</CODE> if enabled, <CODE>false</CODE> if not
     *         enabled.
     */
    public boolean isEnabled() {
        return enabled;
    }
    
    /**
     * Write a string and exception to the log file.
     *
     * @param msg The message to log.
     * @param throwable Exception or error to log.
     */
    public void println(String msg, Throwable throwable) {
        if (enabled) {
            flush();
            channel.write(level, msg, throwable);
        }
    }
}


/**
 * Internal Writer object that interfaces to the LogChannel.
 * Can't be inner class of LogWriter, it must be passed
 * to super constructor.
 */
class ChannelWriter extends Writer {
    // Output collected here
    private StringBuffer buffer = new StringBuffer();

    // Log channel, etr.
    private LogChannel channel;
    private int level;
    private boolean enabled;

    /**
     * Construct a new writer.
     */
    public ChannelWriter(LogChannel logChannel,
                         int logLevel) {
        channel = logChannel;
        level = logLevel;
        enabled = logChannel.isEnabled(level);
    }

    /**
     * Main write method that transfers stuff to the buffer.
     */
    public void write(char[] cbuf, int off, int len) {
        buffer.append(cbuf, off, len);
    }

    /**
     * Flush the buffer to the log file.
     */
    public void flush() {
        if (enabled && (buffer.length() > 0)) {
            synchronized(lock) {
                channel.write(level, buffer.toString());
                buffer.setLength(0);
            }
        }
    }

    /**
     * Close is a no-op.
     */
    public void close() {
    }
}

