/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: MimeHeader.java,v 1.8 2000/02/18 07:24:52 shawn Exp $
 *-----------------------------------------------------------------------------
 */




package com.lutris.mime;
import java.util.*;
import java.io.*;

/**
 * Represents a generic parsed Mime header.  Specific header types, such
 * as <code>Content-Type</code> are represented by classes derived from
 * this header.  The constructor for this class parses only parameters
 * according to the rules specified in RFC2045 Section 5.1 Page 11.
 * It is left to derived classes to parse type-specifiec information
 * such as the Mime type (<code>Content-Type</code>) or content disposition
 * (<code>Content-Disposition</code>).
 */
public class MimeHeader {
    /**
     * Contains the parsed parameters for this hash table.
     */
    private Hashtable	myParams;
    
    /**
     * Contains the type-independent value (if any) for the header.
     */
    private String	myValue;
    
    /**
     * Contains the Mime header type value, defined as the characters
     * before the first colon (':') character.
     */
    private String	myHeaderType;
    
    /**
     * The raw, unparsed header line passed to the constructor.
     */
    private String	rawHeaderLine;
    
    /**
     * Constructs a generic <code>MimeHeader</code> object and initializes
     * its internal parameter list.
     */
    public
    MimeHeader(String headerLine)
    {
	myParams = new Hashtable();
	rawHeaderLine = headerLine;
	int pos = headerLine.indexOf(':');
	if (pos < 0) {
	    myHeaderType = null;
	} else {
	    myHeaderType = headerLine.substring(0, pos).trim().toLowerCase();
	    headerLine = headerLine.substring(pos+1);
	}
	pos = headerLine.indexOf(';');
	if (pos < 0) {
	    // Handle potential comment.
	    pos = headerLine.indexOf('(');
	    if (pos < 0) {
		myValue = headerLine.trim();
	    } else { 
		myValue = headerLine.substring(0, pos).trim();
	    }
	} else { 
	    myValue = headerLine.substring(0, pos).trim();
	}
    }
    
    /** 
     * Used by derived classes to put a parameter into the internal
     * parameter value holder.
     * 
     * @param name	The name of the parameter value to add.
     * @param value	The value to add.
     */
    protected
    void putParameter(String name, String value)
    {
	name = name.toLowerCase();
	String[] oldvals = (String[]) myParams.get(name);
	int num = oldvals == null ? 0 : oldvals.length;
	String[] newvals = new String[num+1];
	for (int i=0; i < num; i++) {
	    newvals[i] = oldvals[i];
	    oldvals[i] = null;
	}
	newvals[num] = value;
	oldvals = null;
	myParams.put(name, newvals);
    }
    
    /**
     * Returns the "value" of this header.  This is defined to be the
     * trimmed substring between the first colon (':') character and
     * the first semicolon (';') character. For example, observe the
     * following header line:
     * <PRE>
     * 	Content-Type: multipart/form-data; boundary="000572A9AD4"
     * </PRE>
     * The <b>value</b> of this header is <i>multipart/form-data</i>,
     * the <b>header type</b> is <i>Content-Type</i>, and the first
     * <b>parameter</b> is named <i>boundary</i> and has the value
     * <i>00572A9AD4</i>.
     * 
     * @return	The <b>value</b> of this Mime header.
     */
    public
    String getValue()
    {
	return myValue;
    }
    
    /**
     * Returns the <b>header type</b> of this header.  This is defined to
     * be the substring from the first character up to first colon (':')
     * character.  For example, observe the following header line:
     * <PRE>
     * 	Content-Type: multipart/form-data; boundary="000572A9AD4"
     * </PRE>
     * The <b>value</b> of this header is <i>multipart/form-data</i>,
     * the <b>header type</b> is <i>Content-Type</i>, and the first
     * <b>parameter</b> is named <i>boundary</i> and has the value
     * <i>00572A9AD4</i>.
     * 
     * @return	The <b>header type</b> of this Mime header.
     */
    public
    String getHeaderType()
    {
	return myHeaderType;
    }
    
    /**
     * Returns the string passed to the constructor without any parsing.
     * 
     * @returns	The unparsed Mime header line in its entirety.
     */
    public
    String getRawHeaderLine()
    {
	return rawHeaderLine;
    }

    /**
     * Returns a single <code>String</code> value representing the
     * last occurence of the <b>parameter</b> named by <code>name</code>.
     * For example, observe the following header line:
     * <PRE>
     * 	Content-Type: multipart/form-data; boundary="000572A9AD4"
     * </PRE>
     * The <b>value</b> of this header is <i>multipart/form-data</i>,
     * the <b>header type</b> is <i>Content-Type</i>, and the first
     * <b>parameter</b> is named <i>boundary</i> and has the value
     * <i>00572A9AD4</i>.
     * 
     * @param	name	The name of the <b>parameter</b> to return.
     * @return	The value of the last occurence of the named
     * 		<b>parameter</b> or <code>null</code> if not found.
     */
    public
    String getParameter(String name)
    {
	name = name.toLowerCase();
	String[] vals = (String[]) myParams.get(name);
	if (vals == null) return null;
	if (vals.length < 1) return null;
	return vals[vals.length - 1];
    }
    
    /**
     * Returns an array of <code>String</code> containing all occurences
     * of values named by <code>name</code>.
     * For example, observe the following header line:
     * <PRE>
     * 	Content-Type: multipart/form-data; boundary="000572A9AD4"
     * </PRE>
     * The <b>value</b> of this header is <i>multipart/form-data</i>,
     * the <b>header type</b> is <i>Content-Type</i>, and the first
     * <b>parameter</b> is named <i>boundary</i> and has the value
     * <i>00572A9AD4</i>.
     * 
     * @param	name	The name of the <b>parameter</b> to return.
     * @return	Array containing all values for the named 
     * 		<b>parameter</b> or <code>null</code> if not found.
     */
    public String[] getParameters(String name)
    {
	name = name.toLowerCase();
	String[] vals = (String[]) myParams.get(name);
	if (vals == null) return null;
	if (vals.length < 1) return null;
	return vals;
    }
    
    /**
     * Returns an enumeration of all of the parameter names parsed from
     * the current Mime header.
     * 
     * @return An enumeration of the names of all parameters parsed from
     * 		the Mime header.
     */
    public Enumeration getParameterNames()
    {
	return myParams.keys();
    }
    
    private static
    void skipComment(PushbackReader input)
    throws IOException
    {
	int ch;
	while ((ch = input.read()) >= 0) if (ch == ')') return;
    }
    
    private final
    String parseToken(PushbackReader input, boolean trim)
    throws IOException
    {
	StringBuffer buf = new StringBuffer();
	int ch;
	if (trim) {
	    while ((ch = input.read()) >= 0) if ((ch > ' ') && (ch < 0xff)) {
		input.unread(ch);
		if (ch == '(') {
		    skipComment(input);
		}
		else 
		    break;
	    }
	}
        while ((ch = input.read()) >= 0) {
	    switch (ch) {
		case '=': case ';':
		    input.unread(ch);
		    return buf.toString();
		case '(':
		    skipComment(input);
		    break;
		default:
		    if ((ch > ' ') && (ch < 0xff))
			buf.append((char)ch);
		    else {
			input.unread(ch);
			return buf.toString();
		    }
		    break;
	    }
	}
	return buf.toString();
    }
    
    private final
    String parseQuotedString(PushbackReader input)
    throws IOException
    {
	StringBuffer buf = new StringBuffer();
	int ch;
        while ((ch = input.read()) >= 0) {
	    switch (ch) {
		case '"':
		    return buf.toString();
            
		case '\\':
		    if ((ch = input.read()) >= 0) {
                if ((ch!='\\')&&(ch!='"')) {
    			    buf.append('\\');                                    
                }
			    buf.append((char)ch);
		    } else {
			    buf.append('\\');
			return buf.toString();
		    }
		    break;
            
		default:
		    buf.append((char)ch);
		    break;
	    }
	}
	return buf.toString();
    }
    
    /**
     * Parses the a given string into <i>key=parameter</i> pairs in 
     * accordance with RFC2045 Section 5.1 Page 11.  Quoted strings
     * are parsed in accordance with the RFC822 definition of the
     * "<code>quoted-string</code>" grammatic construct.  According
     * to RFC822, quoted strings are strings, enclosed in double-quote
     * marks, which contain any character except LF or double-quote.
     * The backslash character is used to literally quote the following
     * character.  Octal or hex code backquoting is not supported.
     */
    protected
    void parseParameters(String headerLine)
    {
	try {
	    StringReader reader = new StringReader(headerLine);
	    PushbackReader input = new PushbackReader(reader);
	    int ch;
	    String key, value;
	    while (true) {
	        key = parseToken(input, true);
		ch = input.read();
		if (ch < 0) break;
		switch (ch) {
		    case '=':
			switch (ch = input.read()) {
			    case '"':
				value = parseQuotedString(input);
				if (key.length() > 0) {
				    putParameter(key, value);
				}
				break;
			    default:
				input.unread(ch);
				value = parseToken(input, false);
				if (key.length() > 0) {
				    putParameter(key, value);
				}
				break;
			}
			while (((ch = input.read()) >= 0) && (ch != ';')) {
			    if (ch == '(') {
				skipComment(input);
			    }
			}
			break;
		    case ';':
			// Not a key=value pair.
			break;
		    default:
			// Garbled input.
			return;
		} // outer switch
	    } // outer while
	    return;
	}
	catch (IOException e) {
	    return;
	}
    }
}

