/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: CircularQueue.java,v 1.5 2000/02/18 07:24:54 shawn Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.util;

/*
 * Circular queue data structure implemented with an array
 * of references to objects.
 * It is the responsibility of the user to synchronize access to the queue.
 *
 * @version $Revision: 1.5 $
 * @author Mark Sanguinetti
 */

public class CircularQueue {

    /** Array of references to the objects being queued. */
    protected Object queue[] = null;

    /** The array index for the next object to be stored in the queue. */
    protected int sIndex;

    /** The array index for the next object to be removed from the queue. */
    protected int rIndex;

    /** Number of objects currently stored in the queue. */
    protected int count;

    /** The number of objects in the array (Queue size + 1). */
    protected int qSize;

    /**
     * Creates a circular queue of size s (s objects).
     *
     * @param s The maximum number of elements to be queued.
     */
    public CircularQueue(int s) { 
        qSize = s + 1;
        sIndex = 0;
        rIndex = qSize; 
        count = 0;
        queue = new Object[qSize];
    }

    /**
     * Stores an object in the queue.
     *
     * @param x The object to be stored in the queue.
     * @return true if successful, false otherwise.
     * @exception ArrayIndexOutOfBoundsException 
     */
    public boolean put(Object x) throws ArrayIndexOutOfBoundsException {

        if ((sIndex + 1 == rIndex) ||
            ((sIndex + 1 == qSize ) && (rIndex == 0))) {
            // queue is full
            return false;
        } else {
            // insert object into queue.
            queue[sIndex++] = x;
            count++;
            if (sIndex == qSize) { 
                // loop back
                sIndex = 0;
            }
        }
        return true;
    }

    /**
     * Removes an object from the queue. 
     *
     * @return a reference to the object being retrieved.
     * @exception ArrayIndexOutOfBoundsException 
     */
    public Object get() throws ArrayIndexOutOfBoundsException {

        if (rIndex == qSize) {
            // loop back 
            rIndex = 0;
        }
        if (rIndex == sIndex) {
            // queue is empty
            return null; 
        } else  {
            // return object
            count--;
            Object obj = queue[rIndex];
            queue[rIndex] = null;
            rIndex++;
            return obj;
        }
    }

    /**
     * Returns the total number of objects stored in the queue.
     * 
     * @return The total number of objects in the queue.
     */
    public int getCount() {
        return count;
    }

    /**
     * Checks to see if the queue is empty.
     *
     * @return true if queue is empty, false otherwise.
     */
    public boolean isEmpty() {
        return (count == 0 ? true : false);
    }
} 

