/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: ConfigException.java,v 1.10 2000/02/18 07:24:55 shawn Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.util;

/**
 * Exception class thrown by class <code>Config</code>.  If a syntax
 * error is found in the configuration input stream, or if a component
 * expects a data format for which a configuration element cannot be
 * converted, then this exception is thrown to indicate the error.
 * 
 * @see		Config
 * @version	$Revision: 1.10 $
 * @author	John Marco
 * @since	Soda1.0
 */
public class ConfigException extends KeywordValueException {
    /**
     * The <code>reason</code> field may contain this value to indicate
     * that the cause of the current exception is unknown.
     * 
     * @see ConfigException#reason
     */
    public static final int 	UNKNOWN = 0;

    /**
     * The <code>reason</code> field may contain this value to indicate
     * that a syntax error in the configuration input file or stream
     * caused this exception.
     * 
     * @see ConfigException#reason
     */
    public static final int 	NOT_FOUND = 1;

    /**
     * The <code>reason</code> field may contain this value to indicate
     * that a syntax error in the configuration input file or stream
     * caused this exception.
     * 
     * @see ConfigException#reason
     */
    public static final int 	SYNTAX = 2;
    
    /**
     * The <code>reason</code> field may contain this value to indicate
     * that a component expects a different number of elements than
     * are provided in the configuration file for a particular key.
     * 
     * @see ConfigException#reason
     */
    public static final int 	COUNT = 3;

    /**
     * The <code>reason</code> field may contain this value to indicate
     * that a configuration element could not be converted from its
     * internal string form to the requested type.
     * 
     * @see ConfigException#reason
     */
    public static final int 	FORMAT = 4;

    /**
     * Internal constant to indicate the highest valid <code>reason</code>
     * code.
     * 
     * @see ConfigException#reason
     */
    private static final int 	MAX_REASON = 4;
    
    /**
     * Indicates the cause (if known) of the current exception.
     * 
     * @see ConfigException#UNKNOWN
     * @see ConfigException#SYNTAX
     * @see ConfigException#COUNT
     * @see ConfigException#FORMAT
     */
    public int reason;

    /**
     * Creates a new ConfigException object with no informational string
     * and a <code>reason</code> field of <code>UNKNOWN</code>.
     * 
     * @see ConfigException#reason
     */
    public
    ConfigException()
    {
	super("Unknown reason");
	reason = UNKNOWN;
    }

    /**
     * Creates a new ConfigException object with the specified informational
     * string and a <code>reason</code> field of <code>UNKNOWN</code>.
     * 
     * @param s Informational string to store in the exception object.
     * 
     * @see ConfigException#reason
     */
    public
    ConfigException(String s)
    {
	super(s);
	reason = UNKNOWN;
    }
    
    /**
     * Creates a new ConfigException object with the specified informational
     * string and the specified <code>reason</code> field.
     * 
     * @param s Informational string to store in the exception object.
     * @param r Reason code to store in the <code>reason</code> field.
     * 
     * @see ConfigException#reason
     */
    public
    ConfigException(int r, String s)
    {
	super(s);
	if ((r < 0) || (r > MAX_REASON)) reason = UNKNOWN;
	else reason = r;
    }
}
