/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: QuotedString.java,v 1.8 2000/02/18 07:24:57 shawn Exp $
 *-----------------------------------------------------------------------------
 */




package com.lutris.util;

/**
 * Static convenience class for parsing various types of quoted strings.
 */
public final class QuotedString {
    /**
     * Parse a C style quoted string.  If the first character is a quote,
     * then all characters up to a closing quote or end-of-string are
     * gathered into a new String.  The '\' character has the same 
     * semantics as in C.  That is, it literally quotes the next
     * character.  Also, the four character sequence "\<code>ddd</code>" is 
     * converted to the character represented by the octal value
     * <code>ddd</code>.
     * 
     * @param s		The string from which the quoted string is to be
     * 			parsed.
     * @return		The parsed, quoted string.
     */
    public static final
    String parseCString(String s)
    {
	return parseCString(s.toCharArray(), 0);
    }

    /**
     * Parse a C style quoted string.  If the first character is a quote,
     * then all characters up to a closing quote or end-of-string are
     * gathered into a new String.  The '\' character has the same 
     * semantics as in C.  That is, it literally quotes the next
     * character.  Also, the four character sequence "\<code>ddd</code>" is 
     * converted to the character represented by the octal value
     * <code>ddd</code>.
     * 
     * @param s		The string from which the quoted string is to be
     * 			parsed.
     * @param offset	The index into the string at which parsing is 
     * 			to begin.
     * @return		The parsed, quoted string.
     */
    public static final
    String parseCString(String s, int offset)
    {
	return parseCString(s.toCharArray(), offset);
    }
    
    /**
     * Parse a C style quoted string.  If the first character is a quote,
     * then all characters up to a closing quote or end-of-string are
     * gathered into a new String.  The '\' character has the same 
     * semantics as in C.  That is, it literally quotes the next
     * character.  Also, the four character sequence "\<code>ddd</code>" is 
     * converted to the character represented by the octal value
     * <code>ddd</code>.
     * 
     * @param c		The character array  from which the quoted string
     * 			is to be parsed.
     * @return		The parsed, quoted string.
     */
    public static final
    String parseCString(char[] c)
    {
	return parseCString(c, 0);
    }
    
    /**
     * Parse a C style quoted string.  If the first character is a quote,
     * then all characters up to a closing quote or end-of-string are
     * gathered into a new String.  The '\' character has the same 
     * semantics as in C.  That is, it literally quotes the next
     * character.  Also, the four character sequence "\<code>ddd</code>" is 
     * converted to the character represented by the octal value
     * <code>ddd</code>.
     * 
     * @param c		The character array  from which the quoted string
     * 			is to be parsed.
     * @param offset	The index into the string at which parsing is 
     * 			to begin.
     * @return		The parsed, quoted string.
     */
    public static final
    String parseCString(char[] c, int offset)
    {
	int pos, end, len;
	len = c.length;
	pos = offset;
	
	if (pos > len) return null;
	if (pos == len) return "";
	
	if (c[pos] == '"') {
	    StringBuffer b = new StringBuffer();
	    while (++pos < len) {
		switch (c[pos]) {
		    case '"':
			return new String(b);
		    case '\\':
			if (++pos >= len) {
			    b.append('\\');
			    return new String(b);
			}
			if (c[pos] == '0') {
			    if ((len - pos) >= 3) {
				String xs = new String(c, pos, 3);
				int i=-1;;
				try { i = Integer.parseInt(xs, 8); }
				catch (Throwable t) { i = -1; }
				if ((i >= 0) && (i <= 255)) {
				    b.append((char)i);
				    pos += 2;
				} else {
				    b.append(c[pos]);
				}
			    } else {
				b.append(c[pos]);
			    }
			} else {
			    b.append(c[pos]);
			}
			break;
		    default:
			b.append(c[pos]);
			break;
		}
	    }
	    return new String(b);
	} else {
	    end = pos;
	    while ((end < len) && (c[end] > ' ')) end++;
	    if ((end - pos) <= 0) return "";
	    return new String(c, pos, end - pos);
	}
    }
}
