/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: TmpDir.java,v 1.9 2000/02/18 07:24:57 shawn Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.util;

import	java.io.*;

/**
 * Class for creating and manage a temporary directory.
 *
 * @version	$Revision: 1.9 $
 * @author	Mark Diekhans
 * @since	Jolt1.0
 */
class TmpDir {
  File dir;

  /**
   * Create a tmp directory.
   *
   * @param dirPath Path of directory to generate the temporary file in.
   * @param baseName Prefix name of the directory.
   * @exception java.io.IOException if unable to create the directory.
   */
  public TmpDir (String path,
                 String baseName)
      throws IOException {
    int dirNum = 0;

    /*
     * Generate a unique directory.
     */
    while (true) {
      dir = new File (path, baseName + "." + System.currentTimeMillis () +
                      "." + dirNum + ".tmp");
      if (dir.mkdirs ()) {
        break;
      }
      if (!dir.exists ()) {
        /*
         * Didn't fail because it already exists, must be an access error.
         */
        throw new IOException ("no permission to create directory " +
                                dir.getAbsolutePath ());
      }
      dirNum++;
      if (dirNum >= 25) {
        throw new IOException ("failed to create tmp directory named in the " +
                               "form \"" + dir.getAbsolutePath () +
                               "\" after " + dirNum + " trys");

      }
    }
  }

  /**
   * Get the File object describing the directory.
   *
   * @return Reference to the file object.
   */
  public File file () {
    return dir;
  }
  
  /**
   * Recursive delete files.
   */
  void recursiveDelete (File dirPath) {
    String [] ls = dirPath.list ();

    for (int idx = 0; idx < ls.length; idx++) {
      File file = new File (dirPath, ls [idx]);
      if (file.isDirectory ())
        recursiveDelete (file);
      file.delete ();
    }
  }


  /**
   * Delete the directory and its contents
   *
   * @exception java.io.IOException if unable to remove the directory.
   */
  public void delete ()
      throws IOException {
    recursiveDelete (dir);
    if (dir.exists ()) {
      throw new IOException ("Unable to delete directory hierarchy \"" +
                             dir.getAbsolutePath () + "\"");
    }
  }

  /**
   * Create the specified directory in the tmp directory if it doesn't exist.
   *
   * @param path Path of directory relative to the tmp directory.
   * @return A File naming the new directory.
   * @exception java.io.IOException if unable to create the directory.
   */
  public File mkdirs (String path)
      throws IOException {
    File newDir = new File (dir.getAbsolutePath (), path);
    newDir.mkdirs ();
    if (!newDir.exists ()) {
      throw new IOException ("Unable to create directory \"" +
                             dir.getAbsolutePath () + "\"");
    }
    return newDir;
  }
}
