/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: QueryBuilder.java,v 1.6.2.1 2000/04/05 20:32:54 jay Exp $
 *-----------------------------------------------------------------------------
 */




package com.lutris.dods.builder.generator.query;

import com.lutris.appserver.server.*;
import com.lutris.appserver.server.sql.*;
import java.sql.*;
import java.util.*;
import java.math.BigDecimal;
import java.text.SimpleDateFormat;

import com.lutris.dods.builder.generator.dataobject.GenericDO;


/**
 * <PRE>
 *	A QueryBuilder is a helper-object used by the xxxQuery classes.
 *
 *	Stand-alone usage:
 *
 *		QueryBuilder qb;
 *		qb = new QueryBuilder( "cats" );
 *			or
 *		qb = new QueryBuilder( "cats", "name, age" );
 *			...
 *		qb.addWhereClause( "owner.name", "Smith",
 *					QueryBuilder.NOT_NULL,
 *					QueryBuilder.EXACT_MATCH );
 *		qb.addWhereClause( "paws", 3, QueryBuilder.GREATER_THAN );
 *              qb.addReferentialWhereClause( "cat.ownerId", "owner.Id" );
 *		ResultSet rs = qb.executeQuery( conn );
 *			...
 *		qb.reset();
 *
 * </PRE>
 *
 * @author Jay Gunter
 * @version $Revision: 1.6.2.1 $
 */
public class QueryBuilder implements Query {

// Special values for whereClauses:
private static final String	OR			= " OR ";
private static final String	OPEN_PAREN		= " ( ";
private static final String	CLOSE_PAREN		= " ) ";

// Comparison operator passed as 3rd arg to addWhereClause( int|float|double )
//
public static final String	EQUAL			= "=";
public static final String	NOT_EQUAL		= "!=";
public static final String	LESS_THAN		= "<";
public static final String	LESS_THAN_OR_EQUAL	= "<=";
public static final String	GREATER_THAN		= ">";
public static final String	GREATER_THAN_OR_EQUAL	= ">=";

// Null-acceptability indicator passed as 3rd arg to addWhereClause( String )
//
public static final boolean	NULL_OK			= true;
public static final boolean	NOT_NULL		= false;

// Wild-card          indicator passed as 4th arg to addWhereClause( String )
//
public static final boolean	EXACT_MATCH		= true;
public static final boolean	NOT_EXACT		= false;

// Default inexact string match keyword (e.g. can be MATCHES or LIKE)
public static final String DEFAULT_MATCHES_KEYWORD = "MATCHES";

// Default inexact string wildcard (e.g. can be * or %)
public static final String DEFAULT_WILDCARD = "*";

// See debug() method.
static private boolean debugAllSQL = false;
private boolean debugSQL = false;

/*
The DODS (Data Object Design Studio) development tool
generates data-layer source code which uses QueryBuilder objects
to perform database queries.
This is the original (1st, old) purpose of QueryBuilder.

QueryBuilder can also be used by application developers
to perform straight SQL queries.  Certain static final data members
of the DO classes generated by DODS are used to specify the tables and columns
for the query.  This allows developers to write straight SQL queries
but still enjoy the compile-time checking.
This is the 2nd (new) purpose of QueryBuilder.

================= Usage by DODS ===============================================

1st, old, String-based style (used by DODS-generated Query classes):

    Query creation:

	The ctor takes a String tablename
	(and optional String list of comma-separated column names).
	E.g.:

	    // In PersonQuery class:
	    QueryBuilder qb = new QueryBuilder( "person", "person.LastName" );

    addWhereClause() methods:

	A where-clause is added by calling one of the addWhereClause() methods
	which takes a String column name a String value for comparison.
	Because table and column names are Strings, the application code
	can become out-of-sync with a changing database schema,
	and the developer gets no warning from the compiler.
	This scheme is ok for use by DODS since the application code
	(DO, Query classes) are (re)generated every time the
	schema (the DODS project file) is changed.
	However, this scheme is non-ideal for manual coding.
	E.g.:

	    // In PersonQuery class:
	    qb.addWhereClause( "FirstName", "Bob", QueryBuilder.EQUAL );

    Obtaining results:

	The results of the query are returned as a JDBC ResultSet.
	Column values are retrieved by their postion within each returned row,
	or by the String name of the column in the returned row.
	Again, no hope for compile-time checking.
	E.g.:

	    // PersonQuery class creates a DBQuery object to run the query:
	    DBQuery dbQuery = Enhydra.getDatabaseManager().createQuery();
	    dbQuery.query( this );  // invokes PersonQuery.executeQuery();
	    // PersonQuery class extends Query and implements executeQuery():
	    ResultSet rs = qb.executeQuery( connection );
	    while ( rs.next() ) {
		String lastname = rs.getString( 1 );
	    }

================= Usage by Application Developers =============================

2nd, new, Object Oriented style (used by application developers):

    Query creation:

	The ctor takes a Vector of RDBColumn objects.
	Each RDBColumn object knows which RDBTable it belongs to.
	In each DO class generated by DODS, there is a static RDBColumn object
	for each column in database table for the DO.
	E.g.:

	    Vector fields = new Vector();
	    fields.addElement( PersonDO.LastName );
	    QueryBuilder qb = new QueryBuilder( fields );

    addWhere() methods:

	A where-clause is added by passing an RDBColumn object and a value.
	Because objects (not Strings) are used to specify columns in the query,
        errors in hand-written application code can be caught by the compiler.
	E.g.:

	    qb.addWhere( PersonDO.FirstName, "Bob" );

    Obtaining results:

	The results of the query are returned as a series of RDBRow objects.
	Each RDBRow contains an RDBColumnValue object for each returned column.
	An RDBColumnValue object is retrieved from the RDBRow using
	a one of the RDBColumn objects passed to the QueryBuilder ctor.
	The value is extracted from an RDBColumnValue object via its
	getString or getInteger (...) method.
	These methods throw an exception if the application attempts to
	extract an incorrect type for the RDBColumnValue.
	E.g.:

	    RDBRow row;
	    while ( null != ( row = qb.getNextRow() ) ) {
		String lastname = row.get( PersonDO.LastName ).getString();
	    }
	    qb.close();  // always release the resources

    Note:
	The templates used by the DODS sourceGenerator_Query class
	could be converted to use this 2nd style.
*/

/**
 * Construct a QueryBuilder object that will return specified fields of a table.
 * Style 2 (see above.)
 *
 * @param fields Vector of RDBColumn objects,
 * one for each column to be returned.  Fields can refer to different tables.
 */
public QueryBuilder( Vector fields ) {
    reset();
    RDBColumn[] array = new RDBColumn[ fields.size() ];
    fields.copyInto( array );
    initFromFieldList( array );
}

/**
 * Construct a QueryBuilder object that will return specified fields of a table.
 * Style 2 (see above.)
 *
 * @param fields array of RDBColumn objects,
 * one for each column to be returned.  Fields can refer to different tables.
 */
public QueryBuilder( RDBColumn[] fields ) {
    reset();
    initFromFieldList( fields );
}

/**
 * Initializer for Style 2 ctors.
 * Style 2 (see above.)
 *
 * @param fields array of RDBColumn objects,
 * one for each column to be returned.  Fields can refer to different tables.
 */
private RDBColumn[] selectedFields;
private void initFromFieldList( RDBColumn[] fields ) {
    selectedFields = fields;
    selectClause = "";
    String glue = "";
    for ( int i = 0; i < fields.length; i++ ) {
	selectClause += glue + fields[i].getFullColumnName();
	storeTableName( mainTableName = fields[i].getTableName() );
	glue = ", ";
    }
}

/**
 * Returns an array of RDBRow objects.
 *
 * @return array of RDBRow objects.
 * @exception ATException If the data is not accessible
 * @exception BusinessRuleException If x is invalid.
 * @see getNextRow
 * @author Jay Gunter
 */
public RDBRow[] getRows()
throws QueryException {
    Vector v = new Vector();
    RDBRow row;
    while ( null != ( row = getNextRow() ) )
	v.addElement( row );
    RDBRow[] ret = new RDBRow[ v.size() ];
    v.copyInto( ret );
    return ret;
}


/**
 * Returns an RDBRow object containing the fields specified in the ctor.
 * On first call, executes the query.
 *
 * @return row the RDBRow representing 1 row from the ResultSet.
 * @exception ATException If the data is not accessible
 * @exception BusinessRuleException If x is invalid.
 * @author Jay Gunter
 */
private DBQuery dbQuery = null;
private ResultSet rs = null;
public RDBRow getNextRow()
throws QueryException {
    if ( null == rs ) {
	try {
	    if ( null == dbQuery ) {
		// May not be null if I later implement
		// a ctor that takes a DBQuery.
		dbQuery = Enhydra.getDatabaseManager().createQuery();
	    }
	} catch ( Exception e ) {
	    throw new QueryException(
		"SQL=[" + sql + "]: Unable to create query", e );
	}
        try {
	    // invokes executeQuery which sets ResultSet rs
	    dbQuery.query( this );
        } catch ( Exception e ) {
	    throw new QueryException(
		"SQL=[" + sql + "]: Unable to run query", e );
        }
	if ( null == rs )
	    throw new QueryException( "No ResultSet for Query." );
    }
    try {
	if ( ! rs.next() ) {
	    dbQuery.release();
	    return null;
	}
    } catch ( Exception e ) {
	throw new QueryException(
	    "SQL=[" + sql + "]: Unable to get query results", e );
    }
    RDBColumnValue[] vals = new RDBColumnValue[ selectedFields.length ];
    for ( int i = 0; i < selectedFields.length; i++ ) {
        try {
	    vals[i] = new RDBColumnValue(
		selectedFields[i], rs.getObject(i+1) );
        } catch ( Exception e ) {
	    throw new QueryException(
		"SQL=[" + sql + "]: Unable to get query result for column "
		+ selectedFields[i], e );
        }
    }
    return new RDBRow( vals );
}


/**
 * WARNING!  This method is disabled and should never be called.
 * It's implementation is forced by the Query interface.
 * Use getNextRow instead().
 */
public Object next(ResultSet rs)
throws SQLException, ObjectIdException {
    return null;
}

///////////////////////////////////////////////////////////////////////////////
/// Constructors used by DODS-generated Query classes. ////////////////////////
///////////////////////////////////////////////////////////////////////////////

/**
 * Construct a QueryBuilder object that will return all fields of a table.
 * Style 1 (see above.)
 *
 * @param table Name of the table on which to perform the SQL 'select'.
 */
public QueryBuilder( String tableName ) {
    this( tableName, tableName + ".*" );
}

/**
 * Construct a QueryBuilder object that will return specified fields of a table.
 * Style 1 (see above.)
 *
 * @param table     Name of the table on which to perform the SQL 'select'.
 * @param fieldList Comma-separated string of field names to retrieve.
 */
public QueryBuilder( String tableName, String fieldList  ) {
    reset();
/*     mainTableName = tableName.toLowerCase(); */
    mainTableName = tableName;
    storeTableName( mainTableName );
    selectClause = fieldList;
}

/**
 * Determine table name of column.
 * For example, we return 'tableX' from the following column specification:
 *     UPPER(tableX.columnY)
 * WARNING: trickier column specifications will cause this method to fail.
 *
 * @param column     A column (field) name, possibly with a leading table name.
 */
private String getTableName( String column ) {
    int offset = 0;
    int end = 0;
    String str = column;
    int index = str.lastIndexOf( "(" );
    if ( -1 != index )
//	str = str.substring( index + 1, 1000 );
	str = str.substring( index + 1);
    index = str.indexOf( ")" );
    if ( -1 != index )
	str = str.substring( 0, index );
    index = str.indexOf( "." );
    if ( -1 == index )
	return null;
    return str.substring( 0, index );
}

/**
 * Store all table names that occur as prefix of column specifications.
 * All table names must appear in the "from" clause.
 *
 * @param column     A column (field) name, possibly with a leading table name.
 */
private void storeTableNameForColumn( String column ) {
    String table = getTableName( column );
    storeTableName( table );
}

/**
 * Store a table name.
 * All table names must appear in the "from" clause.
 *
 * @param table name.
 */
private void storeTableName( String table ) {
    if ( null != table )
        tableNames.put( table.toLowerCase(), "" );
}

/**
 * <PRE>
 * Specify that the next conditional expression (where-clause)
 * should be preceded by OR instead of AND.  (AND is the default).
 * For example:
 *          addWhereClause( "person.fname", "Bob", EQUAL );
 *          addWhereClause( "person.lname", "Smith", EQUAL );
 *          	produces
 *          person.fname = 'Bob' AND person.lname = 'Smith'
 * But:
 *          addWhereClause( "person.fname", "Bob", EQUAL );
 *          addWhereOr();
 *          addWhereClause( "person.lname", "Smith", EQUAL );
 *          	produces
 *          person.fname = 'Bob' OR person.lname = 'Smith'

 * Disallows 'OR OR'.
 * Trailing ORs are removed by prepareStatement().
 *
 * </PRE>
 *
 * @author Jay Gunter
 */
public void addWhereOr() {
    int n = whereClauses.size();
    if ( 0 == n )
	return;
    String prev = (String) whereClauses.lastElement();
    if ( ! OR.equals( prev ) )
	whereClauses.addElement( OR );
}

/**
 * Specify that the next conditional expression (where-clause)
 * should be preceded by '('.
 * Trailing OPEN_PARENs are removed by prepareStatement().
 *
 * @author Jay Gunter
 */
public void addWhereOpenParen() {
    whereClauses.addElement( OPEN_PAREN );
/*
for ( int i = 0; i < whereClauses.size(); i++ ) {
System.err.println("whereClauses.elementAt("+i+")="+whereClauses.elementAt(i));
}
*/
}

/**
 * Close a previous addWhereOpenParen.
 * Disallow 'OR )' and '( )'.
 *
 * @author Jay Gunter
 */
public void addWhereCloseParen() {
    int n = whereClauses.size();
    if ( 0 == n )
	return;
    String prev = (String) whereClauses.lastElement();
    if ( ! ( OR.equals( prev ) || OPEN_PAREN.equals( prev ) ) )
	whereClauses.addElement( CLOSE_PAREN );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, int value, String cmp_op ) {
    addWhereClause( column.getFullColumnName(), value, cmp_op );
}

/**
 * Add a where-clause that tests if a column is equal to a given String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, int value ) {
    addWhere( column, value, QueryBuilder.EQUAL );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   String value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhereLower( RDBColumn column, String value, String cmp_op ) {
    addWhereClause( "LOWER(" + column.getFullColumnName() + ")", value, cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   String value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, String value, String cmp_op ) {
    addWhereClause( column.getFullColumnName(), value, cmp_op );
}

/**
 * Add a where-clause that tests if a column is equal to a given String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   String value to compare against.
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, String value ) {
    addWhere( column, value, QueryBuilder.EQUAL );
}

/**
 * Add a where-clause that compares a column against a Date.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   Date value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, java.sql.Date value, String cmp_op ) {
    addWhereClause( column.getFullColumnName(), value, cmp_op );
}

/**
 * Add a where-clause that compares a column against another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column1, RDBColumn column2, String cmp_op ) {
    addTwoColumnWhereClause(	column1.getFullColumnName(),
				column2.getFullColumnName(),
				cmp_op );
}

/**
 * Add a where-clause that tests if a column is equal to another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column1, RDBColumn column2 ) {
    addWhere( column1, column2, QueryBuilder.EQUAL );
}

/**
 * Add a where-clause that imposes a RIGHT OUTER JOIN to 
 * test if a column is equal to another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 *
 * @author Jay Gunter
 */
public void addWhereRightOuter( RDBColumn column1, RDBColumn column2 ) {
    addTwoColumnWhereClause(
	    column1.getFullColumnName() + "(+)",
	    column2.getFullColumnName(), QueryBuilder.EQUAL );
}

/**
 * Add a where-clause that imposes a LEFT OUTER JOIN to 
 * test if a column is equal to another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 *
 * @author Jay Gunter
 */
public void addWhereLeftOuter( RDBColumn column1, RDBColumn column2 ) {
    addTwoColumnWhereClause(
	    column1.getFullColumnName(),
	    column2.getFullColumnName() + "(+)", QueryBuilder.EQUAL );
}

/**
 * <PRE>
 * Add a where-clause to the SQL command.
 * The where-clause compare the values of two columns
 * using the specified comparison operator.
 * E.g., we want to find cats with too many paws:
 *      column1 = "cat.paws"
 *      column2 = "animalNorms.catPaws"
 *      cmp_op  = QueryBuilder.GREATER_THAN
 * generates the where-clause:
 *      cat.paws > animalNorms.catPaws
 * and remembers the table names (cat, animalNorms)
 * for inclusion in the from-clause.
 * </PRE>
 *
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  Another column to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 */
public void addTwoColumnWhereClause(
			       String column1, String column2, String cmp_op )
{
    storeTableNameForColumn( column1 );
    storeTableNameForColumn( column2 );

    whereClauses.addElement( column1 + " " + cmp_op + " " + column2 );
}

/**
 * <PRE>
 * Add a where-clause to the SQL command.
 * The where-clause compare the values of two columns
 * using the EQUAL operator.
 * E.g., we want to find cats with the normal number of paws:
 *      column1 = "cat.paws"
 *      column2 = "animalNorms.catPaws"
 * generates the where-clause:
 *      cat.paws = animalNorms.catPaws
 * and remembers the table names (cat, animalNorms)
 * for inclusion in the from-clause.
 *
 * </PRE>
 *
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  Another column to compare against.
 */
public void addTwoColumnWhereClause( String column1, String column2 )
{
    addTwoColumnWhereClause( column1, column2, EQUAL );
}

/**
 * <PRE>
 * Add an inter-table reference where-clause to the SQL command.
 * The two columns specified will be compared using the EQUAL operator.
 *
 * column1 and column2 should refer to different tables.
 *
 * E.g.:
 *      column1 = "cat.ownerId"
 *      column2 = "owner.Id"
 * generates the where-clause:
 *      cat.ownerId = owner.Id
 * and remembers the table names for inclusion in the from-clause.
 * </PRE>
 *
 * @param column1  A column (field) name, possibly with a leading table name.
 * @param column2  A column (field) name, possibly with a leading table name.
 * @return Whether a where-clause was generated.
 * True only if column1 and column2 refer to different tables.
 */
public boolean addReferentialWhereClause( String column1, String column2 ) {
    String table1 = getTableName( column1 );
    if ( null == table1 )
        table1 = mainTableName;

    String table2 = getTableName( column2 );
    if ( null == table2 )
        table2 = mainTableName;

    if ( table1.equals( table2 ) )
	return false;

    storeTableName( table1 );
    storeTableName( table2 );

    refWhereClauses.put( column1 + " = " + column2, "" );

    return true;
}

/**
 * Add a where-clause that compares a column against a string value
 * to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 */
public void addWhereClause(	String column, String value, String cmp_op )
{
    storeTableNameForColumn( column );

    whereClauses.addElement( column + " " + cmp_op + " ?" );
    parms.addElement( value );
}

/**
 * Add a where-clause that compares a column against a string value
 * to the SQL command.  The 3rd and 4th params allow checking to be done
 * on the string value before the where-clause is generated.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param nullOk   Whether a value of NULL is acceptable.
 * @param exact    Whether the value must match exactly (no wildcards).
 * @return Whether a where-clause was generated.
 * false if value==null and nullOk==false.
 */
public boolean addWhereClause(	String column, String value,
				boolean nullOk, boolean exact )
{
    storeTableNameForColumn( column );

    String append = "";
    boolean added = false;

    if ( value == null ) {
	if ( nullOk ) {
	    append = " " + column + " is null";
	    added = true;
	}
    } else {
        String pattern = value.trim();
        //if ( pattern.length() > 0 && ! pattern.equals( wildcard ) ) {
        if ( ! pattern.equals( wildcard ) ) {
            if ( exact )
                append = " " + column + " = ?";
            else {
                append = " " + column + " " + matchKeyword + " ?";
                fuzzies.set( parms.size() );
	    }
            added = true;
	    parms.addElement( pattern );
        }
    }
    if ( added )
        whereClauses.addElement( append );

    return added;
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param nullOk   Whether a value of NULL is acceptable.
 * @return Whether a where-clause was generated.
 * false if value==null and nullOk==false.
 */
public boolean addWhereClause(	String column, BigDecimal value,
				boolean nullOk )
{
    storeTableNameForColumn( column );

    String append = "";
    boolean added = false;

    if ( value == null ) {
	if ( nullOk ) {
	    append = " " + column + " is null";
	    added = true;
	}
    } else {
        append = " " + column + " = ?";
        added = true;
	parms.addElement( value );
    }
    if ( added )
        whereClauses.addElement( append );

    return added;
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, int value )
{
    addWhereClause( column, value, "=" );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 */
public void addWhereClause(	String column, int value, String cmp_op )
{
    storeTableNameForColumn( column );

    whereClauses.addElement( column + " " + cmp_op + " ?" );
    parms.addElement( new Integer( value ) );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, float value )
{
    addWhereClause( column, value, "=" );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 */
public void addWhereClause(	String column, float value, String cmp_op )
{
    storeTableNameForColumn( column );

    whereClauses.addElement( column + " " + cmp_op + " ?" );
    parms.addElement( new Float( value ) );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, double value )
{
    addWhereClause( column, value, "=" );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 */
public void addWhereClause(	String column, double value, String cmp_op )
{
    storeTableNameForColumn( column );

    whereClauses.addElement( column + " " + cmp_op + " ?" );
    parms.addElement( new Double( value ) );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, boolean value )
{
    storeTableNameForColumn( column );

    whereClauses.addElement( column + " = ?" );
    parms.addElement( new Integer( value ? 1 : 0 ) );
}

/**
 * Add a trailing clause (order by, group by, etc.) to the SQL command.
 *
 * @param clause     The clause to append to the SQL command.
 */
public void addEndClause( String clause ) {
    endClauses.put( clause, "" );
}

/**
 * Append to the 'select' portion of the SQL command.
 * This has been deprecated.  Use addEndClause instead.
 *
 * @param str     The string to append to the SQL command.
 */
public void add( String str ) {
    addEndClause( str );
}

/**
 * Sets debug flag to print the SQL statement just before its execution.
 * Affects just this instance of QueryBuilder.
 */
public void debug() {
    debugSQL = true;
}

/**
 * Sets debug flag to print the SQL statement just before its execution.
 * Affects all instances of QueryBuilder.
 */
static public void debugAll() {
    debugAllSQL = true;
}

/**
 * Add DISTINCT keyword to query.
 */
private String distinct = "";
public void distinct() {
    distinct = "distinct ";
}

public void addWhere( String fullClause ) {
    whereClauses.addElement( fullClause );
}

/**
 * QueryBuilder.addWhereIn( RDBColumn, QueryBuilder )
 * allows queries like 
 *   select T1X from T1 where T1Y in ( select T2A from T2 where T2B = 1 )
 * with code like
 *   QueryBuilder subQuery = new QueryBuilder( T2DO.T2A );
 *   subQuery.addWhere( T2DO.T2B, 1 );
 *   QueryBuilder mainQuery = new QueryBuilder( T1DO.T1X );
 *   mainQuery.addWhereIn( T1DO.T1Y, subQuery );
 * 
 * @param field T1DO.T1Y in the example above
 * @param qb subQuery in the example above
 * @author Jay Gunter
 */
private Vector subQueries = new Vector();
public void addWhereIn( RDBColumn field, QueryBuilder qb )
throws QueryException {
    if ( this == qb )
        throw new QueryException( 
	    "Recursion detection: subquery cannot be " +
	    "the same object as the main query." );
    storeTableName( field.getTableName() );
/*     addEndClause( " AND " +  */
    addWhere(
        field.getTableName() + "." + field.getColumnName() + 
        " IN ( " + qb.getSQL() + ")" );
    subQueries.addElement( qb );
}

private String commaSeparatedList( String[] values ) {
    String ret = "";
    String comma = "";
    for ( int i = 0; i < values.length; i++ ) {
	ret += comma + values[i];
	comma = ", ";
    }
    return ret;
}

public void addWhereIn( RDBColumn field, String[] values )
throws QueryException {
    if ( null == values || 0 == values.length )
        throw new QueryException( 
	    "IN clause must have at least one value." );
    storeTableName( field.getTableName() );
/*     addEndClause( " AND " +  */
    addWhere(
        field.getTableName() + "." + field.getColumnName() + 
        " IN ( " + commaSeparatedList( values ) + ")" );
}

public String getSQL() {
    // Query begins with SELECT <fields> FROM <tablelist>
    sql = "select " + distinct + selectClause + " from ";
    Enumeration tables = tableNames.keys();
    for ( int i = 0; tables.hasMoreElements(); i++ ) {
	if ( i > 0 )
	    sql += ", ";
	sql += tables.nextElement();
    }

    // Cleanup WHERE clauses.
    // Discard trailing ORs and OPEN_PARENs.
    String chunk = null;
    int numWhereClauses = whereClauses.size();
    for (; numWhereClauses > 0; numWhereClauses-- ) {
	chunk = (String) whereClauses.elementAt( numWhereClauses - 1 );
	if ( ! ( chunk.equals( OR ) || chunk.equals( OPEN_PAREN ) ) )
	    break;
    }

/*
    // Add WHERE clauses to query.
    int parenDepth = 0;
    String glueWord = " WHERE ";
    boolean thisWasCloseParen = false;
    boolean prevWasCloseParen = false;
    for ( int i = 0; i < numWhereClauses; i++ ) {
	chunk = (String) whereClauses.elementAt( i );
System.err.println("chunk ="+chunk );
	sql += glueWord;
	glueWord = " AND ";
	thisWasCloseParen = false;
	if (		chunk.equals( OR ) ) {
	    //glueWord = OR;
	    glueWord = "";
	} else if (	chunk.equals( OPEN_PAREN ) ) {
	    glueWord = "";
	    parenDepth++;
	} else if (	chunk.equals( CLOSE_PAREN ) ) {
	    thisWasCloseParen = true;
	    glueWord = "";
	    if ( parenDepth > 0 ) {
		parenDepth--;
	    } else {
		chunk = "";
	    }
	}
System.err.println("sql ="+sql );
System.err.println("prevWasCloseParen ="+prevWasCloseParen );
System.err.println("thisWasCloseParen ="+thisWasCloseParen );
	if ( prevWasCloseParen && ! thisWasCloseParen )
	    sql += " and ";
        prevWasCloseParen = thisWasCloseParen;
	sql += chunk;
    }
*/

    // Add WHERE clauses to query.
    int parenDepth = 0;
    boolean prevWas_OpenParen_or_OR = true;
    for ( int i = 0; i < numWhereClauses; i++ ) {
	chunk = (String) whereClauses.elementAt( i );
/* System.err.println("chunk ="+chunk ); */
	if ( 0 == i )
	    sql += " WHERE ";
	if (		chunk.equals( OR ) ) {
	    prevWas_OpenParen_or_OR = true;
	} else if (	chunk.equals( OPEN_PAREN ) ) {
            if ( ! prevWas_OpenParen_or_OR )
	        sql += " AND ";
	    prevWas_OpenParen_or_OR = true;
	    parenDepth++;
	} else if (	chunk.equals( CLOSE_PAREN ) ) {
	    if ( parenDepth > 0 ) {
		prevWas_OpenParen_or_OR = false;
		parenDepth--;
	    } else {
		chunk = "";
	    }
	} else { // chunk is actual expression for where-clause
            if ( ! prevWas_OpenParen_or_OR )
	        sql += " AND ";
	    prevWas_OpenParen_or_OR = false;
	}
	sql += chunk;
    }

    // Add any needed trailing CLOSE_PARENs.
    while ( parenDepth-- > 0 )
	sql += CLOSE_PAREN;

    /*
    String glueWord = " where ";
    String chunk = null;
    for ( int i = 0; i < whereClauses.size(); i++ ) {
	sql += glueWord + whereClauses.elementAt( i );
	glueWord = " and ";
    }
    */

    /*
    Enumeration refClauses = refWhereClauses.keys();
    while ( refClauses.hasMoreElements() ) {
	sql += glueWord + refClauses.nextElement();
	glueWord = " and ";
    }
    */

    /*
     * ORDER BY code added by Chris Ryan (cryan@plugged.net.au)
     * The following code adds the ORDER BY 'xxx[, xxx]' to the
     * statement for querying.
     */
    int nr_OrderFields = orderFields.size();

    // ORDER BY required?
    if (nr_OrderFields > 0) {

      // yes, fill out the SQL...
      sql += " ORDER BY ";
      for ( int i = 0; i< nr_OrderFields; i++ ) {
            if ( i > 0 )
              sql += ", ";
            sql += orderFields.elementAt(i)+" "+orderDirections.elementAt(i);
            //sql += orderFields.elementAt(i);
      }

    }

    Enumeration endings = endClauses.keys();
    while ( endings.hasMoreElements() )
	sql += endings.nextElement();

    return sql;
}


public void applyParameters( PreparedStatement ps, int paramCount )
throws SQLException {
    for ( int i = 0; i < parms.size(); i++ ) {
	paramCount++;
        Object    o = parms.elementAt( i );
	if (debugSQL || debugAllSQL) {
    	    System.out.print("Param: " + i);
    	    System.out.println(", Value: " + o);
	}
//System.err.println("PARM " + paramCount + " + " + o );
	if      ( o instanceof String ) {
	    if ( fuzzies.get( i ) ) {
		if (matchKeyword.equals(DEFAULT_MATCHES_KEYWORD)) {
		    ps.setString( paramCount, getRegExpMatch( (String) o ) );
		} else {
		    ps.setString( paramCount, (String)        o );
		}
	    } else {
		ps.setString(paramCount, (String) o );
	    }
        } else if ( o instanceof BigDecimal	) {
		  ps.setBigDecimal( paramCount,  (BigDecimal) o );
	} else if ( o instanceof Integer	) {
		  ps.setInt(        paramCount, ((Integer)    o).intValue() );
	} else if ( o instanceof Float		) {
		  ps.setFloat(      paramCount, ((Float)      o).floatValue() );
	} else if ( o instanceof Double		) {
		  ps.setDouble(     paramCount, ((Double)     o).doubleValue());
	} else if ( o instanceof Boolean	) {
		  ps.setBoolean(    paramCount, ((Boolean)    o).booleanValue() );
    	}
    }
    for ( int i = 0; i < subQueries.size(); i++ ) {
        QueryBuilder subQuery = (QueryBuilder) subQueries.elementAt(i);
	subQuery.applyParameters( ps, paramCount );
    }
}

/**
 * Generate a JDBC PreparedStatement using the values passed 
 * for the where-clauses.
 *
 * @param conn   The DBConnection object to use for the search.
 */
private String sql = null;
private void prepareStatement( DBConnection conn )
throws SQLException {
    getSQL();
    if ( debugSQL || debugAllSQL )
	System.out.println("\nQueryBuilder.prepareStatement sql="+sql+"\n");

    try {
	stmt = conn.prepareStatement( sql );
    } catch ( SQLException se ) { 
	System.err.println("\n\nSQLException: "+se.getMessage() +","+ se.getErrorCode()+","+se.getSQLState() );
        throw se;
    }

    applyParameters( stmt, 0 );
}

/**
 * Return the JDBC PreparedStatement generated by prepareStatement().
 * @return The JDBC PreparedStatement.
 */
public PreparedStatement getStatement() {
    // Would anyone ever want this?
    return stmt;
}

/**
 * Execute the PreparedStatement.
 * @return The ResultSet object generated by JDBC for the search.
 */
/*
public ResultSet executeQuery() {
    conn = new DBConnection();
    return executeQuery( conn );
}
*/

/**
 * Execute the PreparedStatement and return a JDBC ResultSet.
 * @param conn   The DBConnection object to use for the search.
 * @return The ResultSet object generated by JDBC for the search.
 */
public ResultSet executeQuery( DBConnection conn )
		throws SQLException {
    prepareStatement( conn );
    rs = stmt.executeQuery();
    return rs;
}

/**
 * Reset the where-clause information.
 * The same fields specified in the QueryBuilder constructor,
 * can be retrieved for a different set of where-clauses.
 */
public void reset() {
    whereClauses = new Vector(); // keep the select, forget the where-clauses
    parms   = new Vector();	// discard parms for previous where-clause
    fuzzies = new BitSet();	// forget special string handling

    // Clear the ORDER BY Vectors
    orderFields = new Vector();
    orderDirections = new Vector();

    try {
	if ( null != rs )
	    rs.close();
    } catch ( SQLException e ) {
	System.err.println(
	    "QueryBuilder unable to close ResultSet: SQLException=" +
	    e.getMessage() );
    }

    if ( null != conn )	// if this class created the connection, close it.
	conn.close();
    conn    = null;
}

public void close() {
    try {
	if ( null != dbQuery )
	    dbQuery.release();
	if ( null != rs )
	    rs.close();
    } catch ( SQLException e ) {
	System.err.println(
	    "QueryBuilder unable to close ResultSet: SQLException=" +
	    e.getMessage() );
    }
}

/**
 * Returns a regular expression, as supported by "MATCHES"
 * that will will perform a case insensitive match of a substring.
 * @param base   The base String to convert to a regular expression.
 * @return       The generated regular expression.
 */
private String getRegExpMatch( String base )
{
    int len = base.length();
    char [] baseChars = new char[len];
    base.getChars(0, len, baseChars, 0);

    // The result will never exceed 5 * the length of
    // the original string.
    char [] result = new char[len * 5];
    int resultIdx = 0;
    for (int idx = 0; idx < baseChars.length; idx++) {
        if (Character.isSpaceChar(baseChars[idx])) {
            // Convert spaces to ' *'
            if ((resultIdx > 0) && (result[resultIdx-1] != '*')) {
                result[resultIdx++] = ' ';
                //result[resultIdx++] = '*';
                result[resultIdx++] = '%';
            }
        }
        else if (Character.isLetter(baseChars[idx])) {
            // Convert letters to [Aa]
            result[resultIdx++] = '[';
            result[resultIdx++] = Character.toUpperCase(baseChars[idx]);
            result[resultIdx++] = Character.toLowerCase(baseChars[idx]);
            result[resultIdx++] = ']';
        }
        else {
            // Leave everything else as is.
            result[resultIdx++] = baseChars[idx];
        }
    }
    // Match anything at end of line
    result[resultIdx++] = '*';
    return new String(result);
}

private String mainTableName;
private Hashtable tableNames = new Hashtable();
private Hashtable refWhereClauses = new Hashtable();
private Hashtable endClauses = new Hashtable();
private String selectClause;
private Vector whereClauses = new Vector();
private Vector orderFields = new Vector();
private Vector orderDirections = new Vector();
private PreparedStatement stmt;
private Vector parms;
private BitSet fuzzies;
private int param = 0;
private DBConnection conn = null;

/*
 * The following methods made the implementation of the query template easier.
 * They allow one template to handle all Java and database types.
 * They do this by ignoring some of the parameters passed.
 * The addWhereClause() for String     ignores dbtype.
 * The addWhereClause() for byte[]     ignores dbtype.
 * The addWhereClause() for boolean    ignores            nullOk and exact.
 * The addWhereClause() for int        ignores dbtype,    nullOK and exact.
 * The addWhereClause() for float      ignores dbtype,    nullOK and exact.
 * The addWhereClause() for double     ignores dbtype,    nullOK and exact.
 * The addWhereClause() for BigDecimal ignores dbtype            and exact.
 * The addWhereClause() for Date       ignores dbtype            and exact.
 * The addWhereClause() for Time       ignores dbtype            and exact.
 * The addWhereClause() for Timestamp  ignores dbtype            and exact.
 * The addWhereClause() for GenericDO  ignores dbtype            and exact.
 *
 * An alternative solution would be to implement 4 different templates
 * and have SourceGenerator_Query create filters for each template
 * and use the appropriate template according to the type of each Attribute.
 * While that approach would not be wildly complex, it would be less
 * well-contained than implementing the methods below.
 */

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, String value, String dbtype,
				boolean nullOk, boolean exact )
throws QueryException
{
    if ( ! addWhereClause( column, value, nullOk, exact ) )
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, byte[] value, String dbtype,
				boolean nullOk, boolean exact )
throws QueryException
{
/*     String s = new String( value, 0 );	// FIX check this:  hibyte==0? */
    String s = new String( value );
    if ( ! addWhereClause( column, s, nullOk, exact ) )
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, boolean value, String dbtype,
				boolean nullOk, boolean exact )
{
    boolean guess = false;
    if ( dbtype == null ) {
	guess = true;
    } else if ( dbtype.equals( "BIT" ) ) {
	parms.addElement( new Integer( value ? 1 : 0 ) );
    } else if ( dbtype.equals( "INTEGER" ) ) {
	parms.addElement( new Integer( value ? 1 : 0 ) );
    } else if ( dbtype.equals( "CHAR" ) ) {
	parms.addElement( value ? "1" : "0" );
    } else if ( dbtype.equals( "VARCHAR" ) ) {
	parms.addElement( value ? "1" : "0" );
    } else {
	guess = true;
    }

    if ( guess ) {
	addWhereClause( column, value );
    } else {
	storeTableNameForColumn( column );

	whereClauses.addElement( column + " = ?" );
    }
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, int value, String dbtype,
				boolean nullOk, boolean exact )
throws QueryException
{
    addWhereClause( column, value );
    if ( false ) // all addWhereClauses used by template must throw QueryException
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, float value, String dbtype,
				boolean nullOk, boolean exact )
throws QueryException
{
    addWhereClause( column, value );
    if ( false ) // all addWhereClauses used by template must throw QueryException
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, double value, String dbtype,
				boolean nullOk, boolean exact )
throws QueryException
{
    addWhereClause( column, value );
    if ( false ) // all addWhereClauses used by template must throw QueryException
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, BigDecimal value, String dbtype,
				boolean nullOk, boolean exact )
throws QueryException
{
    if ( ! addWhereClause( column, value, nullOk ) )
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Set the Database vendor name.
 * Each vendor has a date format that must be used for searching.
 *
 * @param vendor Name of the database vendor; e.g. "Oracle".
 */
private final String dateFormatInformix = "yyyy-MM-dd HH:mm:ss";
private final String dateFormatSybase   = "MMM dd yyyy hh:mma";
private final String dateFormatOracle   = "dd-MMM-yyyy";
private final String dateFormatMSQL     = "MMM dd yyyy hh:mma"; // FIX
private final String dateFormatStandard = "MMM dd yyyy hh:mma"; // FIX
private final String dateFormatPostgreSQL = "MMM dd yyyy hh:mma";

private       String dateFormat = dateFormatInformix;
private SimpleDateFormat df = new SimpleDateFormat( dateFormat );

public void setDatabaseVendor( String vendor ) {
    if (      vendor.toLowerCase().equals( "informix" ) )
	df = new SimpleDateFormat( dateFormatInformix );
    else if ( vendor.toLowerCase().equals( "sybase" ) )
	df = new SimpleDateFormat( dateFormatSybase );
    else if ( vendor.toLowerCase().equals( "oracle" ) )
	df = new SimpleDateFormat( dateFormatOracle );
    else if ( vendor.toLowerCase().equals( "msql" ) )
	df = new SimpleDateFormat( dateFormatMSQL );
    else if ( vendor.toLowerCase().equals( "standard" ) )
	df = new SimpleDateFormat( dateFormatStandard );
    else if ( vendor.toLowerCase().equals( "postgresql" ) )
	df = new SimpleDateFormat( dateFormatPostgreSQL );
    else {
	System.out.println(
		"ERROR in QueryBuilder!  Unimplemented Database Vendor: " +
		vendor );
    }
}


/**
 * Set the string matching details for inexact string matches
 * See dods.conf for details.
 * Code added by Chris Ryan (cryan@plugged.net.au)
 *
 * @param match_keyword  Whether the DB can use "MATCHES" or "LIKE"
 * @param wildcard  The wildcard character for string searches
 */
private String matchKeyword = DEFAULT_MATCHES_KEYWORD;
private String wildcard = DEFAULT_WILDCARD;
public void setStringMatchDetails( String match_keyword, String wildcard ) {

  this.matchKeyword = match_keyword;
  this.wildcard = wildcard;

}


/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, java.sql.Date value,
				String dbtype, boolean nullOk, boolean exact )
throws QueryException
{
    String strValue = df.format( value );
    if ( ! addWhereClause( column, strValue, nullOk, true ) )
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause that compares a column against a date value
 * to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 */
public void addWhereClause(	String column, java.sql.Date value, String cmp_op )
{
    storeTableNameForColumn( column );

    String strValue = df.format( value );
    whereClauses.addElement( column + " " + cmp_op + " '" + strValue + "' " );
}

/*
public void addWhereClause(	String column, java.sql.Date date, String cmp_op )
{
    storeTableNameForColumn( column );

    whereClauses.addElement( column + " " + cmp_op + " ?" );
    parms.addElement( date );
}
*/

public void addWhereClause(	String column, java.sql.Timestamp date, String cmp_op )
{
    storeTableNameForColumn( column );

    whereClauses.addElement( column + " " + cmp_op + " ?" );
    parms.addElement( date );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, java.sql.Time value,
				String dbtype, boolean nullOk, boolean exact )
throws QueryException
{
    String strValue = df.format( value );
    if ( ! addWhereClause( column, strValue, nullOk, true ) )
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column, java.sql.Timestamp value,
				String dbtype, boolean nullOk, boolean exact )
throws QueryException
{
    String strValue = df.format( value );
    if ( ! addWhereClause( column, strValue, nullOk, true ) )
        throw new QueryException( "Search for null disallowed by caller." );
}

/**
 * Add a where-clause to the SQL command.
 * NOTE: for use by SourceGenerator_Query and writeMemberStuff.template.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 */
public void addWhereClause(	String column,
				GenericDO DO,
				String dbtype, boolean nullOk, boolean exact )
throws QueryException
{
    if ( null == DO ) {
	if ( nullOk )
	    addWhereClause( column, null, nullOk );
	// ignore this call if DO is null and ! nullOk
    } else {
        BigDecimal id = DO.getOId().toBigDecimal();
        if ( ! addWhereClause( column, id, nullOk ) )
	    throw new QueryException( "Search for null disallowed by caller." );
    }
}

/**
 * Add another field to the ORDER BY list.
 * NOTE: is used by SourceGenerator_Query and writeMemberStuff.template.
 *
 * This method added by Chris Ryan (cryan@plugged.net.au) to
 * allow easier access to the ORDER BY part of the SQL query.
 *
 * @param column     The column to order by
 * @param direction  ASCENDING or DESCENDING order
 */
public void addOrderByColumn(String column, String direction) {
  orderFields.addElement(column);
  orderDirections.addElement(direction);
}

/**
 * Add another field to the ORDER BY list.
 * NOTE: is used by SourceGenerator_Query and writeMemberStuff.template.
 *
 * This method added by Chris Ryan (cryan@plugged.net.au) to
 * allow easier access to the ORDER BY part of the SQL query.
 *
 * @param column     The column to order by
 * @param direction  ASCENDING or DESCENDING order
 */
public void addOrderByColumn( RDBColumn column, String direction ) {
  orderFields.addElement( column.getFullColumnName() );
  orderDirections.addElement( direction );
}

} // end QueryBuilder class
