/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: RDBColumnValue.java,v 1.2 2000/01/29 07:01:37 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.dods.builder.generator.query;

import java.math.*;

/**
 * After a Vector of RDBColumn objects is used to construct
 * an instance of QueryBuilder, the QueryBuilder.getNextRow() method
 * is used to return an RDBRow object which contains RDBColumnValue objects
 * representing the column values in the returned row.
 * An instance of RDBColumnValue is returned by the RDBRow.get() method.
 * The value contained in the RDBColumnValue object is extracted
 * using RDBColumnValue.get methods.
 * @see QueryBuilder
 * @author Jay Gunter
 */
public class RDBColumnValue extends RDBColumn {

    /**
     * The constructor for RDBColumnValue is used only by QueryBuilder,
     * never by the application developer.
     * 
     * @return the RDBColumnValue object containing the value for the 
     * RDBColumn specified.
     * @param column the column specified by QueryBuilder during row access.
     * @param val the value for the column in the currently accessed row.
     * @see QueryBuilder 
     * @author Jay Gunter
     */
    public RDBColumnValue( RDBColumn column, Object val ) {
	super( column.table, column.name );
	value = val;
    }

    private Object value;
    /**
     * Return the column value as an Object.
     * Rarely used by the application developer
     * (only when the type of the column value is not known.)
     * 
     * @return the column value as an Object whose type can be interrogated.
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public Object getValue() {
	return value;
    }

    /**
     * Used only by the QueryBuilder.
     * Sets the value for the RDBColumnValue object.
     * 
     * @param o the value as an Object
     * @author Jay Gunter
     */
    public void   setValue( Object o ) {
	value = o;
    }

    /**
     * Retreive column value as a String.
     * 
     * @return the column value as a String.
     * @exception ColumnTypeException If the column value is not a String
     * (if the column type is not CHAR, VARCHAR, VARCHAR2, etc.)
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public String getString()
    throws ColumnTypeException {
	if ( ! ( value instanceof String ) )
	    throw new ColumnTypeException( this, "String" );
	return (String) value;
    }

    /**
     * Retreive column value as an int.
     * 
     * @return the column value as an int.
     * @param nullOk if false and the column in the row contains 
     * a database NULL value, a ColumnNullException is thrown.
     * If true and the column in the row contains a NULL, 0 is returned.
     * @exception ColumnTypeException If the column value is not an Integer
     * (if the column type is not INTEGER, LONG, etc.) or if NULL is not
     * an allowed value (nullOk==false.)
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public int getInteger( boolean nullOk )
    throws ColumnTypeException, ColumnNullException {
	if ( ! ( value instanceof Integer ) )
	    throw new ColumnTypeException( this, "Integer" );
	if ( null == value )
	    if ( ! nullOk ) throw new ColumnNullException( this );
	    else            return 0;
	return ( (Integer) value ).intValue();
    }

    /**
     * Retreive column value as an int.
     * 
     * @return the column value as an int.  If the column value is NULL,
     * 0 is returned.
     * @exception ColumnTypeException If the column value is not an Integer
     * (if the column type is not INTEGER, LONG, etc.)
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public Integer getInteger()
    throws ColumnTypeException {
	if ( ! ( value instanceof Integer ) )
	    throw new ColumnTypeException( this, "Integer" );
	return (Integer) value;
    }

    /**
     * Retreive column value as an double.
     * 
     * @return the column value as an double.
     * @param nullOk if false and the column in the row contains 
     * a database NULL value, a ColumnNullException is thrown.
     * If true and the column in the row contains a NULL, 0 is returned.
     * @exception ColumnTypeException If the column value is not a Double
     * (if the column type is not FLOAT, DOUBLE, NUMERIC, DECIMAL, etc.)
     * or if NULL is not an allowed value (nullOk==false.)
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public double getDouble( boolean nullOk )
    throws ColumnTypeException, ColumnNullException {
	if ( null == value )
	    if ( ! nullOk ) throw new ColumnNullException( this );
	    else            return 0;
	Double d = getDouble();
	return d.doubleValue();
    }

    /**
     * Retreive column value as an double.
     * 
     * @return the column value as an double.  If the column value is NULL,
     * 0 is returned.
     * @exception ColumnTypeException If the column value is not a Double
     * (if the column type is not FLOAT, DOUBLE, NUMERIC, DECIMAL, etc.)
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public Double getDouble()
    throws ColumnTypeException {
	if ( value instanceof Double )
	    return (Double) value;
	if ( value instanceof BigDecimal ) {
	    BigDecimal b = (BigDecimal) value;
	    return (Double) new Double( b.toString() );
	}
	throw new ColumnTypeException( this, "Double/BigDecimal" );
    }

    /**
     * Retreive column value as a BigDecimal.
     * 
     * @return the column value as a BigDecimal.
     * If the column value is NULL, 0 is returned.
     * @exception ColumnTypeException If the column value is not a BigDecimal
     * (if the column type is not NUMERIC, or DECIMAL.)
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public BigDecimal getBigDecimal()
    throws ColumnTypeException {
	if ( ! ( value instanceof BigDecimal ) )
	    throw new ColumnTypeException( this, "BigDecimal" );
	return (BigDecimal) value;
    }

}
