/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: RDBRow.java,v 1.2 2000/01/29 07:01:38 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.dods.builder.generator.query;

/**
 * An instance of RDBRow is returned by the QueryBuilder.getNextRow() method.
 * An RDBRow contains RDBColumnValue objects corresponding to the 
 * RDBColumn objects used to construct the instance of QueryBuilder.
 * in a particular table in the database.
 * @see QueryBuilder
 * @author Jay Gunter
 */
public class RDBRow {

    private RDBColumnValue[] values;
    /**
     * Constructor used only by QueryBuilder, never the application developer.
     * 
     * @param vals the column values in the current row
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public RDBRow( RDBColumnValue[] vals ) {
	if ( null == vals )
	    values = new RDBColumnValue[0];
	else
	    values = vals;
    }

    /**
     * Returns an RDBColumnValue object for a column in the current row.
     * Application developers are encouraged to use the 
     * RDBRow.get(RDBColumn) method instead because specifying the 
     * desired column value using RDBColumn is less error-prone
     * than using an int offset into the columns.
     * 
     * @return the value of the ith column
     * @param i the index into the row of the column value desired.
     * @exception InvalidRowColumnException If i is not a valid column number.
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public RDBColumnValue get( int i ) 
    throws InvalidRowColumnException {
	if ( ! ( 0 <= i && i < values.length ) )
	    throw new InvalidRowColumnException( "" + i );
	return values[i];
    }

    /**
     * Returns an RDBColumnValue object for a column in the current row.
     * This is the preferred method for use by application developers.
     * 
     * @return the value of the specified column
     * @param col the column value desired.
     * @exception InvalidRowColumnException 
     * If col is not a valid column in the row.
     * @see QueryBuilder
     * @author Jay Gunter
     */
    public RDBColumnValue get( RDBColumn col )
    throws InvalidRowColumnException {
	if ( null == col )
	    throw new InvalidRowColumnException( "Null RDBColumn" );
	for ( int i = 0; i < values.length; i++ ) {
	    if ( values[i].equals( col ) )
		return values[i];
	}
	throw new InvalidRowColumnException( col.getFullColumnName() );
    }
}
