/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: ConnectionAllocator.java,v 1.10 2000/01/29 07:01:47 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql;

import com.lutris.appserver.server.LBS;
import java.util.Date;	// Must appear before java.sql.Date
import java.sql.*;


/**
 * Defines the connection allocator. A connection allocator is an integral
 * part of a logical database implementation. It manages a pool of database
 * connections.
 * 
 * @author	Paul Morgan
 * @since	LBS1.8
 * @version	$Revision: 1.10 $
 */
public interface ConnectionAllocator {

    /**
     * Allocate a connection either from the managed pool or create a new.
     *
     * @exception java.sql.SQLException
     *   if a SQL error occures.
     */
    DBConnection allocate () throws java.sql.SQLException;


    /**
     * Return a connection to the pool.  If it is of an old generation,
     * close and drop.
     *
     * @param dbConnection The connection object to return.
     */
    void release (DBConnection dbConnection);


    /**
     * Used to drop a connection from the pool. The connection
     * should not be re-allocated
     *
     * @param dbConnection
     *   The connection object to drop.
     */
    void drop (DBConnection dbConnection);
    

    /**
     * Drop all the connection in the pool immediately.
     */
    void dropAllNow ();


    /**
     * Return the number of currently active (allocated) connections.
     *
     * @return The number of connections.
     */
    public int getActiveCount ();


    /**
     * Return the maximum number of connections active at one time.
     *
     * @return The number of connections.
     */
    public int getMaxCount ();


    /**
     * Return the time when the maximum connection count occured.
     *
     * @return The <CODE>Date</CODE> when the maximum connection
     *  count occured.
     */
    public Date getMaxCountDate ();


    /**
     * Reset the maximum connection count and date.
     */
    public void resetMaxCount ();


    /**
     * Return the number of database requests made on any of the
     * connection allocated from this managed pool. Implementations
     * of the connection allocator must provide a mechanism for
     * <CODE>DBQuery</CODE> and <CODE>DBTransaction</CODE> objects
     * to count requests.
     *
     * @return The number of database requests (queries or transactions).
     */
    public long getRequestCount ();
}
