/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DBQuery.java,v 1.17 2000/01/29 07:01:48 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql;

import java.sql.*;

/**
 * Utility for querying object from a database.  Allocates connections from
 * the database connection pool, ensures the integrity of those
 * connections, and manages result sets after a
 * <A HREF=#query>query</A> is performed.  Returns objects representing
 * data in the database.
 *
 * <P>Example - querying a user:
 * <PRE>
 * 
 * 	import com.lutris.appserver.server.LBS;
 * 	import com.lutris.appserver.server.sql.*;
 *
 *      DBQuery dbQuery =
 *      	LBS.getDatabaseManager().createQuery(DATABASE_NAME);
 *      	
 *      // NOTE: class CustomerQuery implements Query { ... }
 *      CustomerQuery
 *          customerQuery = new CustomerQuery();
 *      String [] loginIds = { "customer1", "customer2" };
 *
 *      try {
 *          for (int idx=0; idx &lt loginIds.length; idx++) {
 *
 *              customerQuery.setQueryLoginId(loginIds[idx]);
 *              dbQuery.query(customerQuery); // query the database
 *
 *              // Print all query results.
 *              CustomerDO customerResult;
 *              while ((customerResult =
 *                     (CustomerDO)dbQuery.next()) != null) {
 *                  System.out.println("Customer name for " + 
 *                                      loginIds[idx] + 
 *                                     " is " + customerResult.getName());
 *              }
 *
 *          }
 *      }
 *      finally {
 *          // 
 *          // Return database connections used by
 *          // this object back to the connection pool
 *          // and free any other resources consumed
 *          // by this object.
 *          // 
 *          dbQuery.release();
 *      }
 *
 * </PRE>
 *
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.17 $
 */
public interface DBQuery {

    /**
     * Query the database.
     *
     * @param q
     *   Query interface via which the query will be executed.
     * @exception SQLException
     *   If a database access error occurs. 
     */
    public void query(Query q)
        throws SQLException;


    /**
     * Returns a new object representing the next result form
     * the query.  
     *
     * @return
     *   New instance of object representing query results.
     *   <I>null</I> is returned when there are no more results.
     * @exception SQLException
     *   If a database access error occurs. 
     * @exception ObjectIdException
     *   If ObjectId was not found.
     */
    public Object next()
	throws SQLException, ObjectIdException;


    /**
     * Frees all resources consumed by this query.
     * Connections are returned to the connection pool.
     * Subsequent queries via this object,
     * will throw an exception.
     */
    public void release();


    /**
     * Exception handler.  This object is should not be
     * used for subsequent queries if this method returns
     * false.
     *
     * @return
     *   True if the exception can be handled and the object is
     *   still valid, false otherwise.
     */
    public boolean handleException(SQLException e);


    /**
     * Method to ensure this object is still valid.
     * Once this object has been released it cannot be
     * used any more.
     * 
     * @exception SQLException
     *   If a database access error occurs. 
     */
    public void validate()
        throws SQLException;
}
