/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DBTransaction.java,v 1.17 2000/01/29 07:01:48 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql;

import java.sql.*;

/**
 * Used to perform database transactions.
 *
 * <P>Example - adding a new user:
 * <PRE>
 * 	 import com.lutris.appserver.server.LBS;
 * 	 import com.lutris.appserver.server.sql.*;
 *
 *       DBTransaction transaction =
 *       	LBS.getDatabaseManager().createTransaction(DATABASE_NAME);
 *
 *       // NOTE: class CustomerDO implements Transaction { ... }
 *       // NOTE: An Object ID is automatically calculated by the constructor.
 *       CustomerDO customer = new CustomerDO();
 *       customer.setFirstName("Santa");
 *       customer.setLastName("Claus");
 *
 *       // ... set all other CustomerFields ...
 *
 *       //
 *       // Now add the new object to the database.
 *       //
 *       try {
 *           transaction.insert(customer);
 *           transaction.commit();
 *           System.out.println("Object ID is " + customer.getOId());
 *       }
 *       catch (SQLException e) {
 *           transaction.rollback();
 *           throw e;
 *       }
 *       finally {
 *           transaction.release();
 *       }
 * </PRE>
 *
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.17 $
 */
public interface DBTransaction {

    /**
     * Method to update an object in the database.
     *
     * @param transaction
     *   Object that implements transaction interface.
     */
    public void update(Transaction transaction);


    /**
     * Method to delete an object in the database.
     *
     * @param transaction
     *   Object that implements transaction interface.
     */
    public void delete(Transaction transaction);


    /**
     * Method to insert an object in the database.
     *
     * @param transaction
     *   Object that implements transaction interface.
     */
    public void insert(Transaction transaction);


    /**
     * Method to commit upates.
     *
     * @exception java.sql.SQLException If a database access error occurs. 
     * @exception DBRowUpdateException If a version error occurs.
     */
    public void commit() throws SQLException, DBRowUpdateException;     


    /**
     * Method to rollback changes.
     *
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     */
    public void rollback() throws SQLException;


    /**
     * Frees all resources consumed by this transaction
     * Connections are returned to the connection pool.
     * Subsequent transactions via this object,
     * will allocate a new set of resources (i.e. connection).
     */
    public void release();


    /**
     * Exception handeler.  This object is should not be
     * used for subsequent queries if this method returns
     * false.
     *
     * @return boolean True if the exception can be handeled
     *   and the object is still valid, false otherwise.
     */
    public boolean handleException(SQLException e);
}
