/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DatabaseManager.java,v 1.13 2000/01/29 07:01:49 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.sql;

import java.util.Date;
import java.sql.*;

/**
 * The database management object interface. This class implementing this
 * interface manages the database connections for one application.
 *
 * @version	$Revision: 1.13 $
 * @author	Paul Morgan
 */
public interface DatabaseManager {

    /**
     * Flag to enable debug logging of queries and transactions.
     */
    boolean debug = false;


    /**
     * Allocate a connection to the specified logic database. The
     * connection should be returned to the pool by calling its 
     * <a href=com.lutris.appserver.server.sql.DBConnection#release>
     * release()</a> function.  A thread will wait if no connections
     * are available.  Interupted exceptions are converted to errors.
     *
     * @param dbName
     *   Logical name of the database to allocate a connection to.
     * @return The allocated connection object.
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     * @exception SQLException
     *   If a SQL error occures.
     */
    public DBConnection allocateConnection (String dbName) 
        throws DatabaseManagerException, SQLException;


    /**
     * Allocate a connection to the default logical database. The
     * connection should be returned to the pool by calling its 
     * <a href=com.lutris.appserver.server.sql.DBConnection#release>
     * release()</a> function.  A thread will wait if no connections
     * are available.  Interupted exceptions are converted to errors.
     *
     * @return The allocated connection object.
     * @exception DatabaseManagerException 
     *   If a nonexistent default logical database name is supplied.
     * @exception SQLException
     *   If a SQL error occures.
     * @see
     *   #setDefaultDatabase
     */
    public DBConnection allocateConnection () 
        throws DatabaseManagerException, SQLException;


    /**
     * Allocate an object id from the specified logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain an object id.
     * @return The allocated unique OID
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     * @exception ObjectIdException 
     *   If a problem (e.g. SQL error) occured in obtaining the OID.
     */
    public ObjectId allocateObjectId (String dbName) 
        throws DatabaseManagerException, ObjectIdException;


    /**
     * Allocate an object id from the specified logical database.
     *
     * @return The allocated connection object.
     * @exception DatabaseManagerException 
     *   If a nonexistent default logical database has been set.
     * @exception ObjectIdException 
     *   If a problem (e.g. SQL error) occured in obtaining the OID.
     * @see
     *   #setDefaultDatabase
     */
    public ObjectId allocateObjectId () 
        throws DatabaseManagerException, ObjectIdException;


    /**
     * Create a transaction object for the specified logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a transaction.
     * @return The transaction
     * @exception DatabaseManagerException 
     *   If a nonexistent or invalid logical database name is supplied.
     * @exception SQLException 
     *   If a problem occured creating the transaction.
     */
    public DBTransaction createTransaction (String dbName) 
        throws DatabaseManagerException, SQLException;


    /**
     * Create a transaction object for the default logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a transaction.
     * @return The transaction
     * @exception DatabaseManagerException 
     *   If a nonexistent default logical database has been set.
     * @exception SQLException 
     *   If a problem occured creating the transaction.
     * @see
     *   #setDefaultDatabase
     */
    public DBTransaction createTransaction () 
        throws DatabaseManagerException, SQLException;


    /**
     * Create a query object for the specified logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a query.
     * @return The query
     * @exception DatabaseManagerException 
     *   If a nonexistent or invalid logical database name is supplied.
     * @exception SQLException 
     *   If a problem occured creating the query.
     */
    public DBQuery createQuery (String dbName) 
        throws DatabaseManagerException, SQLException;


    /**
     * Create a query object for the default logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a query.
     * @return The query
     * @exception DatabaseManagerException 
     *   If a nonexistent default logical database has been set.
     * @exception SQLException 
     *   If a problem occured creating the query.
     * @see
     *   #setDefaultDatabase
     */
    public DBQuery createQuery () 
        throws DatabaseManagerException, SQLException;


    /**
     * Set the default logical database. This should be used with
     * caution, but it makes allocating connections easier.
     *
     * @param dbName
     *   The default logical dabase name.
     * @exception DatabaseManagerException 
     *   If a nonexistent or illegal logical database name is supplied.
     */
    public void setDefaultDatabase(String dbName)
        throws DatabaseManagerException;


    /**
     * Shutdown the database manager. All logical databases will be
     * shutdown and all connection closed.
     */
    public void shutdown();


    //====================================================================
    // The following are primarily for management purposes...
    //====================================================================

    /**
     * Returns the list of managed logical databases.
     * 
     * @return	List of logical database names.
     */
    public String[] getLogicalDatabaseNames();


    /**
     * Returns a description of the logical database type.
     * 
     * @param dbName
     *   The logical database name.
     * @return
     *   A text description of the logical database type.
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     */
    public String getType(String dbName)
	throws DatabaseManagerException;
    

    /**
     * Returns the number of requests made to the database since startup time.
     * 
     * @param dbName
     *   The logical database name.
     * @return
     *   The number of database requests since the server started.
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     */
    public long getRequestCount(String dbName)
	throws DatabaseManagerException;

    
    /**
     * Returns the number of currently active connections for the
     * supplied logical database name.
     * If not implemented, then -1 is returned.
     * 
     * @param dbName
     *   The logical database name.
     * @return
     *   The number of currently active connections.
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     */
    public int getActiveConnectionCount(String dbName)
	throws DatabaseManagerException;
   

    /**
     * Returns the maximum number of concurent connections that existed
     * at any time since this object was created, or
     * <CODE>resetMaxConnectionCount()</CODE> was called.
     * This is a historical highwater mark. 
     * If not implemented, then -1 is returned.
     *
     * @param dbName
     *   The logical database name.
     * @return
     *   The highwater mark for number of connections, or -1.
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     */
    public int getMaxConnectionCount(String dbName)
	throws DatabaseManagerException;


    /**
     * Returns the time when the maximum refered to by 
     * <CODE>maxConnectionCount()</CODE> occured.
     * If not implemented, then null is returned.
     *
     * @param dbName
     *   The logical database name.
     * @return
     *   The Date of when the maximum number of connections occured.
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     */
    public Date getMaxConnectionCountDate(String dbName)
	throws DatabaseManagerException;


    /**
     * Reset the maximum connection count. See
     * <CODE>maxConnectionCount()</CODE>. The highwater mark should be
     * reset to the current number of connections.
     *
     * @param dbName
     *   The logical database name.
     * @exception DatabaseManagerException 
     *   If a nonexistent logical database name is supplied.
     */
    public void resetMaxConnectionCount(String dbName)
	throws DatabaseManagerException;
}
