/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: LogicalDatabase.java,v 1.10 2000/01/29 07:01:50 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql;

import java.util.Date;		// Must appear before java.sql.Date
import java.sql.*;
import com.lutris.util.Config;
import com.lutris.util.ConfigException;

/**
 * Interface for a logical database. Each logical database must
 * provide services for connection and object id allocation.
 *
 * @author	Paul Morgan
 * @since	LBS1.8
 * @version	$Revision: 1.10 $
 */
public interface LogicalDatabase {

    /**
     * Initialize the logical database.
     *
     * @exception ConfigException
     *  if there is an error in the configuration.
     * @exception SQLException
     *  if a SQL error occurs.
     */
    public void init (String dbName, Config dbConfig)
	throws ConfigException, SQLException;

    /**
     * Return a connection to this logical database.
     *
     * @return The connection.
     * @exception SQLException
     *  if a SQL error occurs.
     */
    public DBConnection allocateConnection ()
	throws SQLException;

    /**
     * Return an object id for this logical database.
     *
     * @return The object id.
     * @exception ObjectIdException
     *  if an object id cannot be allocated.
     */
    public ObjectId allocateObjectId ()
	throws ObjectIdException;

    /**
     * Return a transaction for use on this logical database.
     *
     * @return The transaction object.
     * @exception SQLException
     *   if a SQL error occurs.
     */
    public DBTransaction createTransaction ()
	throws SQLException;

    /**
     * Return a query for use on this logical database.
     *
     * @return The query object.
     * @exception SQLException
     *   if a SQL error occurs.
     */
    public DBQuery createQuery ()
	throws SQLException;

    /**
     * Immediately shutdown the logical database. All connections
     * should be closed and any other cleanup performed.
     */
    public void shutdown ();
    

    //===================================================================
    // The following are primaryily for management purposes...
    //===================================================================

    /**
     * Return the symbolic name of this logical database.
     *
     * @return The symbolic name.
     */
    public String getName ();

    /**
     * Return a description of the logical database type.
     *
     * @return The type.
     */
    public String getType ();

    /**
     * Return the number of currently active connections.
     *
     * @return The number of connections.
     */
    public int getActiveConnectionCount ();

    /**
     * Return the maximum number of connections active at one time.
     *
     * @return The number of connections.
     */
    public int getMaxConnectionCount ();

    /**
     * Return the time when the maximum connection count occured.
     *
     * @return The <CODE>Date</CODE> when the maximum connection
     *  count occured.
     */
    public Date getMaxConnectionCountDate ();

    /**
     * Reset the maximum connection count and date.
     */
    public void resetMaxConnectionCount ();

    /**
     * Return the number of database requests.
     *
     * @return The number of database requests (queries or transactions).
     */
    public long getRequestCount ();
}
