/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: ObjectId.java,v 1.12 2000/01/29 07:01:51 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql;

import java.math.*;
import java.io.Serializable;

/**
 * Represents an object id used by LBS data objects. The object id
 * must be unique. The combination of database URL and object id
 * constitutes a GUID. The maximum value of an object id is
 * DECIMAL(19,0)
 *
 * @version	$Revision: 1.12 $
 * @author	Kyle Clark
 * @since	LBS1.7
 * @see		ObjectIdAllocator
 */
public class ObjectId implements Serializable {

    /**
     * The value associated with this object id.
     */
    private BigDecimal value;

    /**
     * The value of one.
     */
    static final public BigDecimal ONE = new BigDecimal("1");

    /**
     * The maximum value that can be associated with an object
     * id in LBS - DECIMAL(19,0).
     */
    static final public BigDecimal MAX = new BigDecimal("9999999999999999999");


    /**
     * Translates a string containing one or more integers
     * of the specified radix into an ObjectID.  The string
     * may not represet a negative number.
     *
     * @param val
     *   The string representation of the number.
     *   The character to digit mapping is provided by
     *   Character.digit()
     * @param radix
     *   Must be between Character.MIN_RADIX(2) and
     *   Character.MAX_RADIX(36).
     * @exception java.lang.NumberFormatException
     *   If the string representation contains invalid characters.
     * @exception ObjectIdException
     *   If val represents a negative number.
     */
    public ObjectId(String val, int radix)
        throws ObjectIdException, NumberFormatException
    {

        value = new BigDecimal(new BigInteger(val, radix), 0);
        if (value.signum() < 0)
            throw new ObjectIdException("Object IDs cannot be negative.");
        if (value.compareTo(MAX) > 0)
            throw new ObjectIdException("Object IDs cannot exceed " + MAX.toString());

    }


    /**
     * Translates a string containing one or more decimal digits
     * into an ObjectID.  
     *
     * @param val
     *   The string representation of the decimal number that
     *   
     *   The character to digit mapping is provided by
     *   Character.digit()
     * @exception java.lang.NumberFormatException
     *   If the string representation contains invalid characters.
     * @exception ObjectIdException
     *   If val represents a negative number.
     */
    public ObjectId(String val)
        throws ObjectIdException, NumberFormatException
    {

        value = new BigDecimal(new BigInteger(val), 0);
        if (value.signum() < 0)
            throw new ObjectIdException("Object IDs cannot be negative.");
        if (value.compareTo(MAX) > 0)
            throw new ObjectIdException("Object IDs cannot exceed " + MAX.toString());

    }


    /**
     * Translates a long into an ObjectID.  
     *
     * @param val
     *   The value to assign to the object id.
     * @exception ObjectIdException
     *   If val is a negative number.
     */
    public ObjectId(long val)
        throws ObjectIdException
    {

        if (val < 0)
            throw new ObjectIdException("Object IDs cannot be negative.");
        value = BigDecimal.valueOf(val);
        if (value.compareTo(MAX) > 0)
            throw new ObjectIdException("Object IDs cannot exceed " + MAX.toString());

    }


    /**
     * Creates and object id whose value is the same as val.
     *
     * @param val
     *   The value to assign to the object id.
     * @exception ObjectIdException
     *   If val is a negative number or the scale of val
     *   is greater than zero.
     */
    public ObjectId(BigDecimal val)
        throws ObjectIdException
    {
        if (val.signum() < 0)
            throw new ObjectIdException("Object IDs cannot be negative.");
        if (val.scale() > 0)
            throw new ObjectIdException("Object IDs cannot have a scale " +
                                        "greater than zero.");
        if (val.compareTo(MAX) > 0)
            throw new ObjectIdException("Object IDs cannot exceed " + MAX.toString());
        value = val;
    }

    /**
     * Returns an object id whose value is (this+val).
     *
     * @param val
     *   The value to add to this object.
     * @exception ObjectIdException
     *   If the result of the addition would result
     *   in an object id that exceeds the
     *   <a href=#MAX>maximum</a> object id
     *   size.
     */
    public ObjectId add(ObjectId val)
        throws ObjectIdException
    {
        return new ObjectId(value.add(val.toBigDecimal()));
    }


    /**
     * Returns an object id whose value is (this+val).
     *
     * @param val
     *   The value to add to this object.
     * @exception ObjectIdException
     *   If the result of the addition would result
     *   in an object id that exceeds the
     *   <a href=#MAX>maximum</a> object id
     *   size.
     */
    public ObjectId add(long val)
        throws ObjectIdException
    {
        return new ObjectId(value.add(BigDecimal.valueOf(val)));
    }



    /**
     * Returns an object id whose value is (this+1)
     *
     * @exception ObjectIdException
     *   If the result of the addition would result
     *   in an object id that exceeds the
     *   <a href=#MAX>maximum</a> object id
     *   size.
     */
    public ObjectId increment()
        throws ObjectIdException
    {
        return new ObjectId(value.add(ONE));
    }


    /**
     * Returns a big decimal representation of the object id.
     */
    public BigDecimal toBigDecimal()
    {
        return value;
    }


    /**
     * Test if this object id is equal to another object id.
     */
    public boolean equals(ObjectId oid)
    {
        return (oid.toBigDecimal().compareTo(value) == 0);
    }

    /**
     * Returns a hash code for this object id. 
     * 
     * @return a hash code for this object id. 
     */
    public int hashCode()
    {
        return toString().hashCode();
    }

    /**
     * String representation of this object id.
     */
    public String toString()
    {
        return value.toString();
    }
}
