/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: InformixDBConnection.java,v 1.9 2000/01/29 07:01:55 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql.informix;

import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;
import com.lutris.logging.*;
import java.sql.*;
import java.util.*;

/**
 * An Informix database connection.
 *
 * @see		DBConnection
 * @author	Kyle Clark
 * @since	LBS1.7
 * @version	$Revision: 1.9 $
 */
public class InformixDBConnection extends StandardDBConnection {

    /**
     * The amound time to wait on a query or transaction.
     */
    private int lockModeWait = 0;

    /**
     * The logical database for this connection.
     */
    protected InformixLogicalDatabase logicalDatabase;

    /**
     * Create a connection to an Informix database.
     *
     * @param url JDBC URL of database.
     * @param user SQL user name.
     * @param password SQL password.
     * @param maxPreparedStatements  Maximum number of preparse statements.
     *  A value of less than zero queries JDBC for the value.
     * @param connectionAllocator The connection allocator that this
     *  connection belongs to.
     * @param logging Specifying <CODE>true</CODE> enables SQL logging.
     * @param generation A generation number used to drop old connection
     *  when they are released.
     * @exception java.sql.SQLException If a connection can't be established.
     */
    protected InformixDBConnection(InformixConnectionAllocator connectionAllocator,
                                   String url, String user, String password,
                                   int maxPreparedStatements,
                                   boolean logging, int generation)
        throws SQLException {
        super(connectionAllocator, url, user, password, maxPreparedStatements,
              logging, generation);
        this.logicalDatabase = connectionAllocator.logicalDatabase;
    }

    /**
     * Specify if you prefer to wait for locks instead of generating
     * errors.
     *
     * @param on
     *   True if you prefer to wait, false otherwise.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     */
    public synchronized void setLockModeToWait(boolean on)
        throws SQLException
    {
        if (on) {
            setLockModeToWait(-1);
        } else {
            setLockModeToWait(0);
        }
    }

    /**
     * Specify if you prefer to wait for locks for the specified
     * amount of time before an exception is thrown.
     *
     * @param seconds
     *   The number of seconds to wait.  If seconds is == 0
     *   the waiting is disabled.  If seconds is less than
     *   0 then wait indefinitly.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     */
    public synchronized void setLockModeToWait(int seconds)
        throws SQLException
    {
        if (lockModeWait != seconds) {
            logDebug("set lock mode to wait " + seconds);            
            if (seconds > 0) {
                execute("SET LOCK MODE TO WAIT " + seconds);
            } else if (seconds == 0) {
                execute("SET LOCK MODE TO NOT WAIT");
            } else {
                execute("SET LOCK MODE TO WAIT");
            }
            lockModeWait = seconds;
        }
    }

}
