/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: InformixDBQuery.java,v 1.9 2000/01/29 07:01:55 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql.informix;

import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;
import com.lutris.logging.*;
import java.sql.*;
import java.util.*;

/**
 * An Informix database query object.
 *
 * @author	Kyle Clark
 * @since	LBS1.7
 * @version	$Revision: 1.9 $
 */
public class InformixDBQuery extends StandardDBQuery {

    /**
     * Amount of time (seconds) to wait on locks before
     * throwing an exception.
     * @see setQueryTimeOut
     */
    private int queryTimeout = 0;

    /**
     * The informix database connection.
     */
    protected InformixDBConnection conn;

    /**
     * Private constructor.
     *
     * @param dbConnection
     *   The database connection to use.
     * @exception SQLException
     *   If a database access error occurs. 
     */
    protected InformixDBQuery(InformixDBConnection dbConnection)
        throws SQLException
    {
        super(dbConnection);
        this.conn = dbConnection;
        try {
            setQueryTimeout(dbConnection.logicalDatabase.getDefaultQueryTimeout());
        } catch (SQLException sqlExcept) {
            handleException(sqlExcept);
            dbConnection.release();
            throw sqlExcept;
        }
    }


    /**
     * Public constructor.  Establishes a connection with the default
     * logical database.  The connection manager must have been
     * initialized with a default logical database name.
     *
     *
     * @exception SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If the default logical database wasn't initialized in the
     *   Database Manager.
     * @see
     *   com.lutris.appserver.server.sql.DatabaseManager#setDefaultDatabase
     */
    public InformixDBQuery()
        throws SQLException, DatabaseManagerException
    {
        this((InformixDBConnection)
             LBS.getApplication().getDatabaseManager().allocateConnection());
    }


    /**
     * Public constructor.  Establishes a connection to the specified
     * logical database name.
     *
     * @param dbName
     *   Logical database name to access.
     * @exception SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If a nonexistant logical database name is supplied.
     */
    public InformixDBQuery(String dbName)
        throws SQLException, DatabaseManagerException
    {
        this((InformixDBConnection)LBS.getApplication().getDatabaseManager().
             allocateConnection(dbName));
    }

    /**
     * Set query time out.  If a query blocks, then it will only
     * block as long as specified here.
     *
     * @param seconds
     *   The amount of time to block.
     * @exception SQLException if a database error occurs.
     */
    public void setQueryTimeout(int seconds) throws SQLException
    {
        try {
            logDebug("set query time out to " + seconds);
            conn.setLockModeToWait(queryTimeout);
            queryTimeout = seconds;
        }
        catch (SQLException e) {
            handleException(e);
            throw e;
        }
    }

}
