/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: InformixDBTransaction.java,v 1.9 2000/01/29 07:01:56 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql.informix;

import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;
import com.lutris.logging.*;
import java.sql.*;
import java.util.*;

/**
 * Informix implementation of SQL database transaction.
 *
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.9 $
 */
public class InformixDBTransaction extends StandardDBTransaction {

    /**
     * The connection to the informix database.
     */
    protected InformixDBConnection conn;

    /**
     * Private constructor. Used if connection is already known.
     *
     * @param conn
     *   The database connection to use.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     */
    protected InformixDBTransaction(InformixDBConnection conn)
        throws SQLException
    {
        super(conn);
        this.conn = conn;
        try {
            setTransactionTimeout(conn.logicalDatabase.
                                  getDefaultTransactionTimeout());
        } catch (SQLException sqlExcept) {
            conn.handleException(sqlExcept);
            conn.release();
            throw sqlExcept;
        }
    }


    /**
     * Public constructor.  Establishes a connection with the default
     * logical database.  
     * The connection manager must have been initialized with
     * a default logical database name.
     *
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If the default logical database wasn't initialized in the
     *   Database Manager.
     * @exception ClassCastException
     *   If the default database manager doesn't allocate connections
     *   of type InformixDBConnection.
     * @see DatabaseManager#setDefaultDatabase
     */
    public InformixDBTransaction()
	throws SQLException, DatabaseManagerException {

        this((InformixDBConnection)
             LBS.getApplication().getDatabaseManager().allocateConnection());

    }

    /**
     * Public constructor.  Establishes a connection with the specified
     * logical database name.
     *
     * @param dbName
     *   Logical database name to access.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     */
    public InformixDBTransaction(String dbName)
	throws SQLException, DatabaseManagerException {

        this((InformixDBConnection)LBS.getApplication().getDatabaseManager().
             allocateConnection(dbName));
    }

    /**
     * Set transaction time out.  If a transaction blocks, then it will only
     * block as long as specified here.
     *
     * @param seconds
     *   The amount of time to block in seconds.  If less than zero then blocks
     *   indefinitly.
     * @exception SQLException if a database error occurs.
     */
    public void setTransactionTimeout(int seconds) throws SQLException {
        try {
            logDebug("set transaction time out to " + seconds);
            conn.setLockModeToWait(seconds);
        } catch (SQLException e) {
            handleException(e);
            throw e;
        }
    }
}
