/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: InformixLogicalDatabase.java,v 1.8 2000/01/29 07:01:56 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql.informix;

import java.util.Date;	// Must be before java.sql.Date
import java.sql.*;
import com.lutris.util.*;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;

/**
 * Represents a logical database with Informix capabilities.
 *
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.8 $
 */
public class InformixLogicalDatabase extends StandardLogicalDatabase {

    /**
     * The amount of time (in seconds) that
     * a query will block before throwing an exception. If <= 0 then the
     * query will not block. Optional, if not specified, then the value
     * defaults to 0.
     */
    protected int defaultQueryTimeout = 0;

    /**
     * The amount of time (in seconds)
     * that a transaction will block before throwing an exception. If
     * <= 0 then the transaction will not block. Optional, if not specified,
     * then the value defaults to 0.
     */
    protected int defaultTransactionTimeout = 0;

    /**
     * Default constructor to configure a single logical database.
     * Note that the <CODE>init()</CODE> method must be called to
     * configure the database.
     */
    public InformixLogicalDatabase () {
        super();
    }

    /**
     * Creates and configures a single logical database.
     *
     * @param dbName
     *   The logical name of the database.  Used to
     * @param dbConfig
     *   The configuration object for this logical Informix database.
     * <p>
     * The configuration data is specified in the section:
     * <CODE>DatabaseManager.DB.<I>dbName</I></CODE>
     * <p>
     * The Informix specific configuration parameters are:
     * <UL>
     * <LI> <CODE>QueryTimeout</CODE> - The amount of time (in seconds) that
     *	a query will block before throwing an exception. If < 0 then the
     *	query will block indefinitly. Optional, if not specified, then the value
     *	defaults to 0.
     * <LI> <CODE>TransactionTimeout</CODE> - The amount of time (in seconds)
     *	that a transaction will block before throwing an exception. If
     *	< 0 then the transaction will block indefinitly. Optional, if not specified,
     *	then the value defaults to 0.
     * </UL>
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @exception SQLException
     *   If a SQL error occurs.
     */
    public InformixLogicalDatabase (String dbName, Config dbConfig)
        throws ConfigException, SQLException {

        super(dbName, dbConfig);
        this.defaultQueryTimeout = dbConfig.getInt("QueryTimeout", 0);
        this.defaultTransactionTimeout =
            dbConfig.getInt("TransactionTimeout", 0);
    }
    

    /**
     * Return the connection allocator.
     *
     * @param connectionConfig
     *   The configuration object for the connection allocator.
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @return
     *   The connection allocator.
     */
    public ConnectionAllocator loadConnectionAllocator(Config connectionConfig)
        throws ConfigException {

	return new InformixConnectionAllocator (this, connectionConfig);

    }


    /**
     * Return the object id allocator.
     *
     * @param objIdConfig
     *   The configuration object for the object id allocator.
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @return
     *   The object id allocator.
     */
    public ObjectIdAllocator loadObjectIdAllocator (Config objIdConfig)
        throws ConfigException {

	return new InformixObjectIdAllocator (this, objIdConfig);

    }


    /**
     * Return a description of the logical database type.
     *
     * @return The type.
     */
    public String getType () {
	return "Informix JDBC";
    }

    /**
     * Returns the configured default amount of time (in seconds) that
     * a query should block before throwing an exception. If < 0 then the
     * query should block indefinitly.
     *
     * @return the default query timeout.
     */
    protected int getDefaultQueryTimeout() {
        return defaultQueryTimeout;
    }

    /**
     * Returns the configured default amount of time (in seconds)
     * that a transaction should block before throwing an exception. If
     * less than zero then the transaction should block indefinitly.
     *
     * @return the default transaction timeout.
     */
    protected int getDefaultTransactionTimeout() {
        return defaultTransactionTimeout;
    }

}
