/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: StandardDBQuery.java,v 1.7 2000/01/29 07:02:05 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql.standard;

import com.lutris.appserver.server.Enhydra;
import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.logging.*;
import com.lutris.util.*;
import java.sql.*;
import java.util.*;

/**
 * Standard implementation of a DBQuery object.
 *
 * @see		DBQuery
 * @author	Kyle Clark
 * @since	LBS1.7
 * @version	$Revision: 1.7 $
 */
public class StandardDBQuery implements DBQuery {

    /**
     * Identifier for this query object.
     */
    private int id;

    /**
     * Next available identifier.
     */
    private static int nextId;

    /**
     * Connection used by this query object.
     */
    private DBConnection conn;

    /**
     * Result set after performing a query.
     */
    private ResultSet resultSet = null;

    /**
     * The interface via which the query will be
     * executed.
     */
    private Query queryInterface = null;

    /**
     * Is this object still active, or has it been released.
     */
    private boolean released = false;

    /**
     * The log channel.
     */
    private LogChannel log;


    /**
     * Construct a query object for use on the supplied dB connection.
     *
     * @param dbConnection
     *   The database connection to use.
     * @exception SQLException
     *   If a database access error occurs.
     */
    protected StandardDBQuery(DBConnection dbConnection)
        throws SQLException {
        id = nextId++;
	logDebug("new instance");
	conn = dbConnection;
    }


//    /**
//     * Public constructor.  Establishes a connection with the default
//     * logical database.  The connection manager must have been
//     * initialized with a default logical database name.
//     *
//     *
//     * @exception SQLException
//     *   If a database access error occurs.
//     * @exception DatabaseManagerException
//     *   If the default logical database wasn't initialized in the
//     *   Database Manager.
//     * @see
//     *   com.lutris.appserver.server.sql.DatabaseManager#setDefaultDatabase
//     */
//    public StandardDBQuery()
//        throws SQLException, DatabaseManagerException
//    {
//        this(LBS.getApplication().getDatabaseManager().allocateConnection());
//    }
//
//
//    /**
//     * Public constructor.  Establishes a connection to the specified
//     * logical database name.
//     *
//     * @param dbName
//     *   Logical database name to access.
//     * @exception SQLException
//     *   If a database access error occurs.
//     * @exception DatabaseManagerException
//     *   If a nonexistant logical database name is supplied.
//     */
//    public StandardDBQuery(String dbName)
//        throws SQLException, DatabaseManagerException
//    {
//        this(LBS.getApplication().getDatabaseManager().allocateConnection(dbName));
//    }


    /**
     * Query the database.
     *
     * @param q
     *   Query interface via which the query will be executed.
     * @exception SQLException
     *   If a database access error occurs.
     */
    public synchronized void query(Query q)
        throws SQLException
    {
        logDebug("execute query");
        validate();
	conn.incrRequestCount();
        queryInterface = q;
        try {
            if (resultSet != null)
                resultSet.close();
            resultSet = queryInterface.executeQuery(conn);
        }
        catch (SQLException e) {
            handleException(e);
            throw e;
        }
    }


    /**
     * Returns a new object representing the next result form
     * the query.
     *
     * @return
     *   New instance of object representing query results.
     *   <I>null</I> is returned when there are no more results.
     * @exception SQLException
     *   If a database access error occurs.
     * @exception ObjectIdException
     *   If ObjectId was not found.
     */
    public Object next() throws SQLException, ObjectIdException {
        logDebug("get next result");
        validate();
        try {
            return queryInterface.next(resultSet);
        } catch (SQLException e) {
            handleException(e);
            throw e;
        }
    }


    /**
     * Frees all resources consumed by this query.
     * Connections are returned to the connection pool.
     * Subsequent queries via this object,
     * will throw an exception.
     */
    public synchronized void release() {
        logDebug("release");
        try {
            validate();
        } catch (SQLException except) {
            throw new FatalExceptionError(except);
        }

        try {

            //
            // Free all resources consumed by this connection
            // and return the connection to the connection
            // pool.
            //
            // NOTE: if a connection cannot handle and exception,
            // then it is automatically dropped from the connection
            // pool and closed.   This means that the connection
            // should no longer be used.
            //

            SQLException sqlEx = null;

            try {
                if (resultSet != null)
                    resultSet.close();
            }
            catch (SQLException e) {
                conn.handleException(e);
            }
            try {
                conn.reset();
            }
            catch (SQLException e) {
                conn.handleException(e);
            }

            //
            // Release the connection.
            //
            conn.release();

        }
        finally {
            released = true;
            resultSet = null;
            queryInterface  = null;
            conn = null;
        }
    }


    /**
     * Exception handler.  This object is should not be
     * used for subsequent queries if this method returns
     * false.
     *
     * @return
     *   True if the exception can be handled and the object is
     *   still valid, false otherwise.
     */
    public synchronized boolean handleException(SQLException e)
    {
        logDebug("handle exception");
        return conn.handleException(e);
    }


    /**
     * Method to ensure this object is still valid.
     * Once this object has been released it cannot be
     * used any more.
     *
     * @exception SQLException
     *   If a database access error occurs.
     */
    public void validate()
        throws SQLException
    {
        if (released)
            throw new SQLException("Cannot access this object " +
                                   "once it has been released.");
    }


    /**
     * Finalizer.
     * If this object has not been <A HREF=#release>released</A>,
     * this method ensures that garbage collection does so.
     */
    protected void finalize()
    {
	if (!released)
	    release();
    }


    /**
     * Logging. For debuging only, since it effects all Query objects.
     *
     * @param str
     *   The data to log.
     */
    protected void logDebug(String str) {
        if (Enhydra.getDatabaseManager().debug) {
	    if (log == null) log = Enhydra.getLogChannel();
            log.write(Logger.DEBUG, "DBQuery[" + id + "]: " + str);
        }
    }
}
