/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: StandardLogicalDatabase.java,v 1.6 2000/01/29 07:02:06 markd Exp $
 *-----------------------------------------------------------------------------
 */





package	com.lutris.appserver.server.sql.standard;

import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.util.*;
import java.util.Date;	// Must be before java.sql.Date
import java.sql.*;


/**
 * Represents a logical database. Each logical database has a connection
 * allocator and a object id allocator. The standard implementation of
 * a local database is used when the database is described as of
 * type <I>Standard</I> in
 * <CODE>DatabaseManager.DB.<I>dbname</I>.ClassType</CODE>
 * <P>
 * The configuration data is specified in the section:
 * <CODE>DatabaseManager.DB.<I>dbName</I></CODE>
 * <P>
 * The following is a description of the sub fields:
 * <UL>
 * <LI> <B><CODE>JdbcDriver</CODE></B> -
 *	The JDBC driver to use to access that database. Manditory.
 *	E.g. "intersolv.jdbc.sequelink.SequeLinkDriver"
 * <LI> <B><CODE>Connection</CODE></B> -
 *      This <I>section</I> is passed onto the connection allocator. See
 *     <A HREF="com.lutris.appserver.server.sql.StandardConnectionAllocator"</A>
 *	for details on connection allocator sub fields.
 * <LI> <B><CODE>ObjectId</CODE></B> -
 *      This <I>section</I> is passed onto the object id allocator. See
 *	<A HREF="com.lutris.appserver.server.sql.ObjectIdAllocator"</A>
 *	for details on object id allocator sub fields.
 * </UL>
 *
 * @author	Paul Morgan
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.6 $
 */
public class StandardLogicalDatabase implements LogicalDatabase {

    /**
     * Database connection allocator.
     */
    protected ConnectionAllocator connectionAllocator;

    /**
     * Object id manager.
     */
    protected ObjectIdAllocator objectIdAllocator;

    /**
     * Driver object object.  Important: a reference to this must be
     * maintained or the driver class might be garbage collected out from
     * under us.
     */
    protected Driver driver;

    /**
     * The symbolic name of this logical database.
     */
    protected String dbName;


    /**
     * Default constructor to configure a single logical database.
     * Note that the <CODE>init()</CODE> method must be called to
     * configure the database.
     */
    public StandardLogicalDatabase () {
    }

    /**
     * Creates and configures a single logical database.
     *
     * @param dbName
     *   The logical name of the database.  Used to
     * @param dbConfig
     *   The configuration object for this logical database.
     *   get config parameters.
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @exception SQLException
     *   If a SQL error occurs.
     */
    public StandardLogicalDatabase (String dbName, Config dbConfig)
        throws ConfigException, SQLException {
	    
	    this.init(dbName, dbConfig);
    }

    /**
     * Initialize a single logical database. Used after the default
     * constructor is called.
     *
     * @param dbName The logical name of the database.  Used to
     * @param dbConfig The configuration object for this logical database.
     *  get config parameters.
     * @exception ConfigException If there is an error in
     *   the configuration file.
     * @exception SQLException If a SQL error occurs.
     */
    public void init (String dbName, Config dbConfig)
        throws ConfigException, SQLException {

	this.dbName = dbName;

	/**
	 * Read configuration information.
	 */
	Config objIdConfig, connectionConfig;
	try {
	    objIdConfig = (Config) dbConfig.getSection ("ObjectId");
	    connectionConfig = (Config) dbConfig.getSection ("Connection");
	} catch (KeywordValueException except) {
	    throw new ConfigException("Invalid ObjectId or Connection sections in config file.");
	}
        String jdbcDriver = dbConfig.getString ("JdbcDriver");

	/**
         * Load JDBC driver; report an SQLException on failure.
         * No-op if already loaded.
         */
        try {
	    // We must use the applications class loader to load JDBC classes.
// FIX: the following has problems with oracle jdbc driver????
//	    ClassLoader appClassLoader =
//		LBS.getApplication().getClass().getClassLoader();
//	    Class driverClass = appClassLoader.loadClass(jdbcDriver);
	    Class driverClass = Class.forName(jdbcDriver);
	    driver = (Driver) driverClass.newInstance ();
        }
        catch (java.lang.ClassNotFoundException except) {
            throw new SQLException ("can't load JDBC driver class: " +
                                    except.getMessage ());
        }
        catch (java.lang.InstantiationException except) {
            throw new SQLException ("can't instantiate JDBC driver class: " +
                                    except.getMessage ());
        }
        catch (java.lang.IllegalAccessException except) {
            throw new SQLException ("can't instantiate JDBC driver class: " +
                                    except.getMessage ());
        }

        ConnectionAllocator connectionAllocator =
	    loadConnectionAllocator (connectionConfig);

	ObjectIdAllocator objectIdAllocator =
	    loadObjectIdAllocator (objIdConfig);

	this.connectionAllocator = connectionAllocator;
	this.objectIdAllocator = objectIdAllocator;
    }
    

    /**
     * Return the connection allocator.
     *
     * @param connectionConfig
     *   The configuration object for the connection allocator.
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @return
     *   The connection allocator.
     */
    public ConnectionAllocator
	loadConnectionAllocator (Config connectionConfig)
        throws ConfigException {
	    
	return (ConnectionAllocator)
	    new StandardConnectionAllocator (this, connectionConfig);
    }


    /**
     * Return the object id allocator.
     *
     * @param objIdConfig
     *   The configuration object for the object id allocator.
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @return
     *   The object id allocator.
     */
    public ObjectIdAllocator
	loadObjectIdAllocator (Config objIdConfig)
        throws ConfigException {
	    
	return (ObjectIdAllocator)
	    new StandardObjectIdAllocator (this, objIdConfig);
    }


    /**
     * Return the connection allocator object associated with this
     * logical database.
     *
     * @return
     *   The connection allocator.
     * @exception SQLException
     *   if SQL error occurs allocating connection.
     */
    public DBConnection allocateConnection ()
	throws SQLException {

	return connectionAllocator.allocate();
    }


    /**
     * Return an object id for this logical database.
     *
     * @return
     *   The object id.
     * @exception ObjectIdException
     *   if an error occurs allocating an object id.
     */
    public ObjectId allocateObjectId ()
	    throws ObjectIdException {

	return objectIdAllocator.allocate();
    }


    /**
     * Return a transaction for use on this logical database.
     *
     * @return The transaction object.
     * @exception SQLException
     *   if a SQL error occurs.
     */
    public DBTransaction createTransaction ()
	throws SQLException {

	return new StandardDBTransaction (connectionAllocator.allocate());
    }


    /**
     * Return a query for use on this logical database.
     *
     * @return The query object.
     * @exception SQLException
     *   if a SQL error occurs.
     */
    public DBQuery createQuery ()
	throws SQLException {

	return new StandardDBQuery (connectionAllocator.allocate());
    }


    /**
     * Immediately shutdown the logical database. This is
     * normally only called by the database manager on shutdown.
     */
    public void shutdown () {
	connectionAllocator.dropAllNow ();
	driver = null;
    }


    /**
     * Return the symbolic name of this logical database.
     *
     * @return The symbolic name.
     */
    public String getName () {
	return dbName;
    }


    /**
     * Return a description of the logical database type.
     *
     * @return The type.
     */
    public String getType () {
	return "Standard JDBC";
    }


    /**
     * Return the number of currently active connections.
     *
     * @return The number of connections.
     */
    public int getActiveConnectionCount () {
	return connectionAllocator.getActiveCount();
    }


    /**
     * Return the maximum number of connections active at one time.
     *
     * @return The number of connections.
     */
    public int getMaxConnectionCount () {
	return connectionAllocator.getMaxCount();
    }


    /**
     * Return the time when the maximum connection count occured.
     *
     * @return The <CODE>Date</CODE> when the maximum connection
     *  count occured.
     */
    public Date getMaxConnectionCountDate () {
	return connectionAllocator.getMaxCountDate();
    }


    /**
     * Reset the maximum connection count and date.
     */
    public void resetMaxConnectionCount () {
	connectionAllocator.resetMaxCount();
    }


    /**
     * Return the number of database requests.
     *
     * @return The number of database requests (queries or transactions).
     */
    public long getRequestCount () {
	return connectionAllocator.getRequestCount();
    }
}
