/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: StandardAppUtil.java,v 1.24 2000/03/08 09:04:54 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server;

import java.net.*;
import java.util.StringTokenizer;
import com.lutris.util.*;
import com.lutris.http.*;
import com.lutris.appserver.server.session.*;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.xml.xmlc.html.*;
import org.w3c.dom.html.*;
import org.w3c.dom.*;
import javax.servlet.http.Cookie;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;


/**
 * Static methods used by <CODE>StandardApplication</CODE>.  These mostly
 * are used to manage sessions.
 *
 * @version	$Revision: 1.24 $
 * @author	Mark Diekhans
 * @author	Richard Kunze <kunze@opus5.de>
 */
public class StandardAppUtil {

    /*
     * Constant used for encoding the session in a url rewrite.
     */
    public static final String ENHYDRA_SESSION_ID_URL = "jsessionid";
    public static final String ENHYDRA_SESSION_ID_COOKIE = "JSESSIONID";

    /**
     * Prevent instantiation.
     */
    private StandardAppUtil() {
    }

    /**
     * Get internet peer address.
     *
     * @param comms The presentations manager comms object with
     *  all request/response information in it.
     * @return The peer address.
     * @exception ApplicationException If address can't be obtained.
     */
    //FIXME: no longer needed, but might make a nice general util function.
    public static InetAddress[] getPeerAddress(HttpPresentationComms comms)
            throws ApplicationException {

        InetAddress[] peer = new InetAddress[1];
        String remoteHost;
        try {
            remoteHost = comms.request.getRemoteHost();
        } catch (HttpPresentationException except) {
            throw new ApplicationException(except);
        }
        if (remoteHost == null) {
            throw new ApplicationException("Can't find remote host for request.");
        }
        if (remoteHost != null) {
            try {
                peer[0] = InetAddress.getByName(remoteHost);
            } catch (UnknownHostException except) {
                throw new ApplicationException("Can't find remote address for \"" + remoteHost + "\"", except);
            }
        }
        return peer;
    }

    /**
     * Search for a session cookie with the specified application name.
     *
     * @param applicationName Application name that is saved in the cookie.
     * @param cookies Array of cookies for the current request.
     * @return The session key decoded from the cookie, or null if
     *  the session was not found.
     * @deprecated
     */
    private static String findSessionKey(String applicationName,
                                         Cookie[] cookies) {
	return findSessionKey(cookies);
    }

    /**
     * Search for a session cookie with the name jsessionid.
     *
     * @param applicationName Application name that is saved in the cookie.
     * @param cookies Array of cookies for the current request.
     * @return The session key decoded from the cookie, or null if
     *  the session was not found.
     * @deprecated
     */
    private static String findSessionKey(Cookie[] cookies) {
        if (cookies == null) {
            return null;
        }
	for (int i=0; i<cookies.length; i++) {
	    if (cookies[i].getName().equalsIgnoreCase(ENHYDRA_SESSION_ID_COOKIE)) {
                return cookies[i].getValue();
	    }
	}
        return null;
    }

    /**
     * Encode a URL with a session identifier.
     *
     * @param url The url to encode.
     * @param id The session identifier to encode with the url.
     */
    public static String encodeUrl(String url, String id) {
        String pre = url;
        String post = "";
        int pos;
        // Need to put the parameters before a local anchor
        if ((pos = url.indexOf('?')) != -1) {
            pre  = url.substring(0, pos);
            post = url.substring(pos);
        }
	return pre + ';' + ENHYDRA_SESSION_ID_URL + '=' + id + post;
    }

    /**
     * Determine if a URL references a presentation object. Used for
     * URL rewriting to test if a session ID should be appended to a
     * given URL. Takes the cheap way out at the moment in that it
     * simply returns true iff (if and only if, for the
     * non-mathematicians :-) the URL contains the character sequence
     * <code>.po</code>.
     * <br>
     * TODO: Implement a better check
     *
     * @param url The url to check
     * @return True if the URL points to a PO, false otherwise
     */
    public static boolean pointsToPO(String url) {
        int idx = url.indexOf(".po");
        if (idx == -1) {
            return false;
        }
        if (idx == url.length()-3) {
            return true;
        }
        switch (url.charAt(idx+3)) {
            case '?':
            case '#':
                return true;
            case '/':
                // A bit more difficult: Need to decide if it's in the
                // the server or URI part of the URL.
                try {
                    URL u = new URL(url);
                    int i = u.getFile().indexOf(".po");
                    return (i != -1);
                } catch (MalformedURLException e) {
                    // url doesn't contain a schema and/or server
                    // part, so it must be relative. OK, found a PO.
                    return true;
                }
            default:
                return false;
        }
    }


    /**
     * Determine if a session exists for this request.  Does not
     * validate that a session is logged in.
     *
     * @param comms The presentations manager comms object with
     *  all request/response information in it.
     * @return The session object or null if a session cookie or
     *  matching session is not found.
     * @exception ApplicationException If address can't be obtained.
     */
    public static Session getRequestSession(HttpPresentationComms comms)
            throws ApplicationException {
        Session s = null;
        String sessionKey = null;
        Cookie[] cookies;
	StandardSessionManager sessionManager =
	    (StandardSessionManager)comms.application.getSessionManager();

        try {
	    if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_ALWAYS)) {
		cookies = comms.request.getCookies();
		sessionKey = findSessionKey(ENHYDRA_SESSION_ID_COOKIE,
					    cookies);
		if (sessionKey != null) {
		    // These are set so that application programmers
		    // can determine the state of session cookies if not
                    // using automatic url encoding
		    comms.request.setRequestedSessionIdFromCookie(true);
		    comms.request.setRequestedSessionIdFromUrl(false);
		    comms.response.setSessionIdCookieRequired(true);
		    comms.response.setSessionIdEncodeUrlRequired(false);
		}
            }

	    if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER) &&
		sessionKey == null) {

                //
                // This needs to be redone.
                //
//                if (comms.request.getHttpServletRequest() instanceof EnhydraRequest ) {
//                    EnhydraRequest er = (EnhydraRequest) comms.request.getHttpServletRequest();
//                }

		// Check if sessionKey is encoded as a parameter in the path
		// info (the form is http://localhost/foo.po;jsessionid=XYZ
                String pathInfo = comms.request.getPathInfo();
		StringTokenizer tokens =
		    new StringTokenizer(pathInfo, "/\\;=", false);
		while (tokens.hasMoreTokens()) {
		    String param = tokens.nextToken();
		    if (param.equalsIgnoreCase(ENHYDRA_SESSION_ID_URL)) {
			sessionKey = tokens.nextToken();
			break;
		    }
		}

        	if (sessionKey != null) {
                    // These are set so that application programmers
                    // can determine the state of session cookies if not
                    // using automatic url encoding
        	    comms.request.setRequestedSessionIdFromUrl(true);
        	    comms.request.setRequestedSessionIdFromCookie(false);
		    comms.response.setSessionIdEncodeUrlRequired(true);
		    comms.response.setSessionIdCookieRequired(false);
        	}
            }
        } catch (HttpPresentationException except2) {
            throw new ApplicationException(except2);
        }
	if (sessionKey == null) {
	    return null;
	}

	if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER)) {
	    comms.response.setSessionKey(sessionKey);
	    comms.response.setSessionManager(comms.application.getSessionManager());
	}

        try {
            s = comms.application.getSessionManager().getSession(
		    Thread.currentThread(), sessionKey);
        } catch (SessionException e) {
            throw new ApplicationException(e);
        }
        return s;
    }


    /**
     * Add a session ID to links, forms and frames. This is a convenience
     * function making it easier to pass session IDs to the client
     * without using cookies. The following edits are performed on the
     * DOM:
     *
     * <ul> <li> Every link (<code>&lt;A&gt;</code> and
     * <code>&lt;AREA&gt;</code>) node whose <code>href</code>
     * attribute contains the string <code>.po</code> is changed to
     * include the session ID as parameter.
     * <li> Every <code>&lt;frame&gt;</code> node whose <code>src</code>
     * attribute contains the string <code>.po</code> is changed to
     * include the session ID as parameter.
     * <li> A hidden input field
     * with the session ID is added as first child of every
     * <code>&lt;FORM&gt;</code> node whose <code>action</code>
     * attribute contains the string <code>.po</code> </ul>
     *
     * @param comms An HTTP communications object, contains the
     * session to use.
     * @param htmlObject An HTMLObject to which session ids should be added.
     */
    public static void addSession(HttpPresentationComms comms,
                                  HTMLDocument htmlObject) {
	StandardSessionManager sessionManager =
	    (StandardSessionManager)comms.application.getSessionManager();
	if (sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER)) {
	    return;
	}
	addSession(htmlObject.getFirstChild(),
		   comms.session.getSessionKey());
    }

    /**
     * Recursively walk the DOM and add a session ID to links, forms
     * and frames.
     * The following edits are performed on the DOM:
     * <ul>
     * <li> Every link (<code>&lt;A&gt;</code> and
     * <code>&lt;AREA&gt;</code>) node whose <code>href</code>
     * attribute contains the string <code>.po</code> is changed to
     * include the session ID as parameter.
     * <li> Every <code>&lt;frame&gt;</code> node whose <code>src</code>
     * attribute contains the string <code>.po</code> is changed to
     * include the session ID as parameter.
     * <li> A hidden input field
     * with the session ID is added as first child of every
     * <code>&lt;FORM&gt;</code> node whose <code>action</code>
     * attribute contains the string <code>.po</code>
     * </ul>
     * @param node The DOM node to start at.
     * @param sessionKey The session key.
     */
    public static void addSession(Node node,
				  String sessionKey) {

        // No need to check further on text nodes.
        // Saves a few CPU cycles for every text node in a DOM...
        if (node instanceof Text) {
            return;
        }

        Node child = node.getFirstChild();
        while (child != null) {
            addSession(child, sessionKey);
            child = child.getNextSibling();
        }
        String url;

        // Keep the checks sorted in order of average occurrence of
        // the respective DOM elements in typical HTML pages.
        // XXX: Maybe check for a few widely used HTML Elements first,
        // and immediately return if found?
        if (node instanceof HTMLAnchorElement) {
            HTMLAnchorElement a = (HTMLAnchorElement)node;
            url = a.getHref();
            if (url != null && pointsToPO(url)) {
                a.setHref(encodeUrl(url, sessionKey));
            }
        } else if (node instanceof HTMLFormElement) {
            HTMLFormElement form = (HTMLFormElement)node;
            url = form.getAction();
            if (url != null && pointsToPO(url)) {
                form.setAction(encodeUrl(url, sessionKey));
            }
        } else if (node instanceof HTMLFrameElement) {
            HTMLFrameElement frame = (HTMLFrameElement)node;
            url = frame.getSrc();
            if (url != null && pointsToPO(url)) {
                frame.setSrc(encodeUrl(url, sessionKey));
            }
        } else if (node instanceof HTMLAreaElement) {
            HTMLAreaElement a = (HTMLAreaElement)node;
            url = a.getHref();
            if (url != null && pointsToPO(url)) {
                a.setHref(encodeUrl(url, sessionKey));
            }
        }
    }

    /**
     *  This is a rarely used utility function. Give a raw ServletRequest,
     *  the session cookie used to look up the session object that would be
     *  used with the request if it were passed into the Enhydra framework.
     *  Returns null if no cookie or session is found; a new session will
     *  not be created.
     *  Currently this is only used to support the debugger.
     *
     *  @param request
     *    The raw request.
     *  @param application
     *    The application that will be servicing the request.
     *  @return
     *    The session object that would be used if the request were
     *    passed into the Enhydra framework, or null if not found (no new
     *    session will be created).
     */
    public static Session getRequestSession(ServletRequest request,
                                            Application application)
            throws ApplicationException {
        if (application == null) {
            return null;
        }
        Cookie[] cookies;
        Session s;
        HttpServletRequest req;
        try {
            req = (HttpServletRequest)request;
        } catch (ClassCastException e) {
            return null;
        }
        cookies = req.getCookies();
        String sessionKey = findSessionKey(cookies);
        if (sessionKey == null) {
            return null;
        }
        SessionManager sm = application.getSessionManager();
        if (sm == null) {
            return null;
        }
        try {
            s = sm.getSession(Thread.currentThread(), sessionKey);
        } catch (SessionException e) {
            throw new ApplicationException(e);
        }
        return s;
    }


    /**
     * Associate a new <CODE>StandardSession<CODE> to a client cookie.
     *
     * @param comms The presentations manager comms object with
     *  all request/response information in it.
     * @exception ApplicationException If address can't be obtained.
     */
    public static void bindSessionToClient(HttpPresentationComms comms)
            throws ApplicationException {

        StandardSession session = (StandardSession)comms.session;
	StandardSessionManager sessionManager =
	    (StandardSessionManager)session.getSessionManager();
	if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_ALWAYS)) {
	    try {
		Cookie cookie = new Cookie(ENHYDRA_SESSION_ID_COOKIE,
					   session.getSessionKey());
		cookie.setPath(comms.request.getApplicationPath());
		comms.response.addCookie(cookie);
	    } catch (HttpPresentationException except) {
		throw new ApplicationException(except);
	    }
	}
	if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER)) {
	    comms.response.setSessionKey(comms.session.getSessionKey());
	    comms.response.setSessionManager(comms.application.getSessionManager());
	}
    }
}
