/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: CopyFilePresentation.java,v 1.11 2000/01/29 07:02:10 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.httpPresentation;

import java.io.*;
import com.lutris.util.*;
import com.lutris.classloader.*;

/**
 * Presentation for file data.  This should not be cached, as it copies
 * the input stream and closes it.
 */
class CopyFilePresentation implements HttpPresentation {
    private InputStream input;
    private int contentLength;
    private String mimeType;

    private static final int FILE_BUFFER_SIZE = 4096;

    /*
     * Copy the input stream to the HTTP output stream.
     */
    private void copyFile(HttpPresentationComms comms)
        throws FilePresentationException {

        try {
            byte [] buffer = new byte[FILE_BUFFER_SIZE];
            HttpPresentationOutputStream output = comms.response.getOutputStream();
            while (true) {
                int readLen = input.read(buffer, 0, FILE_BUFFER_SIZE);
                if (readLen < 0) {
                    break;
                }
                output.write(buffer, 0, readLen);
            }
        } catch (Exception except) {
            throw new FilePresentationException(except);
        }
    }

    /**
     * Construct a new copy file presentation, given an input stream.
     */
    protected CopyFilePresentation(ClassLoader classLoader,
                                   String urlPath,
                                   String fileMimeType) 
            throws FilePresentationException {
        mimeType = fileMimeType;

        /*
         * Get the stream from the classloader.  If we are using the
         * Lutris classloader, we maybe able to get the file size from
         * the classloader.
         */
        try {
            if (classLoader instanceof MultiClassLoader) {
                Resource resource = ((MultiClassLoader)classLoader).getResourceAsIs(urlPath);
                if (resource != null) {
                    input = resource.getInputStream();
                    contentLength = (int)resource.getSize();
                }
            } else {
                input = classLoader.getResourceAsStream(urlPath);
                contentLength = -1;
            }
        } catch (IOException except) {
            throw new FilePresentationException("Error accessing  \"" + urlPath + "\"", except);
        }
        if (input == null) {
            throw new FilePresentationException
                (new FileNotFoundException ("File \"" + urlPath
                                            + "\" not found on application class path"));
        }
    }
    
    /**
     * Finalizer to close input file if we haven't done it already.
     */
    protected void finalize() throws Throwable {
        if (input != null) {
            input.close();
        }
        super.finalize();
    }

    /**
     * Entry point for copy file presenation.
     */
    public void run(HttpPresentationComms comms)
            throws FilePresentationException {
        try {
            comms.response.setContentType(mimeType);
            comms.response.setContentLength(contentLength);
            try {
                if (!comms.request.getMethod().equals("HEAD")) {
                    copyFile(comms);
                }
            } finally {
                input.close();
                input = null;
            }
        } catch (FilePresentationException except) {
            throw except;
        } catch (Exception except) {
            throw new FilePresentationException(except);
        }
    }
}
