/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: HttpPresentationIOException.java,v 1.6 2000/01/29 07:02:10 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.httpPresentation;
import java.io.IOException;
import com.lutris.util.*;


/**
 * IOException derived class that is thrown when the presentation manager
 * encounters an I/O error when talking to a client.  This allows the methods
 * to determine how the error should be handled, even though it was generated
 * by a call that occured in presentation code.
 */
public class HttpPresentationIOException extends IOException {
    /**
     * Static method that does the approriate conversion from an IOException
     * to a HttpPresentationIOException If the exception is already an
     * HttpPresentationIOException, it is simple rethrown.  If it is another type of
     * IOException, it is converted to a HttpPresentationIOException.
     *
     * @param except The exception to rethrow or encapsulate.
     * @return An exception ready to throw.
     */
    public static IOException rethrow(IOException except)
            throws IOException {
        if (except instanceof HttpPresentationIOException) {
            return except;
        } else {
            return new HttpPresentationIOException(except);
        }
    }


    /**
     * Flag that indicates that this came from an a I/O operation
     * rather than a generated exception.
     */
    private boolean trueIOException = false;

    /**
     * Construct a new exception from an existing IOException.
     *
     * @param except The exception to encapsulate.
     */
    public HttpPresentationIOException(IOException except) {
        super(except.getMessage());
        trueIOException = true;
    }

    /**
     * Construct a new exception with a specific message.
     *
     * @param msg The error message
     */
    public HttpPresentationIOException(String msg) {
        super(msg);
    }

    /**
     * Did this exception come from an a I/O operation
     * rather than a generated exception.
     * @return True if its an actual I/O exception. 
     */
    public boolean isTrueIOException() {
        return trueIOException;
    }

    /**
     * Given a exception or error, determine if its generated by a HTTP
     * client I/O exception.  This handles ChyainedExceptions that might
     * contain a client I/O exception.
     *
     * @param except Exception to check.
     * @return <CODE>true</CODE> if this is cause by the client socket,
     *  <CODE>false</CODE> if not.
     */
    public static boolean isClientIOException(Throwable except) {
        Throwable scan = except;
        while (scan != null) {
            if ((scan instanceof HttpPresentationIOException)
                && ((HttpPresentationIOException)scan).isTrueIOException()) {
                return true;
            }
            if (scan instanceof ChainedException) {
                scan = ((ChainedException)scan).getCause();
            } else {
                return false;
            }
        }
        return false;
    }

}
