/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: HttpPresentationResponse.java,v 1.21 2000/03/04 04:46:41 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.httpPresentation;

import com.lutris.appserver.server.session.SessionManager;
import javax.servlet.http.Cookie;
import org.w3c.dom.html.HTMLDocument;

/**
 * Object passed to presentation objects that is used to generate HTTP
 * responses.
 *
 * @author	Mark Diekhans
 * @version	$Revision: 1.21 $
 */
public interface HttpPresentationResponse {

    /**
     * Sets the content length for this response.
     *
     * @param len the content length
     */
    public void setContentLength(int len)
	throws HttpPresentationException;

    /**
     * Sets the content type for this response.
     *
     * @param type the content's MIME type
     */
    public void setContentType(String type)
	throws HttpPresentationException;

    /**
     * Returns an output stream for writing response data.
     */
    public HttpPresentationOutputStream getOutputStream()
	throws HttpPresentationException;

    /**
     * Adds the specified cookie to the response. It can be called multiple
     * times to set more than one cookie.
     *
     * @param cookie - The Cookie to return to the client.
     */
    public void addCookie(Cookie cookie)
	throws HttpPresentationException;

    /**
     * Returns true if the response message header has a field with
     * the specified name.
     *
     * @param name the header field name
     */
    public boolean containsHeader(String name)
	throws HttpPresentationException;

    /**
     * Sets the status code and message for this response.
     *
     * @param sc the status code
     * @param sm the status message
     */
    public void setStatus(int sc, String sm)
	throws HttpPresentationException;

    /**
     * Sets the status code and a default message for this response.
     *
     * @param sc the status code
     */
    public void setStatus(int sc)
	throws HttpPresentationException;

    /**
     * Adds a field to the response header with a given name and
     * value.  If the field had already been set, the new value
     * overwrites the previous one.  The containsHeader method can be
     * used to test for the presence of a header before setting its
     * value.
     * 
     * @param name the header field name
     * @param value the header field value
     */
    public void setHeader(String name, String value)
	throws HttpPresentationException;

    /**
     * Adds a field to the response header with a given name and
     * integer value.  If the field had already been set, the new
     * value overwrites the previous one.  The containsHeader method
     * can be used to test for the presence of a header before setting
     * its value.
     *
     * @param name the header field name
     * @param value the header field integer value
     */
    public void setIntHeader(String name, int value)
	throws HttpPresentationException;

    /**
     * 
     * Adds a field to the response header with a given name and
     * date-valued field.  The date is specified in terms of
     * milliseconds since the epoch.  If the date field had already
     * been set, the new value overwrites the previous one.  The
     * containsHeader method can be used to test for the presence of a
     * header before setting its value.
     * 
     * @param name the header field name
     * @param value the header field date value
     */
    public void setDateHeader(String name, long date)
	throws HttpPresentationException;

    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.
     *
     * @param sc the status code
     * @param msg the detail message
     */
    public void sendError(int sc, String msg)
	throws HttpPresentationException;

    /**
     * Sends an error response to the client using the specified
     * status code and a default message.
     *
     * @param sc the status code
     */
    public void sendError(int sc)
	throws HttpPresentationException;

    //Omitted: sendRedirect(String location)

    //Omitted: public abstract String encodeUrl(String url)

    //Omitted: public abstract String encodeRedirectUrl(String url)

    /**
     * Called at the end of processing a response to force any cached headers to
     * be written and buffers flushed.  This maybe a no-op if no buffering is
     * implemented.  This is normally not called by a client.
     */
    public void flush ()
	throws HttpPresentationException;

    /**
     * Utility method to output an HTML page. The appropriate headers are
     * set for MIME type and to disable caching of the HTML by the broswer.
     */
    public void writeHTML(String html) throws HttpPresentationException;

    /**
     * Utility method to output an HTML page from a DOM object. The 
     * appropriate headers are set for MIME type and to disable caching
     * of the HTML by the broswer.
     *
     * @param doc
     *        The DOM object to be returned as response
     */
    public void writeHTML(HTMLDocument doc) throws HttpPresentationException;

    /**
     * Set the output character encoding.
     *
     * @param enc character encoding 
     */
    public void setEncoding(String enc);
    
    /**
     * Get the output character encoding.
     *
     * @return the current encoding 
     */
    public String getEncoding();

    /**
     * Sets the current session key for this response
     *
     * @param sessionKey
     *        The current sessionKey
     **/
    public void setSessionKey(String sessionKey);

    /**
     * Sets the current application name for this response
     *
     * @param appName
     *        The current application name
     **/
    public void setSessionManager(SessionManager sessionManager);

    /**
     * Indicates whether client response requires a sessionId cookie
     * @return true if client response requires a sessionId cookie
     *         false otherwise
     */
    public boolean isSessionIdCookieRequired()
	throws HttpPresentationException; 

    /*
     * set boolean flag for sessionId response cookie.  Indicates whether 
     * client response requires a sessionId cookie
     * @param isFromCookie boolean flag 
     */
    public void setSessionIdCookieRequired(boolean sessionIdCookie) 
	throws HttpPresentationException; 

    /**
     * Indicates whether client response requires url encoding for
     * sessionId
     * @return true if client response requires url encoding for
     *         sessionId; false otherwise
     */
    public boolean isSessionIdEncodeUrlRequired()
	throws HttpPresentationException; 

    /*
     * set boolean flag url encoding for sessionId.  Indicates whether 
     * client response requires url encoding for sessionId
     * @param sessionIdUrl boolean flag 
     */
    public void setSessionIdEncodeUrlRequired(boolean sessionIdUrl) 
	throws HttpPresentationException; 
    
    /*
     * SERVER status codes; see RFC 1945.
     */

    /**
     * Status code (200) indicating the request succeeded normally.
     */
    public static final int SC_OK = 200;

    /**
     * Status code (201) indicating the request succeeded and created
     * a new resource on the server.
     */
    public static final int SC_CREATED = 201;

    /**
     * Status code (202) indicating that a request was accepted for
     * processing, but was not completed.
     */
    public static final int SC_ACCEPTED = 202;

    /**
     * Status code (204) indicating that the request succeeded but that
     * there was no new information to return.
     */
    public static final int SC_NO_CONTENT = 204;

    /**
     * Status code (301) indicating that the resource has permanently
     * moved to a new location, and that future references should use a
     * new URI with their requests.
     */
    public static final int SC_MOVED_PERMANENTLY = 301;

    /**
     * Status code (302) indicating that the resource has temporarily
     * moved to another location, but that future references should
     * still use the original URI to access the resource.
     */
    public static final int SC_MOVED_TEMPORARILY = 302;

    /**
     * Status code (304) indicating that a conditional GET operation
     * found that the resource was available and not modified.
     */
    public static final int SC_NOT_MODIFIED = 304;

    /**
     * Status code (400) indicating the request sent by the client was
     * syntactically incorrect.
     */
    public static final int SC_BAD_REQUEST = 400;

    /**
     * Status code (401) indicating that the request requires HTTP
     * authentication.
     */
    public static final int SC_UNAUTHORIZED = 401;

    /**
     * Status code (403) indicating the server understood the request
     * but refused to fulfill it.
     */
    public static final int SC_FORBIDDEN = 403;

    /**
     * Status code (404) indicating that the requested resource is not
     * available.
     */
    public static final int SC_NOT_FOUND = 404;

    /**
     * Status code (500) indicating an error inside the HTTP service
     * which prevented it from fulfilling the request.
     */
    public static final int SC_INTERNAL_SERVER_ERROR = 500;

    /**
     * Status code (501) indicating the HTTP service does not support
     * the functionality needed to fulfill the request.
     */
    public static final int SC_NOT_IMPLEMENTED = 501;

    /**
     * Status code (502) indicating that the HTTP server received an
     * invalid response from a server it consulted when acting as a
     * proxy or gateway.
     */
    public static final int SC_BAD_GATEWAY = 502;

    /**
     * Status code (503) indicating that the HTTP service is
     * temporarily overloaded, and unable to handle the request.
     */
    public static final int SC_SERVICE_UNAVAILABLE = 503;
}
