/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: HttpSerialized.java,v 1.7 2000/01/29 07:02:11 markd Exp $
 *-----------------------------------------------------------------------------
 */




package com.lutris.appserver.server.httpPresentation;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  This class provides two utility methods for presentation objects
 *  that want to read and/or write Java serialized objects. <P>
 *
 *  Reading is accomplished by having the client issue a POST request,
 *  with the serialized data as the body of the request. <P>
 *
 *  Writing is accomplished by simply sending out the serialized bytes.
 *
 * @see java.io.Serializable
 */
public class HttpSerialized {

    /**
     *  The mime type used to send and recieve Java serialized objects.
     */
    public static final String serializedMimeType = 
                                             "application/java-serialized";

    /**
     *  Read in a Java object from a POST request. The object is returned.
     *  
     *  @param request
     *   HTTP communication request.
     *  @return 
     *   The java object sent in the POST request.
     *  @exception HttpPresentationException 
     *   If the request is not of type POST, or the content type is not
     *   correct, or there is an IO or serialization error.
     */
    public static Object readSerializedObject(
                             HttpPresentationRequest request)
                             throws HttpPresentationException {
        /*
         *  Is this a POST request?
         */
        String method = request.getMethod();
        if ((method == null) || 
            !method.equalsIgnoreCase("POST")) 
            throw new HttpPresentationException("Request method is not POST.");
        /*
         *  Is the content type correct?
         */
        String type = request.getContentType();
        if ((type == null) ||
           !type.equalsIgnoreCase(serializedMimeType))
            throw new HttpPresentationException("POSTed data is not of type " +
                                                serializedMimeType + ".");
        /*
         *  Try to read in the object.
         */
        Object result = null;
        try {
            HttpPresentationInputStream hpis = request.getInputStream();
            ObjectInputStream ois = new ObjectInputStream(hpis);
            result = ois.readObject();
        } catch (InvalidClassException e) {
            throw new HttpPresentationException(
                          "Invalid class sent: " + e.getMessage(), e);
        } catch (StreamCorruptedException e) {
            throw new HttpPresentationException(
                          "Invalid data sent: " + e.getMessage(), e);
        } catch (OptionalDataException e) {
            throw new HttpPresentationException(
                          "Invalid data sent: " + e.getMessage(), e);
        } catch (ClassNotFoundException e) {
            throw new HttpPresentationException(
                          "Invalid class sent: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new HttpPresentationException(
                          "IO error reading object: " + e.getMessage(), e);
        }
        /*
         *  Success!
         */
        return result;
    }

    /**
     *  Read in a Java object from a POST request. The object is returned.
     *  
     *  @param comms
     *   HTTP communications object. Contains objects and interfaces to read
     *   the request and send a response.
     *  @return 
     *   The java object sent in the POST request.
     *  @exception HttpPresentationException 
     *   If the request is not of type POST, or the content type is not
     *   correct, or there is an IO or serialization error.
     */
    public static Object readSerializedObject(
                             HttpPresentationComms comms)
                             throws HttpPresentationException {
        return readSerializedObject(comms.request);
    }

    /**
     *  Return a serialized Java object to the client. 
     *
     *  @param comms
     *   HTTP Response object. interface to send a response.
     * @param object 
     *  The object to return to the client.
     * @exception HttpPresentationException 
     *  If there is an error serializing the object, or an error in the
     *  underlying calls to response.
     */
    public static void writeSerializedObject(
                           HttpPresentationResponse response,
                           Serializable object)
                           throws HttpPresentationException {
        /*
         * First try to serialize the object, so we can catch any exceptions
         * up front, and so we know the length.
         */
        byte[] objectData = null;
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(object); 
            objectData = baos.toByteArray();
        } catch (InvalidClassException e) {
            throw new HttpPresentationException(
                          "Invalid class: "  + e.getMessage(), e);
        } catch (NotSerializableException e) {
            throw new HttpPresentationException(
                          "Invalid class: "  + e.getMessage(), e);
        } catch (IOException e) {
            throw new HttpPresentationException(
                          "Error serializing object: " + e.getMessage(), e);
        }
        if (objectData == null)
            throw new HttpPresentationException(
                         "Error serializing object: no data generated.");
        /*
         *  Set up the response.
         */
        response.setContentType(serializedMimeType);
        response.setContentLength(objectData.length);
        /*
         *  Transmit the object data.
         */
        HttpPresentationOutputStream hpos = response.getOutputStream();
        try {
            hpos.write(objectData);
        } catch (IOException e) {
            throw new HttpPresentationException(
                         "Error sending object: " + e, e);
        }
        /*
         * Success!
         */
    }

    /**
     *  Return a serialized Java object to the client. 
     *
     *  @param comms
     *   HTTP communications object. Contains objects and interfaces to read
     *   the request and send a response.
     * @param object 
     *  The object to return to the client.
     * @exception HttpPresentationException 
     *  If there is an error serializing the object, or an error in the
     *  underlying calls to response.
     */
    public static void writeSerializedObject(
                           HttpPresentationComms comms,
                           Serializable object)
                           throws HttpPresentationException {
        writeSerializedObject(comms.response,object);
    }
    
}

