/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: PageUnauthorizedException.java,v 1.8 2000/01/29 07:02:11 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.httpPresentation;


/**
 * Throwing this class will cause a 401 Unauthorized response. 
 *
 * @see com.lutris.http.BasicAuth
 * @author Andy John
 */
public class PageUnauthorizedException extends RuntimeException {
    /**
     * The relm to ask for. This string is used by the browser in the
     * prompt for username and password. It should identify to the 
     * user which system they are logging in to, i.e. which username
     * and password to use.
     */
    private String relm;

    /**
     * Default title.
     */
    private static final String defaultHtmlTitle =
        "Unauthorized access";

    /**
     * Default text message to display.
     */
    private static final String defaultHtmlText =
        "<B>Client is not authorized to access this page</B>";

    /**
     * HTML title to display. (not yet configurable).
     */
    private String htmlTitle = defaultHtmlTitle;

    /**
     * HTML Text message to display.
     */
    private String htmlText = defaultHtmlText;

    /**
     * Create a new unauthorized exception.
     *
     * @param relm The relm to ask for. This string is used by the browser
     * in the prompt for username and password. It should identify to the 
     * user which system they are logging in to, i.e. which username   
     * and password to use. For example, if relm is XXX, then Netscape
     * Navigator will prompt the user with "Enter the username for
     * XXX at <host>:<port>".
     */
    public PageUnauthorizedException(String requestedRelm) {
        super(defaultHtmlTitle);
        relm = requestedRelm;
    }
 
 
    /**
     * Create a new unauthorized exception.
     *
     * @param relm The relm to ask for. This string is used by the browser
     * in the prompt for username and password. It should identify to the 
     * user which system they are logging in to, i.e. which username   
     * and password to use. For example, if relm is XXX, then Netscape
     * Navigator will prompt the user with "Enter the username for
     * XXX at <host>:<port>".
     * @param htmlDisplayText Text to display on the page.  If null,
     *  a default will be used.
     */
    public PageUnauthorizedException(String requestedRelm,
                                     String htmlDisplayText) {
        super(defaultHtmlTitle);
        relm = requestedRelm;
        if (htmlDisplayText != null) {
            htmlText = htmlDisplayText;
        }
    }
    
 
 
   /**
     * Get the relm that was passed in to the constructor.
     * This is used as part of the prompt to the user by the browser.
     *
     * @return The relm string. 
     */
    public String getRelm() {
        return relm;
    }

   /**
     * Get the title to use for the HTML page.
     *
     * @return The title string.
     */
    public String getHtmlTitle() {
        return htmlTitle;
    }

   /**
     * Get the text to display in the client.
     *
     * @return The HTML text.
     */
    public String getHtmlText() {
        return htmlText;
    }

}
