/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: AbstractFilterableServlet.java,v 1.8 2000/02/21 23:50:38 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.httpPresentation.servlet;

import java.util.Vector;
import java.io.IOException;
import javax.servlet.ServletException;
import javax.servlet.Servlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.enhydra.servlet.filter.Filter;


/**
 *  This class demonstrates the responsibilities of a
 *  <CODE>FilterableServlet</CODE>. Because most servlets will need to 
 *  extend the base class <CODE>HttpServlet</CODE>, they will not be able
 *  to extend this class. Therefore this class is not expected to be used,
 *  it is merely an instructional reference. That is why it does not implement
 *  the full <CODE>Servlet</CODE> interface, just the extensions in 
 *  <CODE>FilterableServlet</CODE> <CODE>service()</CODE> method.
 *
 * @see com.lutris.appserver.server.httpPresentation.servlet.FilterableServlet
 */
public abstract class AbstractFilterableServlet
                      extends HttpServlet
                      implements FilterableServlet {

    /**
     *  The list of current filters.
     */
    private Vector filters = new Vector();


    public AbstractFilterableServlet() {
    }


    /**
     *  Add a filter to the end of the list of filters.
     *
     *  @param filter
     *    The filter to add.
     */
    public void addFilter(Filter filter) {
        synchronized (filters) {
            filters.addElement(filter);
        }
    }

    /**
     *  Remove a filter from the list of current filters. 
     *  Silently fails if the filter is not in the list.
     *
     *  @param filter
     *    The filter to remove.
     */
    public void removeFilter(Filter filter) {
        synchronized (filters) {
            filters.removeElement(filter);
        }
    }


    /**
     *  Handle a service request. Use the first filter to wrap the request,
     *  response, and servlet (this) objects. This will return new objects.
     *  Then use the second filter to wrap those, again returning new
     *  objects. Then use the fourth filter to wrap those etc....
     *  Then finally call <CODE>service()</CODE> on the outermost results.
     *
     *  @param request
     *    The original servlet request object.
     *  @param response
     *    The original servlet response object.
     */
    public void service(HttpServletRequest request,
                        HttpServletResponse response) 
                throws ServletException, IOException {
        /*
         * This is synchronized for safety, but in real life you wouldn't want
         * to do this as it would introduce too much slowdown in such a
         * critical bottlneck area of the code. You would be prepared for the
         * array to change out from under you. Adding/removing filters is
         * a very rare operation, so it's ok.
         */
        synchronized (filters) {
            int size = filters.size();
            if (size == 0) {
                serviceDirect(request, response);
            } else {
		HttpPresentationServletGlue glue =
			new HttpPresentationServletGlue(this);
		HttpServletRequest reqResult = request;
		HttpServletResponse resResult = response;
                for (int i=0; i<size; i++) {
                    Filter f = (Filter) filters.elementAt(i);
                    reqResult = f.wrapRequest(reqResult, resResult);
                    resResult = f.wrapResponse(reqResult, resResult);
                }
                glue.service(reqResult, resResult);
            } 
        }
    }



    /**
     *  Once the filters are done, you don't want to call service(), or it
     *  would result in an infinite loop of filter processing. So you call this
     *  instead. It does the real work that service() would normally do, 
     *  without applying the filters. <P>
     *
     *  This is left an abstract method because this is where you would do
     *  your real work if you subclassed this. However this class is 
     *  most likely never going to be extended; it is meant to be used as
     *  a code reference to see how filters work.
     *
     *  @param request
     *    The request object.
     *  @param response
     *    The response object.
     */
    public abstract void serviceDirect(HttpServletRequest request,
                                       HttpServletResponse response)
                         throws ServletException, IOException;


}

