/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: ServletHttpPresentationResponse.java,v 1.31.2.3 2000/04/06 01:40:47 peterj Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server.httpPresentation.servlet;

import java.util.Enumeration;
import java.io.IOException;
import java.io.PrintWriter;
import javax.servlet.http.*;
import com.lutris.http.debug.*;
import com.lutris.util.StringEnum;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.StandardAppUtil;
import com.lutris.appserver.server.session.SessionManager;
import com.lutris.appserver.server.session.SessionException;
import org.w3c.dom.html.HTMLDocument;
import com.lutris.xml.xmlc.html.HTMLFormatter;
import org.enhydra.xml.io.Encoding;

/**
 * Object passed to presentation objects that is used to generate HTTP
 * responses.
 */
public class ServletHttpPresentationResponse
    implements HttpPresentationResponse, DebugResponse {

    /*
     * Servlet objects.
     */
    private HttpServletResponse response;
    private HttpPresentationOutputStream outputStream = null; // Lazy allocation

    /*
     * Default encoding of the VM
     */
    private static final String defaultEncoding = System.getProperty("file.encoding");

    /*
     * Encoding for output
     */
    private String encoding = defaultEncoding;

    /*
     * The current session key
     */
    private String sessionKey = null;

    /*
     * The current session manager
     */
    private SessionManager sessionManager = null;

    /*
     * Boolean flag indicating if response requires a sessionId cookie
     */
    private boolean sessionIdCookie = true;

    /*
     * Boolean flag indicating if response requires url encoding for sessionId
     */
    private boolean sessionIdUrl = true;

    /**
     * Construct an object associated with a servlet response.
     *
     * @param response Servlet response object that this object will front-end.
     */
    protected ServletHttpPresentationResponse(HttpServletResponse response) {
        this.response = response;
    }

    /**
     * Sets the content length for this response.
     *
     * @param len the content length
     */
    public void setContentLength(int len)
	throws HttpPresentationException {

        response.setContentLength(len);
    }


    /**
     * Sets the content type for this response.
     *
     * @param type the content's MIME type
     */
    public void setContentType(String type)
	throws HttpPresentationException {

        response.setContentType(type);
    }


    /**
     * Returns an output stream for writing response data.
     */
    public HttpPresentationOutputStream getOutputStream()
	throws HttpPresentationException {

        if (outputStream == null) {
            try {
                outputStream = new ServletHttpPresentationOutputStream
			(this, response.getOutputStream());
            } catch (IOException except) {
                // Want to throw a HttpPresentationIOException to record the fact that
                // the client disconnected, but want to maintain the HttpPresentationException
                // API.
                throw new HttpPresentationException(new HttpPresentationIOException(except));
            }
        }
        return outputStream;
    }


    /**
     * Adds the specified cookie to the response. It can be called multiple
     * times to set more than one cookie.
     *
     * @param cookie - The Cookie to return to the client.
     */
    public void addCookie(Cookie cookie)
	throws HttpPresentationException {
        response.addCookie(cookie);
    }


    /**
     * Returns true if the response message header has a field with
     * the specified name.
     * @param name the header field name
     */
    public boolean containsHeader(String name)
	throws HttpPresentationException {

        return response.containsHeader(name);
    }


    /**
     * Sets the status code and message for this response.
     *
     * @param sc the status code
     * @param sm the status message
     */
    public void setStatus(int sc, String sm)
	throws HttpPresentationException {

        response.setStatus(sc, sm);
    }


    /**
     * Sets the status code and a default message for this response.
     *
     * @param sc the status code
     */
    public void setStatus(int sc)
	throws HttpPresentationException {

        response.setStatus(sc);
    }


    /**
     * Adds a field to the response header with a given name and
     * value.  If the field had already been set, the new value
     * overwrites the previous one.  The containsHeader method can be
     * used to test for the presence of a header before setting its
     * value.
     *
     * @param name the header field name
     * @param value the header field value
     */
    public void setHeader(String name, String value)
	throws HttpPresentationException {

        response.setHeader(name, value);
    }


    /**
     * Adds a field to the response header with a given name and
     * integer value.  If the field had already been set, the new
     * value overwrites the previous one.  The containsHeader method
     * can be used to test for the presence of a header before setting
     * its value.
     *
     * @param name the header field name
     * @param value the header field integer value
     */
    public void setIntHeader(String name, int value)
	throws HttpPresentationException {

        response.setIntHeader(name, value);
    }


    /**
     * Adds a field to the response header with a given name and
     * date-valued field.  The date is specified in terms of
     * milliseconds since the epoch.  If the date field had already
     * been set, the new value overwrites the previous one.  The
     * containsHeader method can be used to test for the presence of a
     * header before setting its value.
     *
     * @param name the header field name
     * @param value the header field date value
     */
    public void setDateHeader(String name, long date)
	throws HttpPresentationException {

        response.setDateHeader(name, date);
    }


    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.
     *
     * @param sc the status code
     * @param msg the detail message
     */
    public void sendError(int sc, String msg)
	throws HttpPresentationException {

	try {
	    response.sendError(sc, msg);
	} catch (IOException except) {
            // Want to throw a HttpPresentationIOException to record the fact that
            // the client disconnected, but want to maintain the HttpPresentationException
            // API.
            throw new HttpPresentationException(new HttpPresentationIOException(except));
	}
    }

    /**
     * Sends an error response to the client using the specified
     * status code and a default message.
     *
     * @param sc the status code
     */
    public void sendError(int sc)
	throws HttpPresentationException {

	try {
            response.sendError(sc);
	} catch (IOException except) {
            // Want to throw a HttpPresentationIOException to record the fact that
            // the client disconnected, but want to maintain the HttpPresentationException
            // API.
            throw new HttpPresentationException(new HttpPresentationIOException(except));
	}
    }

    /**
     * Called at the end of processing a response to force any cached
     * headers to be written and buffers flushed.
     */
    public void flush() throws HttpPresentationException {

	try {
            // Before version 3.0 of enhydra this was getOutputStream().flush()
            // the getOutStream() can throw an exception if the writer was used
            // instead of the outputstream.
            response.flushBuffer();
        } catch (IOException except) {
            // Want to throw a HttpPresentationIOException to record the fact that
            // the client disconnected, but want to maintain the HttpPresentationException
            // API.
            throw new HttpPresentationException(new HttpPresentationIOException(except));
        }
    }

    /**
     * Encodes a url with session id for use with cookieless browsers.
     *
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     */
    public String encodeUrl(String url) {
	return response.encodeUrl(url);
    }

    /**
     * Encodes a url with session id for sendRedirect for use with
     * cookieless browsers.
     *
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     *
     * @see #sendRedirect
     * @see #encodeUrl
     */
    public String encodeRedirectUrl(String url) {
	return response.encodeUrl(url);
    }




    /*
     * The following methods are for the debugging interface...
     */

    /**
     * Returns the size of the request entity data, or -1 if not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the content length.
     */
    public int getContentLength() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return -1;
    }


    /**
     * Returns the Internet Media Type of the response, or null if not known
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the content type.
     */
    public String getContentType() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return null;
    }


    /**
     * Returns the status code for the response, or -1 if not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the status code.
     */
    public int getStatusCode() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return -1;
    }


    /**
     * Returns the status message for the response, or null if not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the status message.
     */
    public String getStatusMessage() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return null;
    }


    /**
     * Returns an array of cookies to be set by this response, or
     * a 0 length array if the information is not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return The array of cookies found in this response.
     */
    public Cookie[] getCookies() {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
	return new Cookie[0];
    }


    /**
     * Returns the values of the specified header for the response as
     * an array of strings, or a 0 length array if the named header does
     * not exist.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @param name the case-insensitive header field name
     * @return the value of the supplied header.
     */
    public String getHeader(String name) {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
	return "";
    }


    /**
     * Returns an array of strings representing the header names
     * for this response. This does not have return cookies - use
     * getCookies().
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return name the case-insensitive header field name
     */
    public Enumeration getHeaderNames() {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
	return new StringEnum(new String[0]);
    }


    /**
     * Returns the total number of bytes that make up this response.
     * This is not supported for Jolt Presentations and is necessary
     * to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the total number of bytes in the response.
     */
    public int getTotalBytes() {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
        return -1;
    }

    /**
     * Returns null.
     * Normally this function returns the data written to the output
     * stream associated with this request.
     */
    public String getResponseData() {
        return null;
    }

    /**
     * Set the output character encoding.
     *
     * @param enc character encoding
     */
    public void setEncoding(String enc) {
        encoding = enc;
    }

    /**
     * Get the output character encoding.
     */
    public String getEncoding() {
        return encoding;
    }

    /**
     * Sets the current session key for this response
     *
     * @param sessionKey
     *        The current sessionKey
     **/
    public void setSessionKey(String sessionKey) {
	this.sessionKey = sessionKey;
    }

    /**
     * Sets the session manager.
     **/
    public void setSessionManager(SessionManager sessionManager) {
	this.sessionManager = sessionManager;
    }

    /**
     * Indicates whether client response requires a sessionId cookie
     * @return true if client response requires a sessionId cookie
     *         false otherwise
     */
    public boolean isSessionIdCookieRequired()
	throws HttpPresentationException {
	return sessionIdCookie;
    }

    /*
     * set boolean flag for sessionId response cookie.  Indicates whether
     * client response requires a sessionId cookie
     * @param isFromCookie boolean flag
     */
    public void setSessionIdCookieRequired(boolean sessionIdCookie)
	throws HttpPresentationException {
	this.sessionIdCookie = sessionIdCookie;
    }

    /**
     * Indicates whether client response requires url encoding for
     * sessionId
     * @return true if client response requires url encoding for
     *         sessionId; false otherwise
     */
    public boolean isSessionIdEncodeUrlRequired()
	throws HttpPresentationException {
	return sessionIdUrl;
    }

    /*
     * set boolean flag url encoding for sessionId.  Indicates whether
     * client response requires url encoding for sessionId
     * @param sessionIdUrl boolean flag
     */
    public void setSessionIdEncodeUrlRequired(boolean sessionIdUrl)
	throws HttpPresentationException {
	this.sessionIdUrl = sessionIdUrl;
    }

    /**
     * Write of HTML bytes to the output steam, setting approriate
     * headers, etc.
     */
    private void writeHTMLBytes(byte[] bytes,
                                String mimeEncoding)
        throws HttpPresentationException {

	try {
            if (mimeEncoding != null) {
                setContentType("text/html; charset=" + mimeEncoding);
            } else {
                setContentType("text/html");
            }

            setContentLength(bytes.length);

            // Disable caching of page.
            setHeader("Cache-Control", "no-cache");
            setHeader("Expires", "0");

            getOutputStream().write(bytes);
            getOutputStream().flush();
        } catch (IOException except) {
            throw new HttpPresentationException(except);
        }
    }

    /**
     * Utility method to output an HTML page. The appropriate headers are
     * set for MIME type and to disable caching of the HTML by the broswer.
     */
    public void writeHTML(String html) throws HttpPresentationException {
        byte bytes[];

        // Determine encoding to use.
        String javaEncoding = null;
        if (encoding != null) {
            // Will return null if its invalid
            javaEncoding = Encoding.xml2JavaEncoding(encoding);
        }

        try {
            if (javaEncoding != null) {
                bytes = html.getBytes(javaEncoding);
            } else {
                bytes = html.getBytes();
            }
        } catch (IOException except) {
            throw new HttpPresentationException(except);
        }
        writeHTMLBytes(bytes, encoding);
    }

    /**
     * Utility method to output an HTML page from a DOM object. The
     * appropriate headers are set for MIME type and to disable caching
     * of the HTML by the broswer.
     *
     * @param doc
     *        The DOM object to be returned as response
     */
    public void writeHTML(HTMLDocument doc) throws HttpPresentationException {
	if (sessionIdUrl && (sessionManager != null) && (sessionKey != null)) {
	    try {
		if (sessionManager.sessionExists(sessionKey)) {
		    StandardAppUtil.addSession(doc, sessionKey);
		}
	    } catch (SessionException e) {
                throw new HttpPresentationException(e);
	    }
	}
	writeHTML(HTMLFormatter.toDocument(doc));
    }

}
