/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: HttpDebug.java,v 1.9 2000/01/29 07:02:12 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.acme.gui.presObjs;

import java.util.Vector;
import java.util.Enumeration;
import javax.servlet.http.Cookie;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.http.*;
import com.lutris.html.*;
import com.lutris.util.FatalExceptionError;

/**
 * This class is used for debugging.   Now only used for httpPresentation
 * tests.
 *
 * @author Kyle Clark
 */
public class HttpDebug {

    private static int tableBorderWidth = 2;
    private static int tablePercentWidth = 100;
    private static String [] httpRequestHeader = {
        "Accept",
        "Accept-Charset",
        "Accept-Encoding",
        "Accept-Language",
        "Authorization",
        "From",
        "Host",
        "If-Modified-Since",
        "If-Match",
        "If-None-Match",
        "If-Range",
        "If-Unmodified-Since",
        "Max-Forwards",
        "Proxy-Authorization",
        "Range",
        "Referer",
        "User-Agent" };

    private static String [] httpResponseHeader = {
        "Age",
        "Allow",
        "Content-Base",
        "Content-Encoding",
        "Content-Language",
        "Content-Length",
        "Content-Location",
        "Content-MD5",
        "Content-Range",
        "Content-Type",
        "ETag",
        "Expires",
        "Last-Modified",
        "Location",
        "Proxy-Authenticate",
        "Public",
        "Retry-After",
        "Server",
        "Vary",
        "Warning",
        "WWW-Authenticate"
    };

    private static String [] httpRedirectHeader = httpResponseHeader;

    /**
     * Private constructor, prevent instantiation.
     */
    private HttpDebug()
    {
    }

    /**
     * Output an set of HTML table of a presentation object's arguments
     * and other state information.
     * @param comms The comms object for the presentation.  The HTML will
     *  be written to the response.
     */
    public static void displayComms(HttpPresentationComms comms) {
        try {
            HttpPresentationOutputStream out = comms.response.getOutputStream();
            out.println("<HR><P>");
            out.println(getRequestTables(comms.request));
            out.println("<P>");
            out.println(getCGIEnvironmentTable(comms.request));
            out.println("<P><HR>");
        } catch (Exception except) {
            throw new FatalExceptionError(except);
        }
    }

    /**
     * Generates tables containing request debug information.
     * Debug information includes request headers and cookies.
     *
     * @param request Object from which debug data is compiled.
     * @exception Exception Throws all exceptions.
     */
    public static
        String getRequestTables(HttpPresentationRequest request)
        throws Exception
    {
        HtmlTable tbl = new HtmlTable();
        HtmlTableRow row;
        HtmlTableDataCell td;

        tbl.setBorderWidth(tableBorderWidth*2);
        tbl.setPercentWidth(tablePercentWidth); 

        addRequestHeaders(request, httpRequestHeader, tbl);
        addRequestCookies(request, tbl);

        return tbl.toString();
    }

    /**
     * Generates a table containing all CGI environment information.
     * The table is in html format.
     *
     * @param request Object from which CGI data is compiled.
     * @exception Exception Throws all exceptions.
     */
    public static
        String getCGIEnvironmentTable(HttpPresentationRequest request)
        throws Exception
    {
        HtmlTable tbl = new HtmlTable();
        tbl.setBorderWidth(tableBorderWidth);
        tbl.setPercentWidth(tablePercentWidth); 
        addCGIEnvironmentData(request, tbl);
        return tbl.toString();
    }

    /**
     * This method adds rows to a table that contain the
     * CGI environment data.
     *
     * @param request Object from which CGI data is compiled.
     * @param tbl Table to which to add the data.
     * @exception Exception Throws all exceptions.
     */
    protected static
        void addCGIEnvironmentData(HttpPresentationRequest request,
                                   HtmlTable tbl)
        throws Exception
    {
        HtmlTableRow row;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        row = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1>CGI Environment</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        row.addCell(th);
        row.setColSpan(2);
        row.setBackgroundColor("#999999");
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Name</B>");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        td = new HtmlTableDataCell("<B>Value</B>");
        td.setHorizontalAlignment(td.ALIGN_LEFT);
        row.addCell(td);
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("DOCUMENT_URI");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getRequestURI()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("SERVER_NAME");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getServerName()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("SERVER_PROTOCOL");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getProtocol()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("SERVER_PORT");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(Integer.toString(request.getServerPort())));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("REQUEST_METHOD");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getMethod()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("PATH_INFO");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getPathInfo()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("PATH_TRANSLATED");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getPathTranslated()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("APPLICATION");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getApplicationPath()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("PRESENTATION");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getPresentationObjectPath()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("RELATIVE PRESENTATION");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getPresentationObjectRelativePath()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("QUERY_STRING");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getQueryString()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("REMOTE_HOST");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getRemoteHost()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("REMOTE_ADDR");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getRemoteAddr()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("AUTH_TYPE");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getAuthType()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("REMOTE_USER");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getRemoteUser()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("CONTENT_TYPE");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(request.getContentType()));
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("CONTENT_LENGTH");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        row.addCell(new HtmlTableDataCell(Integer.toString(request.getContentLength())));
        tbl.addRow(row);
    }

    /**
     * Adds request headers data rows to a table.
     *
     * @param request Object from which http request header data is compiled.
     * @param httpRequestHeader List of header names that should be displayed.
     * @param tbl Table to which to add the data.
     * @exception Exception Throws all exceptions.
     */
    private static void addRequestHeaders(HttpPresentationRequest request,
                                          String [] httpRequestHeader,
                                          HtmlTable tbl)
        throws Exception {
        HtmlTableRow row;
        HtmlTableDataCell td;
	HtmlTableHeaderCell th;

        row = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1>Request Headers</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        row.addCell(th);
        row.setColSpan(2);
        row.setBackgroundColor("#999999");
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Name</B>");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        td = new HtmlTableDataCell("<B>Value</B>");
        td.setHorizontalAlignment(td.ALIGN_LEFT);
        row.addCell(td);
        tbl.addRow(row);

        for (int idx=0; idx<httpRequestHeader.length; idx++) {
            row = new HtmlTableRow();
            td = new HtmlTableDataCell(httpRequestHeader[idx]);
            td.setHorizontalAlignment(td.ALIGN_CENTER);
            row.addCell(td);
            String val = request.getHeader(httpRequestHeader[idx]);
            if (val != null) {
                td = new HtmlTableDataCell(val);
                row.addCell(td);
                tbl.addRow(row);
            }
        }
    }

    /**
     * Generates a table containing http request headers.
     * Only headers that are set are displayed.
     * The table is in html format.
     *
     * @param request from which http request header data is compiled.
     * @param httpRequestHeader List that that specifies the
     *   set of headers that should be displayed.
     * @exception Exception Throws all exceptions.
     */
    private static String getRequestHeaderTable(HttpPresentationRequest request,
                                                String [] httpRequestHeader)
        throws Exception {
        HtmlTable tbl = new HtmlTable();
        tbl.setBorderWidth(tableBorderWidth);
        tbl.setPercentWidth(tablePercentWidth); 
        addRequestHeaders(request, httpRequestHeader, tbl);
        return tbl.toString();
    }

    /**
     * Generates a table containing http request header values.
     * The table is in html format.  Only headers that have values
     * are displayed.
     *
     * @param request Object from which http request header data is compiled.
     * @exception Exception Throws all exceptions.
     */
    private static String getRequestHeaderTable(HttpPresentationRequest request)
        throws Exception {
        return getRequestHeaderTable(request, httpRequestHeader);
    }


    /**
     * Adds request cookie setting to table.
     * The table is in html format.
     *
     * @param request Object from which http request header data is compiled.
     * @param tbl Table to which to add the data.
     * @exception Exception Throws all exceptions.
     */
    private static void addRequestCookies(HttpPresentationRequest request, HtmlTable tbl)
        throws Exception{
        HtmlTableRow row;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        row = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1>Request Cookies</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        row.addCell(th);
        row.setColSpan(2);
        row.setBackgroundColor("#999999");
        tbl.addRow(row);

        row = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Name</B>");
        td.setHorizontalAlignment(td.ALIGN_CENTER);
        row.addCell(td);
        td = new HtmlTableDataCell("<B>Value</B>");
        td.setHorizontalAlignment(td.ALIGN_LEFT);
        row.addCell(td);
        tbl.addRow(row);

        Cookie [] cookies = request.getCookies();
        if (cookies != null) {
            for (int i=0; i<cookies.length; i++) {
                row = new HtmlTableRow();
                td = new HtmlTableDataCell(cookies[i].getName());
                td.setHorizontalAlignment(td.ALIGN_CENTER);
                row.addCell(td);
                td = new HtmlTableDataCell(cookies[i].getValue());
                td.setHorizontalAlignment(td.ALIGN_LEFT);
                row.addCell(td);
                tbl.addRow(row);
            }
        }
    }

    /**
     * Generates a table containing http request cookie values.
     * The table is in html format.
     *
     * @param request Object from which http request header data is compiled.
     * @exception Exception Throws all exceptions.
     */
    private static String getRequestCookieTable(HttpPresentationRequest request)
        throws Exception {
        HtmlTable tbl = new HtmlTable();
        tbl.setBorderWidth(tableBorderWidth);
        tbl.setPercentWidth(tablePercentWidth); 
        addRequestCookies(request, tbl);
        return tbl.toString();
    }
}
