/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: BasicSession.java,v 1.16.2.1 2000/04/04 00:31:42 peterj Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server.session;

import java.util.*;
import java.net.*;
import java.io.Serializable;
import javax.servlet.http.*;

import org.apache.tomcat.core.*;

import com.lutris.util.*;
import com.lutris.appserver.server.user.User;
import com.lutris.appserver.server.*;

/**
 * The standard implementation of the StandardSession interface.
 * The session reflects and instance of a client interacting
 * with the application.  A BasicSession maybe associated
 * with a user or exist in a unauthenticated state.  The
 * session manager associated with a BasicSession object must be
 * a StandardSessionManager or a derivation.
 *
 * @version $Revision: 1.16.2.1 $
 * @author  John Marco
 * @author  Shawn McMurdo
 */
public class BasicSession implements StandardSession, HttpSession, Serializable{

    // Intentionally transient
    protected transient StandardSessionManager sessionManager = null;

    protected User user = null;
    private String sessionKey = null;
    private long timeCreated;       // Milliseconds.
    private long timeLastUsed;      // Milliseconds.
    private long timeExpires;       // Milliseconds.
    private long maxIdleTime;       // Milliseconds.
    private long maxNoUserIdleTime; // Milliseconds.

    /**
     * The application specific data for this session.
     */
    public SessionData data;

    // The number of references to this session.
    private int refCount = 0;

    /**
     * Need the following constructor in order to serialize?
     */
    public BasicSession() {}

    /**
     * Construct a new session.  Only called by
     * <CODE>StandardSessionManager</CODE>.
     *
     * @param sessionManager
     *   The session manager that will manager this session.
     * @param sessionKey
     *   The unique session key associated with the session.
     */
    protected BasicSession(StandardSessionManager sessionManager,
                           String sessionKey) {
        this(sessionManager, sessionKey, new SessionData());
        initHttpSession();
    }

    /**
     * Construct a new session.  Only called by
     * <CODE>StandardSessionManager</CODE>.
     *
     * @param sessionManager
     *   The session manager that will manager this session.
     * @param sessionKey
     *   The unique session key associated with the session.
     * @param sessionData
     *   The session data to associate with this session.
     */
    protected BasicSession(StandardSessionManager sessionManager,
                           String sessionKey, SessionData sessionData) {
        this.sessionManager = sessionManager;
        this.sessionKey = sessionKey;
        /*
         * Set session information.
         */
        long now = System.currentTimeMillis();
        timeCreated = now;
        timeLastUsed = now;
        long lifeTime = sessionManager.getMaxSessionLifeTime();
        if (lifeTime > 0) {
            // Convert timeExpires from delta seconds to an absolute
            // millisecond-based date.
            timeExpires = now + (lifeTime * 1000);
        } else {
            // Not using this feature.
            timeExpires = 0;
        }
        maxIdleTime = sessionManager.getMaxSessionIdleTime() * 1000;
        maxNoUserIdleTime = sessionManager.getMaxNoUserSessionIdleTime() * 1000;
        this.data = sessionData;
    }

    /**
     * Set the "last used" timestamp to the current time, resetting the
     * idle period.
     */
    public void touch() {
        timeLastUsed = System.currentTimeMillis();
    }

    /**
     * Obtain the user associated with this session.
     *
     * @return The user object or <CODE>null</CODE> if the
     * session is not logged in.
     */
    public User getUser() {
        return user;
    }

    /**
     * Set the user associated with this session.  This will register the
     * user with the <CODE>SessionManager</CODE>. <P>
     *
     * If it is neccessary to prevent a user from logging on multiple times,
     * this can be accomplished by synchronizing on the
     * <CODE>Sessionmanager</CODE> object and enquiring about the number of
     * users associated with a session.  It is then possible to delete other
     * sessions before adding a new session.
     *
     * @param user
     *   the user object to associate with the session.
     * @exception SessionException
     *   if the user cannot be associated with the session.
     */
    public void setUser(User sessionUser) throws SessionException {
        if (user != null) {
            sessionManager.unregisterUser(this);
        }
        user = sessionUser;
        if (sessionUser != null) {
            sessionManager.registerUser(this);
        }
    }

    /**
     * Remove the user association with this session.  This will unregister the
     * user with the <CODE>SessionManager</CODE>.
     *
     * @exception SessionException
     *   if an error occurs.
     */
    public void clearUser() throws SessionException {
        if (user != null) {
            sessionManager.unregisterUser(this);
            user = null;
        }
    }

    /**
     * Obtain the unique key associated with this session.
     *
     * @return A String containing the key for this session.
     */
    public String getSessionKey() {
        return sessionKey;
    }

    /**
     * Obtain the session manager associated with this session.
     *
     * @return The session manager object.
     */
    public SessionManager getSessionManager() {
        return sessionManager;
    }

    /**
     * Obtain the application specific data for this session.
     *
     * @return A SessionData object containing application specific
     *   data for this session.
     */
    public SessionData getSessionData() {
        return data;
    }

    /**
     * Obtain the creation time for this object.  The time is in
     * milliseconds since Midnight, Jan 1, 1970 (epoch).
     *
     * @return The creation time since epoch for this object.
     */
    public long getTimeCreated() {
        return timeCreated;
    }

    /**
     * Obtain the time of last use for this object.  The time is in
     * milliseconds since Midnight, Jan 1, 1970 (epoch).
     *
     * @return The time of last use since epoch for this object.
     */
    public long getTimeLastUsed() {
        return timeLastUsed;
    }

    /**
     * Obtain the time of expiry for this object.  The time is in
     * milliseconds since Midnight, Jan 1, 1970 (epoch).
     * Returns zero (or negative) if there is no expiration date.
     * Note: this is a maximum lifespan for the session, regardless of activity.
     *
     * @return The time of expiry since epoch for this object.
     */
    public long getTimeExpires() {
        return timeExpires;
    }

    /**
     * Set the time of expiry for this object.  The time is in
     * milliseconds since Midnight, Jan 1, 1970 (epoch).
     * Set this value to zero (or less) to indicate that the session does not
     * have a maximum age.
     *
     * @param timeLastUsed The time of expiry since epoch.
     */
    public void setTimeExpires(long timeExpires) {
        this.timeExpires = timeExpires;
    }

    /**
     * Obtain the maximum idle time for this object. Zero (or negative)
     * indicates that sessions may be idle indefinetly.
     *
     * @return The maximum number of milliseconds this session
     *      may be idle.
     */
    public long getMaxIdleTime() {
        return maxIdleTime;
    }

    /**
     * Set the maximum idle time for this object. Set this to zero
     * (or negative) to disable idle checking.
     *
     * @param maxIdleTime The maximum number of milliseconds this
     *  session may be idle, or zero (or negative) to allow sessions to be idle
     *  indefinetly.
     */
    public void setMaxIdleTime(long maxIdleTime) {
        this.maxIdleTime = maxIdleTime;
    }

    /**
     * Obtain the maximum idle time when a <CODE>User</CODE> object
     * is not associated with the session. Zero (or negative) indicates that
     * sessions may be idle indefinetly.
     *
     * @return The maximum number of milliseconds this session
     *      may be idle.
     */
    public long getMaxNoUserIdleTime() {
        return maxNoUserIdleTime;
    }

    /**
     * set the maximum idle time when a <CODE>User</CODE> object
     * is not associated with the session. Set this to zero (or negative) to
     * disable idle checking.
     *
     * @param maxIdleTime The maximum number of milliseconds this
     *      session may be idle.
     */
    public void setMaxNoUserIdleTime(long maxIdleTime) {
        this.maxNoUserIdleTime = maxIdleTime;
    }

    /**
     * Returns true if the session is new.  A session is new if it has
     * been created but a client cookie was never used to reference it.
     * In other words, the cookie associated with the session was
     * never submitted by a client.
     *
     * @return true if the session is new.
     */
    public boolean isNew() {
        checkValid();
        return (timeCreated == timeLastUsed);
    }

    /**
     * Returns the number of active references to this object.
     *
     * @return
     *   the number of references.
     */
    public int getRefCount() {
        return refCount;
    }

    /**
     * Returns the number of active references to this object.
     *
     * @return
     *   the number of references.
     */
    public int incrementRefCount() {
        refCount++;
        return refCount;
    }

    /**
     * Returns the number of active references to this object.
     *
     * @return
     *   the number of references.
     */
    public int decrementRefCount() {
        if (refCount > 0) {
            refCount--;
        }
        return refCount;
    }

    /**
     * Prints a string representation of this object.
     *
     * @return
     *   the string representation.
     */
    public String toString() {
        StringBuffer str = new StringBuffer();
        str.append("{");
        str.append("sessionKey=" + sessionKey);
        str.append(", ");
        if (user != null) {
            str.append("user=" + user.getName());
        } else {
            str.append("user=" + user);
        }
        str.append(", ");
        str.append("timeCreated=" + timeCreated);
        str.append(", ");
        str.append("timeLastUsed=" + timeLastUsed);
        str.append(", ");
        str.append("timeExpires=" + timeExpires);
        str.append(", ");
        str.append("maxIdleTime=" + maxIdleTime);
        str.append(", ");
        str.append("maxNoUserIdleTime=" + maxNoUserIdleTime);
        str.append(", ");
        str.append("sessionData=" + data);
        str.append("}");
        return str.toString();
    }



   public HttpSession  getHttpSession() {
     // should be a ref to another object
        return this;
   }




   // The HttpSession interface implementation
   //

    Hashtable attributes;
    boolean valid;
    private Context context;

    private void initHttpSession() {
        attributes = new Hashtable();
        valid = true;
        context = null;
    }

    public void setContext(Context context) {
        this.context = context;
    }

    public long getCreationTime() {
        checkValid();
        return getTimeCreated();
    }

    public String getId() {
        return getSessionKey();
    }

    public long getLastAccessedTime() {
        checkValid();
        return getTimeLastUsed();
    }

    public void setMaxInactiveInterval(int interval) {
        checkValid();
        // convert from seconds to miliseconds
        setMaxIdleTime(interval*1000);
    }


    public int getMaxInactiveInterval() {
        checkValid();
        // convert from miliseconds to seconds
        return (int)(getMaxIdleTime()/1000);
    }


    public HttpSessionContext getSessionContext() {
        checkValid();
	return new SessionContextImpl();
    }


    public Object getAttribute(String name) {
        checkValid();
        Object o = attributes.get(name);
        return attributes.get(name);
    }

    public Object getValue(String name) {
        checkValid();
        return getAttribute(name);
    }


    public Enumeration getAttributeNames() {
        checkValid();
        Hashtable attributesClone = (Hashtable)attributes.clone();
        return (Enumeration)attributesClone.keys();
    }

    public String[] getValueNames() {
        checkValid();
        Enumeration e = getAttributeNames();
        Vector names = new Vector();
        while (e.hasMoreElements()) {
            names.addElement(e.nextElement());
        }
        String[] valueNames = new String[names.size()];
        names.copyInto(valueNames);
        return valueNames;
    }

    public void setAttribute(String name, Object value) {
        checkValid();
 	if (name == null) {
            String msg = "null value";
	    throw new IllegalArgumentException(msg);
	}
        removeAttribute(name);
	if (value != null && value instanceof HttpSessionBindingListener) {
	    HttpSessionBindingEvent e = new HttpSessionBindingEvent(this, name);
	    ((HttpSessionBindingListener)value).valueBound(e);
	}
        attributes.put(name,value);
    }

    public void putValue(String name, Object value) {
        checkValid();
        setAttribute(name, value);
    }

    public void removeAttribute(String name) {
        checkValid();
	if (name == null) {
            String msg = "null value";
	    throw new IllegalArgumentException(msg);
	}

	Object o = attributes.get(name);
	if (o instanceof HttpSessionBindingListener) {
	    HttpSessionBindingEvent e = new HttpSessionBindingEvent(this,name);
	    ((HttpSessionBindingListener)o).valueUnbound(e);
	}
        attributes.remove(name);
    }

    public void removeValue(String name) {
        checkValid();
        removeAttribute(name);
    }

    public void invalidate() {
        checkValid();
	Enumeration enum = attributes.keys();
	while (enum.hasMoreElements()) {
	    String name = (String)enum.nextElement();
	    removeAttribute(name);
	}
	valid = false;
    }


    private void checkValid() {
      if (! valid) {
        String msg = "Session is invalid.";
	throw new IllegalStateException(msg);
      }

    }



}

