/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: PagedSessionHandle.java,v 1.2 2000/01/29 07:02:14 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server.session;
import com.lutris.appserver.server.user.User;
import com.lutris.util.*;
import com.lutris.logging.*;
import java.util.*;

/**
 * A lightweight handle to a session which has been paged to disk.
 * Referenced to the session manger and user associated with the
 * session are not paged.  Instead they are heald by this object
 * until the session and corresponding session data are read
 * back from disk.
 *
 * @see         PagedSessionHome
 * @version	$Revision: 1.2 $
 * @author	Kyle Clark
 */
class PagedSessionHandle {

    /**
     * Keep track if the session object that is paged to disk
     * is new.
     */
    private boolean sessionIsNew;

    /**
     * Keep track of time session was created.
     */
    private long sessionTimeCreate;

    /**
     * The store interface.
     */
    private PersistentStore store;

    /**
     * The session that is being paged.
     */
    private PagedSession session;

    /**
     * The session key.
     */
    private String sessionKey;

    /**
     * Reference to transient data that should be
     * held by the handle.
     */
    private Object[] transientData;
     
    /**
     * Creates an instance of PagedSessionHandle.
     *
     * @param config
     *   configuration settings required by this object.
     * @param session
     *   the session represented by this object.
     */
    PagedSessionHandle(PagedSession session, PersistentStore store) {
        this.session = session;
        this.store = store;

        // Convenience - performance
        this.sessionIsNew = session.isNew();
        this.sessionTimeCreate = session.getTimeCreated();
        this.sessionKey = session.getSessionKey();

    }

    /**
     * Pages the session to the file system.
     *
     * @exception SessionException if an error occurs.
     */
    synchronized void write() throws SessionException {
        try {
            transientData = session.getTransientData();
            store.store(sessionKey, session);
            // delete reference to session
            // can be cleaned up via garbage collector.
            session = null; 
        } catch (Exception e) {
            throw new SessionException("Unable to write session to persistent store.", e);
        }
    }

    /**
     * Reads the session from the file system.  The session is
     * deleted from the file system.  May return null.
     *
     * @return the session.
     * @exception SessionException if an error occurs.
     */
    synchronized PagedSession read() throws SessionException {
        if (session != null) {
            return session;
        }
        try {
            session = (PagedSession)store.remove(sessionKey);
            session.restoreTransientData(transientData);
            return session;
        } catch (Exception e) {
            throw new SessionException("Unable to read session from persistent store.", e);
        }
    }

    /**
     * Deletes the paged session.
     *
     * @exception SessionException if an error occurs.
     */
    synchronized void delete() throws SessionException {
        try {
            store.delete(sessionKey);
        } catch (Exception e) {
            throw new SessionException("Unable to delete session from persistent store.", e);
        }
    }

    /**
     * Returns true if the session referenced by this handle is new.
     *
     * @return
     *   true if the session referenced by this handle is new.
     */
    boolean isNew() {
        return sessionIsNew;
    }

    /**
     * Returns the time the session referenced by this handle
     * was created.
     *
     * @return
     *   the time the session referenced by this handle
     *   was created.
     */
    long getTimeCreated() {
        return sessionTimeCreate;
    }

    /**
     * Returns the session key associated with this object.
     *
     * @return
     *   the session key.
     */
    String getSessionKey() {
        return sessionKey;
    }
     
}

