/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: SessionUtil.java,v 1.2 2000/01/29 07:02:15 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server.session;
import com.lutris.appserver.server.user.User;
import com.lutris.util.*;
import java.util.*;

/**
 * Static methods that can be used by applications to manage sessions.
 *
 * @version	$Revision: 1.2 $
 * @author	Kyle Clark
 */
public class SessionUtil {

    /**
     * Prevent instantiation.
     */
    private SessionUtil() {
    }

    /**
     * Logs a user in.
     *
     * @param user the user to log in.
     * @param session the session to associated with the user.
     * @param multiple if true the user can have more than one session
     *   associated with him/her.  If false, all previous sessions
     *   associated with the user will be removed.
     * @exception SessionException if an error occurs.
     */
    public static void logIn(User user, Session session, boolean multiple)
        throws SessionException {
        SessionManager mgr = session.getSessionManager();
        synchronized(mgr) {
            if (!multiple) {
                Enumeration e = mgr.getSessionKeys(user);
                while (e.hasMoreElements()) {
                    String key = (String)e.nextElement();
                    mgr.deleteSession(key);
                }
            }
            session.setUser(user);
        }
    }

    /**
     * Dissasociates the user from the session and all
     * references to the session are deleted from the
     * session manager.
     *
     * @param session the session associated with the user.
     * @exception SessionException if an error occurs.
     */
    public static void logOut(Session session)
        throws SessionException {
	if (session != null) {
            SessionManager mgr = session.getSessionManager();
            mgr.deleteSession(session);
        }
    }

    /**
     * If a user attempts to login but has not logged
     * out from a previous session then this method can be used
     * to copy data from the previous session into this session
     * and to log out the previous session.
     *
     * @param user the user
     * @param session the session
     * @return true if the user was already logged in and the session
     *   has been resumed.
     * @exception SessionException
     *   if an error occurs.
     */
    public static boolean resumeLogIn(User user, Session session)
        throws SessionException {
        SessionManager mgr = session.getSessionManager();
        try {
            synchronized (mgr) {
                Enumeration e = mgr.getSessionKeys(user);
                if (e.hasMoreElements()) {
                    String key = (String)e.nextElement();
                    if (e.hasMoreElements()) {
                        // More than one login - we don't deal with it.
                        return false;
                    }
                    Session oldSession = mgr.getSession(Thread.currentThread(), key);
                    SessionData dataSrc = oldSession.getSessionData();
                    SessionData dataDst = session.getSessionData();
                    // Copy src to dst
                    String[] k = dataSrc.keys();
                    for (int i=0; i<k.length; i++) {
                        dataDst.set(k[i], dataSrc.get(k[i]));
                    }
                    SessionUtil.logIn(user, session, false);
                    return true;
                }
            }
        } catch (SessionException e) {
            throw e;
        } catch (Exception e) {
            throw new SessionException(e);
        }
        return false;
    }
}
