/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: StandardSessionHome.java,v 1.3 2000/01/29 07:02:15 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server.session;

import com.lutris.util.*;
import com.lutris.logging.*;
import java.util.Enumeration;
import java.net.*;

/**
 * StandardSessionManager uses StandardSessionHome to manage
 * a collection of sessions.  StandardSessionHome acts both as
 * a session factory and as a session repository.  The session
 * manager gains access to instances of sessions via the
 * home (StandardSessionHome) interface.  The session manager
 * dynamically loads the home interface.  The implementation
 * of the home interface that is loaded is specified in the
 * applications configuration file:
 *
 * <ul>
 * <li><code>SessionHome.Class: {class name}</code><p>
 * </ul>
 *
 * StandardSessionHome manages the state of a session.
 * A session exists in either of two states: 'active'
 * or 'passive'.  An 'active' session
 * is one that is actively being referenced by a thread of
 * execution (request).  A 'passive' session is one that
 * is not currently associated with any request (thread).
 * A session in the 'passive' state may be written to persistent store
 * until it becomes 'active'.  When a session becomes active,
 * if it isn't in memory then it can be read from persistent store.<p>
 *
 * @see         StandardSession
 * @see         StandardSessionManager
 * @see         BasicSessionHome
 * @see         PagedSessionHome
 * @see         com.lutris.appserver.server.session.persistent.PersistentSessionHome
 * @version	$Revision: 1.3 $
 * @author	Kyle Clark
 */
public interface StandardSessionHome {

    /**
     * Creates and returns a new session instance.  The
     * session is bound to the specified session key.  The
     * session is also associated with the current thread
     * and is considered in the 'active' state.  The
     * session remains in the 'active' state until the
     * thread puts the session into the 'passive' state..
     * Only this thread will be able to put the
     * session into the 'passive' state.
     *
     * @param sessionKey the key to associate with the session.
     * @return the newly created session.
     * @exception CreateSessionException if the session cannot be
     *   created.
     * @exception DuplicateKeyException if the session cannot
     *   be created because the key is already in use.
     * @exception SessionException if the session cannot
     *   be created for some other reason.
     * @see #passivateSession
     */
    public StandardSession createSession(String sessionKey)
        throws CreateSessionException, DuplicateKeyException, SessionException;

    /**
     * Returns the session bound to the session key.
     * The session must already be in the 'active' state
     * and associated with the current thread,
     * otherwise null is returned.
     *
     * @param sessionKey
     *   the session key for the session.  If the session
     *   doesn't exist or is not is not bound to the current
     *   thread then null is returned.
     * @return
     *   the session.
     * @exception SessionException
     *   if the session cannot be retrieved.
     * @see #getSession(Thread, String)
     */
    public StandardSession getSession(String sessionKey)
        throws SessionException;

    /**
     * Returns the session bound to the specified
     * session key.  The session is put into the 'active' state.
     * The session is also associated with the specified thread.
     * Only this thread will be able to put the session
     * back into the 'passive' state once it is done with the
     * session.
     *
     * @param thread the thread that should be associated with
     *   the session while it is in the active state.  Only this
     *   thread can put the session back into the passive state.
     * @param sessionKey
     *   the session key for the session that will be made
     *   'active' and returned.  If the session doesn't exist
     *   then null is returned.
     * @return
     *   the session.
     * @exception SessionException
     *   if the session cannot be retrieved.
     * @see #passivateSession
     */
    public StandardSession getSession(Thread thread, String sessionKey)
        throws SessionException;

    /**
     * Removes a session from the cache.  If the session
     * doesn't exist the opration is ignored.
     *
     * @param sessionKey
     *   the session key associated with the session.
     * @exception SessionException
     *   if the session cannot be retrieved.
     */
    public void removeSession(String sessionKey) throws SessionException;

    /**
     * Puts a session into the 'passive' state.  A 'passive'
     * session may be made persistent.  Only the thread that
     * put the session into the 'active' state may put
     * the session into the 'passive' state.
     *
     * @param thread the thread that is currently associated
     *   with the session.
     * @param sessionKey
     *   the session key for the session that will be made passive.
     * @exception SessionException
     *   if the session cannot be retrieved.
     */
    public void passivateSession(Thread thread, String sessionKey)
        throws SessionException;

    /**
     * Specifies if a key is currently bound to a session.
     *
     * @param sessionKey
     *   the session key to be tested.
     * @return
     *   true if the session key is in use.
     * @exception SessionException
     *   if the existence of the key cannot be determined.
     */
    public boolean containsKey(String sessionKey) throws SessionException;

    /**
     * Returns the current number of sessions.
     *
     * @return
     *   the 'active' session count.
     * @exception SessionException
     *   if the size cannot be determined
     */
    public int size() throws SessionException;

    /**
     * Returns the current number of sessions that are paged to
     * persistent store.
     *
     * @return
     *   the 'paged' session count.
     * @exception SessionException
     *   if the size cannot be determined
     */
    public int pagedSize() throws SessionException;

    /**
     * Returns an enumeration of the keys for all the sessions.
     *
     * @return
     *   the enumeration of session keys.
     * @exception SessionException
     *   if the session enumeration cannot be retrieved.
     */
    public Enumeration keys() throws SessionException;


    /**
     * Shuts dows the session home.
     */
    public void shutdown();

}




