/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: StandardSessionIdleTimer.java,v 1.19 2000/02/21 21:21:26 peterj Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server.session;

import com.lutris.appserver.server.*;
import com.lutris.logging.*;
import java.security.*;
import java.io.*;
import java.util.*;
import com.lutris.util.*;

/**
 * The idle timer thread for <CODE>StandardSessionManager</CODE>.
 * This thread sleeps in the background, waking up periodically to
 * check for inactive sessions and to terminate any session that has been
 * inactive for too long.
 *
 * @version	$Revision: 1.19 $
 * @author	John Marco
 * @author	Shawn McMurdo
 */
public class StandardSessionIdleTimer extends Thread {

    private StandardSessionIdleHandler idleHandler;
    private long scanInterval;   // In milliseconds.
    private Application app;    // The application context.

    /*
     * Constructor an idle timer object.  Sets the
     * internal session manager reference and check period,
     * but doesn't start the thread.
     *
     * @param manager The session manager to be checked periodically for
     * 		idle sessions.
     * @param app the application context.
     * @param scanIntervalSec Time interval, in seconds, to scan for idle
     *          session.
     */
    public StandardSessionIdleTimer(StandardSessionIdleHandler manager,
                                    Application app,
                                    long scanIntervalSec) {
        idleHandler = manager;
        this.app = app;
	// Convert idle timer scanning interval milliseconds.
	scanInterval = scanIntervalSec * 1000;
        setPriority(MIN_PRIORITY); // For performance
	setDaemon(true); // Don't require this thread to exit.
    }

    /**
     * The main code body of the Idle Timer Thread.  Enters an endless
     * loop that sleeps for a configurable period, periodically waking
     * up to check the Session Manager for idle sessions.
     */
    public void run() {
        if (app != null) {
          Enhydra.register(app);
        }
	while (true) {
	    try {
		sleep(scanInterval);
	    } catch (InterruptedException e) {
                // Ignore.
            }
            try {
                idleHandler.cleanUpIdleSessions();
            } catch (Exception e) {
                LogChannel log = Enhydra.getLogChannel();
                log.write(Logger.ALERT, "Failed to clean up idle sessions: " + e);
                // TODO - print to log
                e.printStackTrace();
            }
	}
        // TODO - Enhydra.unregister();
    }

    /**
     * Shutdown the thread associated with this object.
     */
    public void shutdown() {
        stop();
    }
}

