/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DBUtil.java,v 1.4 2000/01/29 07:02:16 markd Exp $
 *-----------------------------------------------------------------------------
 */

package com.lutris.appserver.server.session.persistent;

import com.lutris.appserver.server.Enhydra;
import com.lutris.appserver.server.session.*;
import com.lutris.appserver.server.sql.*;
import com.lutris.logging.*;
import java.util.*;
import java.sql.*;

/**
 * Utilitity class for performint database operations.
 *
 * @see         StandardSession
 * @see         StandardSessionManager
 * @version	$Revision: 1.4 $
 * @author	Kyle Clark
 */
class DBUtil {

    /**
     * Saves a session to the database.
     *
     * @param session
     *   the session to save.
     * @param dbName the database that should be accessed.  May be
     *   null in which case the default database is used.
     * @return the number of rows that were affected
     *   by this update.
     * @exception SessionException
     *   if an error occurs.
     */
    static int dbUpdate(PersistentSession session, String dbName)
        throws SessionException {
        PersistentSessionDO sessionDO = new PersistentSessionDO(session);
        try {
	    DBTransaction t = createTransaction(dbName);
            try {
                t.update(sessionDO);
                t.commit();
            } catch (Exception e) {
                t.rollback();
		throw e;
            } finally {
                t.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
        return sessionDO.getTransactionRowCount();
    }

    /**
     * Deletes a session from the database.
     *
     * @param sessionKey
     *   the key that identifies the session that will be deleted.
     * @param dbName the database that should be accessed.  May be
     *   null in which case the default database is used.
     * @exception SessionException
     *   if an error occurs.
     */
    static void dbDelete(String sessionKey, String dbName)
        throws SessionException {
        try {
	    DBTransaction t = createTransaction(dbName);
            SessionDeleteDO s = new SessionDeleteDO(sessionKey);
            try {
                t.delete(s);
                t.commit();
            } catch (Exception e) {
                t.rollback();
		throw e;
            } finally {
                t.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Inserts a session into the database.
     *
     * @param session
     *   the session that will be inserted.
     * @param dbName the database that should be accessed.  May be
     *   null in which case the default database is used.
     * @exception SessionException
     *   if an error occurs.
     */
    static void dbInsert(PersistentSession session, String dbName)
        throws SessionException {
        try {
            PersistentSessionDO sessionDO = new PersistentSessionDO(session);
	    DBTransaction t = createTransaction(dbName);
            try {
                t.insert(sessionDO);
                t.commit();
            } catch (Exception e) {
                t.rollback();
		throw e;
            } finally {
                t.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Executes a query of the database.
     *
     * @param query
     *   the object containg the query statement.
     * @param dbName the database that should be queried.  May be
     *   null in which case the default database is used.
     * @exception SessionException
     *   if an erro occurs.
     */
    static Object dbQuery(Query query, String dbName) throws SessionException {
        try {
            DBQuery q = createQuery(dbName);
            try {
                q.query(query);
                return q.next();
            } catch (Exception e) {
                throw e;
            } finally {
                q.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Creates a transaction for the specified DB name.
     *
     * @param dbName the database name.  May be null in
     *   which case the default database is used.
     * @exception DatabaseManagerException if an error occurs.
     * @exception SQLException if an error occurs.
     */
    static DBTransaction createTransaction(String dbName)
        throws DatabaseManagerException, SQLException {
        DBTransaction t;
        if (dbName != null) {
            t = Enhydra.getDatabaseManager().createTransaction(dbName);
        } else {
            t = Enhydra.getDatabaseManager().createTransaction();
        }
        return t;
    }

    /**
     * Creates a query for the specified DB name.
     *
     * @param dbName the database name.  May be null in
     *   which case the default database is used.
     * @exception DatabaseManagerException if an error occurs.
     * @exception SQLException if an error occurs.
     */
    static DBQuery createQuery(String dbName)
        throws DatabaseManagerException, SQLException {
        DBQuery q;
        if (dbName != null) {
            q = Enhydra.getDatabaseManager().createQuery(dbName);
        } else {
            q = Enhydra.getDatabaseManager().createQuery();
        }
        return q;
    }

}




