/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: FileLines.java,v 1.7 2000/01/29 07:02:19 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.jolt.joltc;

import java.util.*;
import java.io.*;

/**
 * Class used to assemble file of generated code.
 * Handles tracking of filename/linenumbers and doing indentation.
 */
class FileLines {
    /*
     * Amount to adjust indent by.
     */
    private static final int INDENT_INCR = 4;
    
    private StringBuffer code = new StringBuffer();
    private int currentIndent = 0;    // Current indentation.
    private String indentPrefix = "";
    private Vector lines = new Vector();

    // Previous location is used when append without a file name/line num
    // being specified.
    FileLine prevLine;

    /**
     * Construction a new file lines buffer.
     */
    public FileLines(String startingFile) {
        setIndentPrefix();

        // Use starting file for lines before first append with line numbers.
        prevLine= new FileLine("", startingFile, 0);
    }

    /**
     * Construction a new file lines buffer.
     */
    public FileLines(File startingFile) {
        this(startingFile.getName());
    }

    /**
     * Set the indention prefix.
     */
    private void setIndentPrefix() {
        StringBuffer prefixBuf = new StringBuffer();
        for (int cnt = 0; cnt < currentIndent; cnt++) {
            prefixBuf.append(' ');
        }
        indentPrefix = prefixBuf.toString();
    }


    /**
     * Increase indentation.
     */
    public void incrIndent() {
        currentIndent += INDENT_INCR;
        setIndentPrefix();
    }

    /**
     * Decrease indentation.
     */
    public void decrIndent() {
        currentIndent -= INDENT_INCR;
        setIndentPrefix();
    }

    /**
     * Add a line, remembering it as the previous line.
     */
    private void addLine(String line,
                         String srcFileName,
                         int srcLineNum) {
        prevLine = new FileLine(indentPrefix + line, srcFileName, srcLineNum);
        lines.addElement(prevLine);
    }

    /**
     * Append a string to the buffer, handling indentation and
     * line number mapping.  A newline is appended to the end of the
     * string.
     */
    public void append(String str,
                       String srcFileName,
                       int srcLineNum) {
        str += "\n";

        int idx = 0;
        int nlIdx;

        while ((idx < str.length()) &&
               ((nlIdx = str.indexOf('\n', idx)) >= 0)) {
            addLine(str.substring(idx, nlIdx+1), srcFileName, srcLineNum);
            srcLineNum++; 
            idx = nlIdx + 1;
        }
    }

    /**
     * Append with out a location.  Used for code that that does not
     * come from a source file.
     */
    public void append(String str) {
        String srcFileName = prevLine.srcFileName;
        // Line number is Not incremented by this routine.
        int srcLineNum = prevLine.srcLineNum;
        str += "\n";

        int idx = 0;
        int nlIdx;

        while ((idx < str.length()) &&
               ((nlIdx = str.indexOf('\n', idx)) >= 0)) {
            addLine(str.substring(idx, nlIdx+1), srcFileName, srcLineNum);
            idx = nlIdx + 1;
        }
    }

    /**
     * Append another FileLines to this one.
     */
    public void append(FileLines srcFile) {
        for (int idx = 0; idx < srcFile.lines.size(); idx++) {
            prevLine = (FileLine)srcFile.lines.elementAt(idx);
            lines.addElement(prevLine);
        }
    }

    /**
     * Get current number of lines.
     */
    int size() {
        return lines.size();
    }

    /**
     * Get code as a string.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        for (int idx = 0; idx < lines.size(); idx++) {
            buf.append(((FileLine)lines.elementAt(idx)).line);
        }
        return buf.toString();
    }
    
    /**
     * Write code to a file.
     */
    public void write(FileOutputStream outStream)
            throws java.io.IOException {
        PrintWriter out = new PrintWriter(outStream);
        for (int idx = 0; idx < lines.size(); idx++) {
            out.print(((FileLine)lines.elementAt(idx)).line);
        }
        out.flush();
    }
    
    /*
     * Print class (for debugging).
     */
    public void dbgPrint() {
        for (int idx = 0; idx < lines.size(); idx++) {
            System.err.println(idx+":"+(FileLine)lines.elementAt(idx));
        }
    }

    /**
     * Build map of generated line numbers to a string of "file:linenum".
     */
    String[] getLineNumMap() {
        String[] map = new String[lines.size()+1];
        map[0] = "???";
        for (int idx = 0; idx < lines.size(); idx++) {
            FileLine loc = (FileLine)lines.elementAt(idx);
            map[idx+1] = loc.srcFileName + ":" + loc.srcLineNum;
        }
        return map;
    }
}
