/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: JoltToJava.java,v 1.8 2000/01/29 07:02:19 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.server.jolt.joltc;

import com.lutris.appserver.server.jolt.parser.*;
import java.io.*;
import java.util.*;

/**
 * Convert a Jolt source file to a object containing the class.
 * All methods of this class are static.
 *
 * @version	$Revision: 1.8 $
 * @author	Mark Diekhans
 * @since	2.0
 */
class JoltToJava {
    /*
     * Class names to use for Presentation, Page and pageArgs
     */
    private static String PRESENTATION_BASE_CLASS =
        "JoltPage";
    private static String PAGE_ACCESS =
        "public";

    /*
     * Standard imports to add.
     */
    private static String PRESENTATION_IMPORT =
        "import com.lutris.appserver.server.httpPresentation.*;";
    private static String JOLT_PRESENTATION_IMPORT =
        "import com.lutris.appserver.server.jolt.joltpo.*;";
    private static String KEYWORD_VALUE_IMPORT =
        "import com.lutris.util.*;";

    /*
     * Comment strings to seperate sections of the generated code.
     */
    private static final String sectionSeparator;
    private static final String subSectionSeparator;

    /**
     * Class constructor.
     */
    static {
        StringBuffer strTmp;
        
        /*
         * Generate comment separators that will be used to make generated
         * code more readable.
         */
        strTmp = new StringBuffer ("/*");
        for (int cnt = 0; cnt < 70; cnt++) {
            strTmp.append ('=');
        }
        strTmp.append ("*/");
        sectionSeparator = strTmp.toString ();
        
        strTmp = new StringBuffer ("/*");
        for (int cnt = 0; cnt < 70; cnt++) {
            strTmp.append ('-');
        }
        strTmp.append ("*/");
        subSectionSeparator = strTmp.toString ();
    }

    /*
     * Prevent instanciation.
     */
    private JoltToJava() {
    }
    
    /**
     * Convert a Jolt HTML file to a Java class.
     *
     * @param pagePackage Package name of the class.
     * @param pageClass Name of class for the page.
     * @param srcFile Jolt HTML source file.
     * @return Object containing all of the Java source and a source line
     *  map information.
     * @exception java.io.FileNotFoundException If the source file or
     *  an include file was not found.
     * @exception java.io.IOException If an error occured reading the source
     *  or an included file was not found.
     * @exception JoltCompileException
     */
    protected static FileLines compile(String pagePackage,
                                       String pageClass,
                                       File srcFile,
                                       File javaFile)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {

        JoltCodeGen code = new JoltCodeGen(srcFile);
        
        FileLines javaSrc = new FileLines(javaFile);
        javaSrc.append("package " + pagePackage + ";\n\n" +
                       PRESENTATION_IMPORT + "\n" +
                       JOLT_PRESENTATION_IMPORT + "\n" +
                       KEYWORD_VALUE_IMPORT + "\n\n");
        
        if (code.importDefs.size() > 0) {
            javaSrc.append(sectionSeparator);
            javaSrc.append(code.importDefs);
        }

        javaSrc.append(PAGE_ACCESS + " class " + pageClass +
                       " extends " + PRESENTATION_BASE_CLASS + " {");

        if (code.htmlStrDefs.size () > 0) {
            javaSrc.append(subSectionSeparator);
            javaSrc.append(code.htmlStrDefs);
        }
            
        if (code.classDefs.size() > 0) {
            javaSrc.append(sectionSeparator);
            javaSrc.append(code.classDefs);
        }
        javaSrc.append(sectionSeparator);
        javaSrc.append("protected void runPage()\n" +
                       "      throws Exception {");
            
        javaSrc.incrIndent();
        javaSrc.append("try {");

        javaSrc.incrIndent();
        javaSrc.append(code.runMethod);
            
        javaSrc.decrIndent();
        javaSrc.append("}");
        if (code.pageCatches.size() > 0) {
            javaSrc.append(code.pageCatches);
        } 
            
        javaSrc.append(" finally {");
        javaSrc.append(code.pageFinally);
        javaSrc.append("}");
        javaSrc.decrIndent();
        javaSrc.append("}");
        javaSrc.append("}");

        return javaSrc;
    }
}  
