/*
 * @(#)CheckAttribsImpl.java   1.3 1999/07/30
 *
 */

package org.w3c.tidy;

/**
 *
 * Check HTML attributes implementation
 *
 * (c) 1998, 1999 (W3C) MIT, INRIA, Keio University
 * See Tidy.java for the copyright notice.
 * Derived from <a href="http://www.w3.org/People/Raggett/tidy">
 * HTML Tidy Release 26 Jul 1999</a>
 *
 * @author  Dave Raggett <dsr@w3.org>
 * @author  Andy Quick <ac.quick@sympatico.ca> (translation to Java)
 * @version 1.0, 1999/05/22
 * @version 1.0.1, 1999/05/29
 * @version 1.1, 1999/06/18 Java Bean
 * @version 1.2, 1999/07/10 Tidy Release 7 Jul 1999
 * @version 1.3, 1999/07/30 Tidy Release 26 Jul 1999
 */

public class CheckAttribsImpl {

    public static class CheckHTML implements CheckAttribs {

        public void check( Lexer lexer, Node node )
        {
            AttVal attval;
            Attribute attribute;

            for (attval = node.attributes; attval != null; attval = attval.next)
            {
                attribute = attval.checkAttribute(lexer, node );

                if (attribute == AttributeTable.attrXmlns)
                    lexer.isvoyager = true;
            }
        }

    };

    public static class CheckSCRIPT implements CheckAttribs {

        public void check( Lexer lexer, Node node )
        {
            Attribute attribute;
            AttVal attval, lang, type;

            lang = null;
            type = null;

            for (attval = node.attributes; attval != null; attval = attval.next)
            {
                attribute = attval.checkAttribute( lexer, node );

                if (attribute == AttributeTable.attrLanguage)
                    lang = attval;
                else if (attribute == AttributeTable.attrType)
                    type = attval;
            }

            if (type == null)
            {
                Report.attrError(lexer, node, "type", Report.MISSING_ATTRIBUTE);

                /* check for javascript */

                if (lang != null)
                {
                    String str = lang.value;
                    if (str.length() > 10)
                        str = str.substring(0, 10);
                    if ( (Lexer.wstrcasecmp(str, "javascript") == 0) ||
                         (Lexer.wstrcasecmp(str, "jscript") == 0) )
                    {
                        type = new AttVal();
                        type.attribute = "type";
                        type.value = "text/javascript";
                        type.next = node.attributes;
                        type.delim = '"';
                        type.dict = AttributeTable.getDefaultAttributeTable().findAttribute(type);
                        node.attributes = type;
                    }
                }
            }
        }

    };

    public static class CheckTABLE implements CheckAttribs {

        public void check( Lexer lexer, Node node )
        {
            AttVal attval;
            Attribute attribute;
            boolean hasSummary = false;

            for (attval = node.attributes; attval != null; attval = attval.next)
            {
                attribute = attval.checkAttribute(lexer, node);

                if (attribute == AttributeTable.attrSummary)
                    hasSummary = true;
            }

            /* suppress warning for missing summary for HTML 2.0 and HTML 3.2 */
            if (!hasSummary && lexer.doctype != Dict.VERS_HTML20 && lexer.doctype != Dict.VERS_HTML32)
            {
                lexer.badAccess |= Report.MISSING_SUMMARY;
                Report.attrError(lexer, node, "summary", Report.MISSING_ATTRIBUTE);
            }
        }

    };

    public static class CheckCaption implements CheckAttribs {

        public void check( Lexer lexer, Node node )
        {
            AttVal attval;
            String value = null;

            for (attval = node.attributes; attval != null; attval = attval.next)
            {
                if ( Lexer.wstrcasecmp(attval.attribute, "align") == 0 )
                {
                    value = attval.value;
                    break;
                }
            }

            if (value != null)
            {
                if (Lexer.wstrcasecmp(value, "left") == 0 || Lexer.wstrcasecmp(value, "right") == 0)
                    lexer.versions &= (short)(Dict.VERS_HTML40_LOOSE|Dict.VERS_FRAMES);
                else if (Lexer.wstrcasecmp(value, "top") == 0 || Lexer.wstrcasecmp(value, "bottom") == 0)
                    lexer.versions &= Dict.VERS_FROM32;
                else
                    Report.attrError(lexer, node, value, Report.BAD_ATTRIBUTE_VALUE);
            }
        }

    };

    public static class CheckIMG implements CheckAttribs {

        public void check( Lexer lexer, Node node )
        {
            AttVal attval;
            Attribute attribute;
            boolean hasAlt = false;
            boolean hasSrc = false;
            boolean hasUseMap = false;
            boolean hasIsMap = false;

            for (attval = node.attributes; attval != null; attval = attval.next)
            {
                attribute = attval.checkAttribute( lexer, node );

                if (attribute == AttributeTable.attrAlt)
                    hasAlt = true;
                else if (attribute == AttributeTable.attrSrc)
                    hasSrc = true;
                else if (attribute == AttributeTable.attrUsemap)
                    hasUseMap = true;
                else if (attribute == AttributeTable.attrIsmap)
                    hasIsMap = true;
            }

            if (!hasAlt)
            {
                lexer.badAccess |= Report.MISSING_IMAGE_ALT;
                Report.attrError(lexer, node, "alt", Report.MISSING_ATTRIBUTE);
            }

            if (!hasSrc)
                Report.attrError(lexer, node, "src", Report.MISSING_ATTRIBUTE);

            if (hasIsMap && !hasUseMap)
                Report.attrError(lexer, node, "ismap", Report.MISSING_IMAGEMAP);
        }

    };

    public static class CheckAREA implements CheckAttribs {

        public void check( Lexer lexer, Node node )
        {
            AttVal attval;
            Attribute attribute;
            boolean hasAlt = false;
            boolean hasHref = false;

            for (attval = node.attributes; attval != null; attval = attval.next)
            {
                attribute = attval.checkAttribute( lexer, node );

                if (attribute == AttributeTable.attrAlt)
                    hasAlt = true;
                else if (attribute == AttributeTable.attrHref)
                    hasHref = true;
            }

            if (!hasAlt)
            {
                lexer.badAccess |= Report.MISSING_LINK_ALT;
                Report.attrError(lexer, node, "alt", Report.MISSING_ATTRIBUTE);
            }
            if (!hasHref)
                Report.attrError(lexer, node, "href", Report.MISSING_ATTRIBUTE);
        }

    };

    public static CheckAttribs getCheckHTML()
    {
        return _checkHTML;
    }

    public static CheckAttribs getCheckSCRIPT()
    {
        return _checkSCRIPT;
    }

    public static CheckAttribs getCheckTABLE()
    {
        return _checkTABLE;
    }

    public static CheckAttribs getCheckCaption()
    {
        return _checkCaption;
    }

    public static CheckAttribs getCheckIMG()
    {
        return _checkIMG;
    }

    public static CheckAttribs getCheckAREA()
    {
        return _checkAREA;
    }


    private static CheckAttribs _checkHTML = new CheckHTML();
    private static CheckAttribs _checkSCRIPT = new CheckSCRIPT();
    private static CheckAttribs _checkTABLE = new CheckTABLE();
    private static CheckAttribs _checkCaption = new CheckCaption();
    private static CheckAttribs _checkIMG = new CheckIMG();
    private static CheckAttribs _checkAREA = new CheckAREA();

}
