/*
 * @(#)Clean.java   1.3 1999/07/30
 *
 */

package org.w3c.tidy;

/**
 *
 * Clean up misuse of presentation markup
 *
 * (c) 1998, 1999 (W3C) MIT, INRIA, Keio University
 * See Tidy.java for the copyright notice.
 * Derived from <a href="http://www.w3.org/People/Raggett/tidy">
 * HTML Tidy Release 26 Jul 1999</a>
 *
 * @author  Dave Raggett <dsr@w3.org>
 * @author  Andy Quick <ac.quick@sympatico.ca> (translation to Java)
 * @version 1.0, 1999/05/22
 * @version 1.0.1, 1999/05/29
 * @version 1.1, 1999/06/18 Java Bean
 * @version 1.2, 1999/07/10 Tidy Release 7 Jul 1999
 * @version 1.3, 1999/07/30 Tidy Release 26 Jul 1999
 */

/*
  Filters from other formats such as Microsoft Word
  often make excessive use of presentation markup such
  as font tags, B, I, and the align attribute. By applying
  a set of production rules, it is straight forward to
  transform this to use CSS.

  Some rules replace some of the children of an element by
  style properties on the element, e.g.

  <p><b>...</b></p> -> <p style="font-weight: bold">...</p>

  Such rules are applied to the element's content and then
  to the element itself until none of the rules more apply.
  Having applied all the rules to an element, it will have
  a style attribute with one or more properties. 

  Other rules strip the element they apply to, replacing
  it by style properties on the contents, e.g.
  
  <dir><li><p>...</li></dir> -> <p style="margin-left 1em">...
      
  These rules are applied to an element before processing
  its content and replace the current element by the first
  element in the exposed content.

  After applying both sets of rules, you can replace the
  style attribute by a class value and style rule in the
  document head. To support this, an association of styles
  and class names is built.

  A naive approach is to rely on string matching to test
  when two property lists are the same. A better approach
  would be to first sort the properties before matching.
*/

public class Clean {

    private int classNum = 1;

    private static StyleProp insertProperty(StyleProp props, String name,
                                            String value)
    {
        StyleProp first, prev, prop;
        int cmp;

        prev = null;
        first = props;

        while (props != null)
        {
            cmp = props.name.compareTo(name);

            if (cmp == 0)
            {
                /* need to merge values */
                break;
            }

            if (cmp > 0) // props.name > name
            {
                /* insert before this */

                prop = new StyleProp(new String(name), new String(value), props);

                if (prev != null)
                    prev.next = prop;
                else
                    first = prop;

                return first;
            }

            prev = props;
            props = props.next;
        }

        prop = new StyleProp(new String(name), new String(value));

        if (prev != null)
            prev.next = prop;
        else
            first = prop;

        return first;
    }

    /*
     Create sorted linked list of properties from style string
     It temporarily places nulls in place of ':' and ';' to
     delimit the strings for the property name and value.
     Some systems don't allow you to null literal strings,
     so to avoid this, a copy is made first.
    */
    private static StyleProp createProps(StyleProp prop, String style)
    {
        int name_end;
        int value_end;
        int value_start = 0;
        int name_start = 0;
        boolean more;

        name_start = 0;
        while (name_start < style.length())
        {
            while (name_start < style.length() &&
                       style.charAt(name_start) == ' ')
                ++name_start;

            name_end = name_start;

            while (name_end < style.length())
            {
                if (style.charAt(name_end) == ':')
                {
                    value_start = name_end + 1;
                    break;
                }

                ++name_end;
            }

            if (name_end > style.length() || style.charAt(name_end) != ':')
                break;

            while (value_start < style.length() &&
                       style.charAt(value_start) == ' ')
                ++value_start;

            value_end = value_start;
            more = false;

            while (value_end < style.length())
            {
                if (style.charAt(value_end) == ';')
                {
                    more = true;
                    break;
                }

                ++value_end;
            }

            prop = insertProperty(prop,
                                  style.substring(name_start, name_end),
                                  style.substring(value_start, value_end));

            if (more)
            {
                name_start = value_end + 1;
                continue;
            }

            break;
        }

        return prop;
    }

    private static String createPropString(StyleProp props)
    {
        String style = "";
        int len;
        StyleProp prop;

        /* compute length */

        for (len = 0, prop = props; prop != null; prop = prop.next)
        {
            len += prop.name.length() + 2;
            len += prop.value.length() + 2;
        }

        for (prop = props; prop != null; prop = prop.next)
        {
            style = style.concat(prop.name);
            style = style.concat(": ");

            style = style.concat(prop.value);

            if (prop.next == null)
                break;

            style = style.concat("; ");
        }

        return style;
    }

    /*
      create string with merged properties
    */
    private static String addProperty(String style, String property)
    {
        StyleProp prop;

        prop = createProps(null, style);
        prop = createProps(prop, property);
        style = createPropString(prop);
        return style;
    }

    private String gensymClass(String tag)
    {
        String str;

        str = "c" + classNum;
        classNum++;
        return str;
    }

    private String findStyle(Lexer lexer, String tag, String properties)
    {
        Style style;

        for (style = lexer.styles; style != null; style=style.next)
        {
            if (style.tag.equals(tag) &&
                style.properties.equals(properties))
                return style.tagClass;
        }

        style = new Style(new String(tag), gensymClass(tag),
                          new String(properties), lexer.styles);
        lexer.styles = style;
        return style.tagClass;
    }

    /*
     Find style attribute in node, and replace it
     by corresponding class attribute. Search for
     class in style dictionary otherwise gensym
     new class and add to dictionary.

     Assumes that node doesn't have a class attribute
    */
    private void style2Rule(Lexer lexer, Node node)
    {
        AttVal av;
        String classname;

        for (av = node.attributes; av != null; av = av.next)
        {
            if (av.attribute.equals("style"))
                break;
        }

        /* if style attribute already exists then append property */

        if (av != null)
        {
            classname = findStyle(lexer, node.element, av.value);
            av.attribute = new String("class");
            av.value = new String(classname);
        }
    }

    /* create style element using rules from dictionary */
    private static void createStyleElement(Lexer lexer, Node doc)
    {
        Node node, html, head;
        Style style;
        AttVal av;

        if (lexer.styles == null)
            return;

        node = new Node(Node.StartTag, null, 0, 0, new String("style"));
        node.implicit = true;

        /* insert type attribute */
        av = new AttVal(null, null, '"',
                        new String("type"),
                        new String("text/css"));
        av.dict = AttributeTable.getDefaultAttributeTable().findAttribute(av);
        node.attributes = av;

        lexer.txtstart = lexer.lexsize;

        for (style = lexer.styles; style != null; style = style.next)
        {
            lexer.addCharToLexer(' ');
            lexer.addStringLiteral(style.tag);
            lexer.addCharToLexer('.');
            lexer.addStringLiteral(style.tagClass);
            lexer.addCharToLexer(' ');
            lexer.addCharToLexer('{');
            lexer.addStringLiteral(style.properties);
            lexer.addCharToLexer('}');
            lexer.addCharToLexer('\n');
        }

        lexer.txtend = lexer.lexsize;

        Node.insertNode(node,
                        new Node(Node.TextNode,
                                 lexer.lexbuf,
                                 lexer.txtstart,
                                 lexer.txtend));

        /*
         now insert style element into document head

         doc is root node. search its children for html node
         the head node should be first child of html node
        */

        for (html = doc.content; html != null; html = html.next)
        {
            if (html.tag == TagTable.tagHtml)
            {
                head = html.content;
                Node.insertNode(head, node);
                break;
            }
        }
    }

    /* ensure bidirectional links are consistent */
    private static void fixNodeLinks(Node node)
    {
        if (node.prev != null)
            node.prev.next = node;
        else
            node.parent.content = node;

        if (node.next != null)
            node.next.prev = node;
        else
            node.parent.last = node;
    }

    /*
     used to strip child of node when
     the node has one and only one child
    */
    private static void stripOnlyChild(Node node)
    {
        Node child;

        child = node.content;
        node.content = child.content;
        node.last = child.last;
        child.content = null;

        for (child = node.content; child != null; child = child.next)
            child.parent = node;
    }

    /* used to strip font start and end tags */
    private static void discardContainer(Node element, MutableObject pnode)
    {
        Node node;
        Node parent = element.parent;

        if (element.content != null)
        {
            element.last.next = element.next;

            if (element.next != null)
            {
                element.next.prev = element.last;
                element.last.next = element.next;
            }
            else
                parent.last = element.last;

            if (element.prev != null)
            {
                element.content.prev = element.prev;
                element.prev.next = element.content;
            }
            else
                parent.content = element.content;

            for (node = element.content; node != null; node = node.next)
                node.parent = parent;

            pnode.setObject(element.content);
        }
        else
        {
            if (element.next != null)
                element.next.prev = element.prev;
            else
                parent.last = element.prev;

            if (element.prev != null)
                element.prev.next = element.next;
            else
                parent.content = element.next;

            pnode.setObject(element.next);
        }

        element.next = null;
        element.content = null;
    }

    /*
     Add style property to element, creating style
     attribute as needed and adding ; delimiter
    */
    private static void addStyleProperty(Node node, String property)
    {
        AttVal av;

        for (av = node.attributes; av != null; av = av.next)
        {
            if (av.attribute.equals("style"))
                break;
        }

        /* if style attribute already exists then insert property */

        if (av != null)
        {
            String s;

            s = addProperty(av.value, property);
            av.value = s;
        }
        else /* else create new style attribute */
        {
            av = new AttVal(node.attributes, null, '"',
                            new String("style"),
                            new String(property));
            av.dict = AttributeTable.getDefaultAttributeTable().findAttribute(av);
            node.attributes = av;
        }
    }

    /*
      Create new string that consists of the
      combined style properties in s1 and s2

      To merge property lists, we build a linked
      list of property/values and insert properties
      into the list in order, merging values for
      the same property name.
    */
    private static String mergeProperties(String s1, String s2)
    {
        String s;
        StyleProp prop;

        prop = createProps(null, s1);
        prop = createProps(prop, s2);
        s = createPropString(prop);
        return s;
    }

    private static void mergeStyles(Node node, Node child)
    {
        AttVal av;
        String s1, s2, style;

        for (s2 = null, av = child.attributes; av != null; av = av.next)
        {
            if (av.attribute.equals("style"))
            {
                s2 = av.value;
                break;
            }
        }

        for (s1 = null, av = node.attributes; av != null; av = av.next)
        {
            if (av.attribute.equals("style"))
            {
                s1 = av.value;
                break;
            }
        }

        if (s1 != null)
        {
            if (s2 != null)  /* merge styles from both */
            {
                style = mergeProperties(s1, s2);
                av.value = style;
            }
        }
        else if (s2 != null)  /* copy style of child */
        {
            av = new AttVal(node.attributes, null, '"',
                            new String("style"),
                            new String(s2));
            av.dict = AttributeTable.getDefaultAttributeTable().findAttribute(av);
            node.attributes = av;
        }
    }

    private static String fontSize2Name(String size)
    {
        String[] sizes =
          {
            new String("50%"),
            new String("60%"),
            new String("80%"),
            null,
            new String("120%"),
            new String("150%"),
            new String("200%")
          };

        if ('0' <= size.charAt(0) && size.charAt(0) <= '6')
        {
            int n = size.charAt(0) - '0';
            return sizes[n];
        }

        if (size.charAt(0) == '-')
            return "smaller";

        return "larger";
    }

    private static void addFontFace(Node node, String face)
    {
        addStyleProperty(node, "font-family: " + face);
    }

    private static void addFontSize(Node node, String size)
    {
        String value;

        if (size.equals("6") && node.tag == TagTable.tagP)
        {
            node.element = new String("h1");
            TagTable.getDefaultTagTable().findTag(node);
            return;
        }

        if (size.equals("5") && node.tag == TagTable.tagP)
        {
            node.element = new String("h2");
            TagTable.getDefaultTagTable().findTag(node);
            return;
        }

        if (size.equals("4") && node.tag == TagTable.tagP)
        {
            node.element = new String("h3");
            TagTable.getDefaultTagTable().findTag(node);
            return;
        }

        value = fontSize2Name(size);

        if (value != null)
        {
            addStyleProperty(node, "font-size: " + value);
        }
    }

    private static void addFontColor(Node node, String color)
    {
        addStyleProperty(node, "color: " + color);
    }

    private static void addAlign(Node node, String align)
    {
        /* force alignment value to lower case */
        addStyleProperty(node, "text-align: " + align.toLowerCase());
    }

    /*
     add style properties to node corresponding to
     the font face, size and color attributes
    */
    private static void addFontStyles(Node node, AttVal av)
    {
        while (av != null)
        {
            if (av.attribute.equals("face"))
                addFontFace(node, av.value);
            else if (av.attribute.equals("size"))
                addFontSize(node, av.value);
            else if (av.attribute.equals("color"))
                addFontColor(node, av.value);

            av = av.next;
        }
    }

    /*
        Symptom: <p align=center>
        Action: <p style="text-align: center">
    */
    private static void textAlign(Lexer lexer, Node node)
    {
        AttVal av, prev;

        prev = null;

        for (av = node.attributes; av != null; av = av.next)
        {
            if (av.attribute.equals("align"))
            {
                if (prev != null)
                    prev.next = av.next;
                else
                    node.attributes = null;

                if (av.value != null)
                {
                    addAlign(node, av.value);
                }

                break;
            }

            prev = av;
        }
    }

    /*
       The clean up rules use the pnode argument to return the
       next node when the orignal node has been deleted
    */

    /*
        Symptom: <dir> <li> where <li> is only child
        Action: coerce <dir> <li> to <div> with indent.
    */

    private static boolean dir2Div(Lexer lexer, Node node, MutableObject pnode)
    {
        Node child;

        if (node.tag == TagTable.tagDir ||
            node.tag == TagTable.tagUl ||
            node.tag == TagTable.tagOl)
        {
            child = node.content;

            if (child == null)
                return false;

            /* check child has no peers */

            if (child.next != null)
                return false;

            if (child.tag != TagTable.tagLi)
                return false;

            if (!child.implicit)
                return false;

            /* coerce dir to div */

            node.tag = TagTable.tagDiv;
            node.element = new String("div");
            addStyleProperty(node, "margin-left: 2em");
            stripOnlyChild(node);
            return true;

//#if 0
            //Node content;
            //Node last;
            //content = child.content;
            //last = child.last;
            //child.content = null;

            /* adjust parent and set margin on contents of <li> */

            //for (child = content; child != null; child = child.next)
            //{
            //    child.parent = node.parent;
            //    addStyleProperty(child, "margin-left: 1em");
            //}

            /* hook first/last into sequence */

            //if (content != null)
            //{
            //    content.prev = node.prev;
            //    last.next = node.next;
            //    fixNodeLinks(content);
            //    fixNodeLinks(last);
            //}

            //node.next = null;

            /* ensure that new node is cleaned */
            //pnode.setObject(cleanNode(lexer, content));
            //return true;
//#endif
        }

        return false;
    }

    /*
        Symptom: implicit blockquote due to <ul>some text</ul>
        Action: replace by <div> with indent style
    */
    private static boolean blockquote2Div(Lexer lexer, Node node, MutableObject pnode)
    {
        Node child;

        if (node.tag == TagTable.tagBlockquote && node.implicit)
        {
            child = node.content;

            if (child == null)
                return false;

            /* coerce blockquote to div */

            node.tag = TagTable.tagDiv;
            node.element = new String("div");
            addStyleProperty(node, "margin-left: 2em");
            return true;
        }

        return false;
    }


    /*
        Symptom: <center>
        Action: replace <center> by <div style="text-align: center">
    */

    private static boolean center2Div(Lexer lexer, Node node, MutableObject pnode)
    {
        if (node.tag == TagTable.tagCenter)
        {
            if (lexer.configuration.DropFontTags)
            {
                if (node.content != null)
                {
                    Node last = node.last;
                    Node parent = node.parent;

                    discardContainer(node, pnode);

                    node = lexer.inferredTag("br");

                    if (last.next != null)
                        last.next.prev = node;

                    node.next = last.next;
                    last.next = node;
                    node.prev = last;

                    parent.last = last;
                    if (parent.last != null)
                        parent.last = node;

                    node.parent = parent;
                }
                else
                {
                    Node prev = node.prev;
                    Node next = node.next;
                    Node parent = node.parent;
                    discardContainer(node, pnode);

                    node = lexer.inferredTag("br");
                    node.next = next;
                    node.prev = prev;
                    node.parent = parent;

                    if (next != null)
                        next.prev = node;
                    else
                        parent.last = node;

                    if (prev != null)
                        prev.next = node;
                    else
                        parent.content = node;
                }

                return true;
            }
            node.tag = TagTable.tagDiv;
            node.element = new String("div");
            addStyleProperty(node, "text-align: center");
            return true;
        }

        return false;
    }

    /*
        Symptom <div><div>...</div></div>
        Action: merge the two divs

      This is useful after nested <dir>s used by Word
      for indenting have been converted to <div>s
    */
    private static boolean mergeDivs(Lexer lexer, Node node, MutableObject pnode)
    {
        Node child;

        if (node.tag != TagTable.tagDiv)
            return false;

        child = node.content;

        if (child == null)
            return false;

        if (child.tag != TagTable.tagDiv)
            return false;

        if (child.next != null)
            return false;

        mergeStyles(node, child);
        stripOnlyChild(node);
        return true;
    }

    /*
        Symptom: <ul><li><ul>...</ul></li></ul>
        Action: discard outer list
    */

    private static boolean nestedList(Lexer lexer, Node node, MutableObject pnode)
    {
        Node child, list;

        if (node.tag == TagTable.tagUl || node.tag == TagTable.tagOl)
        {
            child = node.content;

            if (child == null)
                return false;

            /* check child has no peers */

            if (child.next != null)
                return false;

            list = child.content;

            if (list == null)
                return false;

            if (list.tag != node.tag)
                return false;

            pnode.setObject(node.next);

            /* move inner list node into position of outer node */
            list.prev = node.prev;
            list.next = node.next;
            list.parent = node.parent;
            fixNodeLinks(list);

            /* get rid of outer ul and its li */
            child.content = null;
            node.content = null;
            node.next = null;

            /*
              If prev node was a list the chances are this node
              should be appended to that list. Word has no way of
              recognizing nested lists and just uses indents
            */

            if (list.prev != null)
            {
                node = list;
                list = node.prev;

                if (list.tag == TagTable.tagUl || list.tag == TagTable.tagOl)
                {
                    list.next = node.next;

                    if (list.next != null)
                        list.next.prev = list;

                    child = list.last;  /* <li> */

                    node.parent = child;
                    node.next = null;
                    node.prev = child.last;
                    fixNodeLinks(node);
                }
            }

            cleanNode(lexer, node);
            return true;
        }

        return false;
    }

    /*
        Symptom: the only child of a block-level element is a
        presentation element such as B, I or FONT

        Action: add style "font-weight: bold" to the block and
        strip the <b> element, leaving its children.

      example:

        <p>
          <b><font face="Arial" size="6">Draft Recommended Practice</font></b>
        </p>

      becomes:

          <p style="font-weight: bold; font-family: Arial; font-size: 6">
            Draft Recommended Practice
          </p>
    */
    private static boolean blockStyle(Lexer lexer, Node node, MutableObject pnode)
    {
        Node child;

        if ((node.tag.model & (Dict.CM_BLOCK | Dict.CM_LIST | Dict.CM_DEFLIST | Dict.CM_TABLE)) != 0)
        {
            /* check for align attribute */
            if (node.tag != TagTable.tagCaption)
                textAlign(lexer, node);

            child = node.content;

            if (child == null)
                return false;

            /* check child has no peers */

            if (child.next != null)
                return false;

            if (child.tag == TagTable.tagB)
            {
                addStyleProperty(node, "font-weight: bold");
                stripOnlyChild(node);
                return true;
            }

            if (child.tag == TagTable.tagI)
            {
                addStyleProperty(node, "font-style: italic");
                stripOnlyChild(node);
                return true;
            }

            if (child.tag == TagTable.tagFont)
            {
                addFontStyles(node, child.attributes);
                stripOnlyChild(node);
                return true;
            }
        }

        return false;
    }

    /* the only child of an inline element such as em */
    private static boolean inlineStyle(Lexer lexer, Node node, MutableObject pnode)
    {
        Node child;

        if (node.tag != TagTable.tagFont && (node.tag.model & Dict.CM_INLINE) != 0)
        {
            child = node.content;

            if (child == null)
                return false;

            /* check child has no peers */

            if (child.next != null)
                return false;

            if (child.tag == TagTable.tagB)
            {
                addStyleProperty(node, "font-weight: bold");
                stripOnlyChild(node);
                return true;
            }

            if (child.tag == TagTable.tagI)
            {
                addStyleProperty(node, "font-style: italic");
                stripOnlyChild(node);
                return true;
            }

            if (child.tag == TagTable.tagFont)
            {
                addFontStyles(node, child.attributes);
                stripOnlyChild(node);
                return true;
            }
        }

        return false;
    }

    /*
      Replace font elements by span elements, deleting
      the font element's attributes and replacing them
      by a single style attribute.
    */
    private static boolean font2Span(Lexer lexer, Node node, MutableObject pnode)
    {
        AttVal av, style, next;

        if (node.tag == TagTable.tagFont)
        {
            if (lexer.configuration.DropFontTags)
            {
                discardContainer(node, pnode);
                return false;
            }

            /* if FONT is only child of parent element then leave alone */
            if (node.parent.content == node
                && node.next == null)
                return false;

            addFontStyles(node, node.attributes);

            av = node.attributes;
            style = null;

            while (av != null)
            {
                next = av.next;

                if (av.attribute.equals("style"))
                {
                    av.next = null;
                    style = av;
                }

                av = next;
            }

            node.attributes = style;

            node.tag = TagTable.tagSpan;
            node.element = new String("span");

            return true;
        }

        return false;
    }

    /*
      Applies all matching rules to a node.
    */
    private static Node cleanNode(Lexer lexer, Node node)
    {
        Node next = null;
        MutableObject o = new MutableObject();
        boolean b = false;

        for (next = node; node.isElement(); node = next)
        {
            o.setObject(next);

            b = dir2Div(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            b = blockquote2Div(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            b = nestedList(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            b = center2Div(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            b = mergeDivs(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            b = blockStyle(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            b = inlineStyle(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            b = font2Span(lexer, node, o);
            next = (Node)o.getObject();
            if (b)
                continue;

            break;
        }

        return next;
    }

    private static Node createStyleProperties(Lexer lexer, Node node)
    {
        Node child;

        if (node.content != null)
        {
            for (child = node.content; child != null; child = child.next)
            {
                child = createStyleProperties(lexer, child);
            }
        }

        return cleanNode(lexer, node);
    }

    private void defineStyleRules(Lexer lexer, Node node)
    {
        Node child;

        if (node.content != null)
        {
            for (child = node.content;
                    child != null; child = child.next)
            {
                defineStyleRules(lexer, child);
            }
        }

        style2Rule(lexer, node);
    }

    public void cleanTree(Lexer lexer, Node doc)
    {
        doc = createStyleProperties(lexer, doc);

        if (!lexer.configuration.DropFontTags)
        {
            defineStyleRules(lexer, doc);
            createStyleElement(lexer, doc);
        }
    }

    /* replace i by em and b by strong */
    public static void emFromI(Node node)
    {
        while (node != null)
        {
            if (node.tag == TagTable.tagI)
            {
                node.element = new String(TagTable.tagEm.name);
                node.tag = TagTable.tagEm;
            }
            else if (node.tag == TagTable.tagB)
            {
                node.element = new String(TagTable.tagStrong.name);
                node.tag = TagTable.tagStrong;
            }

            if (node.content != null)
                emFromI(node.content);

            node = node.next;
        }
    }

}
