/*
 * @(#)Node.java   1.3 1999/07/30
 *
 */

package org.w3c.tidy;

/**
 *
 * Node
 *
 * (c) 1998, 1999 (W3C) MIT, INRIA, Keio University
 * See Tidy.java for the copyright notice.
 * Derived from <a href="http://www.w3.org/People/Raggett/tidy">
 * HTML Tidy Release 26 Jul 1999</a>
 *
 * @author  Dave Raggett <dsr@w3.org>
 * @author  Andy Quick <ac.quick@sympatico.ca> (translation to Java)
 * @version 1.0, 1999/05/22
 * @version 1.0.1, 1999/05/29
 * @version 1.1, 1999/06/18 Java Bean
 * @version 1.2, 1999/07/10 Tidy Release 7 Jul 1999
 * @version 1.3, 1999/07/30 Tidy Release 26 Jul 1999
 */

/*
  Used for elements and text nodes
  element name is null for text nodes
  start and end are offsets into lexbuf
  which contains the textual content of
  all elements in the parse tree.

  parent and content allow traversal
  of the parse tree in any direction.
  attributes are represented as a linked
  list of AttVal nodes which hold the
  strings for attribute/value pairs.
*/

public class Node {

    public static final short RootNode        = 0;
    public static final short DocTypeTag      = 1;
    public static final short CommentTag      = 2;
    public static final short ProcInsTag      = 3;
    public static final short TextNode        = 4;
    public static final short StartTag        = 5;
    public static final short EndTag          = 6;
    public static final short StartEndTag     = 7;
    public static final short AspTag          = 8;

    public Node parent;
    public Node prev;
    public Node next;
    public Node last;
    protected int start;             /* start of span onto text array */
    protected int end;               /* end of span onto text array */
    protected byte[] textarray;      /* the text array */
    public short type;              /* TextNode, StartTag, EndTag etc. */
    public boolean implicit;          /* true if inferred */
    public Dict was;   /* old tag when it was changed */
    public Dict tag;   /* tag's dictionary definition */
    public String element;          /* name (null for text nodes) */
    public AttVal attributes;
    public Node content;

    public Node()
    {
        this(TextNode, null, 0, 0);
    }

    public Node(short type, byte[] textarray, int start, int end)
    {
        this.parent = null;
        this.prev = null;
        this.next = null;
        this.last = null;
        this.start = start;
        this.end = end;
        this.textarray = textarray;
        this.type = type;
        this.implicit = false;
        this.was = null;
        this.tag = null;
        this.element = null;
        this.attributes = null;
        this.content = null;
    }

    public Node(short type, byte[] textarray, int start, int end, String element)
    {
        this.parent = null;
        this.prev = null;
        this.next = null;
        this.last = null;
        this.start = start;
        this.end = end;
        this.textarray = textarray;
        this.type = type;
        this.implicit = false;
        this.was = null;
        this.tag = null;
        this.element = element;
        this.attributes = null;
        this.content = null;
        if (type == StartTag || type == StartEndTag || type == EndTag)
            TagTable.getDefaultTagTable().findTag(this);
    }

    /* used to clone heading nodes when split by an <HR> */
    protected Object clone()
    {
        Node node = new Node();

        node.parent = this.parent;
        node.start = 0;
        node.end = 0;
        node.textarray = this.textarray; // not copied
        node.type = this.type;
        node.implicit = this.implicit;
        node.was = this.was;
        node.tag = this.tag;
        if (this.element != null)
            node.element = new String(this.element);
        if (this.attributes != null)
            node.attributes = (AttVal)this.attributes.clone();
        return node;
    }

    public String getNodeValue()
    {
        String value = null;
        if (this.textarray != null)
        {
            if (this.start <= this.end)
            {
                value = Lexer.getString(this.textarray, this.start, this.end - this.start);
            }
        }
        return value;
    }

    /* default method for checking an element's attributes */
    public void checkAttributes( Lexer lexer )
    {
        AttVal attval;

        for (attval = this.attributes; attval != null; attval = attval.next)
            attval.checkAttribute( lexer, this );
    }

    /* find doctype element */
    public Node findDocType()
    {
        Node node;

        for (node = this.content; 
            node != null && node.type != DocTypeTag; node = node.next);

        return node;
    }

    public void discardDocType()
    {
        Node node;

        node = findDocType();
        if (node != null)
        {
            if (node.prev != null)
                node.prev.next = node.next;
            else
                node.parent.content = node.next;

            if (node.next != null)
                node.next.prev = node.prev;

            node.next = null;
        }
    }

    public static void discardElement(Node element)
    {
        Node parent;

        parent = element.parent;

        if (parent.last == element)
            parent.last = element.prev;

        if (parent.content == element)
            parent.content = element.next;

        if (element.prev != null)
            element.prev.next = element.next;

        if (element.next != null)
            element.next.prev = element.prev;

        element.next = null;
    }

    public static void insertNode(Node element, Node node)
    {
        node.parent = element;
        node.prev = element.last;

        if (element.last != null)
            element.last.next = node;
        else
            element.content = node;

        element.last = node;
    }

    public static void insertNodeBeforeElement(Node element, Node node)
    {
        Node parent;

        parent = element.parent;
        node.parent = parent;
        node.next = element;
        node.prev = element.prev;
        element.prev = node;

        if (node.prev != null)
            node.prev.next = node;

        if (parent.content == element)
            parent.content = node;
    }

    public static void insertNodeAfterElement(Node element, Node node)
    {
        Node parent;

        parent = element.parent;
        node.parent = parent;

        if (parent.last == element)
            parent.last = node;
        else
            node.next = element.next;

        element.next = node;
        node.prev = element;
    }

    public static void trimEmptyElement(Lexer lexer, Node element)
    {
        if (element.content == null &&
            (element.tag != TagTable.tagA || element.attributes == null))
        {
            if (element.type == TextNode ||
                    (element.tag != null &&
                     element.tag != TagTable.tagLayer &&
                    !((element.tag.model & Dict.CM_ROW) != 0)))
            {

                if (element.type != TextNode)
                    Report.warning(lexer, element, null, Report.TRIM_EMPTY_ELEMENT);

                discardElement(element);
            }
        }
    }

    /*
      If last child of element is a text node
      then trim trailing white space character.
    */
    public static void trimSpace(Lexer lexer, Node last)
    {
        byte c;

        if (last != null && last.type == TextNode && last.end > last.start)
        {
            while (last.end > last.start)
            {
                c = lexer.lexbuf[last.end - 1];

                if (c == 160)  /* non breaking space */
                {
                    if (last.parent.tag == TagTable.tagTd ||
                        last.parent.tag == TagTable.tagTh)
                    {
                        if (last.end > last.start + 1)
                            last.end -= 1;
                        else
                            break;
                    }
                    else
                        last.end -= 1;
                }
                else if (c == ' ')
                    last.end -= 1;
                else
                    break;
            }
 
            if (last.end < last.start)
                Report.tidyPrintln(lexer.errout, "TrimSpace: screwed up text node");

            /* if empty string then delete from parse tree */
            if (last.start == last.end)
                trimEmptyElement(lexer, last);
        }
    }

    /*
      This maps 
           <em>hello </em><strong>world</strong>
      to
           <em>hello</em> <strong>world</strong>

      If last child of element is a text node
      then trim trailing white space character
      moving it to after element's end tag.
    */
    public static void trimTrailingSpace(Lexer lexer, Node last)
    {
        byte c;

        if (last != null && last.type == TextNode && last.end > last.start)
        {
            c = lexer.lexbuf[last.end - 1];

            if (c == ' ' || c == 160)
            {
                last.end -= 1;

                if ((last.parent.tag.model & Dict.CM_INLINE) != 0)
                    lexer.insertspace = true;
            }

            /* if empty string then delete from parse tree */
            if (last.start == last.end)
                trimEmptyElement(lexer, last);
        }
    }

    /*
      This maps 
           <p>hello<em> world</em>
      to
           <p>hello <em>world</em>

      Trims initial space, by moving it before the
      start tag, or if this element is the first in
      parent's content, then by discarding the space
    */
    public static void trimInitialSpace(Lexer lexer, Node element, Node text)
    {
        Node prev, node;

        if (text.type == TextNode && lexer.lexbuf[text.start] == ' ')
        {
            if (((element.tag.model & Dict.CM_INLINE) != 0) &&
                element.parent.content != element)
            {
                prev = element.prev;

                if (prev.type == TextNode)
                {
                    if (lexer.lexbuf[prev.end - 1] != (byte)' ')
                        lexer.lexbuf[prev.end++] = (byte)' ';

                    ++element.start;
                }
                else /* create new node */
                {
                    node = new Node();
                    node.start = element.start++;
                    node.end = element.start;
                    lexer.lexbuf[node.start] = (byte)' ';
                    node.prev = prev;
                    prev.next = node;
                    node.next = element;
                    element.prev = node;
                    node.parent = element.parent;
                }
            }

            /* discard the space  in current node */
            ++text.start;
        }
    }

    public boolean isDescendantOf(Dict tag)
    {
        Node parent;

        for (parent = this.parent;
                parent != null; parent = parent.parent)
        {
            if (parent.tag == tag)
                return true;
        }

        return false;
    }

    /*
     the doctype has been found after other tags,
     and needs moving to before the html element
    */
    public static void insertDocType(Lexer lexer, Node element, Node doctype)
    {
        Report.warning(lexer, element, doctype, Report.DOCTYPE_AFTER_TAGS);

        while (element.tag != TagTable.tagHtml)
            element = element.parent;

        insertNodeBeforeElement(element, doctype);
    }

    public static Node findBody(Node root)
    {
        Node node;

        node = root.content;

        while (node != null && node.tag != TagTable.tagHtml)
            node = node.next;

        if (node == null)
            return null;

        node = node.content;

        while (node != null && node.tag != TagTable.tagBody)
            node = node.next;

        return node;
    }

    public boolean isElement()
    {
        return (this.type == StartTag || this.type == StartEndTag ? true : false);
    }

    /*
     unexpected content in table row is moved to just before
     the table in accordance with Netscape and IE. This code
     assumes that node hasn't been inserted into the row.
    */
    public static void moveBeforeTable(Node row, Node node)
    {
        Node table;

        /* first find the table element */
        for (table = row.parent; table != null; table = table.parent)
        {
            if (table.tag == TagTable.tagTable)
            {
                node.prev = table.prev;
                node.next = table;
                table.prev = node;
                node.parent = table.parent;
        
                if (node.prev != null)
                    node.prev.next = node;

                break;
            }
        }
    }

}
