/*
 * @(#)StreamInImpl.java   1.3 1999/07/30
 *
 */

package org.w3c.tidy;

/**
 *
 * Input Stream Implementation
 *
 * (c) 1998, 1999 (W3C) MIT, INRIA, Keio University
 * Derived from <a href="http://www.w3.org/People/Raggett/tidy">
 * HTML Tidy Release 26 Jul 1999</a>
 *
 * @author  Dave Raggett <dsr@w3.org>
 * @author  Andy Quick <ac.quick@sympatico.ca> (translation to Java)
 * @version 1.0, 1999/05/22
 * @version 1.0.1, 1999/05/29
 * @version 1.1, 1999/06/18 Java Bean
 * @version 1.2, 1999/07/10 Tidy Release 7 Jul 1999
 * @version 1.3, 1999/07/30 Tidy Release 26 Jul 1999
 */

import java.io.InputStream;
import java.io.IOException;

public class StreamInImpl extends StreamIn {

    public StreamInImpl(InputStream stream, int encoding, int tabsize)
    {
        this.stream = stream;
        this.pushed = false;
        this.c = (int)'\0';
        this.tabs = 0;
        this.tabsize = tabsize;
        this.curline = 1;
        this.curcol = 1;
        this.encoding = encoding;
        this.state = FSM_ASCII;
        this.endOfStream = false;
    }

    /* read char from stream */
    public int readCharFromStream()
    {
        int n, c, i, count;

        try {
            c = this.stream.read();

            if (c == EndOfStream) {
                this.endOfStream = true;
                return c;
            }

            /*
               A document in ISO-2022 based encoding uses some ESC sequences
               called "designator" to switch character sets. The designators
               defined and used in ISO-2022-JP are:

                "ESC" + "(" + ?     for ISO646 variants

                "ESC" + "$" + ?     and
                "ESC" + "$" + "(" + ?   for multibyte character sets

               Where ? stands for a single character used to indicate the
               character set for multibyte characters.

               Tidy handles this by preserving the escape sequence and
               setting the top bit of each byte for non-ascii chars. This
               bit is then cleared on output. The input stream keeps track
               of the state to determine when to set/clear the bit.
            */

            if (this.encoding == Configuration.ISO2022)
            {
                if (c == 0x1b)  /* ESC */
                {
                    this.state = FSM_ESC;
                    return c;
                }

                switch (this.state)
                {
                case FSM_ESC:
                    if (c == '$')
                        this.state = FSM_ESCD;
                    else if (c == '(')
                        this.state = FSM_ESCP;
                    else
                        this.state = FSM_ASCII;
                    break;

                case FSM_ESCD:
                    if (c == '(')
                        this.state = FSM_ESCDP;
                    else
                        this.state = FSM_NONASCII;
                    break;

                case FSM_ESCDP:
                    this.state = FSM_NONASCII;
                    break;

                case FSM_ESCP:
                    this.state = FSM_ASCII;
                    break;

                case FSM_NONASCII:
                    c |= 0x80;
                    break;
                }

                return c;
            }

            if (this.encoding != Configuration.UTF8)
                return c;

            /* deal with UTF-8 encoded char */

            if ((c & 0xE0) == 0xC0)  /* 110X XXXX  two bytes */
            {
                n = c & 31;
                count = 1;
            }
            else if ((c & 0xF0) == 0xE0)  /* 1110 XXXX  three bytes */
            {
                n = c & 15;
                count = 2;
            }
            else if ((c & 0xF8) == 0xF0)  /* 1111 0XXX  four bytes */
            {
                n = c & 7;
                count = 3;
            }
            else if ((c & 0xFC) == 0xF8)  /* 1111 10XX  five bytes */
            {
                n = c & 3;
                count = 4;
            }
            else if ((c & 0xFE) == 0xFC)       /* 1111 110X  six bytes */
            {
                n = c & 1;
                count = 5;
            }
            else  /* 0XXX XXXX one byte */
                return c;

            /* successor bytes should have the form 10XX XXXX */
            for (i = 1; i <= count; ++i)
            {
                c = this.stream.read();

                if (c == EndOfStream) {
                    this.endOfStream = true;
                    return c;
                }

                n = (n << 6) | (c & 0x3F);
            }
        }
        catch (IOException e) {
            System.err.println("StreamInImpl.readCharFromStream: " + e.toString());
            n = EndOfStream;
        }

        return n;
    }

    public int readChar()
    {
        int c;

        if (this.pushed)
        {
            this.pushed = false;
            c =  this.c;

            if (c == '\n')
            {
                this.curcol = 1;
                this.curline++;
                return c;
            }

            this.curcol++;
            return c;
        }

        this.lastcol = this.curcol;

        if (this.tabs > 0)
        {
            this.curcol++;
            this.tabs--;
            return ' ';
        }
    
        for (;;)
        {
            c = readCharFromStream();

            if (c < 0)
                return EndOfStream;

            if (c == '\n')
            {
                this.curcol = 1;
                this.curline++;
                break;
            }

            if (c == '\t')
            {
                this.tabs = this.tabsize - ((this.curcol - 1) % this.tabsize) - 1;
                this.curcol++;
                c = ' ';
                break;
            }

            /* strip control characters, except for Esc */

            if (c == '\033')
                break;

            if (0 < c && c < 32)
                continue;

            /* watch out for IS02022 */

            if (this.encoding == Configuration.RAW ||
                this.encoding == Configuration.ISO2022)
            {
                this.curcol++;
                break;
            }

            /* produced e.g. as a side-effect of smart quotes in Word */

            if (127 < c && c < 160)
            {
                Report.encodingError((Lexer)this.lexer, Report.WINDOWS_CHARS, c);

                if (c == 150)
                    c = 8211;  /* en dash */
                else if (c == 151)
                    c = 8212;  /* em dash */
                else if (c == 138)
                    c = 352;  /* latin capital letter S with caron */
                else if (c == 154)
                    c = 353;  /* latin small letter s with caron */
                else if (c == 159)
                    c = 376;  /* latin capital letter Y with diaeresis */
                else if (c == 140)
                    c = 338;  /* latin capital ligature OE */
                else if (c == 156)
                    c = 339;  /* latin capital ligature OE */
                else if (c == 153)
                    c = 8482;  /* TM */
                else if (c == 134)
                    c = 8224;  /* dagger */
                else if (c == 135)
                    c = 8225;  /* double dagger */
                else if (c == 137)
                    c = 8240;  /* per mille sign */
                else if (c == 130)
                    c = 8218;  /* single low quotation mark */
                else if (c == 132)
                    c = 8222;  /* double low quotation mark */
                else if (c == 145)
                    c = 8216;  /* single left quotation mark */
                else if (c == 146)
                    c = 8217;  /* single right quotation mark */
                else if (c == 147)
                    c = 8220;  /* double left quotation mark */
                else if (c == 148)
                    c = 8221;  /* double right quotation mark */
                else if (c == 139)
                    c = 8249;  /* single left-pointing angle quotation mark */
                else if (c == 155)
                    c = 8250;  /* single right-pointing angle quotation mark */
                else
                    continue;
            }

            this.curcol++;
            break;
        }

        return c;
    }

    public void ungetChar(int c)
    {
        this.pushed = true;
        this.c = c;

        if (c == '\n')
        {
            --this.curline;
        }

        this.curcol = this.lastcol;
    }

    public boolean isEndOfStream()
    {
        return this.endOfStream;
    }

}
