/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: Admin.java,v 1.36 2000/02/21 23:50:39 markd Exp $
 *-----------------------------------------------------------------------------
 */





package com.lutris.appserver.admin;

import java.util.Hashtable;
import com.lutris.util.*;
import com.lutris.http.BasicAuth;
import com.lutris.http.BasicAuthResult;
import com.lutris.util.Config;
import com.lutris.appserver.server.*;
import com.lutris.appserver.server.httpPresentation.*;
import org.enhydra.servlet.debug.DebugManager;


/**
 * Application class for this specific application.  This implements
 * standard startup and shutdown methods and well as application specific
 * ones. The functionality added to StandardApplication in the
 * implementation of HTTP BasicAuth and the default URL implementation.
 */
public class Admin extends StandardApplication {

    // The user who is allowed to run this application. Used with basic auth.
    private String username, password;
    // For the password prompt. Used with basic auth.
    private String relm = "Multiserver Administration";

    /**
     *  The set of all DebugManager, indexed by servlet id.
     *  There needs to be a centeralized set of them in this Application
     *  object so that there will only be at most one manager per
     *  servlet. If the managers are stored in session data, then there
     *  could be one per logged in user. <BR>
     *  The presentation objects manage the contents of this Hashtable.
     */
    public Hashtable debuggers;

    /**
     *  Has the state of the server been modified? If true, then the
     *  MultiServer's config file needs to be re-written. <BR>
     *  The presentaion objects manage this variable.
     */
    public boolean serverDirty;

    /**
     *  How many requests to buffer in the queue when debugging an
     *  application.
     */
    public int debugQueueSize; 

    /**
     *  When debugging, should we keep a copy of the data written to the net?
     *  This is specified in the admin config file, under the key
     *  "Admin.SaveResponseData".
     */
    public boolean saveResponseData;
 
    /**
     * Start the application.
     * 
     * @param appConfig
     *   Configuration object for this application.
     * @exception ApplicationException
     *   If an error occurs starting the application.
     */
    public void startup(Config appConfig)
	throws ApplicationException {

	/*
         * Mandatory config settings.
         */
        try {
	    username = appConfig.getString("Admin.Username");
	    password = appConfig.getString("Admin.Password");
            debugQueueSize = appConfig.getInt("Admin.DebugQueueSize");
	} catch (ConfigException except) {
	    throw new ApplicationException(except);
	}
        if (debugQueueSize < 1) 
            throw new ApplicationException("Admin.DebugQueueSize must " +
                                           "be at least 1.");
        /*
         * Optional config settings.
         */
        saveResponseData = false;
        try {
            saveResponseData = appConfig.getBoolean("Admin.SaveResponseData");
	} catch (ConfigException except) {
            // It's ok.
        }
        debuggers = new Hashtable();
	serverDirty = false;
        super.startup(appConfig);
    }


    /**
     *  This is called for each request, before it is processed.
     *  Ensure that the user has authenticated with HTTP
     *  Basic-Auth before continuing.
     *
     * @param comms
     *   The http communication object through which all aspect of
     *   a request can be obtained.
     * @return
     *   False to proceed with the request.
     * @exception PageUnauthorizedException
     *   if user does not login in correctly.
     * @exception Exception
     *   if any other error occurs.
     */
    public boolean requestPreprocessor(HttpPresentationComms comms)
            throws Exception {

        super.requestPreprocessor(comms);
        if (comms.session == null) {
            // Session not setup; so its not a presentation object.
            return false;
        }

	/*
	 *  Ensure that the username/password are being sent.
         *  If both the username and password are empty strings, then
         *  do not require a login.
	 */
        BasicAuthResult auth = BasicAuth.getAuthentication(comms.request);
        if (((username.length() != 0) || (password.length() != 0)) &&
             ((auth == null) || 
               !username.equals(auth.username) || 
               !password.equals(auth.password))) {
            /*
             * If they click cancel on the username/password dialog box,
             * then the browser will show the HTML sent along with the
             * unauthorized response code. Normally, the user will enter
             * a username and password, the browser will re-issue the request,
             * and the user will never see this.
             */
            throw new PageUnauthorizedException(relm);
        }
        /*
         * Since they sent the correct username and password, we know they
         * are logged in. If there is no user set in the session, then
         * set it now.
         */
        if (comms.session.getUser() == null)
            comms.session.setUser(AdminUser.sharedUser);
	/*
         *  Proceed normally.
	 */
        return false; 
    }


    /**
     *  Since we are a part of Enhydra, we do not need to print out the
     *  copyright header.
     */
    public void printCopyrightHeader() {
    }


}

