/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DisplayRequests.java,v 1.21 2000/02/21 23:50:40 markd Exp $
 *-----------------------------------------------------------------------------
 */




package com.lutris.appserver.admin.presentation;

import java.net.*;
import java.io.*;
import java.util.Vector;
import com.lutris.appserver.admin.Admin;
import com.lutris.appserver.admin.business.*;
import com.lutris.http.*;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.session.*;
import org.enhydra.servlet.ServletIOException;
import org.enhydra.servlet.debug.*;

/**
 * This is an http presentation object, which also implements the 
 * ServletRecordCallback interface, so it can register its self with the
 * DebugManager. <P>
 * This presentation tracks changes to the queue. In the future it will
 * talk to an applet, telling it to add and remove lines of text (each of
 * which is a URL). For now it just creates a neverending list of
 * anchors in normal HTML. Every time a new transaction is added to the
 * queue in the DebugManager, one more anchor appears, on it's own line.
 *
 * Expected query string inputs:
 *     id           The name of the servlet being debugged.
 *     targetUrl    The presentation object the anchors will point to.
 *     targetFrame  The target (frame name) for the generated anchors.
 *     history      true or false on whether history in queue is displayed.
 * 
 * @author Andrew John
 * @version $Revision: 1.21 $
 */
public class DisplayRequests implements HttpPresentation, 
       ServletRecordCallback {

    private Vector pendingRecords;

    public void run(HttpPresentationComms comms)
            throws IOException, PageRedirectException, Exception {
    //
    // Get the required fields from the query string.
    //
    String targetURL = comms.request.getParameter("targetUrl");
    if (targetURL == null)
        throw new HttpPresentationException("targetUrl not set.");
    String servletID = comms.request.getParameter("id");
    if (servletID == null)
        throw new HttpPresentationException("id (of servlet) not set.");
    String targetFrame = comms.request.getParameter("targetFrame");
    if (targetFrame == null)
        throw new HttpPresentationException("targetFrame not set.");

    //
    // Look up the DebugManager for this servlet.
    //
    if (comms.session == null)
        throw new HttpPresentationException("Null session manager.");
    Admin myApp = (Admin) comms.application;
    DebugManager dbm = (DebugManager) myApp.debuggers.get(servletID);
    if (dbm == null)
        throw new HttpPresentationException("Not debugging the servlet " +
                servletID);
    
    //
    //  Get ready to recieve records.
    //
    pendingRecords = new Vector();

    //
    // Register for callbacks.
    //
    String history = comms.request.getParameter("history");
    HttpPresentationOutputStream out = null;
    dbm.addTransactionCallback(this, history.equals("true"));
    try {
        /*
         * Start the output! This will not stop until the user clicks on
         * the stop button, or they click on one of the anchors.
         */ 
	comms.response.setContentType("text/html");
        out = comms.response.getOutputStream();
        out.println("<HTML>\n<HEAD>\n</HEAD>\n" + 
            "<BODY BACKGROUND=media/BackgroundWide.jpg>\n");
        out.println("<CODE><FONT SIZE=-1><SCRIPT></SCRIPT>\n");
        out.flush();
        synchronized(pendingRecords) {
            while (true) {
                while (!pendingRecords.isEmpty()) {
                    ServletTransactionRecord tr =
                        (ServletTransactionRecord) pendingRecords.firstElement();
                    pendingRecords.removeElement(tr);
                    RequestRecord rr = tr.getRequestRecord();
                    if (rr == null)
                        continue;
                    String status = ""; 
                    ResponseRecord resp = tr.getResponseRecord();
                    if (resp != null) {
                        int sc = resp.getStatusCode();
                        if (sc == -1)
                            sc = 200;
                        if (sc == 200)
                            status = sc + " ";
                        else if ((sc > 200) && (sc <=299))
                            status = "<FONT COLOR=green>" + sc + "</FONT> ";
                        else if ((sc >= 300) && (sc <=399))
                            status = "<FONT COLOR=yellow>" + sc + "</FONT> ";
                        else if ((sc >= 400) && (sc <=499))
                            status = "<FONT COLOR=red>" + sc + "</FONT> ";
                        else 
                             status = "<FONT COLOR=orange>" + sc + "</FONT> ";
                    }
                    int id = tr.getId();
                    String host = rr.getRemoteHost();
                    String url = rr.getPathInfo();
                    String method = rr.getMethod();
                    if (host == null)
                        host = "<I>N/A</I>";
                    else if (host.indexOf(".") > 0)
                        host = host.substring(0, host.indexOf("."));
                    if (host.length() > 9)
                        host = host.substring(0, 6) + "...";
                    if (url == null)
                        url = "&quot;&quot;";
                    else if (url.length() > 16)
                        url = "..." + url.substring(url.length() - 13);
                    if (method == null)
                        method = "<I>N/A</I>";
                    out.println(
                        status + host + " <A HREF=" +
                        targetURL +
                        "?id=" + servletID +
                        "&transaction=" + id + 
                        " TARGET=" + targetFrame + ">" +
                        method + "</A> "  + 
                        url + "<BR>\n");
                    out.println("<SCRIPT>window.scroll(0,32000)</SCRIPT>\n");
                    out.flush();
                }

                /*
                 * If the browser broke off the connection, this should
                 * force a network io exception.
                 */
                comms.request.getInputStream().available();

                /*
                 *  We processed all the pending records. Now wait for more.
                 */ 
                pendingRecords.wait(2000);
            }
        }
    } catch (ServletIOException e) {
        /*
         * This is expected, when the user hits "stop" in the browser.
         * Ignore this exception. All others will pass though and get 
         * caught by the standard mechanisms.
         */
    } finally {
        /*
         * No matter what, do this clean up code.
         */
        dbm.removeTranactionCallback(this);
        out.println("</CODE></FONT></BODY>\n</HTML>\n");
        out.flush();
    }
}


    /*
        For now, just add the new one. Ignore deleting elements until
        we are talking to an applet.
    */
    public void transactionRecordAddEvent(
        ServletTransactionRecord newRecord,
        ServletTransactionRecord deletedRecord) {
        synchronized(pendingRecords) {
            if (newRecord == null)
                return;
            pendingRecords.addElement(newRecord);
            pendingRecords.notifyAll();
        }
    }


    public void clearRecordsEvent() {
    }

}


