/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: LutrisAppTree.java,v 1.34 2000/03/08 02:02:28 markd Exp $
 *-----------------------------------------------------------------------------
 */


package com.lutris.devsys.newApp;

import java.io.*;
import com.lutris.util.*;
import com.lutris.multiServer.MultiServer;
import org.enhydra.servlet.ServletContainer;

/**
 * LutrisAppTree is a convenience mechanism for creating a new
 * Enhydra application development tree.
 * It can be used as a standalone application or by instantiating
 * an instance, calling the various set methods, and then calling
 * the create method.
 * The default config file is newapp.conf.
 * Uses '/' instead of File.separator since the make environment expects it.
 *
 * @author Shawn McMurdo
 */
public class LutrisAppTree {
    // === Data ===

    protected Config config = null;
    protected static final String defaultConfFile = "newapp.conf";
    protected String rootDir = null;
    protected String appName = null;      // As entered by user
    protected String appNameLC = null;    // Lowercase for package names etc..
    protected String appNameUC = null;    // Uppercase for class names etc...
    protected String packageDir = null;
    protected String packageName = null;
    protected String copyrightName = "Lutris Technologies, Inc.";

    protected String enhydraDir = "/usr/local/enhydra";
    protected String javaHome = "/usr/local/jdk";
    protected String javaClassPath = javaHome + "/lib/classes.zip:.";
    protected String FS = "/";
    protected String PS = ":";

    // Argument Names
    public static final String JDDI_OPTION = "jddi";
    public static final String SILENT_OPTION = "silent";

    public static final String APP_NAME = "appName";
    public static final String ROOT_DIR = "rootDir";
    public static final String PACKAGE_DIR = "packageDir";
    public static final String COPYRIGHT_NAME = "copyrightName";
    public static final String ENHYDRA_DIR = "enhydraDir";
    public static final String CONF_FILE = "confFile";
    public static final String MULTISERVER_CONF = "multiserver.conf";
    public static final String MULTISERVER_CONF_IN = "multiserver.conf.in";
    public static final String START_SCRIPT = "start";
    public static final String DASH = "-";

    // File Extensions
    public static final String EXT_SH = ".sh";
    public static final String EXT_CONF = ".conf";
    public static final String EXT_JAVA = ".java";
    public static final String EXT_JHTML = ".jhtml";

    // Directories
    public static final String PRESENTATION = "presentation";
    public static final String BUSINESS = "business";
    public static final String DATA = "data";
    public static final String MEDIA = "media";

    // File Names
    public static final String MAKEFILE = "Makefile";
    public static final String CONFIG_MK = "config.mk";

    // Classes
    public static final String SERVLET_CLASS =
	    "com.lutris.appserver.server.httpPresentation.servlet." +
	    "HttpPresentationServlet";

    // Names
    public static final String ADMIN_ID = "LutrisAdmin";

    // generate JDDI or not
    public static boolean jddi;

    // echo output or not
    public static boolean silent;

    // === Constructors ===

    /**
     * Constructor for use as an instantiated utility class.
     * You must set the app name using setAppName()
     * @see #setAppName
     */
    public LutrisAppTree() {
	setRootDir(".");
    }

    /**
     * Constructor for use as an instantiated utility class.
     * @parameter appName The application name.
     */
    public LutrisAppTree(String appName) {
	setAppName(appName);
	setRootDir(".");
    }

    /**
     * Print out a fatal error and exit.
     */
    private static void fatal(String msg) {
        System.err.println("Error: " + msg);
        System.exit(1);
    }

    /**
     * Print out a fatal error with exception and exit.
     */
    private static void fatal(String msg,
                              Throwable except) {
        System.err.print("Error");
        if (msg != null) {
            System.err.print(": " + msg);
        }
        if (except != null) {
            System.err.print(": " + except.getClass().getName()
                             + ": " + except.getMessage());
        }
        System.err.println();
        System.exit(1);
    }

    /**
     * Constructor for use by main() as a standalone utility or
     * as a way of passing configuration arguments in a single
     * KeywordValueTable.
     * @param argTable The keyword value table of config args.
     */
    public LutrisAppTree(KeywordValueTable argTable) {
	//  This should be changed to take a Config object instead.
	String tmpval = null;

	// appName
	try {
	    tmpval = argTable.getString(APP_NAME);
	} catch (KeywordValueException e) {}
	if (tmpval != null) {
	    setAppName(tmpval);
	}

	// rootDir
	tmpval = null;
	try {
	    tmpval = argTable.getString(ROOT_DIR);
	} catch (KeywordValueException e) {}
	if (tmpval != null) {
	    setRootDir(tmpval);
	}
	if (this.rootDir == null) {
	    setRootDir(".");
	}

	// packageDir
	tmpval = null;
	try {
	    tmpval = argTable.getString(PACKAGE_DIR);
	} catch (KeywordValueException e) {}
	if (tmpval != null) {
	    setPackageDir(tmpval);
	}

	// copyrightName
	tmpval = null;
	try {
	    tmpval = argTable.getString(COPYRIGHT_NAME);
	} catch (KeywordValueException e) {}
	if (tmpval != null) {
	    setCopyrightName(tmpval);
	}

	// enhydraDir
	tmpval = null;
	try {
	    tmpval = argTable.getString(ENHYDRA_DIR);
	} catch (KeywordValueException e) {}
	if (tmpval != null) {
	    setEnhydraDir(tmpval);
	}
    }

    // === Main ===

    /**
     * The entry point for use as a standalone utility.
     * The utility will process command line arguments for data
     * that is required to create the application. Optionaly the data
     * can be stored in a config file (The default file name is newapp.conf)
     * in the form:
     *
     * # comments
     * rootDir = /applications/root
     * enhydraDir = ??
     * ... other data
     *
     * @param argv An array of input parameters.
     */
    public static void main(String[] argv) {
        if (argv.length == 0) {
	    usage();
        }

        KeywordValueTable argTable = null;
        File dcf = new File(defaultConfFile);
        if (dcf.exists()) {
            try {
                FileInputStream confFIS = new FileInputStream(dcf);
                ConfigFile configFile = new ConfigFile(confFIS);
                argTable = configFile.getConfig();
            } catch (Exception e) {
                fatal("reading config file \"" + dcf + "\"", e);
            }
        } else {
            argTable = new KeywordValueTable();
        }
	parseArgs(argv, argTable);
	LutrisAppTree latree = new LutrisAppTree(argTable);
	try {
	    latree.create();
	} catch(NewAppException e) {
	    usage();
	}
    }

    // === Methods ===

    private static void usage() {
	System.err.println("Lutris App Tree usage:");
	System.err.println(
		"java com.lutris.devsys.newApp.LutrisAppTree [args] appname");
	System.err.println("where args can be any of:");
        System.err.println("-jddi");
	System.err.println("-confFile filename");
	System.err.println("-rootDir dirname");
	System.err.println("-packageDir dirname");
	System.err.println("-copyrightName name");
	System.err.println("-enhydraDir dirname");
        System.exit(1);
    }

    /**
     * parseArgs sets values in the KeywordValueTable for elements of the
     * string array.  It also reads any specified config files into the
     * KeywordValueTable.
     * @param argv The command line arguments.
     * @param argTable The KeywordValueTable in which to put values.
     */
    private static void parseArgs(String[] argv, KeywordValueTable argTable) {
	// FIX: this should probably use a generic arg parsing utility class
	for (int i = 0; i < argv.length - 1; i++) {
   	    if (argv[i].equals(DASH + JDDI_OPTION)) {
              jddi = true;
            } else if (argv[i].equals(DASH + SILENT_OPTION)) {
              silent = true;
            } else if (argv[i].equals(DASH + CONF_FILE)) {
		try {
                    FileInputStream confFIS = new FileInputStream(argv[++i]);
                    ConfigFile configFile = new ConfigFile(confFIS);
                    argTable = configFile.getConfig();
		} catch (Exception e) {}
	    } else if (argv[i].equals(DASH + ROOT_DIR)) {
		try {
		    argTable.set(ROOT_DIR, argv[++i]);
		} catch (Exception e) {}
	    } else if (argv[i].equals(DASH + PACKAGE_DIR)) {
		try {
		    String tmpstr = argv[++i];
		    if (argv[i].startsWith("/")) {
			tmpstr = argv[i].substring(1);
		    }
		    argTable.set(PACKAGE_DIR, tmpstr);
		} catch (Exception e) {}
	    } else if (argv[i].equals(DASH + COPYRIGHT_NAME)) {
		try {
		    argTable.set(COPYRIGHT_NAME, argv[++i]);
		} catch (Exception e) {}
	    } else if (argv[i].equals(DASH + ENHYDRA_DIR)) {
		try {
		    argTable.set(ENHYDRA_DIR, argv[++i]);
		} catch (Exception e) {}
	    }
	}
	try {
	    argTable.set(APP_NAME, argv[argv.length - 1]);
	} catch (Exception e) {}
    }

    private String fixFS(String dir) {
	if (FS.equals("\\")) {
	    return dir.replace('\\', '/');
	} else {
	    return dir;
	}
    }

    private String fixDrive(String path) {
	if (path.charAt(1) == ':') {
	    return "//" + path.charAt(0) + path.substring(2);
	} else {
	    return path;
	}
    }

    /**
     * Creates a new Enhydra application development tree.
     */
    public void create() throws NewAppException  {
	if (appName == null) {
            fatal("no app name specified");
	}

	// System properties
	try {
	    FS = System.getProperty("file.separator");
	} catch (SecurityException e) {
	    FS = "/";
	}

        if (!silent) {
          System.out.println("Creating new tree for Enhydra application " +
		appName + " in " + this.rootDir + FS + appName + " ...");
        }

	try {
	    PS = System.getProperty("path.separator");
	} catch (SecurityException e) {
	    PS = ":";
	}

	try {
	    javaHome = System.getProperty("java.home");
	} catch (SecurityException e) {}
	javaHome = fixFS(javaHome);

	try {
	    javaClassPath = System.getProperty("java.class.path");
	} catch (SecurityException e) {
	    javaClassPath = javaHome + "/lib/classes.zip" + PS + ".";
	}
	javaClassPath = fixFS(javaClassPath);

	// appDir
	File appDir = new File(this.rootDir, appName);
	appDir.mkdirs();

	// Makefile
	String[] subdirs = {this.packageDir};
        String preInc = "\n" +
            "#\n" +
            "# build_all is normally the top-level-rule. This makes it so the jar file\n" +
            "# will be created after the build is all done.\n" +
            "#\n" +
            "top_build_all: build_all jar\n\n";
	String postinclude = "clean::\n\trm -rf $(CLASS_OUTPUT) $(OUTPUT)";
	try {
	    writeMakefile(appDir, null, subdirs, null, null, preInc,
                          postinclude);
	} catch (IOException e) {
            fatal("writing Makefile in " + appDir);
	}

	// README
	try {
	    writeReadme(appDir);
	} catch (IOException e) {
	    fatal("writing README in " + appDir, e);
	}

	// config.mk
	try {
	    writeConfigMk(appDir);
	} catch (IOException e) {
	    fatal("writing config.mk in " + appDir, e);
	}

	// packageDir
	File pkgDir = new File(appDir, this.packageDir);
	pkgDir.mkdirs();

	// Makefile
	String[] pkg_subdirs = {"data", "business", "presentation"};
	String[] pkg_classes = {appNameUC};
	String confin = appName + EXT_CONF;
	String confout = "${OUTPUT}/" + appName + EXT_CONF;
	String pkg_preinclude = "MISC_BUILD = conf_file ${OUTPUT}/" + MULTISERVER_CONF + " ${OUTPUT}/" + START_SCRIPT;

	String pkg_postinclude = "conf_file:\n"
            + "\t@[ -d ${OUTPUT} ] || mkdir -p ${OUTPUT}\n"
            + "\tcp " + confin + " " + confout + "\n"
            + "\n"
            + "${OUTPUT}/" + MULTISERVER_CONF + ": " + MULTISERVER_CONF_IN + "\n"
            + "\t@[ -d ${OUTPUT} ] || mkdir -p ${OUTPUT}\n"
            + "\trm -rf ${OUTPUT}/" + MULTISERVER_CONF + "\n"
            + "\tsed -e s%@OUTPUT@%`cd $(OUTPUT);pwd`%g \\\n"
            + "\t    < " + MULTISERVER_CONF_IN + " > ${OUTPUT}/" + MULTISERVER_CONF + "\n"
            + "\n"
            + "${OUTPUT}/" + START_SCRIPT + ": " + START_SCRIPT + "\n"
            + "\t@[ -d ${OUTPUT} ] || mkdir -p ${OUTPUT}\n"
            + "\trm -f ${OUTPUT}/" + START_SCRIPT + "\n"
            + "\tcp " + START_SCRIPT + " ${OUTPUT}/" + START_SCRIPT + "\n"
            + "\tchmod +x ${OUTPUT}/" + START_SCRIPT + "\n"
            + "\n"
            + "clean::\n\trm -f "  + confout + "\n";

	try {
	    writeMakefile(pkgDir, this.packageDir, pkg_subdirs, pkg_classes,
		    null, pkg_preinclude, pkg_postinclude);
	} catch (IOException e) {
	    fatal("writing Makefile in " + pkgDir, e);
	}

	// appName.conf
	try {
	    writeAppConf(pkgDir);
	} catch (IOException e) {
	    fatal("writing " + appName + ".conf in "
                  + pkgDir, e);
	}

        try {
            writeStartScript(pkgDir);
	} catch (IOException e) {
	    fatal("writing start script in " + pkgDir, e);
	}

        try {
            writeMultiserverConf(pkgDir);
	} catch (IOException e) {
	    fatal("writing multiserver config in " + pkgDir, e);
	}

	// appName.java
	try {
	    writeAppJava(pkgDir);
	} catch (IOException e) {
	    fatal("writing " + appNameUC + ".java in "
                  + pkgDir, e);
	}


	// presentation
	File presDir = new File(pkgDir, PRESENTATION);
	presDir.mkdirs();

	// Makefile
	String[] pres_subdirs = {MEDIA};
	String[] pres_joltclasses = null;
        if (jddi) {
          pres_joltclasses = new String[] {"WelcomeJDDI"};
        }

        String[] java_classes = {"Redirect", "Welcome"};
	String pres_packagedir = this.packageDir + "/" + PRESENTATION;
        String xmlcStuff = "HTML_DIR = .\nHTML_CLASSES = WelcomeHTML";
	try {
	    writeMakefile(presDir, pres_packagedir, pres_subdirs, java_classes,
		    pres_joltclasses, xmlcStuff, null);
	} catch (IOException e) {
	    fatal("writing Makefile in " + presDir, e);
	}

        if (jddi) {
          // WelcomeJDDI.jhtml
          try {
            writeWelcomeJhtml(presDir);
          } catch (IOException e) {
    	    fatal("writing WelcomeJDDI.jhtml in " + presDir, e);
          }
        }

        // Redirect.java
        try {
	    writeRedirectJava(presDir);
        } catch (IOException e) {
	    fatal("writing Redirect.java in " + presDir, e);
        }

        // Welcome.html  - XMLC
        try {
            writeWelcomehtml(presDir);
        } catch (IOException e) {
            fatal("writing Welcome.html in " + presDir, e);
        }

        // Welcome.java
        try {
            writeWelcomejava(presDir);
        } catch (IOException e) {
            fatal("writing Welcome.java in " + presDir, e);
        }


	// media
	File mediaDir = new File(presDir, MEDIA);
	mediaDir.mkdirs();

	// Makefile
	String media_packagedir = this.packageDir + "/" + PRESENTATION +
		"/" + MEDIA;

	String media_preinclude = "JAR_INSTALL = Enhydra.gif";
	String media_postinclude = "";


	try {
	    writeMakefile(mediaDir, media_packagedir, null, null, null,
		    media_preinclude, media_postinclude);
	} catch (IOException e) {
	    fatal("writing Makefile in " + mediaDir, e);
	}

	try {
	    writeEnhydraGif(mediaDir);
	} catch (Exception e) {
	    fatal("writing Enhydra.gif in " + mediaDir, e);
	}


	// business
	File busDir = new File(pkgDir, BUSINESS);
	busDir.mkdirs();

	// Makefile
	String bus_packagedir = this.packageDir + "/" + BUSINESS;
	preInc = "CLASSES =";
        try {
	    writeMakefile(busDir, bus_packagedir, null, null,
		    null, preInc, null);
	} catch (IOException e) {
	    fatal("writing Makefile in " + busDir, e);
	}

	// data
	File dataDir = new File(pkgDir, DATA);
	dataDir.mkdirs();

	// Makefile
	String data_packagedir = this.packageDir + "/" + DATA;
	try {
	    writeMakefile(dataDir, data_packagedir, null, null,
		    null, preInc, null);
	} catch (IOException e) {
	    fatal("writing Makefile in " + dataDir, e);
	}

        if (!silent) {
            System.out.println("Done.");
        }
    }

    /**
     * Finds the relative path to the rootDir from the provided directory.
     */
    protected String findRoot(String dir)
    {
	if (dir == null) {
	    return ".";
	}
	String root = "..";
	for (int i = 0; i < dir.length(); i++) {
	    if (dir.charAt(i) == '/') {
	        root = root + "/..";
	    }
	}
	return root;
    }

    /**
     * Writes a copyright message to out, starting each line with prefix.
     * @param out A PrintWriter on which to write the copyright message.
     * @param prefix A String to prepend to each line.
     */
    protected void writeCopyright(PrintWriter out, String prefix) {
        out.println(prefix + " Copyright (C) 2000 " + this.copyrightName +
		"\tAll Rights Reserved.");
    }

    /**
     * Writes a CVS Id to out, starting each line with prefix.
     * @param out A PrintWriter on which to write the copyright message.
     * @param prefix A String to prepend to each line.
     */
    protected void writeCvsId(PrintWriter out, String prefix) {
        // Broken up to void cvs expansion when this file is checked in.
        out.println(prefix + " $" + "Id: $");
    }

    /**
     * Writes a Makefile.
     * @param dir The directory in which to write the Makefile.
     * @param packagedir The path from the root directory to here (PACKAGEDIR).
     * @param subdirs An array of strings listing the subdirectories (SUBDIRS).
     * @param classes An array of strings listing the java files (CLASSES).
     * @param joltclasses An array of strings listing the jhtml files
     *	(JDDI_CLASSES).
     * @param preinclude Anything else to put before the include statement.
     * @param postinclude Anything else to put after the include statement.
     */
    public void writeMakefile(File dir, String packagedir, String[] subdirs,
	    String[] classes, String[] joltclasses, String preinclude,
	    String postinclude) throws IOException {
        File f = new File(dir, MAKEFILE);
        PrintWriter out = new PrintWriter(new FileWriter(f));

	// Header
        out.println("#");
        out.println("#  " + appName);
        out.println("#  Enhydra Application Makefile");
        out.println("#");
	// writeCopyright(out, "#");
        out.println("#");
        out.println();

	// ROOT
	String root = findRoot(packagedir);
        out.println("#");
        out.println("#  Where is the root of the source tree?");
        out.println("#");
        out.println("ROOT =\t" + root);
        out.println();

	// PACKAGEDIR
	if (packagedir != null) {
            out.println("#");
            out.println("#  What package does this directory represent? (use / instead of .)");
            out.println("#");
            out.println("PACKAGEDIR = " + packagedir);
            out.println();
	}

	// SUBDIRS
	if (subdirs != null) {
            out.println("#");
            out.println("# To add new directories, create the directories and add their names");
            out.println("# this list. Be sure to copy a Makefile into each directory. Then add a");
            out.println("# \"../\" to the ROOT variable, and add the directory name to the end of.");
            out.println("# the PACKAGEDIR variable.");
            out.println("#");
            out.println("SUBDIRS = \\");
	    for (int i = 0; i < (subdirs.length - 1); i++) {
                out.println("\t" + subdirs[i] + " \\");
	    }
	    if (subdirs.length > 0) {
                out.println("\t" + subdirs[subdirs.length - 1]);
	    }
            out.println();
	}

	// CLASSES
	if (classes != null) {
            out.println("#");
            out.println("# To add more Java classes, create the .java files and add their names to");
            out.println("# this list (without the .java extension).");
            out.println("#");
            out.println("CLASSES = \\");
	    for (int i = 0; i < (classes.length - 1); i++) {
                out.println("\t" + classes[i] + " \\");
	    }
	    if (classes.length > 0) {
                out.println("\t" + classes[classes.length - 1]);
	    }
            out.println();
	}

	// JDDI_CLASSES
	if (joltclasses != null) {
            out.println("#");
            out.println("# To add more Enhydra JDDI pages, simply create the .jhtml files and");
            out.println("# their names this list (without the .jhtml extension). ");
            out.println("#");
            out.println("JDDI_CLASSES = \\");
	    for (int i = 0; i < (joltclasses.length - 1); i++) {
                out.println("\t" + joltclasses[i] + " \\");
	    }
	    if (joltclasses.length > 0) {
                out.println("\t" + joltclasses[joltclasses.length - 1]);
	    }
            out.println();
	}

	// pre include
	if (preinclude != null) {
            out.println(preinclude);
            out.println();
	}

	// include
        out.println("include $(ROOT)/config.mk");
        out.println();

	// post include
	if (postinclude != null) {
            out.println(postinclude);
            out.println();
	}

        out.close();
    }

    /**
     * Writes the top level README file.
     * @param dir a File which is the directory where the README goes.
     */
    public void writeReadme(File dir) throws IOException {
        File f = new File(dir, "README");
        PrintWriter out = new PrintWriter(new FileWriter(f));

	// Header
        out.println("\n To build and run " + appName + ":");
	out.println();

	// Instructions
        out.println("1) Type \"make\" in this directory.\n");
        out.println("2) The application will be compiled, with the resulting class files put in");
        out.println("   the classes directory. Then a jar file will be created in output/lib.");
        out.println("   the application's config file will be placed in the directory");
        out.println("   output. An Enhydra Multiserver config file will also be placed");
        out.println("   in the output directory, along with a start script.");
        out.println("3) Type \"cd output\".");
        out.println("4) Type \"./start\".");
        out.println("5) Use your web browser to access " + appName + " at:  http://localhost:9000\n\n");
        out.println("NOTES:");
        out.println(" localhost can also be your host name or 127.0.0.1 (best for Win95).\n");
        out.println();
	out.close();
    }

    /**
     * Writes the top level config.mk file.
     * @param dir a File which is the directory where the config.mk goes.
     */
    public void writeConfigMk(File dir) throws IOException
    {
        File f = new File(dir, CONFIG_MK);
        PrintWriter out = new PrintWriter(new FileWriter(f));

	// Header
        out.println("#");
        out.println("# " + appName);
        out.println("# Enhydra Application Makefile Configuration");
        out.println("#");
        out.println("#");
        out.println();

	// ENHYDRA_DIR
        out.println("#");
        out.println("# The directory containing the" +
		" Enhydra distribution.");
        out.println("#");
        out.println("ENHYDRA_DIR = " + enhydraDir);
        out.println();
        out.println();

        out.println("#");
        out.println("# You may need to set other values to override the defaults, for example:");
        out.println("#");
        out.println("# JDKDIR = /usr/local/jdk1.2.2");
        out.println("# JSDKDIR = ...");
        out.println("# SWINGDIR = ...  (optional)");
        out.println("#");
        out.println("# See lib/stdrules.mk in the Enhydra installation for more");
        out.println("# configuration settings.");
        out.println("#");
        out.println();
        out.println();

        // jar file
        out.println("#");
        out.println("# What is the name of the jar file to create?");
        out.println("#");
        out.println("JARNAME = " + appName);
        out.println();

        // XMLC options
        out.println("#");
        out.println("# Generate interfaces and implementions");
        out.println("#");
        out.println("XMLC_HTML_OPTS += -generate both");
        out.println();

        // ENHYDRA_CHECK
        out.println("#");
        out.println("# Do some sanity checking before building.");
        out.println("#");
        out.println("ENHYDRA_CHECK = YES");
        out.println();

	// include
        out.println("include $(ENHYDRA_DIR)/lib/stdrules.mk");
        out.println();

        // Error checking
        out.println("#\n# Only used if ENHYDRA_DIR is not correct.\n#");
        out.println("$(ENHYDRA_DIR)/lib/stdrules.mk:");
        out.println("\t@echo");
        out.println("\t@echo \"Please edit config.mk. ENHYDRA_DIR must point to your installation of Enhydra.\"");
        out.println("\t@echo");

	out.close();
    }

    /**
     * Writes the application config file.
     * @param dir a File which is the directory where the file goes.
     */
    public void writeAppConf(File dir) throws IOException
    {
        File f = new File(dir, appName + EXT_CONF);
        PrintWriter out = new PrintWriter(new FileWriter(f));

	// Header
        out.println("# " + appName);
        out.println("# Enhydra Application Configuration");
        out.println("#");
        out.println();

	// Server.ClassPath
        String appClassPath = "../classes";
        String jarPath = "./lib/" + appName + ".jar";
        out.println("#");
        out.println("# Comma separated CLASSPATH directories and files"
                    + " used by this application.");
        out.println("# Assumes run from the output directory for debugging.");
        out.println("# If you run from the jar you must rebuild after every change to the app.");
        out.println("#");
        out.println(MultiServer.SERVER + "." + ServletContainer.CLASS_PATH +
                    "[] = " + appClassPath);
        out.println("#" + MultiServer.SERVER + "." + ServletContainer.CLASS_PATH +
                    "[] = " + jarPath);
        out.println();

	// Server.AppClass
        out.println("#");
        out.println("# The fully qualified name of the application class.");
        out.println("#");
        out.println(MultiServer.SERVER + "." + ServletContainer.APP_CLASS +
		" = " + this.packageName + "." + appNameUC );
        out.println();

	// Server.PresentationPrefix
        out.println("#");
        out.println("# Prefix used to derive presentation object class");
	out.println("# names and paths from URLs.");
        out.println("# Assumes run from the output directory for debugging.");
        out.println("#");
        out.println(MultiServer.SERVER + "." +
                    ServletContainer.PRESENTATION_PREFIX + " = \"" +
                    this.packageDir + "/" + PRESENTATION + "\"");
        out.println();

        // Server.AutoReload
        out.println("#");
        out.println("# Flag to indicate that application classes and resources should be reloaded");
        out.println("# automatically if ANY file in the CLASSPATH changes.  NOTE:  This is a");
        out.println("# debugging option and may slow down the performance of your application.");
        out.println("# The CLASSPATH should NOT contain any directories (or underlying directories)");
        out.println("# that contain constantly changing files, i.e. a log file.  For example, do");
        out.println("# NOT include the application's output directory in the CLASSPATH because it");
        out.println("# contains the application log files in the underlying log directory.");
        out.println("#");
        out.println("Server.AutoReload = false");
        out.println();

	// sessionManager.Lifetime
        out.println("#");
        out.println("# Maximum number of minutes a user session can last.");
        out.println("#");
        out.println("SessionManager.Lifetime = 60");
        out.println();

	// SessionManager.MaxIdle
        out.println("#");
        out.println("# Maximum number of minutes a user may be"
		+ " idle before being logged off.");
        out.println("#");
        out.println("SessionManager.MaxIdleTime = 2");
        out.println();

	//
        out.println("#");
        out.println("# If the URL \"/\" for this application is accessed, the user will be");
        out.println("# redirected to this URL. This should be a relative URL.");
        out.println("#");
        out.println("Application.DefaultUrl = \"Welcome.po\"");
        out.println();

	out.close();
    }


    /**
     * Writes the multiserver config file.
     * @param dir a File which is the directory where the file goes.
     */
    public void writeMultiserverConf(File dir) throws IOException
    {
        File f = new File(dir, MULTISERVER_CONF_IN);
        PrintWriter out = new PrintWriter(new FileWriter(f));

        out.println("#");
        out.println("#  Simple Enhydra Multiserver config file that runs one application.");
        out.println("#  See multiserver.config in the Enhydra installation for a");
        out.println("#  description of the Multiserver config file format.");
        out.println("#");
        out.println();
        out.println("#");
        out.println("# Where the config files are.");
        out.println("#");
        out.println("Server.ConfDir = @OUTPUT@");
        out.println();
        out.println("#");
        out.println("# Set up logging.");
        out.println("# For more information, try adding DEBUG or CLASSLOAD.");
        out.println("#");
        out.println("Server.LogFile = @OUTPUT@/" + appName + ".log");
        out.println("Server.LogToFile[] = EMERGENCY, ALERT, CRITICAL, ERROR, WARNING, INFO");
        out.println("Server.LogToStderr[] = EMERGENCY, ALERT, CRITICAL, ERROR, WARNING, INFO");
        out.println();
        out.println();
        out.println();
        out.println("#");
        out.println("# Specify only one application (no graphical administration).");
        out.println("#");
        out.println("Application." + appName + ".ConfFile = " + appName + ".conf");
        out.println("Application." + appName + ".Description = \"An Enhydra Application\"");
        out.println("Application." + appName + ".Running = yes");
        out.println();
        out.println();
        out.println("#");
        out.println("# Open up a port.");
        out.println("#");
        out.println("Connection.p9000.Type = http");
        out.println("Connection.p9000.Port = 9000");
        out.println();
        out.println();
        out.println("#");
        out.println("# Connect the port to the application.");
        out.println("#");
        out.println("Channel.p9000.aChan.Servlet = " + appName);
        out.println("Channel.p9000.aChan.Url = /");
        out.println("Channel.p9000.aChan.Enabled = yes");
        out.println();
        out.println();
        out.println();
        out.println("#");
        out.println("# ----------------------------------------------------------------------");
        out.println("#");
        out.println("# If you want to use the Multiserver's graphical administration program,");
        out.println("# first copy it's config file to the directory Server.confDir is set to:");
        out.println("#");
        out.println("#     cp " + enhydraDir + "/apps/multiserverAdmin.conf @OUTPUT@");
        out.println("#");
        out.println("# Then uncomment the following lines, re-run the Multiserver (./start),");
        out.println("# and connect to port 9001. The default username is \"admin\" and the default");
        out.println("# password is \"lutris\". This can be changed by editing the local copy of");
        out.println("# multiserverAdmin.conf.");
        out.println("#");
        out.println("# If you are going to be adding other applications, other servlets, or");
        out.println("# other connections, it is highly recommended that you use the graphical");
        out.println("# admin rather than edit this file by hand (remember to save your changes");
        out.println("# by clicking on the \"Save State\" button!).");
        out.println("#");
        out.println("# Application.MultiserverAdmin.ConfFile = multiserverAdmin.conf");
        out.println("# Application.MultiserverAdmin.Description = \"The Enhydra Multiserver Management Console.\"");
        out.println("# Application.MultiserverAdmin.Running = yes");
        out.println("#");
        out.println("# Connection.p9001.Type = http");
        out.println("# Connection.p9001.Port = 9001");
        out.println("#");
        out.println("# Channel.p9001.aChan.Servlet = MultiserverAdmin");
        out.println("# Channel.p9001.aChan.Url = /");
        out.println("# Channel.p9001.aChan.Enabled = yes");
        out.println("#");
        out.println("# ----------------------------------------------------------------------");
        out.println();

	out.close();
    }


    public void writeStartScript(File dir) throws IOException
    {
        File f = new File(dir, START_SCRIPT);
        PrintWriter out = new PrintWriter(new FileWriter(f));

        out.println("#!/bin/sh\n");
        out.println("echo \"***\"");
        out.println("echo \"***  Connect to http://`hostname`:9000/\"");
        out.println("echo \"***\"\n");
        out.println(enhydraDir + "/bin/multiserver ./multiserver.conf");
	out.close();
    }


    /**
     * Writes the application main file.
     * @param dir a File which is the directory where the file goes.
     */
    public void writeAppJava(File dir) throws IOException
    {
        File f = new File(dir, appNameUC + EXT_JAVA);
        PrintWriter out = new PrintWriter(new FileWriter(f));

	// Header
        out.println("/*");
        out.println(" * " + appName);
        out.println(" * A sample Enhydra Application");
        out.println(" * ");
        out.println(" */");
	out.println();

	// package/imports
	out.println("package " + this.packageName + ";");
	out.println();
	out.println("import com.lutris.appserver.server.*;");
	out.println("import com.lutris.appserver.server.httpPresentation.*;");
	out.println("import com.lutris.appserver.server.session.*;");
	out.println("import com.lutris.util.*;");
	out.println();

	// class definition
	out.println("/**");
	out.println(" * The application object.");
	out.println(" * ");
	out.println(" * Application-wide data would go here.");
	out.println(" */");
	out.println("public class " + appNameUC +
		" extends StandardApplication {");
	out.println();

        // methods
        out.println("    /*");
        out.println("     *  A few methods you might want to add to.");
        out.println("     *  See StandardApplication for more details.");
        out.println("     */");
        out.println("    public void startup(Config appConfig) throws ApplicationException {");
        out.println("        super.startup(appConfig);");
	out.println("        //  Here is where you would read application-specific settings from");
        out.println("        //  your config file.");
        out.println("    }");
        out.println("    public boolean requestPreprocessor(HttpPresentationComms comms)");
        out.println("                   throws Exception {");
        out.println("        return super.requestPreprocessor(comms);");
        out.println("    }");
        out.println();



        // toHtml
        out.println("    /**");
        out.println("     * This is an optional function, used only by the Multiserver's graphical");
        out.println("     * administration. This bit of HTML appears in the status page for this");
        out.println("     * application. You could add extra status info, for example");
        out.println("     * a list of currently logged in users.");
        out.println("     *");
        out.println("     * @return HTML that is displayed in the status page of the Multiserver.");
        out.println("     */");
        out.println("    public String toHtml() {");
        out.println("        return \"This is <I>" + appName + "</I>\";");
        out.println("    }");

	// end class definition
	out.println("}");
	out.println();

	out.close();
    }


    /**
     * Writes the WelcomeJDDI.jhtml presentation object source.
     * By default, we don't generate this anymore.
     * @param dir a File which is the directory where the file goes.
     */
    public void writeWelcomeJhtml(File dir) throws IOException
    {
        File f = new File(dir, "WelcomeJDDI" + EXT_JHTML);
        PrintWriter out = new PrintWriter(new FileWriter(f));

	// Head
        out.println("<HTML>");
        out.println("<HEAD>");
        out.println("<TITLE>Welcome to " + appName + "!</TITLE>");
        out.println("</HEAD>");

        // Jolt Section
        out.println();
        out.println("<JDDI JAVAIMPORT>");
        out.println("    import java.util.Date;");
        out.println("</JDDI>");
        out.println();
        out.println("<JDDI JAVADEF>");
        out.println("    /*");
        out.println("     *  This is a sample presentation object using the JDDI compiler.");
        out.println("     *  Page.data is used by the JDDI tags (@...@). See also");
        out.println("     *  page.request.getParameter() for access to the input arguments,");
        out.println("     *  and page.session.getSessionData() for access to the session data.");
        out.println("     */");
        out.println("    static public void getDate(JoltPage page) throws Exception {");
        out.println("       page.data.set(\"now\", new Date().toString());");
        out.println("    }");
        out.println();
        out.println("    /*");
        out.println("     *  For use by showHits() below.");
        out.println("     */");
        out.println("    private static int hitCounter = 0;");
        out.println();
        out.println("    /*");
        out.println("     *  Rather than setting up a JDDI tag, this method directly outputs");
        out.println("     *  HTML to the page.");
        out.println("     *");
        out.println("     *  UNCOMMENT THE CODE TO SEE IT WORK.");
        out.println("     */");
        out.println("     static public void showHits(JoltPage page) throws Exception {");
        out.println("/*");
        out.println("         hitCounter++;");
        out.println("         page.append(\"<P><B><I>This page has been served \" +");
        out.println("                     hitCounter + \" time\");");
        out.println("         if (hitCounter > 1)");
        out.println("             page.append(\"s\");");
        out.println("         page.append(\".</I></B><P>\");");
        out.println("*/");
        out.println("    }");
        out.println("</JDDI>");
        out.println();

	// Body
        out.println("<BODY BGCOLOR=\"#FFFFFF\">");
        out.println("<CENTER><TABLE><TR>");
        out.println("<TD><A HREF=http://www.enhydra.org/>" +
		"<IMG SRC=media/Enhydra.gif WIDTH=200 BORDER=0></A></TD>");
        out.println("<TD><H1>Welcome to " + appName + "!</H1>");
        out.println("<BR><H3>A Simple Enhydra Application</H3></TD></TR>");
        out.println();
        out.println("<JDDI CALL=getDate></JDDI>");
        out.println();
        out.println("<TR><TD COLSPAN=2><CENTER>The current time is (@now@).</CENTER></TD></TR>");
        out.println("<TR><TD COLSPAN=2><CENTER><A HREF=Redirect.po>Sample redirect back to here</A></CENTER>.</TD></TR>");
        out.println("</TABLE>");
        out.println();
        out.println("<P><A HREF=Welcome.po>(XMLC version)</A>");
        out.println();
        out.println("<JDDI CALL=showHits></JDDI>");
        out.println();
        out.println("</CENTER>");
        out.println("</BODY>");
        out.println("</HTML>");
        out.println();

	out.close();
    }

    /**
     * XMLC
     */
    public void writeWelcomehtml(File dir) throws IOException
    {
        File f = new File(dir, "Welcome.html");
        PrintWriter out = new PrintWriter(new FileWriter(f));

	// Head
        out.println("<HTML>");
        out.println("<HEAD>");
        out.println("<TITLE>Welcome to " + appName + "!</TITLE>");
        out.println("</HEAD>");
	// Body
        out.println("<BODY BGCOLOR=\"#FFFFFF\">>");
        out.println();
        out.println("<CENTER><TABLE SUMMARY=\"Welcome.html\"><TR>");
        out.println("<TD><A HREF=http://www.enhydra.org/>" +
		"<IMG SRC=media/Enhydra.gif WIDTH=200 BORDER=0 ALT=\"Enhydra.gif\"></A></TD>");
        out.println("<TD><H1>Welcome to " + appName + "!</H1>");
        out.println("<BR><H3>A Simple Enhydra Application</H3></TD></TR>");
        out.println();
        out.println("<TR><TD COLSPAN=2><CENTER>The current time is <span id=\"Time\">1/1/00 00:00:00 Oh no!</span>.</CENTER></TD></TR>");
        out.println("<TR><TD COLSPAN=2><CENTER><A HREF=Redirect.po>Sample redirect back to here</A></CENTER>.</TD></TR>");
        out.println("</TABLE>");
        out.println();
        if (jddi) {
          out.println("<P><A HREF=WelcomeJDDI.po>(Enhydra JDDI version)</A>");
        }
        out.println("</CENTER>");
        out.println();
        out.println("</BODY>");
        out.println("</HTML>");
        out.println();

	out.close();
    }



    public void writeWelcomejava(File dir) throws IOException
    {
        File f = new File(dir, "Welcome.java");
        PrintWriter out = new PrintWriter(new FileWriter(f));
        out.println("package " + packageName + ".presentation;");
	out.println();
	out.println("import java.util.Date;");
	out.println("import com.lutris.xml.xmlc.*;");
	out.println("import com.lutris.appserver.server.httpPresentation.*;");
	out.println();
	out.println("public class Welcome implements HttpPresentation {");
	out.println();
	out.println();
	out.println("    public void run(HttpPresentationComms comms) ");
	out.println("        throws HttpPresentationException {");
	out.println();
	out.println("        String now = new Date().toString();");
	out.println("        WelcomeHTML welcome = (WelcomeHTML)comms.xmlcFactory.create(WelcomeHTML.class);");
	out.println("        welcome.setTextTime(now);");
	out.println("        comms.response.writeHTML(welcome);");
	out.println("    }");
	out.println();
	out.println("}");

	out.close();
    }



    /**
     * Writes the Redirect.java presentation object source.
     * @param dir a File which is the directory where the file goes.
     */
    public void writeRedirectJava(File dir) throws IOException
    {
        File f = new File(dir, "Redirect.java");
        PrintWriter out = new PrintWriter(new FileWriter(f));

        out.println("/*");
        out.println(" *  This is a sample presentation object written by hand (not using the");
        out.println(" *  Jolt compiler). Since this presentation object only redirects, and does");
        out.println(" *  not emit any HTML, it does not need Jolt.");
        out.println(" *  This type of presentation object could, for example, be the target of");
        out.println(" *  a form's submit button, process the data, and then decide where to send");
        out.println(" *  the user (back or on to the next page).");
        out.println(" */");
        out.println();
        out.println("package " + packageName + ".presentation;");
        out.println();
        out.println("import com.lutris.appserver.server.httpPresentation.*;");
        out.println();
        out.println();
        out.println("public class Redirect implements HttpPresentation {");
        out.println();
        out.println("   /*");
        out.println("    * There is the only function needed in order to be a presentation object.");
        out.println("    */");
        out.println("    public void run(HttpPresentationComms comms) throws Exception {");
        out.println("        throw new ClientPageRedirectException(");
        out.println("                      comms.request.getAppFileURIPath(\"Welcome.po\"));");
        out.println("    }");
        out.println();
        out.println("}");
	out.close();
    }




    void writeEnhydraGif(File dir) throws Exception {
        File f = new File(dir, "Enhydra.gif");
        OutputStream out = new FileOutputStream(f);
        InputStream in;
        in = ClassLoader.getSystemResourceAsStream(
                                     "com/lutris/devsys/newApp/Enhydra.gif");
        if (in == null)
            throw new Exception("Unable to access Enhydra.gif");
        byte[] buf = new byte[1024];
        while (true) {
            int n = in.read(buf);
            if (n < 0)
                break;
            out.write(buf, 0, n);
        }
        in.close();
        out.close();
    }


    /**
     * Sets the application name for the newly created
     * Enhyudra application development tree.
     * If packageDir is not already set, it is set to appName with the
     * first letter lowercased.
     * @param appName The application name.
     * @see #setPackageDir
     */
    public void setAppName(String appName) {
	this.appName = appName;
        appNameLC = appName.substring(0, 1).toLowerCase() +
                    appName.substring(1);
        appNameUC = appName.substring(0, 1).toUpperCase() +
                    appName.substring(1);
	if (this.packageDir == null)
	    setPackageDir(appNameLC);
    }

    /**
     * Sets the root directory under which the new
     * app development directory is created.
     * @param dir The directory path name.
     */
    public void setRootDir(String dir) {
        rootDir = dir;
    }

    /**
     * Sets the package directory which is the top level for source files.
     * @param dir The directory path name.
     */
    public void setPackageDir(String dir) {
	this.packageDir = dir;
	this.packageName = this.packageDir.replace('/', '.');
    }

    /**
     * Sets the name to be used in the copyright message.
     * @param copyrightName The name to use in the copyright message.
     */
    public void setCopyrightName(String copyrightName)
    {
	this.copyrightName = copyrightName;
    }

    /**
     * Sets the class directory where Enhydra is installed.
     * @param dir The directory path name.
     */
    public void setEnhydraDir(String dir)
    {
	enhydraDir = dir;
    }
}



