/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: EnhydraDirectorBufferedOutputStream.java,v 1.3 2000/02/21 23:50:42 markd Exp $
 */

package org.enhydra.servlet.connectionMethods.EnhydraDirector;

import java.io.OutputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;

/**
 * Extension of <code>BufferedOutputStream</code> to allow "reset" of
 * the buffer.
 * <P>
 * A servlet container's <code>ServletResponse</code> object should 
 * not go into the "committed" (data sent) state until the headers
 * have actually been sent over the wire.  This allows a servlet to
 * specify a large buffer size and be able to reset or change headers
 * well after the first byte of data has been written to the 
 * <code>ServletOutputStream</code>.
 * <P>
 * This class has a <code>reset()</code> that, after checking for
 * committed state of the connection, discards any buffered data 
 * and self destructs, dropping any reference to its internally
 * held output stream.  Once <code>reset()</code> has been called,
 * this object must be discarded as it is then useless.  A new
 * instance should be created in its place.
 */
public class EnhydraDirectorBufferedOutputStream extends BufferedOutputStream
{
    /**
     * Internal reference for for this stream's underlying connection.
     */
    private EnhydraDirectorConnection conn;

    /**
     * Creates a new instance with the specified connection context,
     * <code>OutputStream</code>, and buffer size.
     * 
     * @param os         The output stream.
     * @param size       Buffer size.
     * @param connection The connection context.
     */
    public 
    EnhydraDirectorBufferedOutputStream(OutputStream os, int size,
				EnhydraDirectorConnection connection)
    {
	super(os, size);
	conn = connection;
    }

    /**
     * Discards buffered data and drops reference to the output stream.
     * <P>
     * If the current connection has not sent any data over the wire,
     * this method drops any buffered data, as well as the output stream
     * reference, to ensure that this instance makes no further calls
     * to the underlying stream.  It is expected that the caller of
     * this method will also discard any reference to this 
     * <code>EnhydraDirectorBufferedOutputStream</code> object.  Once 
     * <code>reset()</code> has been called, this stream becomes 
     * unusable.
     * 
     * @exception IllegalStateException If called after data has been sent.
     */
    public
    void reset()
    throws IllegalStateException
    {
	if (conn.isCommitted()) {
	    throw new IllegalStateException("Data already sent: cannot reset");
	}
	out = null;
	count = 0;
	buf = null;
	try { close(); } catch (Exception e) {}
    }

    /**
     * Flushes outgoing data, then flushes the contained output stream.
     * 
     * @exception IOException If there is an error sending data.
     */
    public
    void flush()
    throws IOException
    {
	if (out != null) {
	    super.flush();
	    out.flush();
	}
    }
}

