/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: HttpHeader.java,v 1.4 2000/02/21 23:50:44 markd Exp $
 */

package org.enhydra.servlet.connectionMethods.http;
import java.io.IOException;

/**
 * Class to store a single HTTP header.
 */
class HttpHeader {
    /**
     * Header name.  This field is not case-adjusted.
     */
    private String name;

    /**
     * A single header value.  If the header has multiple values,
     * this will be null
     */
    private String value;

    /**
     * An array of header value.  This may or may not be null if
     * the header has a single value.
     */
    private String[] values;

    /**
     * Flag that indicates if this header has been written.
     */
    private boolean written = false;

    /**
     * Construct a new header field.
     *
     * @param hname The header name.
     * @param hvalue The header value.
     */
    protected HttpHeader(String hname, String hvalue) {
        name = hname;
        value = hvalue;
    }

    /**
     * Set the header value to be a single string.
     *
     * @param hvalue The new header value.
     */
    protected synchronized void setValue(String hvalue) {
        value = hvalue;
        values = null;
    }

    /**
     * Add an additional header value.
     *
     * @param hvalue The header value to add
     */
    protected synchronized void addValue(String hvalue) {
        if (values == null) {
            values = new String[2];
            values[0] = value;
        } else {
            String[] newValues = new String[values.length+1];
            System.arraycopy(values, 0, newValues, 0, values.length);
        }
        values[values.length-1] = value;
        value = null;
    }

    /**
     * Get the header name.
     */
    protected synchronized String getName() {
        return name;
    }

    /**
     * Get the header value as a single string.  If the header
     * has multiple values, the first one is returned.
     */
    protected synchronized String getValue() {
        if (value != null) {
            return value;
        } else {
            return values[0];
        }
    }

    /**
     * Get the header value as an array of values.
     */
    protected synchronized String[] getValues() {
        if (values == null) {
            values = new String[1];
            values[0] = value;
        }
        return values;
    }

    /**
     * Has the header been written to the output stream.
     */
    protected boolean isWritten() {
        return written;
    }

    /**
     * Get a integer header value.
     *
     * @param hname The header name.
     * @return The header value.
     * @exception NumberFormatException If value of header can't be
     *  converted to an int.
     */
    protected int getInt() {
        return Integer.parseInt(value);
    }

    /**
     * Compare the header name to a name, ignoring case.
     * @param hname The name to compare against.
     * @return true if the names match, false otherwise.
     */
    protected boolean equals(String hname) {
        return name.equalsIgnoreCase(hname);
    }

    /**
     * Write a header to an output stream.
     * @param out Output stream to write.
     * @exception IOException If an write error occured.
     */
    protected void write(HttpOutputStream out) throws IOException {
        if (value != null) {
            out.print(name);
            out.print(": ");
            out.println(value);
        } else {
            for (int idx = 0; idx < values.length; idx++) {
                out.print(name);
                out.print(": ");
                out.println(values[idx]);
            }
        }
        written = true;
    }
}
