/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: HttpHeaders.java,v 1.4 2000/02/21 23:50:44 markd Exp $
 */

package org.enhydra.servlet.connectionMethods.http;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Hashtable;
import java.io.IOException;


/**
 * Class to store headers.  This stores either a header or an array of
 * headers, Depending if one or more instances of a header has been
 * encountered.  A hash table is not kept, as the order of writing
 * headers seems to tickle bugs in certain browsers.
 */
public class HttpHeaders {
    /**
     * Initial size of header array.
     */
    private static final int INIT_HEADER_TBL_SIZE = 16;

    /**
     * Max size of the table after a clear.
     */
    private static final int MAX_EMPTY_HEADER_TBL_SIZE = 32;

    /**
     * Number of entries currently in header table.
     */
    private int numHeaderEntries = 0;

    /**
     * Table of header objects.
     */
    private HttpHeader[] headers = new HttpHeader[INIT_HEADER_TBL_SIZE];



    /**
     * Enumeration over the header names.
     */
    private class HeaderNameEnumeration implements Enumeration {
        private int savedNumHeaderEntries = numHeaderEntries;
        private HttpHeader[] savedHeaders = headers;
        private int nextIdx = 0;

        /**
         * Tests if this enumeration contains more elements. 
         *
         * @return true if this enumeration contains more elements;
         *  false otherwise.
         */
        public boolean hasMoreElements() {
            return (nextIdx < savedNumHeaderEntries);
        }

        /**
         * Get the next element of this enumeration.
         *
         * @return The next element of this enumeration. 
         * @exception NoSuchElementException 
         */
        public Object nextElement() throws NoSuchElementException {
            if (nextIdx >= savedNumHeaderEntries) {
                throw new NoSuchElementException("no more HTTP headers");
            }
            return savedHeaders[nextIdx++].getName();
        }
    }

    /**
     * Search for a header in the table.
     * @param hname The name of the header to search for.
     * @return The header object, or null if not found
     */
    protected synchronized HttpHeader get(String hname) {
        for (int idx = 0; idx < numHeaderEntries; idx++) {
            if (headers[idx].equals(hname)) {
                return headers[idx];
            }
        }
        return null;
    }

    /**
     * Create a new header.
     */
    private void createHeader(String hname, String hvalue) {
        if (numHeaderEntries == headers.length) {
            HttpHeader[] newHeaders = new HttpHeader[headers.length * 2];
            System.arraycopy(headers, 0, newHeaders, 0, headers.length);
            headers = newHeaders;
        }
        headers[numHeaderEntries++] = new HttpHeader(hname, hvalue);
    }

    /**
     * Add a header.  If the header already exists, add a new value;
     * otherwise create a new object.
     *
     * @param hname The header name.
     * @param hvalue The header value.
     */
    protected synchronized void add(String hname, String hvalue) {
        HttpHeader header = get(hname);
        if (header == null) {
            createHeader(hname, hvalue);
        } else {
            header.addValue(hvalue);
        }
    }

    /**
     * Add a header.  If the header already exists, overwrite the
     * existing value.
     *
     * @param hname The header name.
     * @param hvalue The header value.
     */
    protected synchronized void set(String hname, String hvalue) {
        HttpHeader header = get(hname);
        if (header == null) {
            createHeader(hname, hvalue);
        } else {
            header.setValue(hvalue);
        }
    }

    /**
     * Determine if a header exists.
     * @param hname The header name.
     * @return true if the header exists, false if it doesn't
     */
    protected boolean exists(String hname) {
        return (get(hname) != null);
    }

    /**
     * Get a header value.
     * @param hname The header name.
     * @return The header value or null if it does not exist.
     */
    protected String getValue(String hname) {
        HttpHeader header = get(hname);
        if (header == null) {
            return null;
        } else {
            return header.getValue();
        }
    }

    /**
     * Return array of header values for a header that might have
     * been sent multiple times.
     * @param hname The header name.
     * @return The header value array or null if it does not exist.
     */
    public String[] getValues(String hname) {
        HttpHeader header = get(hname);
        if (header == null) {
            return null;
        } else {
            return header.getValues();
        }
    }

    /**
     * Get a integer header.
     * @param hname The header name.
     * @return The header value or -1 if it does not exist.
     * @exception NumberFormatException If value of header can't be
     *  converted to an int.
     */
    protected int getInt(String hname) {
        HttpHeader header = get(hname);
        if (header == null) {
            return -1;
        } else {
            return header.getInt();
        }
    }

    /**
     * Get an enumeration of the headers.
     * @return An enumeration of HttpHeader.
     */
    public synchronized Enumeration getNames() {
        return new HeaderNameEnumeration();
    }

    /**
     * Write specified headers to an output stream in the specified order.
     * @param headerOrder List of headers to write in the order they are
     *  to be written.
     * @param out Output stream to write.
     * @exception IOException If an write error occured.
     */
    protected synchronized void write(HttpOutputStream out,
                                      String[] headerOrder) throws IOException {
        for (int idx = 0; idx < headerOrder.length; idx++) {
            HttpHeader header = get(headerOrder[idx]);
            if ((header != null) && !header.isWritten()) {
                header.write(out);
            }
        }
    }
    /**
     * Write headers to an output stream that have not already been written.
     * @param out Output stream to write.
     * @exception IOException If an write error occured.
     */
    protected synchronized void write(HttpOutputStream out) throws IOException {
        for (int idx = 0; idx < numHeaderEntries; idx++) {
            if (!headers[idx].isWritten()) {
                headers[idx].write(out);
            }
        }
    }

    /**
     * Clear headers.
     */
    protected synchronized void clear() {
        // Don't let table grow too large
        if (headers.length > MAX_EMPTY_HEADER_TBL_SIZE) {
            headers = new HttpHeader[INIT_HEADER_TBL_SIZE];
        } else {
            for (int idx = 0; idx < numHeaderEntries; idx++) {
                headers[idx] = null;
            }
        }
        numHeaderEntries = 0;
    }
}
 
