/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: DOMFormatter.java,v 1.9.2.3 2000/04/04 05:29:48 markd Exp $
 */

package org.enhydra.xml.io;

import com.lutris.util.FatalExceptionError;
import java.io.Writer;
import java.io.StringWriter;
import java.io.ByteArrayOutputStream;
import java.io.OutputStreamWriter;
import java.io.OutputStream;
import java.io.IOException;
import java.io.File;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.SerializerFactory;
import org.apache.xml.serialize.DOMSerializer;
import org.apache.xml.serialize.Serializer;
import org.apache.xml.serialize.Method;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Node;

/**
 * Class for formatting DOMs into a documents.  The class handles XML and
 * HTML pages, looking up the approriate formatting routines based on the
 * document type.  It has an associated OutputOptions object that controls the
 * formatting.  Instances are reusable and multithreaded.
 */
public class DOMFormatter {
    /**
     * Output options.
     */
    private OutputOptions options;

    /**
     * Construct a new XML formatter with the specified
     * output options.
     * @param outputOptions The OutputOptions object that  
     */
    public DOMFormatter(OutputOptions outputOptions) {
        options = outputOptions;
        if (options == null) {
            options = new OutputOptions();
        }
    }
    
    /**
     * Construct a new XML formatter with default output options.
     */
    public DOMFormatter() {
        this(null);
    }

    /**
     * Get the output options.
     */
    public OutputOptions getOutputOptions() {
        return options;
    }
    
    /**
     * Set the output options.
     */
    public void setOutputOptions(OutputOptions outputOptions) {
        options = outputOptions;
    }
    
    /**
     * Create the formatter for the document base on the document
     * type.
     */
    private Formatter getFormatter(Node node) {
        // Use Xercers code to make up our minds.
        String method = OutputFormat.whichMethod(node.getOwnerDocument());
        if (method == Method.XML) {
            return new XMLFormatter();
        } else if (method == Method.HTML) {
            return new HTMLFormatter();
        } else if (method == Method.XHTML) {
            throw new XMLIOError("XHTML not implemented");
        } else {
            throw new XMLIOError("unsupported formatting method: " + method);
        }
    }

    /**
     * Format a document or any node and its children into a string.
     */
    public String toString(Node node) {
        Formatter formatter = getFormatter(node);
        StringWriter writer = new StringWriter();
        try {
            formatter.write(options, node, writer);
        } catch (IOException except) {
            // Should never happen on String output
            throw new XMLIOError(except);
        }
        return writer.toString();
    }

    /**
     * Format a document or any node and its children into a byte array using
     * the current encoding.
     */
    public byte[] toBytes(Node node) throws IOException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        write(node, out);
        return out.toByteArray();
    }

    /**
     * Output a document or any node and its children to a Writer.
     */
    public void write(Node node,
                      Writer writer) throws IOException {
        Formatter formatter = getFormatter(node);
        formatter.write(options, node, writer);
    }

    /**
     * Output a document or any node and its children to a OutputStream.
     */
    public void write(Node node,
                      OutputStream out) throws IOException {
        Formatter formatter = getFormatter(node);
        String encoding = options.getJavaEncoding();
        if (encoding == null) {
            encoding = formatter.getDefaultJavaEncoding();
        }
        Writer writer = new OutputStreamWriter(out, encoding);
        formatter.write(options, node, writer);
        writer.flush();
    }

    /**
     * Output a document or any node and its children to a File.
     */
    public void write(Node node,
                      File out) throws IOException {

        OutputStream outStream
            = new BufferedOutputStream(new FileOutputStream(out));
        try {
            write(node, outStream);
        } finally {
            outStream.close();
        }
    }
}
