/*
 * Enhydra Java Application Server
 * The Initial Developer of the Original Code is Lutris Technologies Inc.
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc.
 * All Rights Reserved.
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License.
 *
 * $Id: HTMLFormatter.java,v 1.5.2.5 2000/04/02 00:56:50 markd Exp $
 */

package org.enhydra.xml.io;

import java.util.Hashtable;
import java.io.Writer;
import java.io.IOException;
import org.w3c.dom.*;
import org.w3c.dom.html.*;
import org.enhydra.xml.xmlc.html.HTMLEntities;

/**
 * Class for formatting an HTML DOM into an HTML page.
 */
class HTMLFormatter implements Formatter {
    /**
     * List and table of element classes where closing elements are forbidden
     * (or in the <P> case, just not desired).
     */
    private static final String noCloseTagsList[] = {
        "AREA", "BASE", "BASEFONT", "BR", "COL", "FRAME", "HR", "IMG",
        "INPUT", "ISINDEX", "LINK", "META", "PARAM", "P"
    };
    private static final Hashtable noCloseTags = new Hashtable();

    /**
     * List of attributes that do not have values.  If the attribute
     * exists and is empty, we output it without a value.
     */
    private static final String noValueAttrsList[] = {
        "checked", "compact", "declare", "defer", "disabled", "ismap",
        "nohref", "noresize", "noshade", "nowrap", "readonly", "selected"
    };
    private static final Hashtable noValueAttrs = new Hashtable();

    /**
     * Platform line separator.
     */
    private static final String newLine;

    /**
     * The output options.
     */
    private OutputOptions options;

    /**
     * Static constructor.
     */
    static {
        newLine = System.getProperty("line.separator");
        if (newLine == null) {
            throw new XMLIOError("System property line.separator not found");
        }

        for (int idx = 0; idx < noCloseTagsList.length; idx++) {
            noCloseTags.put(noCloseTagsList[idx], noCloseTagsList[idx]);
        }
        for (int idx = 0; idx < noValueAttrsList.length; idx++) {
            noValueAttrs.put(noValueAttrsList[idx], noValueAttrsList[idx]);
        }
    }

    /**
     * Write an attribute value.  Encode quoting character and
     * expand entities.
     */
    private void writeAttributeValue(String value,
                                     Writer out) throws IOException {
        out.write("='");
        for (int idx = 0; idx < value.length(); idx++) {
            char ch = value.charAt(idx);
            if (ch == '\'') {
                out.write("&#39;");
            } else {
                textCharToString(value.charAt(idx), out);
            }
        }
        out.write("'");
    }

    /**
     * Append an attribute and value to a buffer.  If the attribute value
     * contains a single quote, then encode it.  This is also done at compile
     * time, however this is needed to handle values set at runtime.
     */
    private void writeAttribute(Attr attr,
                                Writer out) throws IOException {
        String name = attr.getName();
        String value = attr.getValue();

        out.write(" ");
        out.write(name);

        /*
         * Attributes that don't normally have values only get them if
         * one was explictly supplied.
         */
        if ((value.length() > 0)
            || !noValueAttrs.containsKey(name.toLowerCase())) {
            writeAttributeValue(value, out);
        }
    }

    /**
     * Output a Comment node.
     */
    private void commentToString(Comment comment,
                                 Writer out) throws IOException {
        out.write("<!--");
        out.write(comment.getData());
        out.write("-->");
    }

    /**
     * Output an element node.  This optionally corrects problem cases for
     * browsers:
     *
     * <UL>
     * <LI> ID attributes are dropped from SPAN tags.  This cause Internet
     * Explorer 4.0 to get confused on keep-alive connections.
     * </UL>
     */
    private void elementToString(Element elem,
                                 Writer out) throws IOException {

        String tagName = elem.getTagName();
        boolean dropId = options.getDropHtmlSpanIds()
            && tagName.equalsIgnoreCase("span");

        // open tag start
        out.write('<');
        out.write(tagName);

        // Attributes.
        NamedNodeMap attrs = elem.getAttributes();
        if (attrs != null) {
            for (int idx = 0; idx < attrs.getLength(); idx++) {
                Attr attr = (Attr)attrs.item(idx);
                if (!(dropId && attr.getName().equalsIgnoreCase("id"))) {
                    writeAttribute(attr, out);
                }
            }
        }
        out.write('>');

        nodeChildrenToString(elem, out);

        // Output end tag when legal.
        if (!noCloseTags.containsKey(elem.getTagName())) {
            out.write("</");
            out.write(tagName);
            out.write('>');
        }
    }

    /**
     * Output an EntityReference node.
     */
    private void entityReferenceToString(EntityReference entityRef,
                                         Writer out) throws IOException {
        out.write('&');
        out.write(entityRef.getNodeName());
        out.write(';');
    }

    /**
     * Output a text character to the approriate representation.
     */
    private void textCharToString(char textChar,
                                  Writer out) throws IOException {
        String charStr =

                // This range test improves the speed of this routine about 20%
                (textChar >= 'A' && textChar <= 'z') ? null :

                HTMLEntities.charToEntity(textChar);

        if (charStr != null) {
            out.write(charStr);
        } else {
            out.write(textChar);
        }
    }

    /**
     * Output a Text node.
     */
    private void textToString(Text text,
                              Writer out) throws IOException {
        String data = text.getData();
        int len = data.length();
        for (int idx = 0; idx < len; idx++) {
            textCharToString(data.charAt(idx), out);
        }
    }

    /**
     * Recursively convert children of a DOM node hierarchy to a string.
     */
    private void nodeChildrenToString(Node node,
                                      Writer out) throws IOException {
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            writeDom(child, out);
        }
    }

    /**
     * Recursively wreite a DOM hierarchy.
     */
    private void writeDom(Node node,
                          Writer out) throws IOException {
        switch (node.getNodeType()) {
        case Node.COMMENT_NODE:
            commentToString((Comment)node, out);
            break;
        case Node.DOCUMENT_FRAGMENT_NODE:
            nodeChildrenToString(node, out);
            break;
        case Node.DOCUMENT_NODE:
            nodeChildrenToString(node, out);
            break;
        case Node.ELEMENT_NODE:
            elementToString((Element)node, out);
            break;
        case Node.ENTITY_REFERENCE_NODE:
            entityReferenceToString((EntityReference)node, out);
            break;
        case Node.TEXT_NODE:
            textToString((Text)node, out);
            break;
        case Node.CDATA_SECTION_NODE:
            out.write(((CDATASection)node).getData());
            break;
        case Node.DOCUMENT_TYPE_NODE :
        case Node.NOTATION_NODE:
        case Node.PROCESSING_INSTRUCTION_NODE:
        case Node.ENTITY_NODE:
        default:
            throw new XMLIOError("Node of this type not supported in HTML: "
                                 + node.getClass().getName());
        }
    }

    /**
     * Get the default Java encoding.
     * @see Formatter#getDefaultEncoding
     */
    public String getDefaultJavaEncoding() {
        return "ISO8859_1";
    }

    /**
     * Format the document to the specified writer.
     * @see Formatter#write
     */
    public void write(OutputOptions outputOptions,
                      Document document,
                      Writer writer) throws IOException {
        options = outputOptions;
        writeDom(document, writer);
        writer.write(newLine);
    }

    /**
     * Format a Node and children to the specified writer.
     * @see Formatter#write
     */
    public void write(OutputOptions outputOptions,
                      Node node,
                      Writer writer) throws IOException {
        options = outputOptions;
        writeDom(node, writer);
        writer.write(newLine);
    }
}
