/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: OutputOptions.java,v 1.8.2.2 2000/03/15 08:51:39 markd Exp $
 */
package org.enhydra.xml.io;

import java.util.Hashtable;
import org.apache.xerces.readers.MIME2Java;

/**
 * Object that specifis how a XML file will be formatted.
 */
public class OutputOptions {
    /**
     * The Java encoding to use.
     */
    private String javaEncoding = null;

    /*
     * The XML encoding to use.
     */
    private String xmlEncoding = null;

    /**
     * Should the document be indented?
     */
    private boolean indentingEnabled = false;

    /**
     * Number of characters to indent.
     */
    private int indentSize = 4;

    /**
     * Line width to break long lines if indenting is enabled.
     */
    private int lineWidth = 72;

    /**
     * Is space preserved where not otherwise specified by
     * the document?
     */
    private boolean preserveSpace = true;

    /**
     * Don't add the encoding to the XML header.
     */
    private boolean omitEncoding = false;

    /**
     * MS Internet Explorer 4.0 get very confused about keep-alive connections
     * if HTML SPAN tags have ID attributes.  Until we find another workaround,
     * this removes IDs from SPAN tags.
     */
    private boolean dropHtmlSpanIds = false;

    /**
     * Get the Java encoding.
     * @return The encoding or null if not specified.
     */
    public String getJavaEncoding() {
        return javaEncoding;
    }

    /**
     * Set the encoding using the Java encoding name.  This sets the
     * corresponding XML encoding name.
     *
     * @param newJavaEncoding the new Java encoding or null to set to default.
     * @exception XMLIOError if the encoding can't be converted to a XML
     *  encoding.
     */
    public void setJavaEncoding(String newJavaEncoding) {
        if (newJavaEncoding == null) {
            javaEncoding = null;
            xmlEncoding = null;
        } else {
            String newXmlEncoding = Encoding.java2XmlEncoding(newJavaEncoding);
            if (newXmlEncoding == null) {
                throw new XMLIOError("Java encoding \"" + newJavaEncoding
                                     + "\" doesn't map to a XML encoding");
            }
            javaEncoding = newJavaEncoding;
            xmlEncoding = newXmlEncoding;
        }
    }

    /**
     * Get the XML encoding.
     * @return The encoding or null if not specified.
     */
    public String getXmlEncoding() {
        return xmlEncoding;
    }

    /**
     * Set the encoding using the XML encoding name.  This sets the
     * corresponding Java encoding name.
     *
     * @param newXmlEncoding the new XML encoding or null to set to default.
     * @exception XMLIOError if the encoding can't be converted to a Java
     *  encoding.
     */
    public void setXmlEncoding(String newXmlEncoding) {
        if (newXmlEncoding == null) {
            javaEncoding = null;
            xmlEncoding = null;
        } else {
            String newJavaEncoding = Encoding.xml2JavaEncoding(newXmlEncoding);
            if (newJavaEncoding == null) {
                throw new XMLIOError("XML encoding \"" + newXmlEncoding
                                     + "\" doesn't map to a Java encoding");
            }
            javaEncoding = newJavaEncoding;
            xmlEncoding = newXmlEncoding;
        }
    }

    /**
     * Set both the XML and Java encodings.  This allows setting the 
     * encodings were a mapping is not defined.
     *
     * @param newJavaEncoding the new Java encoding or null to set to default.
     * @param newXmlEncoding the new XML encoding or null to set to default.
     * @exception XMLIOError If one of the encodings, but not both are null.
     */
    public void setXmlEncoding(String newJavaEncoding,
                               String newXmlEncoding) {
        if ((newJavaEncoding == null) && (newXmlEncoding != null)
            || (newJavaEncoding != null) && (newXmlEncoding == null)) {
            throw new XMLIOError("If one of the Java or XML encodings is null, the other must be null");
        }
        javaEncoding = newJavaEncoding;
        xmlEncoding = newXmlEncoding;
    }

    /**
     * Get the maximum value of a character in the specified XML encoding.
     */
    private int maxXmlCharacterValue(String encoding) {
        // FIXME: not used..
        String encodingUp = encoding.toUpperCase();

        // FIXME: This is very incomplete...
        if (encodingUp.equals("ASCII")
            || encodingUp.equals("US-ASCII")) {
            return 0x7F;
        }
        if (encodingUp.startsWith("ISO-8859")) {
            return 0xFF;
        }
        return 0xFFFF;
    }
    

    /**
     * Get indentation enabled state.
     * @return true if enable, false to disable.  The default is disabled.
     */
    public boolean getIndenting() {
        return indentingEnabled;
    }

    /**
     * Enable or disable indentation.
     * @param enable true to enable, false to disable.
     */
    public void setIndenting(boolean enable) {
        indentingEnabled = enable;
    }

    /**
     * Get indentation size.
     * @return Number of characters to indent at each level.
     */
    public int getIndentSize() {
        return indentSize;
    }

    /**
     * Set indentation size.
     * @param size Number of characters to indent at each level.
     */
    public void setIndentSize(int size) {
        indentSize = size;
        if (indentSize < 0) {
            indentSize = 0;
        }
    }

    /**
     * Get the width at while long lines are split when indentation
     * is enabled.  The default is 72.
     */
    public int getLineWidth() {
        return lineWidth;
    }

    /**
     * Set the width at while long lines are split when indentation
     * is enabled.
     * @param width the new line width
     */
    public void setLineWidth(int width) {
        lineWidth = width;
        if (lineWidth < 0) {
            lineWidth = 0;
        }
    }

    /**
     * Get the default space-preservation flag.
     * @return true if preserving space where not otherwise specified by
     *  the document, false otherwise.  The default is true.
     */
    public boolean getPreserveSpace() {
        return preserveSpace;
    }

    /**
     * Set the default space-preservation flag.
     * @param perserve true if preserving space where not otherwise specified
     *  by the document
     */
    public void setPreserveSpace(boolean preserve) {
        preserveSpace = preserve;
    }

    /**
     * Get flag indicating if encoding should be omitted from the XML header.
     */
    public boolean getOmitEncoding() {
        return omitEncoding;
    }

    /**
     * Set flag indicating if encoding should be omitted from the XML header.
     * This is provided as a hack for WML.  Several devices need ASCII
     * encoding but can't handle the header.
     */
    public void setOmitEncoding(boolean omit) {
        omitEncoding = omit;
    }

    /**
     * Get the drop HTML SPAN element ids flag.
     */
    public boolean getDropHtmlSpanIds() {
        return dropHtmlSpanIds;
    }

    /**
     * Set the drop HTML SPAN element ids flag.
     */
    public void setDropHtmlSpanIds(boolean drop) {
        dropHtmlSpanIds = drop;
    }

    /**
     * Convert to a string for debugging.
     */
    public String toString() {
        return "javaEncoding=" + javaEncoding
            + ", xmlEncoding=" + xmlEncoding
            + ", indentingEnabled=" + indentingEnabled
            + ", indentSize=" + indentSize
            + ", lineWidth=" + lineWidth 
            + ", preserveSpace=" + preserveSpace
            + ", omitEncoding=" + omitEncoding
            + ", dropHtmlSpanIds=" +  dropHtmlSpanIds;
    }
}
