/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLCStdFactory.java,v 1.1 2000/02/15 07:28:31 markd Exp $
 */

package org.enhydra.xml.xmlc;

import com.lutris.logging.LogChannel;
import com.lutris.logging.LogWriter;
import com.lutris.logging.Logger;

/**
 * XMLC standard factory class. Used for creating instances of
 * XMLC-generated classes.  This is a default factory used when
 * no special support, such as auto-recompilation is required.
 */
public class XMLCStdFactory implements XMLCFactory {
    /**
     * Class loader that is used to load the first reference to a class.
     */
    private ClassLoader defaultClassLoader;

    /**
     * Log writer for XMLC informative logging or null if not enabled.
     */
    private LogWriter xmlcLog;

    /**
     * Log writer for XMLC debug logging or null if not enabled.
     */
    private LogWriter xmlcDebugLog;

    /**
     * Log writer for XMLC error logging or null if no logging.
     */
    private LogWriter errorLog;

    /**
     * Constructor.
     * @param classLoader Classloader used to load classes when a class name
     *  is specified.  If null, the system classload is used.
     * @param logChannel The log channel to use to log events and errors.  If
     *  null, no logging is done.  Level XMLC is used to log reloading and
     *  recompilation.  Level ERROR is used for errors.
     */
    public XMLCStdFactory(ClassLoader classLoader,
                          LogChannel logChannel) {
        defaultClassLoader = classLoader;
        if (logChannel != null) {
            if (logChannel.isEnabled(XMLC_LOG_LEVEL)) {
                xmlcLog = logChannel.getLogWriter(XMLC_LOG_LEVEL);
            }
            if (logChannel.isEnabled(XMLC_DEBUG_LOG_LEVEL)) {
                xmlcDebugLog = logChannel.getLogWriter(XMLC_DEBUG_LOG_LEVEL);
            }
            errorLog = logChannel.getLogWriter(Logger.ERROR);
        }
    }

    /**
     * Handle all errors, logging if enabled.  Also handles logging
     * error output of the compiler.
     * 
     * @param className Class being loaded.
     * @param except Can be any error or exception.
     * @return Returns runtime exceptions to be rethrown.  Could throw
     * everything, but catch blocks appear to fall through to the compiler so
     * we would have to put in dummy returns.
     */
    private Error handleError(String className,
                              Throwable except) {
        if (errorLog != null) {
            errorLog.println("Factory creation of an XMLC class failed: \""
                             + className + "\"", except);
        }
        
        // Errors just continue on.
        if (except instanceof Error) {
            return (Error)except;
        }
        
        // All other exceptions get wrapped.
        return new XMLCError("Couldn't load XMLC class", except);
    }

    /**
     * Do actualy work of creating a new object.
     */
    private XMLObject doCreate(Class xmlcBasedClass)
            throws ClassNotFoundException, IllegalAccessException, InstantiationException {
        if (xmlcBasedClass.isInterface()) {
            // Convert interface to class name.
            return doCreate(xmlcBasedClass.getName() + "Impl",
                            defaultClassLoader);
        } else {
            if (!XMLObject.class.isAssignableFrom(xmlcBasedClass)) {
                throw new XMLCError("class \"" + xmlcBasedClass.getName()
                                    + "\" does not implement \""
                                    + XMLObject.class.getName() + "\"");
            }
            return (XMLObject)(xmlcBasedClass.newInstance());
        }
    }

    /**
     * Do actualy work of creating a new object given a name.
     */
    private XMLObject doCreate(String xmlcClassName,
                               ClassLoader classLoader)
            throws ClassNotFoundException, IllegalAccessException, InstantiationException {
        if (classLoader != null) {
            return doCreate(classLoader.loadClass(xmlcClassName));
        } else {
            return doCreate(Class.forName(xmlcClassName));
        }
    }

    /**
     * Create an instance of a XMLC-generated class or a class derived from
     * one, given either a class or an interface object, using the associated
     * classloader
     */
    public XMLObject create(String xmlcClassName) {
        try {
            return doCreate(xmlcClassName, defaultClassLoader);
        } catch (Throwable except) {
            throw handleError(xmlcClassName, except);
        }
    }

    /**
     * Create an instance of a XMLC-generated class or a class derived
     * from one, given the name of the class, using the associated
     * classloader
     */
    public XMLObject create(Class xmlcBasedClass) {
        try {
            if (xmlcBasedClass.isInterface()) {
                // Convert interface to class name.
                return doCreate(xmlcBasedClass.getName() + "Impl",
                                xmlcBasedClass.getClassLoader());
            } else {
                return doCreate(xmlcBasedClass);
            }
        } catch (Throwable except) {
            throw handleError(xmlcBasedClass.getName(), except);
        }
    }
}
