/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLCUtil.java,v 1.4.2.1 2000/03/15 08:51:40 markd Exp $
 */

package org.enhydra.xml.xmlc;

import org.w3c.dom.*;
import java.io.*;
import java.util.Vector;
import org.enhydra.xml.xmlc.compiler.JavaLang;

/**
 * Utility methods for dealing with DOMs.
 */
public abstract class XMLCUtil {
    /**
     * Option to printNode to print the contents of Comment nodes.
     */
    public static final int PRINT_COMMENT = 0x01;

    /**
     * Option to printNode to print the contents of Text nodes.
     */
    public static final int PRINT_TEXT = 0x02;

    /**
     * Option to printNode to print the contents of Text nodes.
     */
    public static final int PRINT_CDATA = 0x04;

    /**
     * Option to printNode to print the contents of DocumentType nodes.
     */
    public static final int PRINT_DOCUMENTTYPE = 0x08;

    /**
     * All Print options.
     */
    public static final int PRINT_ALL = 0xff;

    /**
     * Default print options.
     */
    public static final int PRINT_DEFAULT = 
        (PRINT_COMMENT | PRINT_TEXT | PRINT_CDATA);

    /**
     * Find an attribute of a node by name.
     *
     * @param node Node who's attribute is desired.
     * @param name The name of the desired attribute.
     * @return The attribute or null if its not found.
     */
    public static Attr getAttributeByName(Node node, String name) {
        NamedNodeMap nodeMap = node.getAttributes();
        if (nodeMap == null) {
            return null;
        } else {
            return (Attr)nodeMap.getNamedItem(name);
        } 
    }
    /**
     * Find the first text descendent node of an element.
     * This recursively looks more than one level to search
     * for text in font nodes, etc.  If a text node is not
     * found, return null
     *
     * @param node The starting node for the search.
     * @return The text node or null if not found.
     */
    public static Text findFirstText(Node node) {
        if (node instanceof Text) {
            return (Text)node;
        }
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            Text text = findFirstText(child);
            if (text != null) {
                return text;
            }
        }
        return null;
    }

    /**
     * Find the first text descendent node of an element.
     * This recursively looks more than one level to search
     * for text in font nodes, etc.  If a text node is not
     * found, generate an error.
     *
     * @param node The starting node for the search.
     * @return The text node.
     * @exception XMLCError If a text node was not found.
     */
    public static Text getFirstText(Node node) {
        Text text = findFirstText(node);
        if (text == null) {
            String msg = "No child text mode found for element";
            Attr id = getAttributeByName(node, "id");
            if (id != null) {
                msg += "; id=\"" + id.getValue() + "\"";
            }
            throw new XMLCError(msg);
        }
        return text;
    }

    /**
     * Recursively search for an element by id starting at a node.
     * Assumes id attribute is named "id".  With XMLC, it normally
     * better to access elements in a compiled objects directly with
     * the generated access methods.  This is provided for working on
     * cloned pieces of the document or other tasks outside of the XMLC
     * generated object.
     *
     * @param id The element id to find.  Case is ignored.
     * @param node The node to start the search at.
     * @return A point the element if found, otherwise null.
     */
    public static Element getElementById(String id, Node node) {
        if (node instanceof Element) {
            Element elem = (Element)node;
            Attr nodeId = getAttributeByName(elem, "id");
            if ((nodeId != null) && nodeId.getValue().equalsIgnoreCase(id)) {
                return (Element)node;
            }
        }

        // Search children
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            Element childElem = getElementById(id, child);
            if (childElem != null) {
                return childElem;
            }
        }
        return null;
    }

    /**
     * Recursively search for an required element by id starting at a node.
     *
     * @param id The element id to find.  Case is ignored.
     * @param node The node to start the search at.
     * @return A point the element if found.
     * @exception XMLCError Thrown if the element is not found.
     * @see #getElementById
     */
    public static Element getRequiredElementById(String id, Node node) {
        Element elem = getElementById(id, node);
        if (elem == null) {
            // Not found, generate an error. 
            String msg = "No element found for id \"" + id + "\"";
            Attr nodeId = getAttributeByName(node, "id");
            if (nodeId != null) {
                msg += "; starting at id=\"" + nodeId.getValue() + "\"";
            }
            throw new XMLCError(msg);
        } else {
            return elem;
        }
    }

    /**
     * Print indentation.
     */
    private static void printIndent(int indent,
                                    PrintWriter out) {
        for (int i = 0; i < indent; i++) {
            out.print("    ");
        }
    }

    /*
     * Print a label and value if the value is not null or empty.
     */
    static void printLabelValue(String label, String value, PrintWriter out) {
        if ((value != null) && (value.length() > 0)) {
            out.print(" ");
            out.print(label);
            out.print("=");
            out.print(value);
        }
    }

    /**
     * Generate a order list of attributes, based on attribute name.
     */
    private static Attr[] sortAttrs(NamedNodeMap attrList) {
        Vector attrs = new Vector();

        // Crude insertion sort.
        for (int i = 0; i < attrList.getLength(); i++) {
            Node node = attrList.item(i);
            String name = node.getNodeName();
            // Find place to insert.
            int j;
            for (j = 0; j < attrs.size(); j++) {
                if (name.compareTo(((Node)attrs.elementAt(j)).getNodeName()) <= 0) {
                    break;
                }
            }
            attrs.insertElementAt(node, j);
        }
        Attr[] retAttrs = new Attr[attrList.getLength()];
        attrs.copyInto(retAttrs);
        return retAttrs;
    }

    /**
     * Generate string of attribute names and values, sorted by name.
     *
     * @param node Get the attributes of this node.
     * @param all If true, get all attributes, otherwise get only specified
     *  ones.
     * @return A printable string.
     */
    public static String dumpAttrs(Node node, boolean all) {
        NamedNodeMap attrList = node.getAttributes();
        if (attrList == null) {
            return "";
        }
        StringBuffer buf = new StringBuffer();
        Attr[] attrs = sortAttrs(attrList);

        for (int idx = 0; idx < attrs.length; idx++) {
            if (idx > 0) {
                buf.append(" ");
            }
            if (attrs[idx].getSpecified() || all) {
                buf.append(attrs[idx].getName() + "='"
                           + attrs[idx].getValue() + "'");
            }
        }
        return buf.toString();
    }

    /**
     * Print information about a Document Node.
     */
    private static void printDocumentInfo(Document document,
                                          int options,
                                          int indent,
                                          PrintWriter out) {
        if ((options & PRINT_DOCUMENTTYPE) != 0) {
            DocumentType docType = document.getDoctype();
            if (docType == null) {
                out.println("null DocumentType");
            } else {
                out.println();
                printIndent(indent+1, out);
                out.println("Document type =>");
                recursivePrintNode(docType, options, indent+2, out);
            }
        } else {
            out.println();
        }
    }

    /**
     * Print information about a DocumentType Node.
     */
    private static void printDocumentTypeInfo(DocumentType documentType,
                                              int options,
                                              int indent,
                                              PrintWriter out) {
        printLabelValue("name", documentType.getName(), out);

        if ((options & PRINT_DOCUMENTTYPE) != 0) {
            out.println();
            printIndent(indent+1, out);
            NamedNodeMap entities = documentType.getEntities();
            out.println("Entities =>");
            if (entities != null) {
                for (int idx = 0; idx < entities.getLength(); idx++) {
                    recursivePrintNode(entities.item(idx), options,
                                       indent+2, out);
                }
            }

            printIndent(indent+1, out);
            NamedNodeMap notations = documentType.getNotations();
            out.println("Notations =>");
            if (notations != null) {
                for (int idx = 0; idx < notations.getLength(); idx++) {
                    recursivePrintNode(notations.item(idx), options,
                                       indent+2, out);
                }
            }
        }
    }

    /**
     * Print information about an Entity Node.
     */
    private static void printEntityInfo(Entity entity,
                                        int options,
                                        PrintWriter out) {
        printLabelValue("name", entity.getNodeName(), out);
        printLabelValue("value", entity.getNodeValue(), out);
        printLabelValue("notationName", entity.getNotationName(), out);
        printLabelValue("systemId", entity.getSystemId(), out);
        printLabelValue("publicId", entity.getPublicId(), out);
        out.println();
    }

    /**
     * Print information about an EntityReference Node.
     */
    private static void printEntityReferenceInfo(EntityReference entityRef,
                                                 int options,
                                                 PrintWriter out) {
        printLabelValue("name", entityRef.getNodeName(), out);
        printLabelValue("value", entityRef.getNodeValue(), out);
        out.println();
    }

    /**
     * Print information about a Notation Node.
     */
    private static void printNotationInfo(Notation notation,
                                          int options,
                                          PrintWriter out) {
        printLabelValue("systemId", notation.getSystemId(), out);
        printLabelValue("publicId", notation.getPublicId(), out);
        out.println();
    }

    /**
     * Print information about a ProcessingInstruction Node.
     */
    private static void printProcessingInstructionInfo(ProcessingInstruction pi,
                                                       int options,
                                                       PrintWriter out) {
        printLabelValue("target", pi.getTarget(), out);
        printLabelValue("data", pi.getData(), out);
        out.println();
    }

    /**
     * Print information about an Element Node.
     */
    private static void printElementInfo(Element element,
                                         int options,
                                         PrintWriter out) {
        out.print(" ");
        out.print(element.getTagName());
        String attrStr = dumpAttrs(element, false);
        if (attrStr.length() > 0) {
            out.print(": ");
            out.print(attrStr);
        }
        out.println();
    }
    
    /**
     * Print information about CDATASection nodes.
     */
    private static void printCDATASectionInfo(CDATASection cdata,
                                              int options,
                                              PrintWriter out) {
        if ((options & PRINT_CDATA) != 0) {
            out.print(" ");
            out.print(cdata.getData());
        }
        out.println();
    }
    
    /**
     * Print information about Comment nodes.
     */
    private static void printCommentInfo(Comment comment,
                                         int options,
                                         PrintWriter out) {
        if ((options & PRINT_COMMENT) != 0) {
            out.print(" ");
            out.print(comment.getData());
        }
        out.println();
    }
    
    /**
     * Print information about Text nodes.
     */
    private static void printTextInfo(Text text,
                                      int options,
                                      PrintWriter out) {
        if ((options & PRINT_TEXT) != 0) {
            out.print(" ");
            out.print(text.getData());
        }
        out.println();
    }
    
    /**
     * Recursively print a node.
     */
    private static void recursivePrintNode(Node node,
                                           int options,
                                           int indent,
                                           PrintWriter out) {
        printIndent(indent, out);
        
        // Get class name (without package).  Make a crude attempt to convert
        // it to the interface name.
        String className = JavaLang.simpleClassName(node.getClass().getName());
        if (className.endsWith("Impl")) {
            className = className.substring(0, className.length()-4);
        }

        out.print(className);
        out.print(":");
        switch (node.getNodeType()) {
        case Node.ATTRIBUTE_NODE:
            // Never handled directly here.
            break;
        case Node.ELEMENT_NODE:
            printElementInfo((Element)node, options, out);
            break;
        case Node.CDATA_SECTION_NODE:
            printCDATASectionInfo((CDATASection)node, options, out);
            break;
        case Node.COMMENT_NODE:
            printCommentInfo((Comment)node, options, out);
            break;
        case Node.DOCUMENT_FRAGMENT_NODE:
            out.println();
            break;
        case Node.DOCUMENT_NODE:
            printDocumentInfo((Document)node, options, indent, out);
            break;
        case Node.DOCUMENT_TYPE_NODE:
            printDocumentTypeInfo((DocumentType)node, options, indent, out);
            break;
        case Node.ENTITY_NODE:
            printEntityInfo((Entity)node, options, out);
            break;
        case Node.ENTITY_REFERENCE_NODE:
            printEntityReferenceInfo((EntityReference)node, options, out);
            break;
        case Node.NOTATION_NODE:
            printNotationInfo((Notation)node, options, out);
            break;
        case Node.PROCESSING_INSTRUCTION_NODE:
            printProcessingInstructionInfo((ProcessingInstruction)node, options, out);
            break;
        case Node.TEXT_NODE:
            printTextInfo((Text)node, options, out);
            break;
        default:
            out.println("unknown node type");
        }

        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            recursivePrintNode(child, options, indent+1, out);
        }
    }

    /**
     * Print a node and its children.
     *
     * @param msg A message to print at the start.  If null or empty,
     *  don't print a message.
     * @param node Top of the tree.
     * @param options Set of print options.
     * @param out Output writer.
     */
    public static void printNode(String msg, Node node,
                                 int options,
                                 PrintWriter out) {
        if ((msg != null) && (msg.length() > 0)) {
            out.println(msg + ":");
        }
        if (node == null) {
            out.println(" null node");
        } else {
            recursivePrintNode(node, options, 1, out);
        }
    }

    /**
     * Print a node an its children.
     *
     * @param msg A message to print at the start.  If null or empty,
     *  don't print a message.
     * @param node Top of the tree.
     * @param out Output writer.
     */
    public static void printNode(String msg, Node node,
                                 PrintWriter out) {
        printNode(msg, node, PRINT_DEFAULT, out);
        out.flush();
    }

    /**
     * Print a node an its children.
     *
     * @param msg A message to print at the start.  If null or empty,
     *  don't print a message.
     * @param node Top of the tree.
     * @param out Output stream.
     */
    public static void printNode(String msg, Node node,
                                 OutputStream out) {
        PrintWriter writer = new PrintWriter(out);
        printNode(msg, node, PRINT_DEFAULT, writer);
        writer.flush();
    }

    /**
     * Copy a node and its children from one document to another.
     *
     * @param srcNode The node to copy.
     * @param destDocument The destination document.  The node will 
     *  belong to this document but will not be inserted in it.
     * @return The new node.
     * @deprecated use Document.importNode
     * @see Document#importNode
     */
    public static Node copyNode(Node srcNode, Document destDocument) {
        return destDocument.importNode(srcNode, true);
    }

    /**
     * Replace a node with a given one from another document.
     * @param srcElement The node to clone and insert 
     * @param destElement The node to be replaced
     * @return the new node that replaces the destination node
     */
    public static Node replaceNode(Node srcNode, Node destNode) {
        Node node = destNode.getOwnerDocument().importNode(srcNode, true);
        Node parent = destNode.getParentNode(); 
        if (parent!=null){
            parent.replaceChild(node,destNode);
        }
        return node;
    }

}
