/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLObject.java,v 1.7 2000/03/02 07:47:26 markd Exp $
 */

package org.enhydra.xml.xmlc;

import org.w3c.dom.*;

/**
 * Interface for all compiled XML objects.
 */
public interface XMLObject extends Document {
    /**
     * The name of the class field that is used to identify an XMLC
     * generated class.  The field will contain a reference to the 
     * Class object for the class.  This is used to find the XMLC generated
     * class in an inheritance hierarchy.
     */
    public static final String XMLC_GENERATED_CLASS_FIELD_NAME
        = "XMLC_GENERATED_CLASS";

    /**
     * Field name in the generated class that contains the 
     * name of the source file.  This is used to find
     * the source file for recompilation.  It is a relative
     * file path, normally the name of the file within the
     * package containing the class.
     */
    public static final String XMLC_SOURCE_FILE_FIELD_NAME
        = "XMLC_SOURCE_FILE";

    /**
     * Generated method to build the DOM.  This is normally call by the
     * constructor.  However, one of the generated constructors takes a boolean
     * flag to indicate if this should be called automatically.  This method
     * may also be called to reinitialize the object to its empty state.
     */
    public void buildDocument();

    /**
     * Set the delegate object.  Delegation is used to support automatic
     * recompilation of documents into XMLC objects.  If the delegate is
     * not null, the methods of the delegate are called to handle most
     * of the methods of this object.
     *
     * @param delegate The delegate object, or null to remove delegation.
     * The delegate must implement the same interface as the derived, generate
     * object.
     */
    public void setDelegate(XMLObject delegate);

    /**
     * Initialize the fields used by the generated access methods from the
     * current state of the document.  Missing DOM element ids will result in
     * their acccess method returning <CODE>null</CODE>.
     */
    public void syncAccessMethods();

    /**
     * Get the delegate.
     * @return The delegate object, or null if there is no delegate. 
     */ 
    public XMLObject getDelegate();

    /**
     * Get the actual document object.  One should normally just use the
     * XMLObject methods to access the Document functionality, this is for the
     * initialization of derived objects.
     */
    public Document getDocument();

    /**
     * Get the encoding specified in the document.  Note that this is the
     * symbolic name of the XML encoding, which is not the same as the
     * Java encoding names.
     *
     * @return the encoding or null if not encoding was explicitly
     *  specified.
     * @see org.enhydra.xml.io.DOMFormatter
     * @see org.enhydra.xml.io.OutputOptions
     */
    public String getEncoding();

    /**
     * Convert the document to a string representation of the document, that
     * is a string containing XML.  The results are parsable into the same
     * DOM hierarchy.  The formatting provided by this method does not
     * begin to cover all of the issues involved with publishing
     * a XML document, such as character encoding. Use the
     * {@link see org.enhydra.xml.io.DOMFormatter DOMFormatter} class if
     * more options are required.
     *
     * @return A string containing the full XML.
     * @see org.enhydra.xml.io.DOMFormatter
     */
    public String toDocument();
}
