/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLObjectImpl.java,v 1.8 2000/02/24 09:22:22 markd Exp $
 */

package org.enhydra.xml.xmlc;

import com.lutris.util.ChainedException;
import org.w3c.dom.*;
import java.io.StringWriter;
import org.enhydra.xml.io.DOMFormatter;

/**
 * Base class for all compiled XML objects.
 * @see XMLObject
 */
public abstract class XMLObjectImpl implements XMLObject, Document {
    /**
     * Default XML formatting object for use by toDocument().  Initialized
     * on first use and shared
     */
    private static DOMFormatter formatter = null;

    /**
     * The document the object is associated with.
     */
    private Document document;

    /**
     * The encoding specified in the document.
     */
    private String encoding;

    /**
     * The delegate object.
     */
    private XMLObject delegate;

    /**
     * Constructor. The setDocument() method must be called to
     * associate a document with this object.
     */
    protected XMLObjectImpl() {
    }

    /**
     * Set the DOM document associated with this object.  This is
     * used by buildDocument() to set the new document.  It is done
     * separately from the constructor to allow buildDocument() to
     * not be called immediatly.
     */
    protected void setDocument(Document doc) {
        document = doc;
    }

    /**
     * Set the encoding.
     */
    protected void setEncoding(String enc) {
        encoding = enc;
    }

    /**
     * @see XMLObject#getDocument
     */
    public Document getDocument() {
        if (delegate != null) {
            return delegate.getDocument();
        } else {
            return document;
        }
    }

    /**
     * @see XMLObject#getEncoding
     */
    public String getEncoding() {
        if (delegate != null) {
            return delegate.getEncoding();
        } else {
            return encoding;
        }
    }

    /**
     * @see XMLObject#setDelegate
     */ 
    public void setDelegate(XMLObject delegate) {
        this.delegate = delegate;
    }

    /**
     * @see XMLObject#getDelegate
     */ 
    public XMLObject getDelegate() {
        return delegate;
    }

    /**
     * Check that cloneNode on an entire document is done with the
     * deep option.
     */
    protected void cloneDeepCheck(boolean deep) {
        if (!deep) {
            throw new XMLCError("Must use deep clone when cloning entire document");
        }
    }

    /**
     * Clone the entire document.  Derived objects should override this
     * to get the correct derived type.  Cloning with deep being false
     * is not allowed.
     *
     * @see org.w3c.dom.Node#cloneNode
     */
    abstract public Node cloneNode(boolean deep);

    /**
     * @see org.w3c.dom.Document#getDoctype
     */
    public DocumentType getDoctype() {
        if (delegate != null) {
            return delegate.getDoctype();
        } else {
            return document.getDoctype();
        }
    }

    /**
     * @see org.w3c.dom.Document#getImplementation
     */
    public DOMImplementation getImplementation() {
        if (delegate != null) {
            return delegate.getImplementation();
        } else {
            return document.getImplementation();
        }
    }

    /**
     * @see org.w3c.dom.Document#getDocumentElement
     */
    public Element getDocumentElement() {
        if (delegate != null) {
            return delegate.getDocumentElement();
        } else {
            return document.getDocumentElement();
        }
    }

    /**
     * @see org.w3c.dom.Document#importNode
     */
    public Node importNode(Node importedNode, 
                           boolean deep)
        throws DOMException {
        if (delegate != null) {
            return delegate.importNode(importedNode, deep);
        } else {
            return document.importNode(importedNode, deep);
        }
    }

    /**
     * @see org.w3c.dom.Document#createElement
     */
    public Element createElement(String tagName)
        throws DOMException {
        if (delegate != null) {
            return delegate.createElement(tagName);
        } else {
            return document.createElement(tagName);
        }
    }

    /**
     * @see org.w3c.dom.Document#createElementNS
     */
    public Element createElementNS(String namespaceURI, 
                                   String qualifiedName)
        throws DOMException {
        if (delegate != null) {
            return delegate.createElementNS(namespaceURI, qualifiedName);
        } else {
            return document.createElementNS(namespaceURI, qualifiedName);
        }
    }

    /**
     * @see org.w3c.dom.Document#createDocumentFragment
     */
    public DocumentFragment createDocumentFragment() {
        if (delegate != null) {
            return delegate.createDocumentFragment();
        } else {
            return document.createDocumentFragment();
        }
    }

    /**
     * @see org.w3c.dom.Document#createTextNode
     */
    public Text createTextNode(String data) {
        if (delegate != null) {
            return delegate.createTextNode(data);
        } else {
            return document.createTextNode(data);
        }
    }

    /**
     * @see org.w3c.dom.Document#createComment
     */
    public Comment createComment(String data) {
        if (delegate != null) {
            return delegate.createComment(data);
        } else {
            return document.createComment(data);
        }
    }

    /**
     * @see org.w3c.dom.Document#createCDATASection
     */
    public CDATASection createCDATASection(String data)
        throws DOMException {
        if (delegate != null) {
            return delegate.createCDATASection(data);
        } else {
            return document.createCDATASection(data);
        }
    }

    /**
     * @see org.w3c.dom.Document#createProcessingInstruction
     */
    public ProcessingInstruction createProcessingInstruction(String target, 
                                                             String data)
        throws DOMException {
        if (delegate != null) {
            return delegate.createProcessingInstruction(target, data);
        } else {
            return document.createProcessingInstruction(target, data);
        }
    }

    /**
     * @see org.w3c.dom.Document#createAttribute
     */
    public Attr createAttribute(String qualifiedName)
        throws DOMException {
        if (delegate != null) {
            return delegate.createAttribute(qualifiedName);
        } else {
            return document.createAttribute(qualifiedName);
        }
    }

    /**
     * @see org.w3c.dom.Document#createAttributeNS
     */
    public Attr createAttributeNS(String namespaceURI, 
                                  String qualifiedName)
        throws DOMException {
        if (delegate != null) {
            return delegate.createAttributeNS(namespaceURI, qualifiedName);
        } else {
            return document.createAttributeNS(namespaceURI, qualifiedName);
        }
    }

    /**
     * @see org.w3c.dom.Document#createEntityReference
     */
    public EntityReference createEntityReference(String name)
        throws DOMException {
        if (delegate != null) {
            return delegate.createEntityReference(name);
        } else {
            return document.createEntityReference(name);
        }
    }

    /**
     * @see org.w3c.dom.Document#getElementsByTagName
     */
    public NodeList getElementsByTagName(String tagname) {
        if (delegate != null) {
            return delegate.getElementsByTagName(tagname);
        } else {
            return document.getElementsByTagName(tagname);
        }
    }

    /**
     * @see org.w3c.dom.Document#getElementsByTagNameNS
     */
    public NodeList getElementsByTagNameNS(String namespaceURI, 
                                           String localName) {
        if (delegate != null) {
            return delegate.getElementsByTagNameNS(namespaceURI, localName);
        } else {
            return document.getElementsByTagNameNS(namespaceURI, localName);
        }
    }

    /**
     * @see org.w3c.dom.Document#getElementById
     */
    public Element getElementById(String elementId) {
        if (delegate != null) {
            return delegate.getElementById(elementId);
        } else {
            return document.getElementById(elementId);
        }
    }

    /**
     * @see org.w3c.dom.Node#getNodeName()
     */
    public String getNodeName() {
        if (delegate != null) {
            return delegate.getNodeName();
        } else {
            return document.getNodeName();
        }
    }

    /**
     * @see org.w3c.dom.Node#getNodeValue()
     */
    public String getNodeValue()
        throws DOMException {
        if (delegate != null) {
            return delegate.getNodeValue();
        } else {
            return document.getNodeValue();
        }
    }

    /**
     * @see org.w3c.dom.Node#setNodeValue
     */
    public void setNodeValue(String nodeValue)
        throws DOMException {
        if (delegate != null) {
            delegate.setNodeValue(nodeValue);
        } else {
            document.setNodeValue(nodeValue);
        }
    }

    /**
     * @see org.w3c.dom.Node#getNodeType
     */
    public short getNodeType() {
        if (delegate != null) {
            return delegate.getNodeType();
        } else {
            return document.getNodeType();
        }
    }

    /**
     * @see org.w3c.dom.Node#getParentNode
     */
    public Node getParentNode() {
        if (delegate != null) {
            return delegate.getParentNode();
        } else {
            return document.getParentNode();
        }
    }

    /**
     * @see org.w3c.dom.Node#getChildNodes
     */
    public NodeList getChildNodes() {
        if (delegate != null) {
            return delegate.getChildNodes();
        } else {
            return document.getChildNodes();
        }
    }

    /**
     * @see org.w3c.dom.Node#getFirstChild
     */
    public Node getFirstChild() {
        if (delegate != null) {
            return delegate.getFirstChild();
        } else {
            return document.getFirstChild();
        }
    }

    /**
     * @see org.w3c.dom.Node#getLastChild
     */
    public Node getLastChild() {
        if (delegate != null) {
            return delegate.getLastChild();
        } else {
            return document.getLastChild();
        }
    }

    /**
     * @see org.w3c.dom.Node#getPreviousSibling
     */
    public Node getPreviousSibling() {
        if (delegate != null) {
            return delegate.getPreviousSibling();
        } else {
            return document.getPreviousSibling();
        }
    }

    /**
     * @see org.w3c.dom.Node#getNextSibling
     */
    public Node getNextSibling() {
        if (delegate != null) {
            return delegate.getNextSibling();
        } else {
            return document.getNextSibling();
        }
    }

    /**
     * @see org.w3c.dom.Node#getAttributes
     */
    public NamedNodeMap getAttributes() {
        if (delegate != null) {
            return delegate.getAttributes();
        } else {
            return document.getAttributes();
        }
    }

    /**
     * @see org.w3c.dom.Node#getOwnerDocument
     */
    public Document getOwnerDocument() {
        if (delegate != null) {
            return delegate.getOwnerDocument();
        } else {
            return document.getOwnerDocument();
        }
    }

    /**
     * @see org.w3c.dom.Node#insertBefore
     */
    public Node insertBefore(Node newChild, 
                             Node refChild)
        throws DOMException {
        if (delegate != null) {
            return delegate.insertBefore(newChild, refChild);
        } else {
            return document.insertBefore(newChild, refChild);
        }
    }

    /**
     * @see org.w3c.dom.Node#replaceChild
     */
    public Node replaceChild(Node newChild, 
                             Node oldChild)
        throws DOMException {
        if (delegate != null) {
            return delegate.replaceChild(newChild, oldChild);
        } else {
            return document.replaceChild(newChild, oldChild);
        }
    }

    /**
     * @see org.w3c.dom.Node#removeChild
     */
    public Node removeChild(Node oldChild)
        throws DOMException {
        if (delegate != null) {
            return delegate.removeChild(oldChild);
        } else {
            return document.removeChild(oldChild);
        }
    }

    /**
     * @see org.w3c.dom.Node#appendChild
     */
    public Node appendChild(Node newChild)
        throws DOMException {
        if (delegate != null) {
            return delegate.appendChild(newChild);
        } else {
            return document.appendChild(newChild);
        }
    }

    /**
     * @see org.w3c.dom.Node#normalize
     */
    public void normalize() {
        if (delegate != null) {
            delegate.normalize();
        } else {
            document.normalize();
        }
    }

    /**
     * @see org.w3c.dom.Node#supports(String, String)
     */
    public boolean supports(String feature, 
                            String version) {
        if (delegate != null) {
            return delegate.supports(feature, version);
        } else {
            return document.supports(feature, version);
        }
    }

    /**
     * @see org.w3c.dom.Node#getNamespaceURI
     */
    public String getNamespaceURI() {
        if (delegate != null) {
            return delegate.getNamespaceURI();
        } else {
            return document.getNamespaceURI();
        }
    }

    /**
     * @see org.w3c.dom.Node#getPrefix
     */
    public String getPrefix() {
        if (delegate != null) {
            return delegate.getPrefix();
        } else {
            return document.getPrefix();
        }
    }

    /**
     * @see org.w3c.dom.Node#setPrefix
     */
    public void setPrefix(String prefix) {
        if (delegate != null) {
            delegate.setPrefix(prefix);
        } else {
            document.setPrefix(prefix);
        }
    }

    /**
     * @see org.w3c.dom.Node#getLocalName
     */
    public String getLocalName() {
        if (delegate != null) {
            return delegate.getLocalName();
        } else {
            return document.getLocalName();
        }
    }

    /**
     * @see org.w3c.dom.Node#hasChildNodes
     */
    public boolean hasChildNodes() {
        if (delegate != null) {
            return delegate.hasChildNodes();
        } else {
            return document.hasChildNodes();
        }
    }

    /**
     * @see XMLObject#toDocument
     */
    public String toDocument() {
        // Create formatter if needed (no synchronization necessary)
        if (formatter == null) {
            formatter = new DOMFormatter();
        }

        if (delegate != null) {
            return formatter.toString(delegate);
        } else {
            return formatter.toString(this);
        }
    }

    /**
     * Generated function to synchronize the fields used by the access methods.
     * This syncronizes just the node and is not recursive.
     */
    abstract protected void syncWithDocument(Node node);

    /**
     * Recursively synchronize access methods.
     */
    private void syncAccessMethods(Node node) {
        syncWithDocument(node);
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            syncAccessMethods(child);
        }
    }

    /**
     * @see XMLObject#syncAccessMethods
     */
    public void syncAccessMethods() {
        if (delegate != null) {
            delegate.syncAccessMethods();
        } else {
            syncAccessMethods(document);
        }
    }

    /**
     * Old method to initialize the fields used by the generated access
     * methods from the current state of the document.  This method was
     * poorly named and is deprecated.
     *
     * @deprecated Use <CODE>syncAccessMethods()</CODE> instead.
     * @see syncAccessMethods
     */
    public void initFields() {
        syncAccessMethods();
    }
}
