/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: ClassEntry.java,v 1.8 2000/02/26 19:12:36 markd Exp $
 */

package org.enhydra.xml.xmlc.autocomp;

import org.enhydra.xml.xmlc.XMLCError;
import com.lutris.classloader.MultiClassLoader;
import com.lutris.classloader.Resource;
import com.lutris.classloader.LocalDirResource;
import com.lutris.logging.LogWriter;
import java.io.File;
import java.io.FileNotFoundException;

/**
 * Object used to hold and manage information about a XMLC-generate class.
 * This manages only the actually generated class, not a derived class.  This
 * object also serves as a lock for operations on the class.
 */
class ClassEntry implements ClassInfo, MultiClassLoader.ClassFilter {
    /**
     * Class object for originally loaded class.
     */
    private Class originalClass;

    /**
     * File associated with originally loaded class.
     */
    private File originalClassFile;

    /**
     * Most current version of the class.
     */
    private Class currentClass;

    /**
     * The last-modified time when the class was loaded.
     */
    private long currentClassLoadLastModified;

    /**
     * File associated with currenty loaded class.
     */
    private File currentClassFile;

    /**
     * Handler used to check for out-of-date classes and used to
     * recompile them.
     */
    private RecompilationHandler recompHandler;

    /**
     * Factory we are associated with.
     */
    private XMLCAutoCompFactory factory;

    /**
     * Log writer for informative messages or null no logging.
     */
    private LogWriter xmlcLog;

    /**
     * Log writer for debug messages or null no logging.
     */
    private LogWriter xmlcDebugLog;

    /**
     * Get the last modified time from a file, generate an exception
     * if it doesn't exist.
     */
    protected long getLastModified(File file) throws FileNotFoundException {
        long time = file.lastModified();
        if (time == 0L) {
            throw new FileNotFoundException(file.getPath());
        }
        return time;
    }

    /**
     * Constructor.  
     *
     * @param xmlcClass Class generated by XMLC.
     * @param recompilationHandler Object to handle checking for out-of-date
     *  class files and recompiling classes.  Maybe null.
     */
    public ClassEntry(Class xmlcClass,
                      RecompilationHandler recompilationHandler,
                      XMLCAutoCompFactory xmlcFactory) {
        originalClass = xmlcClass;
        currentClass = xmlcClass;
        recompHandler = recompilationHandler;
        factory = xmlcFactory;
        xmlcLog = xmlcFactory.getLogWriter();
        xmlcDebugLog = xmlcFactory.getDebugLogWriter();
        
        // Get the File from the class loader.
        Exception getExcept = null;
        try {
            originalClassFile
                = factory.getResourceLoader().getClassFile(xmlcClass);
            currentClassFile = originalClassFile;
            currentClassLoadLastModified = getLastModified(originalClassFile);
        } catch (FileNotFoundException except) {
            getExcept = except;
        } catch (ClassNotFoundException except) {
            getExcept = except;
        }
        if (getExcept != null) {
            // This should never be able to happen, as we already have
            // the class
            throw new XMLCAutoCompException("Unable to find file for class \""
                                            + xmlcClass.getName() + "\"", getExcept);
        }
    }

    /**
     * Get the class object for originally loaded class.
     */
    public Class getOriginalClass() {
        return originalClass;
    }

    /**
     * Get the File associated with originally loaded class.
     */
    public File getOriginalClassFile() {
        return originalClassFile;
    }

    /**
     * Get the current version of the class.
     */
    public Class getCurrentClass() {
        return currentClass;
    }

    /**
     * Get File associated with currenty loaded class.
     */
    public File getCurrentClassFile() {
        return currentClassFile;
    }

    /**
     * Check class modification time.
     */
    private boolean classOutOfDate() {
        long fileLastModified;
        try {
            fileLastModified = getLastModified(currentClassFile);
        } catch (FileNotFoundException except) {
            throw new XMLCAutoCompException(except);
        }
        
        if (xmlcDebugLog != null) {
            xmlcDebugLog.println("class is "
                                 + ((fileLastModified > currentClassLoadLastModified) ? "out-of-date" : "current")
                                 + " with class file:\n"
                                 + "class file: "
                                 + fileLastModified + ": "
                                 + currentClassFile.getAbsolutePath() + "\n"
                                 + "class: " + currentClassLoadLastModified + ": "
                                 + currentClass.getName());
        }
        return (fileLastModified > currentClassLoadLastModified);
    }

    /**
     * Load the class in a new classloader.  The requested class is the only
     * class this loader will load.  All others are passed to the main class
     * loader.
     * @return true if the class was actually reloaded.
     */
    private synchronized boolean reloadClass() {

        if (classOutOfDate()) {
            if (xmlcLog != null) {
                xmlcLog.println("Reloading " + originalClass.getName());
            }

            // Resource loader is secondary of new MultiClassLoader.
            ResourceLoader resourceLoader = factory.getResourceLoader();

            // Setup classloader with filter and original as secondary loader.
            MultiClassLoader loader
                = new MultiClassLoader(null, resourceLoader.getClassLoader(),
                                       factory.getLogChannel());
            loader.setClassPath(resourceLoader.getClassPath());
            loader.addClassFilter(this);

            // Load in this class loader and get File object.
            try {
                currentClass = loader.loadClass(originalClass.getName(), true);
                currentClassFile = MultiResourceLoader.getClassFile(loader,
                                                                    currentClass);
            } catch (ClassNotFoundException except) {
                throw new XMLCAutoCompException(except);
            }
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Recompile the class with synchronization.
     * @return true if the class was actually recompiled.
     */
    private synchronized boolean recompileClass() {
        // Recompile if still out-of-date.
        if (recompHandler.isOutOfDate(this)) {
            if (xmlcLog != null) {
                xmlcLog.println("Recompiling " + originalClass.getName());
            }
            recompHandler.recompile(this);

            // Check that we actually did something
            if (recompHandler.isOutOfDate(this)) {
                throw new XMLCAutoCompException("recompilation didn't updated class file "
                                                + originalClass.getName());
            }
        
            return true;
        } else {
            return false;
        }
    }

    /**
     * Get the current version of the class if the originally loaded class is
     * out-of-date.  The current class, either the original class or a
     * previously reloaded class will be check to determine if its current.
     * If a recompilation handler is set, then the class file is compared to
     * its source file modification time.  If its out of date, the class
     * will be recompiled. If the loaded class is out-of-date relative to the
     * class file, it is loaded in its own class loaded and returned.  If an
     * updated class is has previously been loaded, it will be returned.
     * This class only synchronizes if necessary to update the class.
     *
     * @param factory The factory the entry is associated with.
     * @return The current class, either the original class or the new one
     *  loaded in its own class loader.
     * @exception XMLCAutoCompException Thrown on failure to recompiler or
     *  reload.
     */
    public Class loadCurrentClass(XMLCAutoCompFactory factory) {
        // Recompile if class file is out-of-date with source file.
        if ((recompHandler != null)
            && recompHandler.isOutOfDate(this)) {
            if (recompileClass()) {
                factory.incrRecompileCount();
            }
        }

        // Reload class if its out-of-date
        if (classOutOfDate()) {
            if (reloadClass()) {
                factory.incrReloadCount();
            }
        }
        
        return currentClass;
    }

    /**
     * Class filter check.
     *
     * @see MultiClassLoader.ClassFilter
     */
    public int loadCheck(String checkClassName) {
        if (originalClass.getName().equals(checkClassName)) {
            return MUST_LOAD;
        } else {
            return DONT_LOAD;
        }
    }
}
