/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: Recompiler.java,v 1.8.2.1 2000/04/04 05:29:49 markd Exp $
 */

package org.enhydra.xml.xmlc.autocomp;

import org.enhydra.xml.xmlc.XMLCException;
import com.lutris.logging.LogWriter;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.File;
import java.net.URL;
import java.util.Vector;

/**
 * Class to recompile a XMLC-generated class.
 */
class Recompiler {
    /**
     * Factory we are associated with.
     */
    private XMLCAutoCompFactory factory;

    /**
     * Information about the class.
     */
    private ClassInfo classInfo;

    /**
     * Source file being recompiled.
     */
    private File srcFile;

    /**
     * Log writer for XMLC debug logging or null if not enabled.
     */
    private LogWriter xmlcDebugLog;

    /**
     * Constructor.
     */
    public Recompiler(ClassInfo classInfo,
                      File srcFile,
                      XMLCAutoCompFactory factory) {
        this.classInfo = classInfo;
        this.srcFile = srcFile;
        this.factory = factory;
        xmlcDebugLog = factory.getDebugLogWriter();
    }

    /**
     * Get the meta-data file path.  The name the interface name for
     * class with .xmlc extension.
     */
    private String metaDataFileName(Class currentClass)
            throws XMLCAutoCompException {
        // Convert class to file path, drop Impl and add extension.
        String fileName = currentClass.getName().replace('.', '/');
        if (!fileName.endsWith("Impl")) {
            throw new XMLCAutoCompException("Class name doesn't end in \"Impl\"");
        }
        return fileName.substring(0, fileName.length()-4) + ".xmlc";
    }

    /**
     * Load the XMLC meta-data file.
     * FIXME: This should be part of metadata package..
     */
    private CompileParams loadMetaData()
            throws IOException, XMLCException, XMLCAutoCompException {
        // Load file throught class loader.
        Class currentClass = classInfo.getCurrentClass();
        String fileName = metaDataFileName(currentClass);

        ClassLoader loader = factory.getResourceLoader().getClassLoader();
        InputStream in = loader.getResourceAsStream(fileName);
        if (in == null) {
            throw new FileNotFoundException("Can't find file on class path: \""
                                            + fileName + "\"");
        }
        try {
            return CompileParams.loadRecompileParams(fileName, in);
        } finally {
            in.close();
        }
    }

    /**
     * Create a new -classpath argument to javac using the current
     * classpath.
     * FIXME: Should really be much smarter about class path, walking
     * class loaders (parent and secondary), determining the actually
     * class path that will be followed.
     */
    private void setClassPath(CompileParams params) {
        ResourceLoader loader = factory.getResourceLoader();
        String opt = "-classpath";
        String pathSepar = "";
        StringBuffer newPath = new StringBuffer();

        URL[] classPath = loader.getClassPath();
        for (int idx = 0; idx < classPath.length; idx++) {
            newPath.append(pathSepar);
            newPath.append(classPath[idx].getFile());
            pathSepar = File.pathSeparator;
        }
        classPath = loader.getSystemClassPath();
        for (int idx = 0; idx < classPath.length; idx++) {
            newPath.append(pathSepar);
            newPath.append(classPath[idx].getFile());
            pathSepar = File.pathSeparator;
        }

        params.remoteJavacOption(opt);
        params.addJavacOption(opt, newPath.toString());
    }

    /**
     * Parse a file path into an array of its components.
     */
    private String[] parseFileName(File file) {
        //FIXME: Use JDK 1.2 get*File() methods
        File absFile = new File(file.getAbsolutePath());
        Vector pathVec = new Vector();
        
        while (absFile != null) {
            pathVec.addElement(absFile.getName());
            String parent = absFile.getParent();
            if (parent == null) {
                absFile = null;
            } else {
                absFile = new File(absFile.getParent());
            }
        }
        String[] path = new String[pathVec.size()];
        
        int idx = 0;
        for (int vidx = pathVec.size()-1; vidx >= 0; vidx--) {
            path[idx++] = (String)pathVec.elementAt(vidx);
        }
        return path;
    }

    /**
     * Convert a class name and class file path to a destination
     * directory for the java compile.
     */
    private File getDestDir(Class classObj,
                            File srcFile) {
        String[] className = JavaLang.parseClassName(classObj.getName());
        String[] path = parseFileName(new File(srcFile.getAbsolutePath()));  //FIXME: use 1.2 get*File

        if (path.length < className.length) {
            throw new XMLCAutoCompException("Class name has more elements that its directory, can form a destDir: \""
                                            + classObj.getName() + "\" and \"" + srcFile.getAbsolutePath()
                                            + "\"");
        }
        
        File destDir = new File("");
        int numElements = path.length - className.length;
        for (int idx = 0; idx < numElements; idx++) {
            destDir = new File(destDir, path[idx]);
        }
        return destDir;
    }

    /**
     * Load the meta-data and set parameters.
     */
    private CompileParams getMetaData()
            throws IOException, XMLCException, XMLCAutoCompException {
        CompileParams params = loadMetaData();

        // Path to source file
        params.setSourceDocument(srcFile);
        
        // Root for class output; also use as tmp dir for generated source.
        File destDir = getDestDir(classInfo.getOriginalClass(),
                                  classInfo.getOriginalClassFile());
        params.setDestDir(destDir);
        params.setSourceOut(destDir);

        // Must generate implementation of interface.
        params.setGenerate(params.GENERATE_IMPLEMENTATION);

        // Must use current CLASSPATH
        setClassPath(params);

        params.complete();
        return params;
    }

    /**
     * Run compilation steps.
     */
    private void doCompile(CompileParams params,
                           ErrorReporter errorReporter)
        throws IOException, XMLCException { 

        Parse parser = new Parse(errorReporter, null);
        XMLCDocument xmlcDoc = parser.parse(params);

        EditDOM domEditor = new EditDOM();
        domEditor.addURLEdits(params.getURLEdits());
        domEditor.addDeleteClasses(params.getDeleteElementClasses());
        domEditor.edit(xmlcDoc);

        CodeGenerator codeGen = new CodeGenerator(params, xmlcDoc, null);
        codeGen.generateJavaSource(params, null);

        // Enable verbose option if we are doing debug logging, otherwise
        // nothing gets logged!
        boolean oldVerbose = params.getVerbose();
        if (xmlcDebugLog != null) {
            params.setVerbose(true);
        }
        try {
            Javac javac = new Javac(params.getJavac());
            javac.setOptions(Javac.COLLECT_STDOUT|Javac.COLLECT_STDOUT
                             |Javac.DUMP_STDERR_ON_FAIL);
            javac.compile(params, errorReporter, xmlcDebugLog);
        } finally {
            params.setVerbose(oldVerbose);
            if (!params.getKeep()) {
                params.getJavaClassSource().delete();
            }
        }
    }
     
    /**
     * Generate message text about compiled failing.
     */
    private String getFailedMsg() {
        return "Recompile of XMLC generated class: "
            + classInfo.getCurrentClass().getName() 
            + " failed:";
    }

    /**
     * Compile the source file.
     */
    public void compile() throws XMLCAutoCompException { 

        CompileParams params;
        try {
            params = getMetaData();
        } catch (IOException except) {
            throw new XMLCAutoCompException(getFailedMsg(), except);
        } catch (XMLCException except) {
            throw new XMLCAutoCompException(getFailedMsg(), except);
        }

        ErrorReporter errorReporter = new ErrorReporter();

        XMLCAutoCompException compExcept = null;

        // Mostly a lot of trouble to report errors.
        try {
            doCompile(params, errorReporter);
        } catch (Exception except) {
            compExcept = new XMLCAutoCompException(getFailedMsg(), except);
        }

        if (errorReporter.getErrorCnt() > 0) {
            if (compExcept == null) {
                // Generate an exception if one isn't pending
                compExcept = new XMLCAutoCompException(getFailedMsg());
            }
            compExcept.setErrorOutput(errorReporter.getOutput());
        }
        if (compExcept != null) {
            throw compExcept;
        }
    }
}
