/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: ResourceLoaderFactory.java,v 1.2 2000/03/02 22:31:16 markd Exp $
 */

package org.enhydra.xml.xmlc.autocomp;

import java.util.Hashtable;

/**
 * Factory for creating a ResourceLoader wrapper class for a particular class
 * loader.  This takes a class loader class name and returns a instance
 * of the appropriate ResourceLoader.
 *
 * The class is pre-configured with ResourceLoader mappings for the
 * following class loaders:
 * <UL>
 * <LI> <code>com.lutris.classloader.MultiClassLoader</code>
 * <LI> <code>org.apache.tomcat.util.NetworkClassLoader</code>
 * </UL>
 *
 * @see ResourceLoader
 * @see MultiResourceLoader
 * @see TomcatResourceLoader
 */
public class ResourceLoaderFactory {
    /**
     * Mapping of class loader class name to resource loader class
     * name or class.
     */
    private static Hashtable loaderTable = new Hashtable();

    /*
     * Initialize with pre-defined resource loaders.
     */
    static {
        registerResourceLoader("com.lutris.classloader.MultiClassLoader",
                               "org.enhydra.xml.xmlc.autocomp.MultiResourceLoader");
        registerResourceLoader("org.apache.tomcat.util.NetworkClassLoader",
                               "org.enhydra.xml.xmlc.autocomp.TomcatResourceLoader");
    }

    /**
     * Disallow instantiation.
     */
    private ResourceLoaderFactory() {
    }

    /**
     * Register a resource loader class namefor a class loader.
     *
     * @param classLoaderName The class name of the class loader.
     * @param resourceLoaderName The class name of the resource loader
     *  It must implement {@link ResourceLoader ResourceLoader} and 
     *  contain a public constructor taking no arguments.  This resource loader
     *  must be loaded by the system class loader.
     *
     * @see ResourceLoader
     */
    public static void registerResourceLoader(String classLoaderName,
                                              String resourceLoaderName) {
        loaderTable.put(classLoaderName, resourceLoaderName);
    }


    /**
     * Register a resource loader class for a class loader.
     *
     * @param classLoaderName The class name of the class loader.
     * @param resourceLoaderClass The class for the resource loader.
     *  It must implement {@link ResourceLoader ResourceLoader} and 
     *  contain a public constructor taking no arguments.  This resource loader
     *  can be loaded by any class loader.
     *
     * @see ResourceLoader
     */
    public static void registerResourceLoader(String classLoaderName,
                                              Class resourceLoaderClass) {
        loaderTable.put(classLoaderName, resourceLoaderClass);
    }

    /**
     * Get a resource loader class for a class loader.
     */
    private static Class getResourceLoaderClass(ClassLoader classLoader) {
        Object classObj = loaderTable.get(classLoader.getClass().getName());
        if (classObj == null) {
            throw new XMLCAutoCompException("No XMLC ResourceLoader defined for class loader \""
                                            + classLoader.getClass().getName() + "\"");
        }
        Class resourceLoaderClass;
        try {
            if (classObj instanceof String) {
                resourceLoaderClass = ResourceLoaderFactory.class.forName((String)classObj);
            } else {
                resourceLoaderClass = (Class)classObj;
            }
        } catch (Throwable except) {
            throw new XMLCAutoCompException("Can't load ResourceLoader class for "
                                            + classLoader.getClass().getName() + "\"",
                                            except);
        }
        return resourceLoaderClass;
    }

    /**
     * Create a ResourceLoader associated with a specified class loader.
     *
     * @param classLoader The class loader to create a ResourceLoader for.
     *  The ResourceLoader class must be registered with the factory.
     * @exception XMLCAutoCompException If there is no registered resource 
     *  loader or the resource loader can't be created.
     */
    public static ResourceLoader create(ClassLoader classLoader) {
        Class resourceLoaderClass = getResourceLoaderClass(classLoader);

        try {
            ResourceLoader resourceLoader
                = (ResourceLoader)resourceLoaderClass.newInstance();
            resourceLoader.setClassLoader(classLoader);
            return resourceLoader;
        } catch (Throwable except) {
            throw new XMLCAutoCompException("Can't create ResourceLoader class for "
                                            + classLoader.getClass().getName() + "\"",
                                            except);
        }
    }
}
