/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: StandardRecompilationHandler.java,v 1.8 2000/03/07 19:36:16 markd Exp $
 */

package org.enhydra.xml.xmlc.autocomp;

import java.util.Hashtable;
import java.io.FileNotFoundException;
import java.io.File;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.XMLCError;
import com.lutris.logging.LogWriter;

/**
 * Standard recompilation handler.  This searches the class path
 * for a source file using the name compiled into the class.
 * The source file must exist when the class is first loaded.
 * A XMLC options file must exist same directory as the class
 * file and having the same basename as the class.  This file
 * contains the information needed to recompile the class.
 */
public class StandardRecompilationHandler implements RecompilationHandler {
    /**
     * Table mapping current class object to a File object.
     */
    private Hashtable srcFiles = new Hashtable();

    /** 
     * Factory we are associated with.
     */
    private XMLCAutoCompFactory factory;

    /**
     * Log writer for XMLC debug logging or null if not enabled.
     */
    private LogWriter xmlcDebugLog;

    /**
     * Tmp directory for recompiler.
     */
    private File tmpDir;

    /**
     * Bind to a factory.  This is done so that loggers, etc
     * can be grabbed from the factory.
     *
     * @param factory Factory that is creating the object.  Useful
     *  for getting logging object.
     */
    public void init(XMLCAutoCompFactory factory) {
        this.factory = factory;
        xmlcDebugLog = factory.getDebugLogWriter();
    }

    /**
     * Get the CLASSPATH-relative file name for the source file assocaited
     * with a class.  This uses reflection to obtain from a compiled-in
     * constant.
     */
    private String getRelativeSourceFileName(Class currentClass) {
        return (String)
            XMLCAutoCompFactory.getClassConstant(currentClass,
                                                 XMLObject.XMLC_SOURCE_FILE_FIELD_NAME);
    }

    /**
     * Search for source file that and add to table.
     */
    private synchronized File findSourceFile(Class currentClass)
        throws XMLCAutoCompException {

        File srcFile = (File)srcFiles.get(currentClass);
        if (srcFile != null) {
            // Added while not synchronizing
            return srcFile;
        }

        // Search for the source file with the class loader.
        String sourcePath = getRelativeSourceFileName(currentClass);
        try {
            srcFile = factory.getResourceLoader().getResourceFile(sourcePath);
        } catch (ClassNotFoundException except) {
            // Not found.
            throw new XMLCAutoCompException("Source file not found for: "
                                            + currentClass.getName()
                                            + ": " + sourcePath, except);
        }
        
        // Got it.
        srcFiles.put(currentClass, srcFile);
        return srcFile;
    }

    /**
     * Get the source file, search if needed.  Avoids synchronization unless
     * needed.
     */
    private File getSourceFile(Class currentClass)
        throws XMLCAutoCompException {
        File srcFile = (File)srcFiles.get(currentClass);
        if (srcFile != null) {
            return srcFile;
        }
        // Load it..
        return findSourceFile(currentClass);
    }

    /**
     * @see RecompilationHandler#isOutOfDate
     */
    public boolean isOutOfDate(ClassInfo classInfo)
        throws XMLCAutoCompException {

        File srcFile = getSourceFile(classInfo.getCurrentClass());
        if (srcFile == null) {
            return false;  // not found
        }
        File currentClassFile = classInfo.getCurrentClassFile();
        boolean isOutOfDate =
            (srcFile.lastModified() > currentClassFile.lastModified());
        if (xmlcDebugLog != null) {
            xmlcDebugLog.println("class file is " + (isOutOfDate ? "out-of-date" : "current")
                                 + " with source:\n"
                                 + "source file: "
                                 + srcFile.lastModified() + ": "
                                 + srcFile.getAbsolutePath() + "\n"
                                 + "class file: "
                                 + currentClassFile.lastModified() + ": "
                                 + currentClassFile.getAbsolutePath());
        }
        return isOutOfDate;
    }

    /**
     * @see StandardRecompilationHandler#recompile
     */
    public void recompile(ClassInfo classInfo)
        throws XMLCAutoCompException {
        File srcFile =  getSourceFile(classInfo.getCurrentClass());
        if (srcFile == null) {
            throw new XMLCError("recompiled called when class is know not to exist");
        }

        if (xmlcDebugLog != null) {
            xmlcDebugLog.println("Recompiling out-of-date class: " + classInfo.getCurrentClass());
        }
        Recompiler recompiler = new Recompiler(classInfo, srcFile, factory);
        recompiler.compile();
    }
}
