/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: TomcatResourceLoader.java,v 1.1 2000/03/02 22:31:16 markd Exp $
 */

package org.enhydra.xml.xmlc.autocomp;

import java.io.File;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;
import com.lutris.classloader.MultiClassLoader;
import org.enhydra.xml.xmlc.autocomp.ResourceLoader;
import org.enhydra.xml.xmlc.autocomp.XMLCAutoCompException;
import org.apache.tomcat.util.NetworkClassLoader;

/*
 * Observed NetworkClassLoader issues:
 *   - The class loader does not search the classpath in order.
 *   - Caching of contents of zip might not be wise, there could
 *     be large files in a zip.  This also results in classes
 *     from zips being in memory twice.  Ack, it actually copies
 *     the entire zip to memory.
 *   - The decision to close the jar files should be revisited.
 *     Opening a jar file is maybe expensive due to reading the
 *     directory.  It's not clear if this is a good trade off
 *     for allowing jar files to be modified, especially since
 *     the loader will not pick up the modifications and its not
 *     clear what will happen if the loader opens a jar file
 *     at the same time its being updated.
 *   - Errors are ignored in many places, making debugging of
 *     failures difficult.
 *   - If it implemented NetworkClassLoader.findResource(), life
 *     would be much better.
 */

// FIXME: This doesn't take into account parent loader.
// FIXME: need to get useful message when file is in zip.

/**
 * ResourceLoader class for the Tomcat classloader.  This is not
 * ideal, as it doesn't take into account the parent loader or
 * chances in the classpath.  This works for classloader
 * <code>oorg.apache.tomcat.util.NetworkClassLoader</code> and
 * descendent <code>org.apache.tomcat.core/ServletClassLoader</code>.
 */
public class TomcatResourceLoader implements ResourceLoader {
    /**
     * Class loader we are associated with.
     */
    private NetworkClassLoader classLoader;

    /**
     * Cached class path to search, in the order that the NetworkClass
     * loader will search it.
     */
    private URL[] searchClassPath;

    /**
     * Set the class loader this class is associated with.
     */
    public void setClassLoader(ClassLoader loader) {
        if (!(loader instanceof NetworkClassLoader)) {
            throw new XMLCAutoCompException("class loader is not an instanceof "
                                            + NetworkClassLoader.class.getName()
                                            + " ( " + loader.getClass().getName()
                                            + ")");
        }
        classLoader = (NetworkClassLoader)loader;
    }

    /**
     * Get the class path associate with this resource loader.
     */
    public URL[] getClassPath() {
        Enumeration pathEnum = classLoader.getURLs();
        Vector pathVec = new Vector();
        while (pathEnum.hasMoreElements()) {
            pathVec.addElement(pathEnum.nextElement());
        }
        URL[] pathUrls = new URL[pathVec.size()];
        pathVec.copyInto(pathUrls);
        return pathUrls;
    }

    /**
     * Get the system class path.
     */
    public URL[] getSystemClassPath() {
        return MultiClassLoader.getSystemClassPath();
    }

    /**
     * Get the class loader associated with this resource loader.
     */
    public ClassLoader getClassLoader() {
        return classLoader;
    }

    /**
     * Initialize the search path.
     */
    private synchronized void initSearchPath() {
        if (searchClassPath != null) {
            return;  // Initialize while synchronizing
        }
        URL[] sysPath = getSystemClassPath();
        URL[] netPath = getClassPath();
        searchClassPath = new URL[sysPath.length + netPath.length];
        
        int idx = 0;
        for (int j = 0; j < sysPath.length; j++) {
            searchClassPath[idx++] = sysPath[j];
        }
        for (int j = 0; j < netPath.length; j++) {
            searchClassPath[idx++] = netPath[j];
        }
    }
    
    /**
     * Check a URL for a resource.  We only handle files, not zips or
     * remotes, which makes things easier.
     */
    private File checkForResource(URL url, 
                                  String name) {
        if (!url.getProtocol().equals("file")) {
            return null;
        }
        File file = new File(url.getFile(), name);
        if (file.exists()) {
            return file;
        } else {
            return null;
        }
    }

    /**
     * Find a resource using the NetworkClassLoader alogrithm.
     */
    private File findResource(String name) throws ClassNotFoundException {
        if (searchClassPath == null) {
            initSearchPath();
        }
        for (int idx = 0; idx < searchClassPath.length; idx++) {
            File file = checkForResource(searchClassPath[idx], name);
            if (file != null) {
                return file;
            }
        }
        throw new ClassNotFoundException(name +
                                         " (would not be found if in a jar/zip)");
    }

    /**
     * Get a File object associated with a class.
     */
    public File getClassFile(Class classObj) throws ClassNotFoundException {
        String name = classObj.getName().replace('.', '/') + ".class";
        return findResource(name);
    }

    /**
     * Get a File object associated with a resource.
     */
    public File getResourceFile(String name) throws ClassNotFoundException {
        return findResource(name);
    }

    /**
     * Load the specified class.
     *
     * @exception ClassNotFoundException if the class could not be load.
     */
    public Class loadClass(String className)
        throws ClassNotFoundException {
        return classLoader.loadClass(className);
    }
}
