/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLCAutoCompFactory.java,v 1.12 2000/03/07 19:36:16 markd Exp $
 */

package org.enhydra.xml.xmlc.autocomp;

import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.XMLCFactory;
import org.enhydra.xml.xmlc.compiler.JavaLang;
import com.lutris.logging.LogChannel;
import com.lutris.logging.LogWriter;
import com.lutris.logging.Logger;
import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.lang.reflect.Field;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

//FIXME: Might want to build a table of the dervied classes and save the
// constructor pointer, saving a trip through reflection.

/**
 * Factory class to create instances of XMLC-generated class with
 * automatic recompilation if the class is out-of-date relative to
 * a source file.  
 * <P>
 * This class can create objects from classes derived from XMLC objects.
 * It uses a constructor that takes a boolean argument which delays
 * the creation of the DOM until the delegate can be set.  If the derived
 * class requires another constructor, this can be approached in two ways.
 * Either deriving a factory that creates objects through the derived
 * constructor or used this factory object inside of the derived class
 * constructor, obtain the delegate with <CODE>getUpdatedClass()<CODE>.
 * Such a constructor would contain:
 * <PRE>
 *    super(false);
 *    setDelegate(factory.getUpdatedClassInstance());
 *    buildDocument();
 * </PRE>
 *
 * If a log channel is supplied, logging maybe performed.  The following
 * log levels are used:
 * <UL>
 * <LI> XMLC - Issue messages when classes are recompiled or reloaded.
 * <LI> XMLC_DEBUG - Debug messages are issues for every class loaded.
 * <LI> ERROR - Any errors creating an object are logged at this level.
 * </UL>
 *
 * @see StandardRecompilationHandler
 * @see ResourceLoader
 * @see ResourceLoaderFactory
 */
public class XMLCAutoCompFactory implements XMLCFactory {
    /**
     * Handler used to recompiled classes.  If null, classes will not be
     * automatically recompiled.
     */
    private RecompilationHandler recompilationHandler;

    /**
     * Resource loader that is used to load the primary class and find
     * the souurce file.  This is a wrapper class around a class
     * loader.
     */
    private ResourceLoader resourceLoader;

    /**
     * Table of ClassEntry objects, indexed by first-loaded, XMLC-generated
     * class object.
     */
    private Hashtable xmlcClassTable = new Hashtable();

    /**
     * Count of class reloads that have occured.
     */
    private int reloadCount = 0;

    /**
     * Count of recompilations that have occured.
     */
    private int recompileCount = 0;

    /**
     * Log channel null if not enabled.
     */
    private LogChannel logChannel;

    /**
     * Log writer for XMLC informative logging or null if not enabled.
     */
    private LogWriter xmlcLog;

    /**
     * Log writer for XMLC debug logging or null if not enabled.
     */
    private LogWriter xmlcDebugLog;

    /**
     * Log writer for XMLC error logging or null if no logging.
     */
    private LogWriter errorLog;

    /**
     * Argument type for object constructor.
     */
    private static final Class[] CONSTRUCTOR_ARG_TYPES = {
        Boolean.TYPE
    };

    /**
     * Argument for object constructor.
     */
    private static final Object[] CONSTRUCTOR_ARGS = {
        new Boolean(false)
    };

    /**
     * Get one of the compiled-in constants in a XMLC-generated class
     * (or subclass).
     */
    protected static Object getClassConstant(Class xmlcBasedClass,
                                             String constName) {
        // Use generated constant to find the class.
        try {
            Field field = xmlcBasedClass.getField(constName);
            return field.get(xmlcBasedClass);
        } catch (NoSuchFieldException except) {
            throw new XMLCAutoCompException("Couldn't find class constant \""
                                            + constName + "\" in class \""
                                            + xmlcBasedClass.getName() 
                                            + "\", maybe be a XMLC generated class or descendent",
                                            except);
        } catch (IllegalAccessException except) {
            throw new XMLCAutoCompException(except);
        }
    }

    /**
     * Constructor.
     * 
     * @param recompiler Recompilation handler used to recompile objects.
     *  If null, classes will not be recompiled.
     * @param resourceLoader ResourceLoader used to load classes and
     *  locate the source file.  This is a wrapper around a class loader
     *  that implements functionality missing in the standard classloader
     *  interface.
     * @param logChannel The log channel to use to log events and errors.  If
     *  null, no logging is done.  Level XMLC is used to log reloading and
     *  recompilation.  Level ERROR is used for errors.
     */
    public XMLCAutoCompFactory(RecompilationHandler recompiler,
                               ResourceLoader resourceLoader,
                               LogChannel logChannel) {
        recompilationHandler = recompiler;
        this.resourceLoader = resourceLoader;
        this.logChannel = logChannel;
        if (logChannel != null) {
            if (logChannel.isEnabled(XMLC_LOG_LEVEL)) {
                xmlcLog = logChannel.getLogWriter(XMLC_LOG_LEVEL);
            }
            if (logChannel.isEnabled(XMLC_DEBUG_LOG_LEVEL)) {
                xmlcDebugLog = logChannel.getLogWriter(XMLC_DEBUG_LOG_LEVEL);
            }
            errorLog = logChannel.getLogWriter(Logger.ERROR);
        }
        
        if (recompilationHandler != null) {
            // Must be done after everthing else is setup.
            recompilationHandler.init(this);
        }
    }

    /**
     * Constructor.
     * @param recompiler Recompilation handler used to recompile objects.
     *  If null, classes will not be recompiled.
     * @param classLoader Classloader used to load classes when a class name
     *  is specified.  This must be one of the class loaders that we have
     *  built-in knowledge of.  The currently supported classloaders are:
     *  <UL>
     *  <LI> com.lutris.classloader.MultiClassLoader
     *  <LI> org.apache.tomcat.util.NetworkClassLoader
     *  </UL>
     *  
     * @param logChannel The log channel to use to log events and errors.  If
     *  null, no logging is done.  Level XMLC is used to log reloading and
     *  recompilation.  Level ERROR is used for errors.
     */
    public XMLCAutoCompFactory(RecompilationHandler recompiler,
                               ClassLoader classLoader,
                               LogChannel logChannel) {
        this(recompiler, ResourceLoaderFactory.create(classLoader),
             logChannel);
    }

    /**
     * Create a new class entry.
     */
    private synchronized ClassEntry createClassEntry(Class xmlcClass) {
        // Try again now synchronized
        ClassEntry classEntry = (ClassEntry)xmlcClassTable.get(xmlcClass);
        if (classEntry == null) {
            classEntry = new ClassEntry(xmlcClass, recompilationHandler,
                                        this);
            xmlcClassTable.put(xmlcClass, classEntry);
        }
        return classEntry;
    }

    /**
     * Get the current version of a class, given the originally loaded
     * class.  If the class is out-of-date, in respects to the class file,
     * a new class is loaded in its own class loader.  This may also result
     * it the class being recompiled.
     *
     * @param xmlcClass The original instance of a XMLC-generate class.
     *  This class must be loaded by a MultiClassLoader class loader.
     * @return If the supplied class is up-to-date, then it is returned,
     *  otherwise the up-to-date class is returned in a new classloader.
     * @see MultiClassLoader
     */
    private Class loadCurrentClass(Class xmlcClass) {
        // Get the entry for this class, creating a new one if needed.
        ClassEntry classEntry = (ClassEntry)xmlcClassTable.get(xmlcClass);
        if (classEntry == null) {
            classEntry = createClassEntry(xmlcClass);
        }
        
        // Get the most up-to-date class object.
        return classEntry.loadCurrentClass(this);
    }

    /**
     * Find the XMLC generated class in an inheritance hierarchy
     * that might extend the generated class.
     */
    private Class findXMLCGenerateClass(Class xmlcBasedClass) {
        return (Class)getClassConstant(xmlcBasedClass,
                                       XMLObject.XMLC_GENERATED_CLASS_FIELD_NAME);
    }

    /**
     * Get the constructor used in creating the objects.
     */
    private Constructor getConstructor(Class xmlcBasedClass) {
        try {
            return xmlcBasedClass.getConstructor(CONSTRUCTOR_ARG_TYPES);
        } catch (NoSuchMethodException except) {
            throw new XMLCAutoCompException("Class \"" + xmlcBasedClass.getName()
                                            + "\" doesn't have required constructor: \"public "
                                            + JavaLang.simpleClassName(xmlcBasedClass)
                                            + "(boolean buildDOM)");
        }
    }

    /**
     * Create an object from a XMLC-generate class or subclass.
     * Don't call buildDocument();
     */
    private XMLObject createObject(Class xmlcBasedClass) {
        try {
            Constructor constructor = getConstructor(xmlcBasedClass);
            XMLObject obj = (XMLObject)constructor.newInstance(CONSTRUCTOR_ARGS);
            return obj;
        } catch (IllegalAccessException except) {
            throw new XMLCAutoCompException(except);
        } catch (InstantiationException except) {
            throw new XMLCAutoCompException(except);
        } catch (InvocationTargetException except) {
            throw new XMLCAutoCompException(except);
        }
    }

    /**
     * Given a class that is either a XMLC-generate class or a class derived
     * from one, determine if the generate class is out-of-date; if so
     * recompile and return the new class as the delegate.
     * This method is useful when building derived factory classes.
     *
     * @param xmlcBasedClass The XMLC-generate class or a subclass.
     * @return The new version of the generated class, or null if the class
     *  is not out of date.
     * @exception XMLCAutoCompException If an error occurs obtaining the class.
     */
    public Class loadUpdatedClass(Class xmlcBasedClass) {
        Class xmlcClass = findXMLCGenerateClass(xmlcBasedClass);
        Class currentClass = loadCurrentClass(xmlcClass);
        if (xmlcClass != currentClass) {
            return currentClass;
        } else {
            return null;
        }
    }

    /**
     * If a XMLC class has been updated, get a new instance of it.
     * Don't build the DOM at this time.
     *
     * @param xmlcBasedClass The XMLC-generate class or a subclass.
     * @return A new instance of the XMLC-generate class (not the
     *  subclass, sutiable for use as a delegate.  If the class is
     *  not out of date, null is returned.
     * @exception XMLCAutoCompException If an error occurs obtaining the class.
     */
    public XMLObject getUpdatedClassInstance(Class xmlcBasedClass) {
        Class updatedClass = loadUpdatedClass(xmlcBasedClass);
        if (updatedClass != null) {
            if (xmlcDebugLog != null) {
                xmlcDebugLog.println("Creating delegate instance of "
                                     + xmlcBasedClass.getName());
            }
            XMLObject obj = createObject(updatedClass);
            return obj;
        } else {
            return null;
        }
    }

    /**
     * Do actual work of creating a new object.
     */
    private XMLObject doCreate(Class xmlcBasedClass)
            throws ClassNotFoundException {
        if (xmlcBasedClass.isInterface()) {
            // Convert interface to class name.
            return create(xmlcBasedClass.getName() + "Impl");
        } else {
            if (!XMLObject.class.isAssignableFrom(xmlcBasedClass)) {
                throw new XMLCAutoCompException("class \"" + xmlcBasedClass.getName()
                                                + "\" does not implement \""
                                                + XMLObject.class.getName() + "\"");
            }
            
            // Create object and obtain a new version of the class as a 
            // delegate if its out-of-date.
            if (xmlcDebugLog != null) {
                xmlcDebugLog.println("Creating instance of "
                                     + xmlcBasedClass.getName());
            }
            XMLObject xmlcObject = createObject(xmlcBasedClass);
            xmlcObject.setDelegate(getUpdatedClassInstance(xmlcBasedClass));
            xmlcObject.buildDocument();
            return xmlcObject;
        }
    }

    /**
     * Handle all errors, logging if enabled.  Also handles logging
     * error output of the compiler.
     * 
     * @param className Class being loaded.
     * @param except Can be any error or exception.
     * @return Returns runtime exceptions to be rethrown.  Could throw
     * everything, but catch blocks appear to fall through to the compiler so
     * we would have to put in dummy returns.
     */
    private RuntimeException handleError(String className,
                                         Throwable except) {
        if (errorLog != null) {
            errorLog.println("Factory creation of an XMLC class failed: \""
                             + className + "\"", except);
        }
        
        // Our runtime exception is thrown as-is, with special logging.
        if (except instanceof XMLCAutoCompException) {
            // Log output of compiler, if we have it.
            XMLCAutoCompException xmlcExcept = (XMLCAutoCompException)except;
            String errorOutput = xmlcExcept.getErrorOutput();
            if ((errorOutput != null) && (errorLog != null)) {
                errorLog.println("Output from XMLC run of Java compiler:\n"
                                 + errorOutput);
            }
            return xmlcExcept;
        }
        
        // Errors just continue on.
        if (except instanceof Error) {
            throw (Error)except;
        }
        
        // All other exceptions get wrapped.
        return new XMLCAutoCompException("Couldn't load XMLC class",
                                         except);
    }

    /**
     * Create an instance of a XMLC-generated class or a class derived
     * from one, given the name of the class.
     * <P>
     * The class must have a constructor that takes a single boolean argument
     * that will disable the immediate build of the DOM if false.  XMLC
     * creates such a constructor in the classes it generates.
     *
     * @param xmlcClassName The name of the XMLC class or derived class.
     * @return The new object.
     * @exception XMLCAutoCompException If an error occurs creating the object.
     * @see #create(Class)
     */
    public XMLObject create(String xmlcClassName) {
        try {
            return doCreate(resourceLoader.loadClass(xmlcClassName));
        } catch (Throwable except) {
            throw handleError(xmlcClassName, except);
        }
    }

    /**
     * Create an instance of a XMLC-generated class or a class derived
     * from one, given either a class or an interface object.  
     * <P>
     * If a class is given, then it should be either the XMLC generated class
     * or a class derived from it.
     * <P>
     * If an interface is given, it is converted to the implementation name
     * by appending "Impl" to the name.  In this case, the implementation
     * must be in the same package. 
     * <P>
     * The class must have a constructor that takes a single boolean argument
     * that will disable the immediate build of the DOM if false.  XMLC
     * creates such a constructor in the classes it generates.
     *
     * @param xmlcBasedClass The Class object for the class or interface.
     * @return The new object.
     * @exception XMLCAutoCompException If an error occurs creating the object.
     */
    public XMLObject create(Class xmlcBasedClass) {
        try {
            return doCreate(xmlcBasedClass);
        } catch (Throwable except) {
            throw handleError(xmlcBasedClass.getName(), except);
        }
    }

    /**
     * Get the ResourceLoader associate this this factory
     */
    public ResourceLoader getResourceLoader() {
        return resourceLoader;
    }

    /**
     * Get the log channel associated with this factory or null if
     * none.
     */
    public LogChannel getLogChannel() {
        return logChannel;
    }

    /**
     * Get the XMLC informative log writer or null if no logging is to
     * be done.
     */
    public LogWriter getLogWriter() {
        return xmlcLog;
    }

    /**
     * Get the XMLC debugging log writer or null if no debug logging is to
     * be done.
     */
    public LogWriter getDebugLogWriter() {
        return xmlcDebugLog;
    }

    /**
     * Increment the count of the number of reloads that have occured.
     */
    protected void incrReloadCount() {
        reloadCount++;
    }

    /**
     * Get the count of the number of reloads that have occured.
     */
    public int getReloadCount() {
        return reloadCount;
    }

    /**
     * Increment the count of the number of recompilations that have occured.
     */
    protected void incrRecompileCount() {
        recompileCount++;
    }

    /**
     * Get the count of the number of recompilations that have occured.
     */
    public int getRecompileCount() {
        return recompileCount;
    }
}
