/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: Option.java,v 1.2 2000/01/22 07:11:31 markd Exp $
 */

package org.enhydra.xml.xmlc.commands.options;

import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCError;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
 * Definition of an option.  The general or specific options are classes derived from
 * this class.  A general class is parameterizes at instantiation and is able to
 * create parse multiple options.  A specific class only parses one option.
 */
public abstract class Option {
    /**
     * The option set the option is associated with.
     */
    protected final OptionSet optionSet;

    /**
     * The option name, including the leading `-'.
     */
    protected final String name;

    /**
     * The number of arguments for the option.
     */
    protected final int numArgs;

    /**
     * Are multiple instances of the option allowed?
     */
    protected final boolean multipleAllowed;

    /**
     * Help message for the option.
     */
    protected String help;

    /**
     * Error messaage for a failed mesage set.
     */
    private static final String SET_FAILED_MSG = "setting an option failed";

    /**
     * Construct a new instance of the object.
     *
     * @param optSet The option set this option is associated with.
     * @param optName Name of the option, including the leading `-'.
     * @param optNumArgs Number of arguments the option takes.
     * @param optMultipleAllowed Can the option be specified multiple times?
     * @param helpMsg A string that is used in generating the help message.
     *  It should include the arguments, a `-', then one line message.  The
     *  option name and a space will be prepended.
     */
    public Option(OptionSet optSet, String optName, int optNumArgs,
                  boolean optMultipleAllowed, String helpMsg) {
        optionSet = optSet;
        name = optName;
        numArgs = optNumArgs;
        multipleAllowed = optMultipleAllowed;
        help = helpMsg;
    }

    /**
     * Get the option set containing the option.
     */
    public OptionSet getOptionSet() {
        return optionSet;
    }

    /**
     * Get the option name, including leading `-'.
     */
    public String getName() {
        return name;
    }

    /**
     * Get the number of arguments for a single instance of the option.  If
     * the option repeats, the actual number of arguments is a multiple of
     * this.
     */
    public int getNumArgs() {
        return numArgs;
    }
    
    /**
     * Return true if the option maybe specified multiple times, resulting
     * in a list of values.  Return false if the option can only be specified
     * once.
     */
    public boolean getMultipleAllowed() {
        return multipleAllowed;
    }

    /**
     * Get the help string, including option name.
     */
    public String getHelp() {
        return name + " " + help;
    }

    /**
     * Determine if two options are equal.  Only the name is considered.  It
     * is illegal to have options with the same name and different attributes.
     *
     * @param obj This can be either another Option object or a string.  If
     *  an Option object is supplied, the names are compared.
     * @return True if this object name is the same as the name in
     */
    public boolean equals(Object obj) {
        if (obj instanceof Option) {
            return name.equals(((Option)obj).getName());
        } else if (obj instanceof String) {
            return name.equals((String)obj);
        } else {
            return false;
        }
    }

    /**
     * Get the hashcode for the object.
     */
    public int hashCode() {
        return name.hashCode();
    }

    /**
     * Set an option value using a method that takes one argument.
     * <!--
     * FIXME: This really doesn't need to take an option set,
     * since we already have that, its a hold over, although
     * the long-term approach mybe to decouple the OptionSet from
     * the object that will have the set methods, so this is a 
     * place holder for that object -->
     */
    protected void setOption1(OptionSet optSet, String setMethod,
                              Object value) {
        try {
            Method method = optSet.getClass().getMethod(setMethod,
                                                        new Class[]{value.getClass()});
            method.invoke(optSet, new Object[]{value});
        } catch (NoSuchMethodException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        } catch (IllegalAccessException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        } catch (InvocationTargetException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        }
    }

    /**
     * Set an option value that takes two arguments.
     */
    protected void setOption2(OptionSet optSet, String setMethod,
                              Object value1, Object value2) {
        try {
            Method method = optSet.getClass().getMethod(setMethod,
                                                        new Class[]{value1.getClass(),
                                                                    value2.getClass()});
            method.invoke(optSet, new Object[]{value1, value2});
        } catch (NoSuchMethodException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        } catch (IllegalAccessException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        } catch (InvocationTargetException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        }
    }

    /**
     * Set a boolean option.
     */
    protected void setOption(OptionSet optSet, String setMethod,
                             boolean value) {
        try {
            Method method = optSet.getClass().getMethod(setMethod,
                                                        new Class[]{Boolean.TYPE});
            method.invoke(optSet, new Object[]{new Boolean(value)});
        } catch (NoSuchMethodException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        } catch (IllegalAccessException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        } catch (InvocationTargetException except) {
            throw new XMLCError(SET_FAILED_MSG, except);
        }
    }

    /**
     * Parse an instance of the option.
     *
     * @param args The option's arguments.  Will always have the number of 
     *  arguments defined for an option.
     */
    abstract protected void parse(String[] args) throws XMLCException;
}
