/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: OptionsParser.java,v 1.4.2.1 2000/04/04 05:29:49 markd Exp $
 */

package org.enhydra.xml.xmlc.commands.options;

import org.enhydra.xml.xmlc.XMLCException;
import java.util.Hashtable;
import java.util.Vector;
import java.io.*;

/**
 * Parse XMLC options, including options files.
 */
public class OptionsParser {
    /*
     * Enabled debug tracing.
     */
    private static final boolean DEBUG = false;

     /*
      * Suffix for options files.
      */
    static final private String OPT_FILE_SUFFIX = ".xmlc";

    /*
     * Command line arguments and index to next argument.
     */
    private String[] commandArgs;
    private int commandArgIdx;

    private OptionSet options;       // Options definition table.

    private File[] optionsFiles;     // Ordered list of options files from cmd line.

    /**
     * Check to see if we are at the end of the command line options.
     * Will skip the `--' end of line marker.
     */
    private boolean endOfCmdOptions() {
        if (commandArgIdx >= commandArgs.length) {
            return true;  // No more args
        }
        if (commandArgs[commandArgIdx].equals("--")) {
            // End of options marker.
            commandArgIdx++;
            return true;
        }
        if (!commandArgs[commandArgIdx].startsWith("-")) {
            return true;  // No more -options
        }
        return false;
    }

    /**
     * Generate error for invalid options.
     */
    private void invalidOption(String name, File optFile) throws XMLCException {
        StringBuffer msg = new StringBuffer();
        msg.append("Invalid option \"" + name + "\"");
        if (optFile != null) {
            msg.append(" in " + optFile.getPath());
        }
        msg.append(", valid options are:\n");
        msg.append(options.getOptionsMsg());
        throw new XMLCException(msg.toString());
    }

    /**
     * Parse a single command line option.
     */
    private void parseCmdOption(boolean parseArguments) throws XMLCException {
        if (DEBUG) {
            System.err.println("parseCmdOption: " + commandArgs[commandArgIdx]);
        }
        Option option = options.findOption(commandArgs[commandArgIdx]);
        if (option == null) {
            invalidOption(commandArgs[commandArgIdx], null);
        }
        commandArgIdx++;
        int numArgs= option.getNumArgs();
        if (numArgs > (commandArgs.length-commandArgIdx)) {
            throw new XMLCException("Insufficient number of arguments for option \""
                                    + option.getName() + ": " + option.getHelp());
        }
        if (parseArguments) {
            String[] args = new String[numArgs];
            System.arraycopy(commandArgs, commandArgIdx, args, 0, numArgs);
            option.parse(args);
        }
        commandArgIdx += numArgs;
    }

    /**
     * Parse options.  A preparse can be done to find the positional
     * arguments or a full parse to get the option arguments. This is
     * needed to find the positional arguments.
     */
    private void parseOptions(boolean parseArguments) throws XMLCException {
        commandArgIdx = 0;
        while (!endOfCmdOptions()) {
            parseCmdOption(parseArguments);
        }
    }

    /**
     * Parse an options file line.
     */
    private void parseOptionsFileEntry(File optFile, String[] entry) 
            throws XMLCException {
        Option option = options.findOption(entry[0]);
        if (option == null) {
            invalidOption(entry[0], optFile);
        }
        int numArgs= option.getNumArgs();
        if (entry.length-1 != numArgs) {
            throw new XMLCException("wrong number of arguments for option \""
                                    + option.getName() 
                                    + " in " + optFile.getPath()
                                    + ": " + option.getHelp());
        }
        String[] args = new String[entry.length-1];
        System.arraycopy(entry, 1, args, 0, entry.length-1);
        option.parse(args);
    }

    /**
     * Process options read from a file.
     */
    private void processOptionsFile(File optFile,
                                    OptionFileParser parsedOpts)
        throws XMLCException {
        String[][] opts = parsedOpts.getOptions();
        for (int idx = 0; idx < opts.length; idx++) {
            parseOptionsFileEntry(optFile, opts[idx]);
        }
    }

    /**
     * Parse an options file.
     */
    public void parseOptionsFile(String fileName,
                                 Reader optFile) 
        throws XMLCException, IOException {
        processOptionsFile(new File(fileName),
                           new OptionFileParser(optFile));
    }

    /**
     * Parse an open options file.
     */
    private void parseOptionsFile(File optFile) 
        throws XMLCException, IOException {
        processOptionsFile(optFile, new OptionFileParser(optFile));
    }

    /**
     * Parse the options files.
     */
    private void parseOptionsFiles()
        throws XMLCException, IOException {
        // Parse backwards so first one overrides
        for (int idx = optionsFiles.length-1; idx >= 0; idx--) {
            parseOptionsFile(optionsFiles[idx]);
        }
    }

    /**
     * Parse the positional arguments.
     */
    private void parsePositionalArgs() throws XMLCException {
        // Find *.xmlc options files.
        Vector optFiles = new Vector();
        while ((commandArgIdx < commandArgs.length) 
               && commandArgs[commandArgIdx].endsWith(OPT_FILE_SUFFIX)) {
            optFiles.addElement(new File(commandArgs[commandArgIdx]));
            commandArgIdx++;
        }
        optionsFiles = new File[optFiles.size()];
        optFiles.copyInto(optionsFiles);

        // Pass the rest of the arguments to the option set.
        int len = commandArgs.length-commandArgIdx;
        String[] posArgs = new String[len];
        System.arraycopy(commandArgs, commandArgIdx,
                         posArgs, 0, len);
        options.processPositionalArgs(posArgs);
    }

    /**
     * Constructor.
     *
     * @param optionsSet The set of options to parse, all value are stored in
     * this object.
     */
    public OptionsParser(OptionSet optionSet) {
        options = optionSet;
    }

    /**
     * Parse options.
     *
     * @param args The arguments.
     */
    public void parse(String[] args) 
        throws XMLCException, IOException {
        commandArgs = args;
        commandArgIdx = 0;

        // Parse once to find the positional arguments.
        parseOptions(false);

        // Parse the positional arguments and process the option files.
        parsePositionalArgs();
        parseOptionsFiles();
 
        // Parse the command line for real, which will override options files.
        parseOptions(true);

        // Let parsed object do remaining work.
        options.complete();

    }

    /**
     * Get the parsed options.
     */
    public OptionSet getOptions() {
        return options;
    }
}
