/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLC.java,v 1.8.2.2 2000/04/04 05:29:50 markd Exp $
 */

package org.enhydra.xml.xmlc.commands.xmlc;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.commands.options.*;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import com.lutris.util.ChainedThrowable;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import java.io.*;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.DOMFormatter;

/**
 * XMLC compiler program.  See the user manual for details on the command
 * line.  Will look for a system property `org.enhydra.xml.xmlc.javac' to find
 * the compiler to use if -javac is not specified.
 */
public class XMLC {
    /*
     * Enabled debug tracing.
     */
    private static final boolean DEBUG = true;

    /*
     * Object containing the parsed parameters.
     */
    private CompileParams params;

    /**
     * Enable verbose output.
     */
    private boolean verbose = false;

    /*
     * Document being compiled.
     */
    private Document document;

    /*
     * Error output.
     */
    ErrorReporter errorReporter;

    /*
     * Trace and verbose output.
     */
    private PrintWriter traceOut = new PrintWriter(System.err, true);

    /*
     * XML document management object.
     */
    private XMLCDocument xmlcDoc;

    /*
     * Print out a stack trace with special handling of SAX exceptions.
     */
    private void printStackTrace(Throwable except) {
        except.printStackTrace();

        // If this is a chained exception, then walk the stack
        // looking for a non-chained exception.
        while ((except != null) && (except instanceof ChainedThrowable)) {
            except = ((ChainedThrowable)except).getCause();
        }
        if ((except != null) && (except instanceof SAXException)) {
            Exception except2 = ((SAXException)except).getException();
            if (except2 != null) {
                printStackTrace(except2);
            }
        }
    }
    
    /*
     * Handle an exception, printing the approriate error
     * message.
     */
    public void printException(Throwable except) {
        if (except instanceof XMLCException) {
            System.err.println("Error: " + except.getMessage());
            if (verbose) {
                printStackTrace(except);
            }

            // If this is result of a java.lang exception, print stack.
            Throwable cause = ((XMLCException)except).getCause();
            if ((!verbose) && (cause != null)
                && cause.getClass().getName().startsWith("java.lang.")) {
                printStackTrace(cause);
            }
        } else if ((except instanceof FileNotFoundException)
                   || (except instanceof IOException)) {
            System.err.println("Error: " +except.getClass().getName() + ": "
                               + except.getMessage());
            if (verbose) {
                printStackTrace(except);
            }
        } else {
            System.err.println("Error: " + except.getClass().getName() + ": "
                               + except.getMessage());
            printStackTrace(except);
        }
        System.exit(1);
    }

    /*
     * Parse the page into the DOM and perform various checks and edits.
     */
    private void parsePage()
        throws XMLCException, IOException {
        if (verbose) {
            traceOut.println(">>> parsing " + params.getSourceDocumentDesc());
        }

        Parse parser = new Parse(errorReporter,
                                (params.getParseInfo() ? traceOut : null));
        xmlcDoc = parser.parse(params);
        document = xmlcDoc.getDocument();

        if (params.getPrintInfo()) {
            if (verbose) {
                traceOut.println(">>> printing document info");
            }
            PrintInfo info = new PrintInfo(document, xmlcDoc);
            info.printInfo(traceOut);
        }


        EditDOM domEditor = new EditDOM();
        domEditor.addURLEdits(params.getURLEdits());
        domEditor.addDeleteClasses(params.getDeleteElementClasses());
        domEditor.edit(xmlcDoc);

        if (params.getDump()) {
            if (verbose) {
                traceOut.println(">>> dumping document hierarchy");
            }
            XMLCUtil.printNode("DOM hierarchy", document, 
                               XMLCUtil.PRINT_ALL,
                               traceOut);
        }
    }

    /*
     * Generate the Java source files.
     */
    private void generateJavaSource() throws XMLCException, IOException {
        if (verbose) {
            traceOut.println(">>> generating code");
        }
        CodeGenerator codeGen =
            new CodeGenerator(params, xmlcDoc,
                              (params.getPrintAccessMethods() ? traceOut : null));
        codeGen.generateJavaSource(params,
                                   verbose ? traceOut : null);
    }

    /*
     * Generate method information without generating code.
     */
    private void generateJavaSourceInfo() throws XMLCException, IOException {
        if (verbose) {
            traceOut.println(">>> generating code info without generating source");
        }
        CodeGenerator codeGen =
            new CodeGenerator(params, xmlcDoc,
                              (params.getPrintAccessMethods() ? traceOut : null));
    }

    /*
     * Compile the generate source.
     */
    private void compileJavaSource() throws XMLCException {
        if (verbose) {
            traceOut.println(">>> compiling code");
        }
        Javac javac = new Javac(params.getJavac());
        javac.compile(params,
                      errorReporter,
                      verbose ? traceOut : null);
    }

    /*
     * Write the DOM to a file.
     */
    private void writeDOM() throws IOException {
        String dir = params.getDocumentOutput().getParent();
        if (dir != null) {
            new File(dir).mkdirs();
        }
        OutputOptions options = new OutputOptions();
        options.setXmlEncoding(xmlcDoc.getDocBuilder().getEncoding());
        DOMFormatter formatter = new DOMFormatter(options);
        formatter.write(xmlcDoc.getDocument(),
                        params.getDocumentOutput());
    }

    /*
     * Generate the Jave source and Compile the page.
     */
    private void compilePage()
        throws XMLCException, IOException {
        try {
            generateJavaSource();
            if (!params.getNoCompile()) {
                compileJavaSource();
                if (params.getForRecomp()) {
                    params.save();
                }
            }
        } finally {
            if (!params.getKeep()) {
                if (params.getJavaClassSource() != null) {
                    params.getJavaClassSource().delete();
                }
                if (params.getJavaInterfaceSource() != null) {
                    params.getJavaInterfaceSource().delete();
                }
            }
        }
    }

    /*
     * Print the XMLC version number.
     */
    private void printVersion() {
        System.out.println("Enhydra XMLC version " + XMLCVersion.VERSION);
        System.out.println("See http://www.enhydra.org for latest distribution");
    }

    /*
     * Parse arguments and compile the page.
     */
    public void compile(String[] args)
        throws XMLCException, IOException {
        
        // Setup and document parsing.
        XMLCOptions options = new XMLCOptions();
        OptionsParser optParser = new OptionsParser(options);
        optParser.parse(args);
        params = options; // Use a more restricted class.
        verbose = params.getVerbose();

        if (params.getPrintVersion()) {
            printVersion();
            if (params.getInputSource() == null) {
                return;
            }
        }

        errorReporter = new ErrorReporter(new PrintWriter(System.err, true));
        parsePage();
        if (params.getDocumentOutput() == null) {
            compilePage();
        } else {
            if (params.getPrintAccessMethods()) {
                generateJavaSourceInfo();
            }
            writeDOM();
        }
        if (verbose) {
            traceOut.println(">>> completed");
        }
    }

    /*
     * Parse arguments and compile the page.  Print errors that
     * are thrown.
     */
    private void compileHandleErrors(String[] args) {
        try {
            compile(args);
        } catch (Exception except) {
            printException(except);
        }
        
    }

    /*
     * Program entry point.
     */
    public static void main(String[] args) {
        XMLC xmlc = new XMLC();
        xmlc.compileHandleErrors(args);
    }
}
