/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLCOptions.java,v 1.7.2.1 2000/04/04 05:29:50 markd Exp $
 */

package org.enhydra.xml.xmlc.commands.xmlc;

import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.commands.options.*;
import org.enhydra.xml.xmlc.html.HTMLProprietaryTags;
import java.io.File;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.InputStream;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

/**
 * OptionSet for xmlc command.  Use to parse the command line.
 * This extends CompileParams and sets the parameters from the
 * command line options.
 */
public class XMLCOptions extends CompileParams implements OptionSet {
    /**
     * Usage message.
     */
    private static final String USAGE = "wrong # args: xmlc [options] [optfile.xmlc ...] src.ext";

    /**
     * Table of standard javac flags and options that are directly supported
     * and those that are not written saved to options files created
     * for recompilation.
     */
    private static final String[] STD_JAVAC_ARGS = {
        "-g", "-O", "-d", "-classpath"
    };
    private static final Hashtable stdJavacArgs = new Hashtable();
    private static final String[] RECOMP_SKIP_JAVAC_ARGS = {
        "-d", "-classpath"
    };
    private static final Hashtable recompSkipJavacArgs = new Hashtable();

    static {
        for (int i = 0; i < STD_JAVAC_ARGS.length; i++) {
            stdJavacArgs.put(STD_JAVAC_ARGS[i], STD_JAVAC_ARGS[i]);
        }
        for (int i = 0; i < RECOMP_SKIP_JAVAC_ARGS.length; i++) {
            recompSkipJavacArgs.put(RECOMP_SKIP_JAVAC_ARGS[i], RECOMP_SKIP_JAVAC_ARGS[i]);
        }
    }

    /**
     * Ordered list of Option objects associated with this OptionSet.
     */
    private Vector optionList = new Vector();

    /**
     * Hash table of valid Option objects.
     */
    private Hashtable options = new Hashtable();

    /**
     * Class for directly supported javac flag or option.
     */
    private class JavacStdArg extends Option {
        /**
         * Constructor.
         */
        public JavacStdArg(String option, int numArgs, String optHelp) {
            super(XMLCOptions.this, option, numArgs, false, optHelp);
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args) throws XMLCException {
            if (getNumArgs() == 0) {
                addJavacFlag(getName());
            } else {
                addJavacOption(getName(), args[0]);
            }
        }
    }

    /**
     * Class for arbitrary javac flag or option.
     */
    private class JavacArbitraryArg extends Option {
        /**
         * Constructor.
         */
        public JavacArbitraryArg(String option, int numArgs, String optHelp) {
            super(XMLCOptions.this, option, numArgs, true, optHelp);
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args) throws XMLCException {
            if (getNumArgs() == 1) {
                addJavacFlag(args[0]);
            } else {
                addJavacOption(args[0], args[1]);
            }
        }
    }

    /**
     * Class for -urlmapping option.
     */
    private class UrlMappingOption extends Option {
        /**
         * Constructor.
         */
        public UrlMappingOption () {
            super(XMLCOptions.this,  "-urlmapping", 2, true,
                  "orgURL newURL - Specify a URL to modify and its new value");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args) throws XMLCException {
            addURLEdit(new URLMapping(null, args[0], args[1]));
        }
    }

    /**
     * Class for -urlsetting option.
     */
    private class UrlSettingOption extends Option {
        /**
         * Constructor.
         */
        public UrlSettingOption () {
            super(XMLCOptions.this,  "-urlsetting", 2, true,
                  "id newURL - Specify a URL to set given a tag id");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args) throws XMLCException {
            addURLEdit(new URLMapping(args[0], null, args[1]));
        }
    }

    /**
     * Class for -urlregexpmapping option.
     */
    private class UrlRegExpMappingOption extends Option {
        /**
         * Constructor.
         */
        public UrlRegExpMappingOption () {
            super(XMLCOptions.this, "-urlregexpmapping", 2, true,
                  "regexp replace - Specify a regular expression to match URLs and a replacement pattern");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args) throws XMLCException {
            addURLEdit(new URLRegexpMapping(null, args[0], args[1]));
        }
    }

    /**
     * Class for the implements option.
     */
    private class ImplementsOption extends Option {
        /**
         * Constructor.
         */
        public ImplementsOption() {
            super(XMLCOptions.this, "-implements", 1, true,
                "interface - Specify an interface the generate class will implement");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args) throws XMLCException {
            addWillImplement(args[0]);
        }
    }

    /**
     * Class for the -generate option.
     */
    private class GenerateOption extends Option {
        /**
         * Constructor.
         */
        public GenerateOption() {
            super(XMLCOptions.this, "-generate", 1, true,
                "generate which - Specify type of class to generate (both, interface, implementation, class)");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args) throws XMLCException {
            String typeStr = args[0];
            if (typeStr.equals("both")) {
                setGenerate(CompileParams.GENERATE_BOTH);
            } else if (typeStr.equals("interface")) {
                setGenerate(CompileParams.GENERATE_INTERFACE);
            } else if (typeStr.equals("implementation")) {
                setGenerate(CompileParams.GENERATE_IMPLEMENTATION);
            } else if (typeStr.equals("class")) {
                setGenerate(CompileParams.GENERATE_CLASS);
            } else {
                throw new XMLCException("invalid argument to -generate: \""
                                        + typeStr
                                        + "\", expected one of \"both\", \"interface\", \"implementation\", or \"class\"");
            }
        }
    }

    /**
     * Class for deprecated -html:frameset option.
     * FIXME: delete this after several releases, added in 1.1.
     */
    private class HTMLFrameSetOption extends FlagOption {
        /**
         * Construct a new object.
         */
        public HTMLFrameSetOption() {
            super(XMLCOptions.this, "-html:frameset", "setHtmlFrameSet",
                  "- Deprecated and ignored.");
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have none.
         */
        public void parse(String[] args) throws XMLCException {
            System.err.println("Notice: The -html:frameset option is deprecated and ignored");
        }
    }

    /**
     * Class for options -html:old-class-constants and
     * -html:old-name-constants options.
     * FIXME: delete this after several releases, added in 1.1.
     */
    private class HTMLOldConstantsOption extends FlagOption {
        /**
         * Construct a new object.
         */
        public HTMLOldConstantsOption(String optName,
                                      String optSetMethod,
                                      String optHelp) {
            super(XMLCOptions.this, optName, optSetMethod, optHelp);
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have none.
         */
        public void parse(String[] args) throws XMLCException {
            System.err.println("Warning: The " + getName()
                               + " option will be deprectated in a future release, please update your code");
            super.parse(args);
        }
    }

    /**
     * Class for the -html:addtagset option.
     */
    private class HTMLAddTagSetOption extends Option {
        /**
         * Construct a new object.
         */
        public HTMLAddTagSetOption() {
            super(XMLCOptions.this, "-html:addtagset", 1, true,
                "tagsetname - Add a predefined set of proprietary tags to the list of valid HTML tags.");
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have one.
         */
        public void parse(String[] args) throws XMLCException {
            // Get or create tag table.
            HTMLProprietaryTags tags = createHTMLProprietaryTags();
            tags.addTagSet(args[0]);
        }
    }

    /**
     * Class for the -html:addtag option.
     */
    private class HTMLAddTagOption extends Option {
        /**
         * Construct a new object.
         */
        public HTMLAddTagOption() {
            super(XMLCOptions.this, "-html:addtag", 2, true,
                "tagname flags - Add a proprietary tag to the list of valid HTML tags.");
        }

        /**
         * Parse the flags representing the content model.
         */
        private int parseFlags(String flagsStr) throws XMLCException {
            // Split list at commas and convert to bit set.
            int flags = 0;
            int startIdx = 0;
            while (startIdx < flagsStr.length()) {
                int nextIdx = flagsStr.indexOf(',', startIdx);
                if (nextIdx < 0) {
                    nextIdx = flagsStr.length(); // End of string
                }
                String flagStr = flagsStr.substring(startIdx, nextIdx);
                if (flagStr.equalsIgnoreCase("inline")) {
                    flags |= HTMLProprietaryTags.TAG_CM_INLINE;
                } else if (flagStr.equalsIgnoreCase("block")) {
                    flags |= HTMLProprietaryTags.TAG_CM_BLOCK;
                } else if (flagStr.equalsIgnoreCase("empty")) {
                    flags |= HTMLProprietaryTags.TAG_CM_EMPTY;
                } else if (flagStr.equalsIgnoreCase("opt")) {
                    flags |= HTMLProprietaryTags.TAG_CM_OPT;
                } else {
                    throw new XMLCException("Invaild flag for " + name
                                            + " \"" + flagStr + "\"");
                }
                startIdx = nextIdx+1;
            }
            return flags;
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have one.
         */
        public void parse(String[] args) throws XMLCException {
            // Get or create tag table.
            HTMLProprietaryTags tags = createHTMLProprietaryTags();
            tags.addTag(args[0],parseFlags(args[1]));
        }
    }

    /**
     * Class for the -html:addattr option.
     */
    private class HTMLAddAttrOption extends Option {
        /**
         * Construct a new object.
         */
        public HTMLAddAttrOption() {
            super(XMLCOptions.this, "-html:addattr", 1, true,
                "attrname - Add a proprietary attribute the list of valid HTML attributes.");
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have one.
         */
        public void parse(String[] args) throws XMLCException {
            // Get or create tag table.
            HTMLProprietaryTags tags = createHTMLProprietaryTags();
            tags.addAttribute(args[0]);
        }
    }

    /**
     * Add an option to the table.
     */
    private void addOption(Option opt) {
        optionList.addElement(opt);
        options.put(opt.getName(), opt);
    }

    /**
     * Constructor.  Initializes all option objects.
     */
    protected XMLCOptions() {
        addOption(new StringOption(this, "-parser", "setParser",
                                   "type - Specify the parser to use (tidy, swing, xerces)."));
        addOption(new FlagOption(this, "-keep", "setKeep",
                                 "- Keep the generate java source, do not delete it."));
        addOption(new FlagOption(this, "-nocompile", "setNoCompile",
                                 "- If specified, don't compile the generate java file."));
        addOption(new FlagOption(this, "-info", "setPrintInfo",
                                 "- Print useful information about the page object, such as ids and URLs."));
        addOption(new FlagOption(this, "-dump", "setDump",
                                 "- Dump out the DOM tree for the page."));
        addOption(new StringOption(this, "-class", "setClassName",
                                   "class - Fully qualified class name for the generated class or interface."));
        addOption(new StringVectorOption(this, "-xcatalog", "addXCatalog",
                                         "catalog - XCatalog file to resolve external entities."));
        addOption(new FileOption(this, "-sourceout", "setSourceOut",
                                 "sourceout - Root directory for generated source files."));
        addOption(new FlagOption(this, "-verbose", "setVerbose",
                                 "- Generate useful output about the compiler process."));
        addOption(new FlagOption(this, "-parseinfo", "setParseInfo",
                                 "- Print detailed information about the page parsing."));
        addOption(new FlagOption(this, "-methods", "setPrintAccessMethods",
                                 "- Print the signature of each generated access method."));
        addOption(new HTMLFrameSetOption());

        addOption(new UrlMappingOption());
        addOption(new UrlSettingOption());
	addOption(new UrlRegExpMappingOption());

        addOption(new StringVectorOption(this, "-implements", "addWillImplement",
                                         "interface - Specify an interface the generate class will implement"));

        addOption(new StringVectorOption(this, "-delete-class", "addDeleteElementClasses",
                                         "classname - Delete all elements are in the specified class (element class attribute, not Java class)"));

        addOption(new FileOption(this, "-docout", "setDocumentOutput",
                                 "outfile - Write a static document to the outfile instead of generating a class"));

        addOption(new StringOption(this, "-javac", "setJavac",
                                   "prog - Specify the java compiler to use."));
        addOption(new FileOption(this, "-d", "setDestDir",
                                 "dir - Destintation directory, passed to javac"));
        addOption(new JavacStdArg("-g", 0,
                                  "- Passed to javac"));
        addOption(new JavacStdArg("-O", 0,
                                  "- Passed to javac"));
        addOption(new JavacStdArg("-classpath", 1,
                                  "path - Passed to javac"));
        addOption(new JavacArbitraryArg("-javacflag", 1,
                                        "flag - Aribitrary flag to pass to the javac program; including leading `-' or `+' character"));
        addOption(new JavacArbitraryArg("-javacopt", 2,
                                        "opt value - Aribitrary option and value to pass to the javac program; including leading `-' or `+' character"));
        addOption(new FlagOption(this, "-version", "setPrintVersion",
                                 "- Print the XMLC version number."));
        addOption(new BooleanOption(this, "-validate", "setValidateDocument",
                                    "yes|no - Changes the default documentvalidation mode of the parser"));
        addOption(new StringOption(this, "-domfactory", "setDOMFactoryClassName",
                                   "classname - XMLCDOMFactory class for creating DTD-specific documents"));
        addOption(new StringOption(this, "-extends", "setExtendClassName",
                                   "classname - Class the generated document extends"));
        addOption(new FlagOption(this, "-for-recomp", "setForRecomp",
                                    "Generate support for automatic recompilation of the generated class."));
        addOption(new GenerateOption());
        addOption(new StringOption(this, "-html:encoding", "setHTMLEncoding",
                                   "encoding - Specify the encoding to use when reading a HTML document."));
        addOption(new HTMLAddTagSetOption());
        addOption(new HTMLAddTagOption());
        addOption(new HTMLAddAttrOption());
        addOption(new HTMLOldConstantsOption("-html:old-class-constants",
                                             "setHTMLOldClassConstants",
                                             "- Generate old-style, all upper-case class constants"));
        addOption(new HTMLOldConstantsOption("-html:old-name-constants",
                                             "setHTMLOldNameConstants",
                                             "- Generate old-style, all upper-case name constants"));
    }

    /**
     * Get displayable string of valid options for use in error messages.
     */
    public String getOptionsMsg() {
        StringBuffer msg = new StringBuffer();
        Enumeration opts = getOptions();

        while (opts.hasMoreElements()) {
            Option opt = (Option)opts.nextElement();
            msg.append("    ");
            msg.append(opt.getHelp());
            msg.append("\n");
        }
        return msg.toString();
    }

    /**
     * Get an enumeration of all options in the set.
     */
    public Enumeration getOptions() {
        return optionList.elements();
    }

    /**
     * Lookup an option in the table.
     *
     * @param name The option name, including the leading `-'.
     * @return The Option object or null if name is not found.
     */
    public Option findOption(String name) {
        return (Option)options.get(name);
    }

    /**
     * Parse the positional arguments.
     */
    public void processPositionalArgs(String[] posArgs) throws XMLCException {
        // No filename is required if print version is specified, otherwise
        // must have an input (although it is allowed with -version).
        if (posArgs.length > 1) {
            throw new XMLCException(USAGE);
        }
        if (posArgs.length > 0) {
            setSourceDocument(new File(posArgs[0]));
        }
    }

    /**
     * Default any options in the set that have not been specified.
     * Determination of document type, and hence validation and defaulting of
     * the parser base on document type, is handled outside of this class.
     */
    public void complete() throws XMLCException {
        if (getInputSource() == null) {
            if (getPrintVersion()) {
                return;  // Only print version, no other checking.
            }
            throw new XMLCException(USAGE);
        }
        super.complete();
    }

    /**
     * Output a quoted string. FIXME: tmp, will be deleted
     */
    private static void writeQuoted(PrintWriter out,
                                    String str) {
        int len = str.length();
        out.print('\'');
        for (int i = 0; i < len; i++) {
            char c = str.charAt(i);
            if (c == '\'') {
                out.print('\\');
            }
            out.print(c);
        }
        out.print('\'');
    }

    /**
     * Output an option with zero arguments. FIXME: tmp, will be deleted
     */
    private static void writeOpt0(PrintWriter out,
                                  String opt) {
        out.println(opt);
    }

    /**
     * Output an option with one argument. FIXME: tmp, will be deleted
     */
    private static void writeOpt1(PrintWriter out,
                                  String opt,
                                  String arg1) {
        if (arg1 != null) {
            out.print(opt);
            out.print(" ");
            writeQuoted(out, arg1);
            out.println();
        }
    }

    /**
     * Output a option with two arguments. FIXME: tmp, will be deleted
     */
    private static void writeOpt2(PrintWriter out,
                                  String opt,
                                  String arg1,
                                  String arg2) {
        if (arg1 != null) {
            out.print(opt);
            out.print(" ");
            writeQuoted(out, arg1);
            out.print(" ");
            writeQuoted(out, arg2);
            out.println();
        }
    }


    /**
     * Output an array of an option with one argument.
     * FIXME: tmp, will be deleted
     */
    private static void writeOpts(PrintWriter out,
                                  String opt,
                                  String[] args) {
        for (int i = 0; i < args.length; i++) {
            writeOpt1(out, opt, args[i]);
        }
    }

    /**
     * Write a URLEdit. FIXME: tmp, will be deleted
     */
    private static void writeURLEdit(PrintWriter out,
                                     URLEdit urlEdit) {
        if (urlEdit instanceof URLMapping) {
            URLMapping urlMapping = (URLMapping)urlEdit;
            if (urlMapping.getOriginalURL() == null) {
                // URL setting
                writeOpt2(out, "-urlsetting", urlMapping.getElementId(),
                          urlMapping.getNewURL());
            } else {
                // URL mapping
                if (urlMapping.getElementId() != null) {
                    throw new XMLCError("attempt to save a URLMapping with a element id and original URL: not implemented");
                }
                writeOpt2(out, "-urlmapping", urlMapping.getOriginalURL(),
                          urlMapping.getNewURL());
            }
        } else if (urlEdit instanceof URLRegexpMapping) {
            URLRegexpMapping regexpMapping = (URLRegexpMapping)urlEdit;
            writeOpt2(out, "-urlregexpmapping", regexpMapping.getRegexpSrc(),
                      regexpMapping.getSubstitution());
        }
    }

    /**
     * Convert tag flags to a string.
     */
    private static String getTagFlagsString(int flags) {
        StringBuffer buf = new StringBuffer();
        String sep = "";
        if ((flags & HTMLProprietaryTags.TAG_CM_INLINE) != 0) {
            buf.append(sep);
            buf.append("inline");
            sep = ",";
        }
        if ((flags & HTMLProprietaryTags.TAG_CM_BLOCK) != 0) {
            buf.append(sep);
            buf.append("block");
            sep = ",";
        }
        if ((flags & HTMLProprietaryTags.TAG_CM_EMPTY) != 0) {
            buf.append(sep);
            buf.append("empty");
            sep = ",";
        }
        if ((flags & HTMLProprietaryTags.TAG_CM_OPT) != 0) {
            buf.append(sep);
            buf.append("opt");
            sep = ",";
        }
        return buf.toString();
    }

    /**
     * Write HTMLProprietaryTags. FIXME: tmp, will be deleted
     */
    private static void writeHTMLProprietaryTags(PrintWriter out,
                                                 HTMLProprietaryTags propTags) {
        Enumeration tagSets = propTags.getTagSets();
        while (tagSets.hasMoreElements()) {
            writeOpt1(out, "-html:addtagset", (String)tagSets.nextElement());
        }
        Enumeration tags = propTags.getTags();
        while (tags.hasMoreElements()) {
            HTMLProprietaryTags.Tag tag = (HTMLProprietaryTags.Tag)tags.nextElement();
            if (tag.tagSet == null) {
                writeOpt2(out, "-html:addtag", tag.tagName,
                          getTagFlagsString(tag.contentModel));
            }
        }
        Enumeration attrs = propTags.getAttributes();
        while (attrs.hasMoreElements()) {
            writeOpt1(out, "-html:addattr", (String)attrs.nextElement());
        }
    }

    /**
     * Generate a XMLC options from a CompileParams.  This is used in
     * recompiling the JAVA class.  This is a short-term hack, the long-term
     * solution is to create an XML file.  Only options that make sense
     * for recompilation are saved.
     * <P>
     * Warning: this method will be removed in a future release.
     */
    public static void saveRecompileParams(PrintWriter out,
                                           CompileParams params) {

        writeOpt1(out, "-parser", params.getParser());
        writeOpt1(out, "-class", params.getClassName());

        URLEdit[] urlEdits = params.getURLEdits();
        for (int i = 0; i < urlEdits.length; i++) {
            writeURLEdit(out, urlEdits[i]);
        }

        writeOpts(out, "-implements", params.getWillImplement());
        writeOpts(out, "-delete-class", params.getDeleteElementClasses());

        Boolean validate = params.getValidateDocument();
        if (validate != null) {
            writeOpt1(out, "-validate",
                      validate.booleanValue() ? "yes" : "no");
        }
        writeOpt1(out, "-domfactory", params.getDOMFactoryClassName());
        writeOpt1(out, "-extends", params.getExtendClassName());
        String[] xCatalog = params.getXCatalog();
        for (int idx = 0; idx < xCatalog.length; idx++) {
            writeOpt1(out, "-xcatalog", xCatalog[idx]);
        }

        writeOpt1(out, "-javac", params.getJavac());

        // Javac flags.
        String[] javacFlags = params.getJavacFlags();
        for (int i = 0; i < javacFlags.length; i++) {
            if (stdJavacArgs.containsKey(javacFlags[i])) {
                writeOpt0(out, javacFlags[i]);
            } else {
                writeOpt1(out, "-javacflag", javacFlags[i]);
            }
        }

        // Javac options
        String[] javacOptions = params.getJavacOptions();
        for (int i = 0; i < javacOptions.length; i+=2) {
            if (!recompSkipJavacArgs.containsKey(javacOptions[i])) {
                if (stdJavacArgs.containsKey(javacOptions[i])) {
                    writeOpt1(out, javacOptions[i], javacOptions[i+1]);
                } else {
                    writeOpt2(out, "-javacopt", javacOptions[i], javacOptions[i+1]);
                }
            }
        }
        
        if (params.getHTMLEncoding() != null) {
            writeOpt1(out, "-html:encoding", params.getHTMLEncoding());
        }
        if (params.getHTMLProprietaryTags() != null) {
            writeHTMLProprietaryTags(out, params.getHTMLProprietaryTags());
        }
    }


    /**
     * Load the XMLC options into a CompileParams.This is a short-term hack,
     * the long-term.
     * <P>
     * Warning: this method will be removed in a future release.
     */
    public static CompileParams loadRecompileParams(String optFile,
                                                    InputStream optStream)
        throws IOException, XMLCException {
        XMLCOptions options = new XMLCOptions();
        OptionsParser parser = new OptionsParser(options);

        Reader inReader = new InputStreamReader(optStream);
        try {
            parser.parseOptionsFile(optFile, inReader);
        } finally {
            inReader.close();
        }
        return options;
    }

}
