/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: AccessConsts.java,v 1.4 2000/02/21 01:27:32 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.dom.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;
import org.w3c.dom.*;


/**
 * Data collection and code generation for constants used to access the
 * DOM.
 */
class AccessConsts {
    /**
     * Container for document.
     */
    private XMLCDocument xmlcDoc;

    /*
     * Table of element class names that have been found.
     */
    private Hashtable elementClassNames = new Hashtable();

    /**
     * Table of element class that are not legal Java identifiers.
     */
    private Hashtable invalidElementClassNames = null;

    /*
     * Table of element names that have been found.
     */
    private Hashtable elementNames = new Hashtable();

    /**
     * Table of element names that are not legal Java identifiers.
     */
    private Hashtable invalidElementNames = null;

    /**
     * Generate old-style, all upper-case class constants.
     */
    private boolean oldClassConstants;

    /**
     * Generate old-style, all upper-case name constants.
     */
    private boolean oldNameConstants;

    /**
     * Constructor.
     * @param params Parameters controlling compile.
     * @param xmlcDocObj Must contain parsed document.
     */
    public AccessConsts(CompileParams params,
                        XMLCDocument xmlcDocObj) {
        oldClassConstants = params.getHTMLOldClassConstants();
        oldNameConstants = params.getHTMLOldNameConstants();
        xmlcDoc = xmlcDocObj;
    }
    
    /**
     * Record information about class name constants that are to be created.
     */
    private void processElementClassNames(Element element)
            throws XMLCException {
        String[] names = xmlcDoc.getElementClassNames(element);
        if (names == null) {
            return;  // No class names
        }

        for (int idx = 0; idx < names.length; idx++) {
            if (!JavaLang.legalJavaIdentifier(names[idx])) {
                if (invalidElementClassNames == null) {
                    invalidElementClassNames = new Hashtable();
                }
                invalidElementClassNames.put(names[idx], names[idx]);
            } else if (!elementClassNames.containsKey(names[idx])) {
                elementClassNames.put(names[idx], names[idx]);
            }
        }
    }

    /**
     * Create a field for a single element class name.
     */
    private void createElementClassNameField(IndentPrinter out,
                                             String className,
                                             PrintWriter methodInfoOutput) {
        String fieldName = "CLASS_"
            + (oldClassConstants ? className.toUpperCase() : className);
        String sig = "public static final String " + fieldName;

        String[] def = {
            "/**",
            " * Class attribute constant for element class " + className,
            " */",
            sig + " = " + JavaLang.createStringConst(className) + ";"
        };
        out.printlns(def);

        if (methodInfoOutput != null) {
            methodInfoOutput.println(sig + ";");
        }
    }

    /**
     * Create element class constants.
     */
    private void createElementClassNameFields(IndentPrinter out,
                                              PrintWriter methodInfoOutput) {
        Enumeration names = elementClassNames.keys();
        while (names.hasMoreElements()) {
            String className = (String)elementClassNames.get(names.nextElement());
            createElementClassNameField(out, className, methodInfoOutput);
        }
    }

    /**
     * Record information about element name constants that are to be created.
     */
    private void processElementName(Element element) {
        String name = xmlcDoc.getElementName(element);
        if ((name == null) || (name.length() == 0)) {
            return;  // No name
        }
        if (elementNames.containsKey(name)) {
            return; // Already have it.
        }
        if (!JavaLang.legalJavaIdentifier(name)) {
            if (invalidElementNames == null) {
                invalidElementNames = new Hashtable();
            }
            invalidElementNames.put(name, name);
            return; // Not legal
        }
        elementNames.put(name, name);
    }

    /**
     * Create a field for a single element name.
     */
    private void createElementNameField(IndentPrinter out,
                                        String name,
                                        PrintWriter methodInfoOutput) {
        String fieldName = "NAME_"
            + (oldNameConstants ? name.toUpperCase() : name);
        String sig = "public static final String " + fieldName;

        String[] def = {
            "/**",
            " * Element name constant for " + name,
            " */",
            sig + " = " + JavaLang.createStringConst(name)  + ";"
        };
        out.printlns(def);

        if (methodInfoOutput != null) {
            methodInfoOutput.println(sig + ";");
        }
    }

    /**
     * Create a fields for a element names.
     */
    private void createElementNameFields(IndentPrinter out,
                                             PrintWriter methodInfoOutput) {
        Enumeration names = elementNames.keys();
        while (names.hasMoreElements()) {
            String name = (String)elementNames.get(names.nextElement());
            createElementNameField(out, name, methodInfoOutput);
        }
    }

    /**
     * Process an element, recording information about class constants that
     * will be generated.
     *
     * @param element The element being processed.
     */
    public void processElement(Element element) throws XMLCException {
        processElementClassNames(element);
        processElementName(element);
    }

    /**
     * Generate access contants definitions.
     */
    public void createAccessConstants(IndentPrinter out,
                                      PrintWriter methodInfoOutput) {
        createElementClassNameFields(out, methodInfoOutput);
        createElementNameFields(out, methodInfoOutput);
    }

    /**
     * Print names that didn't have access constants generated due to
     * being illegal Java identifiers.  Sort so tests can be reproduced.
     */
    public void printOmittedConstants(PrintWriter out) {
        if (invalidElementClassNames != null) {
            out.println("Element class name constants not created for these class names");
            out.println("since they are not legal Java identifiers:");
            AccessMethods.outputSortedStrings(out, invalidElementClassNames.keys());
        }
        if (invalidElementNames != null) {
            out.println("Element name constants not created for these names");
            out.println("since they are not legal Java identifiers:");
            AccessMethods.outputSortedStrings(out, invalidElementNames.keys());
        }
    }
}
