/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: AccessMethods.java,v 1.4 2000/02/21 01:27:32 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.dom.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.*;
import org.w3c.dom.*;

/**
 * Data collection and code generation for element access methods.
 */
class AccessMethods {
    /**
     * Container for document.
     */
    private XMLCDocument xmlcDoc;

    /**
     * Information collected about each element id.
     */
    private ElementIdTable elementIdTable;

    /**
     * Constructor.
     * @param xmlcDocObj Must contain parsed document.
     */
    public AccessMethods(XMLCDocument xmlcDocObj) {
        xmlcDoc = xmlcDocObj;
        elementIdTable = new ElementIdTable(xmlcDoc);
    }
    
    /**
     * Get the name of the field that will store an element id.
     */
    private String getElementIdField(String id) {
        return "$element" + id;
    }

    /**
     * Process an element.  If the element has an id, record information
     * and return initialization code for document builder.
     *
     * @param element The element being processed.
     * @param elemVarName The name of a variable that will contain
     *  the element during document construction at the point that
     *  the element is created.
     * @return A Java statement that will initialize the access method
     *  fields, given the variable name.  Returns null if no statement
     *  is needed (element doesn't have an id).
     */
    public String processElement(Element element,
                                 String elemVarName) throws XMLCException {
        String id = elementIdTable.recordElementId(element);

        if (id == null) {
            return null;
        } else {
            // Generate code to initialize the field when DOM is being built.
            return getElementIdField(id) + " = ("
                + elementIdTable.getElementClassName(id) + ")"
                + elemVarName + ";\n";
        }
    }

    /**
     * Create an access method definitions.
     */
    private void createAccessMethod(IndentPrinter out,
                                    String id,
                                    PrintWriter methodInfoOutput,
                                    boolean createBody,
                                    boolean createDelegate) {
        String elementClass = elementIdTable.getElementClassName(id);
        String adjustedId = elementIdTable.adjustElementId(id);
        String fieldName = getElementIdField(id);
        String methodName = "getElement" + adjustedId;
        String sig = "public " + elementClass + " " + methodName + "()";

        String[] header = {
            "/**",
            " * Get the element with id <CODE>" + id + "</CODE>.",
            " * @see " + elementClass,
            " */"
        };

        out.printlns(header);
        if (createBody) {
            out.println(sig + " {");
            out.enter();
            String[] body;
            if (createDelegate) {
                body = new String[] {
                    "if (" + CodeGenerator.DELEGATE_FIELD_NAME + " != null) {",
                    "    return " + CodeGenerator.DELEGATE_FIELD_NAME + "." + methodName + "();",
                    "} else {",
                    "    return " + fieldName +  ";",
                    "}"
                };
            } else {
                body = new String[] {
                    "return " + fieldName +  ";"
                };
            }
            out.printlns(body);
            out.leave();
            out.println("}");
            out.println("private " + elementClass + " " + fieldName + ";");
            out.println();
        } else {
            out.println(sig + ";");
        }

        // Output the signature information, if requested.
        if (methodInfoOutput != null) {
            methodInfoOutput.println(sig + ";");
        }
    }

    /**
     * Create the set method for a text child of an element.
     */
    private void createSetTextMethod(IndentPrinter out,
                                     String id,
                                     PrintWriter methodInfoOutput,
                                     boolean createBody,
                                     boolean createDelegate) {
        String adjustedId = elementIdTable.adjustElementId(id);
        String fieldName = getElementIdField(id);
        String methodName = "setText" + adjustedId;
        String sig = "public void " + methodName + "(String text)";

        String header[] = {
            "/**",
            " * Get the value of text child of element <CODE>" + id + "</CODE>.",
            " * @see org.w3c.dom.Text",
            " */"
        };
        out.printlns(header);
        if (createBody) {
            out.println(sig + " {");
            out.enter();
            String[] body;
            if (createDelegate) {
                body = new String[] {
                    "if (" + CodeGenerator.DELEGATE_FIELD_NAME + " != null) {",
                    "    " + CodeGenerator.DELEGATE_FIELD_NAME + "." + methodName + "(text);",
                    "} else {",
                    "    XMLCUtil.getFirstText(" + fieldName +  ").setData(text);",
                    "}"
                };
            } else {
                body = new String[] {
                    "XMLCUtil.getFirstText(" + fieldName +  ").setData(text);"
                };
            }
            out.printlns(body);
            out.leave();
            out.println("}");
            out.println();
        } else {
            out.println(sig + ";");
        }

        // Output the signature information, if requested.
        if (methodInfoOutput != null) {
            methodInfoOutput.println(sig + ";");
        }
    }

    /**
     * Generate access methods and field definitions.
     */
    public void createAccessMethods(IndentPrinter out,
                                    PrintWriter methodInfoOutput,
                                    boolean createBody,
                                    boolean createDelegate) {
        Enumeration ids = elementIdTable.getIds();
        while (ids.hasMoreElements()) {
            String id = (String)ids.nextElement();
            createAccessMethod(out, id, methodInfoOutput,
                               createBody, createDelegate);
            if (elementIdTable.hasSetText(id)) {
                createSetTextMethod(out, id, methodInfoOutput,
                                    createBody, createDelegate);
            }
        }
    }

    /**
     * Create code for initializing id fields from an enumeration of ids.
     */
    private void createIdFieldInits(IndentPrinter out,
                                    Enumeration ids) {
        while (ids.hasMoreElements()) {
            String id = (String)ids.nextElement();

            out.println("} else if (id.equals(" + JavaLang.createStringConst(id) + ")) {");
            out.enter();
            out.println(getElementIdField(id) + " = ("
                        + elementIdTable.getElementClassName(id) + ")node;");
            out.leave();
        }
    }

    /**
     * Create body of code to set all element access fields to null.
     */
    private void createNullAccessFieldsInit(IndentPrinter out) {
        Enumeration ids = elementIdTable.getIds();
        while (ids.hasMoreElements()) {
            String id = (String)ids.nextElement();
            out.println(getElementIdField(id) + " = null;");
        }
        out.println();
    }

    /**
     * Create body of field initialization method for global id attribute
     * mode.
     */
    private void createGlobalIdModeInit(IndentPrinter out) {
        String idAttr = xmlcDoc.getGlobalIdAttribute();
        out.println("if (node instanceof Element) {");
        out.enter();
        out.println("String id = ((Element)node).getAttribute("
                    + JavaLang.createStringConst(idAttr) + ");");
        out.println("if (id.length() == 0) {");  // Block contains empty statement.
        
        createIdFieldInits(out, elementIdTable.getIds());

        out.println("}");  // Closes last else-if
        out.leave();
        out.println("}");
    }

    /**
     * Create body of field initialization method for multple id attribute
     * mode.
     */
    private void createMultipleIdModeInit(IndentPrinter out) {
        String[] code1 = {
            "if (node instanceof Element) {",
            "    Element element = (Element)node;",
            "    String tag = element.getTagName();"
        };
        out.printlns(code1);
        out.enter();
        
        // For each element tag name, generate code for all ids associated
        // with it.
        Enumeration tags = elementIdTable.getTagNames();
        while (tags.hasMoreElements()) {
            String tag = (String)tags.nextElement();
            String idAttr = elementIdTable.getIdAttributeName(tag);
            out.println("if (tag.equals(" + JavaLang.createStringConst(tag) + ")) {");
            out.enter();
            out.println("String id = element.getAttribute("
                        + JavaLang.createStringConst(elementIdTable.getIdAttributeName(tag)) + ");");
            out.println("if (id.length() == 0) {");  // Block contains empty statement.
            
            // Generate code for each id field for this tag.
            createIdFieldInits(out, elementIdTable.getIds(tag));
            out.println("}");  // Closes last else-if
            out.leave();
            out.println("}");  // Closes tag.equals(xxx) block
        }
        out.leave();
        out.println("}");
    }

    /**
     * Generate access method field initialization function used in clone.
     * There are two versions of the code that can be generated:
     * <UL>
     * <LI> Global id attribute mode.  One attribute names is used for
     *      all ids and is not used anywhere else.  That is,
     *      <CODE>XMLCDocument.getGlobalIdAttribute()</CODE> returns a name.
     * <LI> Multiple id attribute mode.  Multiple attributes are declared
     *      of type ID.
     * </UL>
     */
    public void createAccessMethodSync(IndentPrinter out) {
        String[] header = {
            "/**",
            " * Recursize function to do set access method fields from the DOM.",
            " * Missing ids have fields set to null.",
            " */",
            "protected void syncWithDocument(Node node) {"
        };

        // Create recursive method to do initialization
        out.printlns(header);
        out.enter();
        
        // Generate body to initialize.
        if (xmlcDoc.getGlobalIdAttribute() != null) {
            createGlobalIdModeInit(out);
        } else {
            createMultipleIdModeInit(out);
        }
        
        // Generate recursion down the tree.
        String[] code1 = {
            "Node child = node.getFirstChild();",
            "while (child != null) {",
            "    syncWithDocument(child);",
            "    child = child.getNextSibling();",
            "}"
        };

        out.leave();
        out.println("}");
        out.println();
    }

    /**
     * Lame insertion sort for omitted ids.
     * FIXME: Use sorted collection when we no longer support 1.1.*.
     */
    static private void insertSorted(String id, Vector table) {
        int idx;
        for (idx = 0; idx < table.size(); idx++) {
            if (id.compareTo((String)table.elementAt(idx)) > 0) {
                break;
            }
        }
        table.insertElementAt(id, idx);
    }
    
    /**
     * Output a string of enumerations sorted.
     * FIXME: Use sorted collection when we no longer support 1.1.*.
     */
    static protected void outputSortedStrings(PrintWriter out,
                                              Enumeration enum) {
        // Sort
        Vector table = new Vector();
        while (enum.hasMoreElements()) {
            insertSorted((String)enum.nextElement(), table);
        }

        // Print
        enum = table.elements();
        while (enum.hasMoreElements()) {
            out.println("    " + (String)enum.nextElement());
        }
    }

    /**
     * Print ids that didn't have accessors generated due to
     * being illegal Java identifiers.  We sort them so that
     * the test always work.
     */
    public void printOmittedIds(PrintWriter out) {
        Enumeration ids = elementIdTable.getInvalidElementIds();
        if (ids != null) {
            out.println("Access methods not created for these element IDs");
            out.println("since they are not legal Java identifiers:");
            outputSortedStrings(out, ids);
        }
    }
}
