/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: CodeGenerator.java,v 1.12.2.2 2000/04/04 05:29:51 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.dom.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;

/**
 * Generate code for XML classes and interfaces.
 */
public class CodeGenerator {
    /*
     * List of standard imports for generated classes.
     */
    private static final String[] CLASS_IMPORT_LIST = {
        "org.w3c.dom.*",
        "org.enhydra.xml.xmlc.XMLCUtil",
        "org.enhydra.xml.xmlc.XMLCError",
        "org.enhydra.xml.xmlc.dom.XMLCDomFactory",
    };

    /*
     * List of standard imports for generated interfaces.
     */
    private static final String[] INTERFACE_IMPORT_LIST = {
        "org.w3c.dom.*",
        "org.w3c.dom.html.*",
    };

    /**
     * Name of field containing delegate reference.
     */
    protected static final String DELEGATE_FIELD_NAME = "delegate";


    /**
     * Class name of DOM factory cache.
     */
    private static final String DOM_FACTORY_CACHE_CLASS
        = "org.enhydra.xml.xmlc.dom.XMLCDomFactoryCache";

    /*
     * Name of variable in init function containing DOM factory.
     */
    private static final String DOM_FACTORY_VAR = "domFactory";

    /*
     * Name of variable in init function containing Document.
     */
    private static final String DOCUMENT_VAR = "document";

    /*
     * Name of variable in init function that holds document type object.
     */
    private static final String DOC_TYPE_VAR = "documentType";

    /*
     * Header to place in each generated file.
     */
    private static final String[] WARNING_HEADER = {
        "/*",
        " ************************************",
        " * XMLC GENERATED CODE, DO NOT EDIT *",
        " ************************************",
        " */"
    };

    /**
     * Parameters controlling compile.
     */
    private CompileParams compileParams;

    /*
     * XMLC DOM management object.
     */
    private XMLCDocument xmlcDoc;

    /*
     * Source document.
     */
    private Document document;

    /*
     * Source document type or null if an HTML document.
     */
    DocumentType docType;

    /*
     * If not null, write information about accessor methods to this
     * file.
     */
    private PrintWriter methodInfoOutput;

    /*
     * Access method creator.
     */
    private AccessMethods accessMethods;

    /*
     * Access constants creator.
     */
    private AccessConsts accessConsts;

    /*
     * Code that builds the DOM object.
     */
    private StringBuffer domInitDefs = new StringBuffer();

    /*
     * Table of tmp node variable names.
     */
    private VarNames nodeVarNames = new VarNames("Node", "$node");

    /*
     * Table of tmp Element node variable names.
     */
    private VarNames elementVarNames = new VarNames("Element", "$elem");

    /*
     * Table of tmp attribute variable names.
     */
    private VarNames attrVarNames = new VarNames("Attr", "$attr");

    /**
     * Generate code to construct a node.  Returned string is an R-value
     * to use in an assignment.  Can't construct top-level documents.
     */
    private String getNodeConstructor(Node node) throws XMLCException {
        switch (node.getNodeType()) {
        case Node.ATTRIBUTE_NODE:
            return DOCUMENT_VAR + ".createAttribute("
                + JavaLang.createStringConst(((Attr)node).getName()) + ");";

        case Node.CDATA_SECTION_NODE:
            return DOCUMENT_VAR + ".createCDATASection("
                + JavaLang.createStringConst(((CDATASection)node).getData()) + ");";

        case Node.COMMENT_NODE:
            return DOCUMENT_VAR + ".createComment("
                + JavaLang.createStringConst(((Comment)node).getData()) + ");";

        case Node.DOCUMENT_FRAGMENT_NODE:
        case Node.DOCUMENT_NODE:
        case Node.DOCUMENT_TYPE_NODE:
            throw new XMLCException("Creating node of this type not supported: "
                                    + node.getClass().getName());

        case Node.ELEMENT_NODE:
            return DOCUMENT_VAR + ".createElement("
                + JavaLang.createStringConst(((Element)node).getTagName()) + ");";

        case Node.ENTITY_NODE:
            throw new XMLCException("XMLC bug: entity node found in document");

        case Node.ENTITY_REFERENCE_NODE:
            return DOCUMENT_VAR + ".createEntityReference("
                + JavaLang.createStringConst(((EntityReference)node).getNodeName()) + ");";

        case Node.NOTATION_NODE:
            Notation notation = (Notation)node;
            return DOCUMENT_VAR + ".createNotation("
                + JavaLang.createStringConst(notation.getNodeName()) + ", "
                + JavaLang.createStringConst(notation.getPublicId()) + ", "
                + JavaLang.createStringConst(notation.getSystemId()) + ");";

        case Node.PROCESSING_INSTRUCTION_NODE:
            ProcessingInstruction procInstr = (ProcessingInstruction)node;
            return DOCUMENT_VAR + ".createProcessingInstruction("
                + JavaLang.createStringConst(procInstr.getNodeName()) + ", "
                + JavaLang.createStringConst(procInstr.getData()) + ");";

        case Node.TEXT_NODE:
            return DOCUMENT_VAR + ".createTextNode("
                + JavaLang.createStringConst(((Text)node).getData()) + ");";

        default:
            throw new XMLCException("Unknown node type: " + node.getNodeType());
        }
    }

    /**
     * Generate code to to initialize an attribute.
     */
    private void processAttr(String elementVarName,
                             Attr attr,
                             int level)
            throws XMLCException {
        String attrVarName = attrVarNames.getVarName(level);
        if (attr.getSpecified()) {
            // Create attribute node
            domInitDefs.append(attrVarName + " = "
                               + getNodeConstructor(attr)
                               + "\n");
            domInitDefs.append(elementVarName + ".setAttributeNode(" + attrVarName + ");\n\n");
            // Create attribute children.
            processChildNodes(attrVarName, attr, level+1);
        }
    }

    /**
     * Generate code to to initialize a attribute.
     */
    private void processNodeAttrs(String elementVarName,
                                  Node node,
                                  int level)
            throws XMLCException {

        NamedNodeMap attrList = node.getAttributes();
        if (attrList != null) {
            for (int idx = 0; idx < attrList.getLength(); idx++) {
                processAttr(elementVarName, (Attr)attrList.item(idx), level);
            }
        }
    }

    /**
     * Process a node, generating code to recreate it.
     */
    private void processNode(String parentVarName,
                             Node node,
                             int level)
            throws XMLCException {
        boolean isRootElement = (node == document.getDocumentElement());
        String nodeVarName;

        // Special handling for Elements; allow for attribute setting
        // without requiring (slower) type casting in generate code.
        if (node instanceof Element) {
            nodeVarName = elementVarNames.getVarName(level);
        } else {
            nodeVarName = nodeVarNames.getVarName(level);
        }
        
        // Create object unless root, which already exists.
        if (isRootElement) {
            // Just get the root.
            domInitDefs.append(nodeVarName + " = " + parentVarName
                               + ".getDocumentElement();\n");
        } else {
            // Create and add to parent.
            domInitDefs.append(nodeVarName + " = " + getNodeConstructor(node) + ";\n");
            domInitDefs.append(parentVarName + ".appendChild(" + nodeVarName + ");\n\n");
        }            

        // If this is an element with an special attributes (id, class, etc),
        // create a n access methods.
        if (node instanceof Element) {
            Element element = (Element)node;
            String createCode = accessMethods.processElement(element,
                                                             nodeVarName);
            if (createCode != null) {
                domInitDefs.append(createCode);
            }
            accessConsts.processElement(element);
        }

        // Handle attributes.
        if (node instanceof Element) {
            processNodeAttrs(nodeVarName, node, level);
        }
        processChildNodes(nodeVarName, node, level+1);
    }

    /**
     * Process children nodes.
     */
    private void processChildNodes(String nodeVarName,
                                   Node node,
                                   int level)
            throws XMLCException {
        Node child = node.getFirstChild();
        while (child != null) {
            processNode(nodeVarName, child, level);
            child = child.getNextSibling();
        }
    }
    
    /**
     * Generate code to create the document via the XMLCDomFactory
     * that is being used.  This adds code to the current method.
     */
    private void createDocumentCreation(IndentPrinter out) {
        String domFactoryClass = xmlcDoc.getDomFactory().getClass().getName();
        DocTypeBuilder docTypeBuilder = xmlcDoc.getDocTypeBuilder();

        
        // Create document factory.
        out.println("XMLCDomFactory " + DOM_FACTORY_VAR
                    + " = " + DOM_FACTORY_CACHE_CLASS
                    + ".getFactory(\"" + domFactoryClass + "\");");
        
        // Create document type, if needed.
        String docTypeRef = "null";  // Object to pass to create document.
        if (docType != null) {
            out.println("DocumentType " + DOC_TYPE_VAR
                        + " = " + DOM_FACTORY_VAR
                        + ".createDocumentType("
                        + JavaLang.createStringConst(docTypeBuilder.getDocumentTypeName())
                        + ", " + JavaLang.createStringConst(docTypeBuilder.getPublicId())
                        + ", " + JavaLang.createStringConst(docTypeBuilder.getSystemId())
                        + ", " + JavaLang.createStringConst(docTypeBuilder.getInternalSubset())
                        + ");");
            docTypeRef = DOC_TYPE_VAR;
        }

        // Create the actual document.
        out.print("Document " + DOCUMENT_VAR + " = " + DOM_FACTORY_VAR);
        if (document instanceof HTMLDocument) {
            out.println(".createDocument(null, null);");
        } else {
            out.println(".createDocument("
                        + JavaLang.createStringConst(docTypeBuilder.getDocumentTypeName())
                        + ", " + docTypeRef + ");");
        }

        // Save document information.
        out.println("setDocument(" + DOCUMENT_VAR + ");");
        String encoding = xmlcDoc.getDocBuilder().getEncoding();
        if (encoding != null) {
            out.println("setEncoding("
                        + JavaLang.createStringConst(encoding) + ");");
        }
        out.println();
    }

    /**
     * Generate code to enable or disable DOM error checking.
     */
    private void createSetErrorChecking(IndentPrinter out,
                                        boolean enable) {
        out.println(DOM_FACTORY_VAR + ".setErrorChecking("
                    + DOCUMENT_VAR + ", " + enable + ");");
        out.println();
    }

    /**
     * Create the delegate support methods and fields.  Only
     * used when delegation support is being compiled in.
     */
    private void createDelegateSupport(IndentPrinter out,
                                       String unqualInterfaceName,
                                       boolean createBody) {
        // NOTE: the argument to setDelegate is not the specific interface
        // type since we want it to override the method in the base class.
        // The generated type-cast will catch the wrong type being passed in.
        String[] header = {
            "/**",
            " * Set the delegate object.",
            " * @param delegate The delegate to set.  This must be an object",
            " * implementing " + unqualInterfaceName + ", however the parameter",
            " * is untyped, since this method must override XMLOject.setDelegate()",
            " */",
        };
        String sig = "public void setDelegate("
            + XMLObject.class.getName() + " delegate)";

        out.printlns(header);
        if (createBody) {
            String[] body = {
                sig + " {",
                "    this.delegate = (" + unqualInterfaceName + ")delegate;",
                "    super.setDelegate(delegate);",
                "}",
                "private " + unqualInterfaceName + " delegate;",
                ""
            };
            out.printlns(body);
        } else {
            out.println(sig + ";");
        }
    }

    /**
     * Construct an object, converting a DOM into code
     * fragments.
     *
     * @param params Parameters controlling compile.
     * @param xmlcDoc XMLC DOM object containing the document.
     * @param methodOutput Write information about the generated 
     *  methods to this file if not NULL.
     * @exception XMLCException If an error is detected.
     */
    public CodeGenerator(CompileParams params,
                         XMLCDocument xmlcDoc,
                         PrintWriter methodOutput)
            throws XMLCException {
        this.compileParams = params;
        this.xmlcDoc = xmlcDoc;
        methodInfoOutput = methodOutput;
        accessMethods = new AccessMethods(xmlcDoc);
        accessConsts = new AccessConsts(params, xmlcDoc);
        document = xmlcDoc.getDocument();
        docType = document.getDoctype();

        // Generate code for documents
        processChildNodes(DOCUMENT_VAR, document, 0);

        // Output information about omitted accessors.
        if (methodOutput != null) {
            accessMethods.printOmittedIds(methodOutput);
            accessConsts.printOmittedConstants(methodOutput);
        }
    }

    /**
     * Write the file header, package and imports.
     */
    private void createFileHeader(IndentPrinter out,
                                  String packageName,
                                  String[] standardImports) {
        out.printlns(WARNING_HEADER);

        // Set up package decl and import list
        if (packageName != null) {
            out.println("package " + packageName + ";");
        }
        for (int idx = 0; idx < standardImports.length; idx++) {
            out.println("import " + standardImports[idx] + ";");
        }
        out.println();
    }

    /**
     * Create the static field that identifies the generated class.
     * Used by auto-recompilation to determine the XMLC generated
     * class in an inheritance hierarchy.
     */
    private void createClassIdent(IndentPrinter out,
                                  String unqualClassName) {
        String[] def = {
            "/**",
            " * Field that is used to identify this as an XMLC",
            " * generated class.  Contains an reference to the",
            " * class object.",
            " */",
            "public static final Class " + XMLObject.XMLC_GENERATED_CLASS_FIELD_NAME
            + " = " + unqualClassName + ".class;"
        };
        out.printlns(def);
        out.println();
    }

    /**
     * Create the static field that contains the name of the source file. Used
     * by auto-recompilation to find the file to recompile.
     */
    private void createSourceFileConst(IndentPrinter out,
                                       String relativeSourceFile) {
        String[] def = {
            "/**",
            " * Field containing CLASSPATH relative name of the source file",
            " * that this class was generated from.",
            " */",
            "public static final String " + XMLObject.XMLC_SOURCE_FILE_FIELD_NAME
            + " = " + JavaLang.createStringConst(relativeSourceFile) + ";"
        };
        out.printlns(def);
        out.println();
    }

    /**
     * Generate an implements or extends clause.
     *
     * @param clause The string 'implements' or 'extends'.
     * @param name1 The first name to implement or extend.
     *  Maybe null.
     * @param names List of names to implement or extend.
     *  Maybe null.
     */
    private void createImplExtend(IndentPrinter out,
                                  String clause,
                                  String name1,
                                  String[] names) {
        boolean haveNames = (names != null) && (names.length > 0);
        if ((name1 != null) || haveNames) {
            out.print(" ");
            out.print(clause);
        }
        String separator = " ";
        if (name1 != null) {
            out.print(separator);
            out.print(name1);
            separator = ", ";
        }
        if (haveNames) {
            for (int i = 0; i < names.length; i++) {
                out.print(separator);
                out.print(names[i]);
                separator = ", ";
            }
        }
    }

    /**
     * Create the DOM building function.
     */
    private void createBuildDocument(IndentPrinter out,
                                     boolean createDelegate) {
        String[] header = {
            "/**",
            " * Create document as a DOM and initialize accessor method fields.",
            " */",
            "public void buildDocument() {"
        };
        String[] delegateCheck = {
            "if (" + DELEGATE_FIELD_NAME + " != null) {",
            "    " + DELEGATE_FIELD_NAME + ".buildDocument();",
            "    return;",
            "}"
        };
        out.printlns(header);
        out.enter();

        if (createDelegate) {
            out.printlns(delegateCheck);
        }
        createDocumentCreation(out);

        // Disable error checking to speed DOM build.
        createSetErrorChecking(out, false);

        out.println(nodeVarNames.getVarDefs());
        out.println(elementVarNames.getVarDefs());
        out.println(attrVarNames.getVarDefs());
        out.println();
        out.println(domInitDefs.toString());

        createSetErrorChecking(out, true);
        out.leave();
        out.println("}");
        out.println();
    }

    /**
     * Create the constructors and clone method
     */
    private void createConstructors(IndentPrinter out,
                                    String unqualClassName) {
        // Default constructor
        String[] constr = {
            "/**",
            " * Default constructor.",
            " */",
            "public " + unqualClassName + "() {",
            "    buildDocument();",
            "}"
        };
        out.printlns(constr);
        out.println();

        // Constructor optional building ofthe DOM.
        String[] condConstr = {
            "/**",
            " * Constructor with optional building of the DOM. ",
            " *",
            " * @param buildDOM If false, the DOM will not be built until",
            " * buildDocument() is called by the derived class.  If true, ",
            " * the DOM is built immediatly.",
            " */",
            "public " + unqualClassName + "(boolean buildDOM) {",
            "    if (buildDOM) {",
            "        buildDocument();",
            "    }",
            "}"
        };
        out.printlns(condConstr);
        out.println();

        // Copy constructor
        String[] copyConstr = {
            "/**",
            " * Copy constructor.",
            " * @param src The document to clone.",
            " */",
            "public " + unqualClassName + "(" + unqualClassName + " src) {",
            "    setDocument((Document)src.getDocument().cloneNode(true));",
            "    syncAccessMethods();",
            "}"
        };
        out.printlns(copyConstr);
        out.println();

        // Clone method
        String[] cloneNode = {
            "/**",
            "  * Clone the document.",
            "  * @param deep Must be true, only deep clone is supported.",
            "  */",
            "public Node cloneNode(boolean deep) {",
            "    cloneDeepCheck(deep);",
            "    return new " + unqualClassName + "(this);",
            "}"
        };
        out.printlns(cloneNode);
        out.println();
    }

    /**
     * Generate a class from the information extracted from the DOM.
     *
     * @param outWriter Print stream to write generate source to.
     * @param packageName Package name for the class.  If null, then
     *  the default package will be used.
     * @param unqualClassName Unqualified class name for the generate class.
     * @param unqualInterfaceName Unqualified interface name if an interface
     *  is going to be associated for this class.  If the class is standalone,
     *  this should be null.
     * @param willImplement Interfaces the class will implement, or null if
     *  none.
     * @param relativeSourceFile The file name, relative to the CLASSPATH
     *  on the deployed system, that is the source of the file.
     */
    public void createClass(PrintWriter outWriter,
                            String packageName,
                            String unqualClassName,
                            String unqualInterfaceName,
                            String[] willImplement,
                            String relativeSourceFile) {
        IndentPrinter out = new IndentPrinter(outWriter);

        // Delegate code is created if we are creating an interface.
        boolean createDelegate = (unqualInterfaceName != null);

        createFileHeader(out, packageName, CLASS_IMPORT_LIST);

        // Start out the class
        out.print("public class " + unqualClassName + " extends "
                    + xmlcDoc.getObjectBaseClass());
        createImplExtend(out, "implements", unqualInterfaceName,
                         willImplement);
        out.println(" {");
        out.enter();
        createClassIdent(out, unqualClassName);
        createSourceFileConst(out, relativeSourceFile);

        // Constants, fields and access methods.
        if (unqualInterfaceName == null) {
            accessConsts.createAccessConstants(out, methodInfoOutput);
        }
        accessMethods.createAccessMethods(out, methodInfoOutput,
                                          true, createDelegate);

        // Other methods.
        if (createDelegate) {
            createDelegateSupport(out, unqualInterfaceName, true);
        }
        createBuildDocument(out, createDelegate);
        accessMethods.createAccessMethodSync(out);
        createConstructors(out, unqualClassName);

        // End of class
        out.leave();
        out.println("}");
        out.done();
    }

    /**
     * Generate a interface from the information extracted from the DOM.
     *
     * @param outWriter Print stream to write generate source to.
     * @param packageName Package name for the class.  If null, then
     *  the default package will be used.
     * @param unqualInterfaceName Name of the interface.
     */
    public void createInterface(PrintWriter outWriter,
                                String packageName,
                                String unqualInterfaceName,
                                String[] willImplement) {
        IndentPrinter out = new IndentPrinter(outWriter);

        createFileHeader(out, packageName, INTERFACE_IMPORT_LIST);

        // Start out the class
        out.print("public interface " + unqualInterfaceName);
        createImplExtend(out, "extends", xmlcDoc.getObjectBaseInterface(),
                         willImplement);
        out.println(" {");
        out.enter();

        // Constants and access methods.
        accessConsts.createAccessConstants(out, methodInfoOutput);
        accessMethods.createAccessMethods(out, methodInfoOutput, 
                                          false, true);
        // Other methods
        createDelegateSupport(out, unqualInterfaceName, false);

        // End of class
        out.leave();
        out.println("}");
        out.done();
    }

    /*
     * Generate the Java source files.
     */
    public void generateJavaSource(CompileParams params,
                                   PrintWriter verboseOut)
        throws XMLCException, IOException {
        if (params.getJavaClassSource() != null) {
            // Create directory if it doesn't exist
            String parent = params.getJavaClassSource().getParent();
            if (parent != null) {
                new File(parent).mkdirs();
            }
            if (verboseOut != null) {
                verboseOut.println("    creating class: " + params.getJavaClassSource().getAbsolutePath());
            }
            PrintWriter javaOut = new PrintWriter
                (new BufferedOutputStream(new FileOutputStream(params.getJavaClassSource())));
            try {
                createClass(javaOut, params.getPackageName(),
                            params.getUnqualClassName(),
                            params.getUnqualInterfaceName(),
                            params.getWillImplement(),
                            params.getRelativeSourceFile());
            } finally {
                javaOut.close();
            }
        }
        if (params.getJavaInterfaceSource() != null) {
            if (verboseOut != null) {
                verboseOut.println("    creating interface: " + params.getJavaInterfaceSource().getAbsolutePath());
            }
            // Create directory if it doesn't exist
            String parent = params.getJavaInterfaceSource().getParent();
            if (parent != null) {
                new File(parent).mkdirs();
            }
            PrintWriter javaOut = new PrintWriter
                (new BufferedOutputStream(new FileOutputStream(params.getJavaInterfaceSource())));
            try {
                createInterface(javaOut, params.getPackageName(),
                                params.getUnqualInterfaceName(),
                                params.getWillImplement());
            } finally {
                javaOut.close();
            }
        }
    }
}
