/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: CompileParams.java,v 1.8.2.2 2000/04/04 05:29:51 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.html.HTMLProprietaryTags;
import org.enhydra.xml.io.Encoding;
import java.io.File;
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.Array;
import org.xml.sax.InputSource;

//FIXME: These are all tmp and will go away.
import org.enhydra.xml.xmlc.commands.xmlc.XMLCOptions;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.FileOutputStream;
import java.io.InputStream;

/**
 * Object to hold the parameters used by the compilation steps.
 * It also defines various constants for the parameters.
 * This class is short-term and will be replaced when complete
 * metadata is supported.
 */
public class CompileParams {
    /**
     * System property containing javac.
     */
    private static final String SYS_PROPERTY_JAVAC
        = "org.enhydra.xml.xmlc.javac";

    /**
     * Valid parsers.
     */
    public static final String[] validParsers = {
        Parse.XERCES_PARSER,
        Parse.HTML_SWING_PARSER,
        Parse.HTML_TIDY_PARSER
    };

    /**
     * Generate both interface and implementation.
     */
    public static final int GENERATE_BOTH = 1;

    /**
     * Generate an interface.
     */
    public static final int GENERATE_INTERFACE = 2;

    /**
     * Generate an implementation.
     */
    public static final int GENERATE_IMPLEMENTATION = 3;

    /**
     * Generate a standalone class.
     */
    public static final int GENERATE_CLASS = 4;

    /**
     * Source file to compile.
     */
    private File sourceDocument;

    /**
     * Parser to use.
     */
    private String parser;

    /**
     * Keep generated java source.
     */
    private boolean keep;

    /**
     * If true, don't compile the generate java source.
     */
    private boolean noCompile = false;

    /**
     * Print information about the page.
     */
    private boolean printInfo;

    /**
     * Dump out the DOM tree.
     */
    private boolean dump;

    /*
     * Name of the generated java class source file.
     */
    private File javaClassSource;

    /*
     * Name of the generated java interface source file.
     */
    private File javaInterfaceSource;

    /**
     * Fully qualified class name, or null if not specified.
     */
    private String className;

    /**
     * Source output path or null if not specified.
     */
    private File sourceOut;

    /*
     * List of interfaces the class will implement.
     */
    private Vector willImplement;

    /**
     * Package name, maybe null.
     */
    private String packageName;

    /**
     * Unqualified class name.
     */
    private String unqualClassName;

    /**
     * Unqualified interface name.
     */
    private String unqualInterfaceName;

    /**
     * Verbose output.
     */
    private boolean verbose;

    /**
     * Information about parsing.
     */
    private boolean parseInfo;

    /**
     * Information about access methods that are generated.
     */
    private boolean printAccessMethods;

    /**
     * What type of class to generate.
     */
    private int generate = GENERATE_CLASS;

    /**
     * Generate support for automatic recompilation.
     */
    private boolean forRecomp;

    /**
     * Set the installed system, CLASSPATH-relative source file path.
     * Normalized to be a relative path with '/' as the separator.
     */
    private String relativeSourceFile;

    /**
     * Table of URL edits.
     */
    private Vector urlEdits;

    /**
     * Table of element classes that are to be deleted.
     */
    private Vector deleteElementClasses;

    /**
     * If not null, write a static document to this file instead
     * of generating a java class.
     */
    private File documentOutput;

    /**
     * The default javac to use.  Value is overriden by a system
     * property SYS_PROPERTY_JAVAC.
     */
    private String defaultJavac = "javac";

    /**
     * Java compiler program.
     */
    private String javac;

    /**
     * Print version number.
     */
    private boolean printVersion = false;

    /**
     * Parser options.
     */
    private int parserOptions = 0;

    /**
     * Destination directory (-d option).
     */
    private File destDir;

    /**
     * Package destination dir (destDir + packageDir).
     */
    private File packageDestDir;

    /*
     * Flags for java compiler.  Each element is an argument to the compiler.
     */
    private Vector javacFlags;

    /*
     * Options and arguments for java compiler.  Each element is an
     * argument to the compiler.
     */
    private Vector javacOptions;

    /**
     * Should the parser do validation.  If null, its not specified.
     * Value used to set parserOptions.
     */
    private Boolean validateDocument = null;

    /**
     * Class the generated class should extend.
     */
    private String extendClassName = null;

    /**
     * DOM Document class to use.
     */
    private String domFactoryClassName = null;

    /**
     * HTML encoding.
     */
    private String htmlEncoding;

    /**
     * Proprietary tags and attribute definitions.
     */
    private HTMLProprietaryTags htmlProprietaryTags = null;

    /**
     * Generate old-style HTML class constants.
     */
    private boolean htmlOldClassConstants;

    /**
     * Generate old-style HTML name constants.
     */
    private boolean htmlOldNameConstants;

    /**
     * XCatalog files to use for XML entity resolution.
     */
    private Vector xCatalog;

    /**
     * Constructor.  Checks for system property used to define
     * the default javac.
     */
    public CompileParams() {
        // Get default javac from system properties
        String prop = System.getProperty(SYS_PROPERTY_JAVAC);
        if ((prop != null) && (prop.length() > 0)) {
            defaultJavac = prop;
        }
    }

    /**
     * Utility convert an vector to an array of a specified
     * class.
     */
    private Object[] vectorToArray(Class elementClass,
                                   Vector vector) {
        Object[] array;
        if (vector == null) {
            array = (Object[])Array.newInstance(elementClass, 0);
        } else {
            array = (Object[])Array.newInstance(elementClass, vector.size());
            vector.copyInto(array);
        }
        return array;
    }

    /**
     * Determine if a String is in a String array, ignoring case.
     */
    private boolean containsIgnoreCase(String value,
                                       String[] valueList) {
        for (int idx = 0; idx < valueList.length; idx++) {
            if (value.equalsIgnoreCase(valueList[idx])) {
                return true;
            }
        }
        return false;
    }

    /**
     * Generated a expected one of message from a list of strings.
     */
    private String expectedOneOf(String value,
                                 String[] valueList) {
        StringBuffer buf = new StringBuffer();
        buf.append(", got \"");
        buf.append(value);
        buf.append("\", expected one of: ");
        for (int idx = 0; idx < valueList.length; idx++) {
            if (idx == valueList.length-1) {
                buf.append(", or ");
            } else if (idx > 0) {
                buf.append(", ");
            }
            buf.append(valueList[idx]);
        }
        return buf.toString();
    }

    /**
     * Get the source file.
     */
    public File getSourceDocument() {
        return sourceDocument;
    }

    /**
     * Get a string that describes the input source for use in
     * messages.
     */
    public String getSourceDocumentDesc() {
        if (sourceDocument == null) {
            return "**no input document**";
        } else {
            return sourceDocument.getPath();
        }
    }

    /**
     * Get the InputSource.
     */
    public InputSource getInputSource() {
        if (sourceDocument == null) {
            return null;
        }
        InputSource input = new InputSource();
        input.setSystemId(sourceDocument.getPath());
        if (htmlEncoding != null) {
            input.setEncoding(Encoding.xml2JavaEncoding(htmlEncoding));
        }
        return input;
    }

    /**
     * Set the source file.
     */
    public void setSourceDocument(File sourceDocument) {
        this.sourceDocument = sourceDocument;
    }

    /**
     * Set the installed system, CLASSPATH-relative source file path.
     */
    public void setRelativeSourceFile(String path) {
        relativeSourceFile = path.replace('\\', '/');
    }

    /**
     * Get the installed system, CLASSPATH-relative source file path.
     */
    public String getRelativeSourceFile() {
        return relativeSourceFile;
    }

    /**
     * Get the parser to use.
     */
    public String getParser() {
        return parser;
    }

    /**
     * Set the parser to use.
     */
    public void setParser(String parser) {
        this.parser = parser;
    }

    /**
     * Get keep generated java source flag.
     */
    public boolean getKeep() {
        return keep;
    }

    /**
     * Set keep generated java source flag.
     */
    public void setKeep(boolean keep) {
        this.keep = keep;
    }

    /**
     * Get the no-compiler flag.
     */
    public boolean getNoCompile() {
        return noCompile;
    }

    /**
     * Set the no-compiler flag.
     */
    public void setNoCompile(boolean noCompile) {
        this.noCompile = noCompile;
    }

    /**
     * Get print page information flag
     */
    public boolean getPrintInfo() {
        return printInfo;
    }

    /**
     * Set print page information flag
     */
    public void setPrintInfo(boolean printInfo) {
        this.printInfo = printInfo;
    }

    /**
     * Get the dump DOM tree flag.
     */
    public boolean getDump() {
        return dump;
    }

    /**
     * Get the dump DOM tree flag.
     */
    public void setDump(boolean dump) {
        this.dump = dump;
    }

    /**
     * Get the fully qualified class name, or null if not specified.
     */
    public String getClassName() {
        return className;
    }

    /**
     * Set the fully qualified class name.
     */
    public void setClassName(String className) {
        this.className = className;
    }

    /**
     * Get the path for generated source, or null if not specified.
     */
    public File getSourceOut() {
        return sourceOut;
    }

    /**
     * Set the path for generated source.
     */
    public void setSourceOut(File sourceOut) {
        this.sourceOut = sourceOut;
    }

    /*
     * Get list of interfaces the class will implement.
     */
    public String[] getWillImplement() {
        return (String[])vectorToArray(String.class, willImplement);
    }

    /*
     * Add an interfaces to the list of interfaces the class will implement.
     */
    public void addWillImplement(String inter) {
        if (willImplement == null) {
            willImplement = new Vector();
        }
        willImplement.addElement(inter);
    }

    /**
     * Get the verbose output flag.
     */
    public boolean getVerbose() {
        return verbose;
    }

    /**
     * Set the verbose output flag.
     */
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    /**
     * Get the validate document flag.  This value is also used to set the
     * parser options.
     *
     * @return null if validation is not specified, true if it is require
     *  false document must not be validated
     */
    public Boolean getValidateDocument() {
        return validateDocument;
    }

    /**
     * Set the validate document flag.
     */
    public void setValidateDocument(Boolean validateDocument) {
        this.validateDocument = validateDocument;
    }

    /**
     * Get print information about parsing flag.
     */
    public boolean getParseInfo() {
        return parseInfo;
    }

    /**
     * Set print information about parsing flag.
     */
    public void setParseInfo(boolean parseInfo) {
        this.parseInfo = parseInfo;
    }

    /**
     * Get print information about access methods flag.
     */
    public boolean getPrintAccessMethods() {
        return printAccessMethods;
    }

    /**
     * Set print information about access methods flag.
     */
    public void setPrintAccessMethods(boolean printAccessMethods) {
        this.printAccessMethods = printAccessMethods;
    }

    /**
     * Get type of class to generate.
     */
    public int getGenerate() {
        return generate;
    }

    /**
     * Set type of class to generate.
     */
    public void setGenerate(int generate) {
        this.generate = generate;
    }

    /**
     * Get the for-recomp flag.
     */
    public boolean getForRecomp() {
        return forRecomp;
    }

    /**
     * Set the forRecomp flag.
     */
    public void setForRecomp(boolean forRecomp) {
        this.forRecomp = forRecomp;
    }

    /**
     * Get table of URL edits.
     */
    public URLEdit[] getURLEdits() {
        return (URLEdit[])vectorToArray(URLEdit.class, urlEdits);
    }

    /**
     * Add a URL edit.
     */
    public void addURLEdit(URLEdit urlEdit) {
        if (urlEdits == null) {
            urlEdits = new Vector();
        }
        urlEdits.addElement(urlEdit);
    }

    /**
     * Get table of element classes that are to be deleted.
     */
    public String[] getDeleteElementClasses() {
        return (String[])vectorToArray(String.class, deleteElementClasses);
    }

    /**
     * Add an element classes that to deleted.
     */
    public void addDeleteElementClasses(String elementClass) {
        if (deleteElementClasses == null) {
            deleteElementClasses = new Vector();
        }
        deleteElementClasses.addElement(elementClass);
    }

    /**
     * Get destination directory.
     */
    public File getDestDir() {
        return destDir;
    }

    /**
     * Set base destination directory.
     */
    public void setDestDir(File destDir) {
        this.destDir = destDir;
        packageDestDir = null;
    }

    /**
     * Determinte the package dest directory names.
     */
    private void makePackageDestDir() {
        if (destDir != null) {
            packageDestDir = destDir;
        } else {
            packageDestDir = new File(".");
        }
        if (packageName != null) {
            String[] pnames = JavaLang.parseClassName(packageName);
            for (int i = 0; i < pnames.length; i++) {
                packageDestDir = new File(packageDestDir, pnames[i]);
            }
        }
    }

    /**
     * Get the package destination directory (destDir + packageDir).
     */
    public File getPackageDestDir() {
        // If we have't determinted it yet, do it now.
        if (packageDestDir == null) {
            makePackageDestDir();
        }
        return packageDestDir;
    }

    /**
     * Get the file to write a static document to or null if
     * a static document should not be written.
     */
    public File getDocumentOutput() {
        return documentOutput;
    }

    /**
     * Set the file that a static document is to be written to.
     */
    public void setDocumentOutput(File documentOutput) {
        this.documentOutput = documentOutput;
    }

    /**
     * Get the Java compiler program.
     */
    public String getJavac() {
        return (javac != null) ? javac : defaultJavac;
    }

    /**
     * Set the Java compiler program.
     */
    public void setJavac(String javac) {
        this.javac = javac;
    }

    /**
     * Get print version number flag.
     */
    public boolean getPrintVersion() {
        return printVersion;
    }

    /**
     * Set print version number flag.
     */
    public void setPrintVersion(boolean printVersion) {
        this.printVersion = printVersion;
    }

    /**
     * Get the parser options.
     */
    public int getParserOptions() {
        return parserOptions;
    }

    /**
     * Set the parser options.
     */
    public void getParserOptions(int parserOptions) {
        this.parserOptions = parserOptions;
    }

    /*
     * Get the javac arguments (flags + options), in command line form.
     * The class destDir will be added in by the method so that it
     * might be changed on the fly.
     */
    public String[] getJavacArgs() {
        int len = 0;
        if (destDir != null) {
            len += 2;
        }
        if (javacFlags != null) {
            len += javacFlags.size();
        }
        if (javacOptions != null) {
            len += javacOptions.size();
        }

        String[] args = new String[len];
        int idx = 0;
        if (destDir != null) {
            args[idx++] = "-d";
            args[idx++] = destDir.getPath();
        }
        if (javacFlags != null) {
            for (int i = 0; i < javacFlags.size(); i++) {
                args[idx++] = (String)javacFlags.elementAt(i);
            }
        }
        if (javacOptions != null) {
            for (int i = 0; i < javacOptions.size(); i++) {
                args[idx++] = (String)javacOptions.elementAt(i);
            }
        }

        return args;
    }

    /*
     * Get the javac flags.
     */
    public String[] getJavacFlags() {
        return (String[])vectorToArray(String.class, javacFlags);
    }

    /*
     * Add a javac flag.
     */
    public void addJavacFlag(String arg) {
        if (javacFlags == null) {
            javacFlags = new Vector();
        }
        javacFlags.addElement(arg);
    }

    /*
     * Get the javac options.
     */
    public String[] getJavacOptions() {
        return (String[])vectorToArray(String.class, javacOptions);
    }

    /*
     * Add a javac option and value
     */
    public void addJavacOption(String arg,
                               String value) {
        if (javacOptions == null) {
            javacOptions = new Vector();
        }
        javacOptions.addElement(arg);
        javacOptions.addElement(value);
    }

    /*
     * Remove a javac option if it exists.
     * @returns The old value or null if the option didn't exist.
     */
    public String remoteJavacOption(String arg) {
        if (javacOptions != null) {
            for (int idx = 0; idx < javacOptions.size(); idx+=2) {
                String entry = (String)javacOptions.elementAt(idx);
                if (entry.equals(arg)) {
                    String value = (String)javacOptions.elementAt(idx+1);
                    javacOptions.removeElementAt(idx);  // argument
                    javacOptions.removeElementAt(idx);  // value
                    return value;
                }
            }
        }
        return null;
    }

    /**
     * Get the class the generated class should extend.
     */
    public String getExtendClassName() {
        return extendClassName;
    }

    /**
     * Set the class the generated class should extend.
     */
    public void setExtendClassName(String extendClassName) {
        this.extendClassName = extendClassName;
    }

    /**
     * Get the DOM factory Document class name.
     */
    public String getDOMFactoryClassName() {
        return domFactoryClassName;
    }

    /**
     * Set the DOM factory Document class name.
     */
    public void setDOMFactoryClassName(String domFactoryClassName) {
        this.domFactoryClassName = domFactoryClassName;
    }

    /**
     * Get the HTML source file encoding.
     */
    public String getHTMLEncoding() {
        return htmlEncoding;
    }

    /**
     * Set the HTML source file encoding.
     */
    public void setHTMLEncoding(String encoding) throws XMLCException {
        if ((encoding != null)
            && (Encoding.xml2JavaEncoding(encoding) == null)) {
            throw new XMLCException("Invalid encoding: \""
                                    + encoding + "\"");
        }
        htmlEncoding = encoding;
    }

    /**
     * Get HTML proprietary tags and attribute definitions or null if there
     * are none.
     */
    public HTMLProprietaryTags getHTMLProprietaryTags() {
        return htmlProprietaryTags;
    }

    /**
     * Create an HTML proprietary tags object and return it.  If it
     * already exists, just return it.
     */
    public HTMLProprietaryTags createHTMLProprietaryTags() {
        if (htmlProprietaryTags == null) {
            htmlProprietaryTags = new HTMLProprietaryTags();
        }
        return htmlProprietaryTags;
    }

    /**
     * Get old-style HTML class constants flag.
     */
    public boolean getHTMLOldClassConstants() {
        return htmlOldClassConstants;
    }

    /**
     * Set old-style HTML class constants flag.
     */
    public void setHTMLOldClassConstants(boolean value) {
        htmlOldClassConstants = value;
    }

    /**
     * Get generate old-style HTML name constants flag.
     */
    public boolean getHTMLOldNameConstants() {
        return htmlOldNameConstants;
    }

    /**
     * Set generate old-style HTML name constants flag.
     */
    public void setHTMLOldNameConstants(boolean value) {
        htmlOldNameConstants = value;
    }

    /**
     * Get the package name, or null if not set.
     */
    public String getPackageName() {
        return packageName;
    }

    /**
     * Get the unqualified class name.
     */
    public String getUnqualClassName() {
        return unqualClassName;
    }

    /**
     * Unqualified interface name.
     */
    public String getUnqualInterfaceName() {
        return unqualInterfaceName;
    }

    /*
     * Get the name of the generated java class source file.
     */
    public File getJavaClassSource() {
        return javaClassSource;
    }

    /*
     * Get the name of the generated java interface source file.
     */
    public File getJavaInterfaceSource() {
        return javaInterfaceSource;
    }

    /**
     * Get the XCatalog file names.
     */
    public String[] getXCatalog() {
        return (String[])vectorToArray(String.class, xCatalog);
    }

    /**
     * Set the XCatalog file name.
     */
    public void addXCatalog(String catalog) {
        if (xCatalog == null) {
            xCatalog = new Vector();
        }
        xCatalog.addElement(catalog);
    }

    /**
     * Handle class, package and source file names from other options.
     */
    private void setupClassName() throws XMLCException {
        String baseName;
        // If class does not exist, use file name if legal.
        if (className == null) {
            String srcName = sourceDocument.getName();
            int dotIdx = srcName.lastIndexOf('.');
            if (dotIdx >= 0) {
                className = srcName.substring(0, dotIdx);
            } else {
                className = srcName;
            }
            if (!JavaLang.legalJavaIdentifier(className)) {
                throw new XMLCException("Can't use source file base name \""
                                        + className
                                        + "\" as class name: "
                                        + "not a legal Java identifier");
            }
            baseName = className;
        } else {
            int idx = className.lastIndexOf('.');
            if (idx < 0) {
                packageName = null;
                baseName = className;
            } else {
                packageName = className.substring(0, idx);
                baseName = className.substring(idx+1);
            }
        }

        /* 
         * Support the ability for the generated source to end up in a
         * specified directory. Directory is determined by taking the
         * -sourceout value and appending the packageName.
         */
        File sourceDirectory = sourceOut;
        if (sourceDirectory == null) {
            sourceDirectory = new File(".");
        } else {
            if (packageName != null) {
                String[] packageParts = JavaLang.parseClassName(packageName);
                for (int i = 0; i < packageParts.length; i++) {
                    sourceDirectory = new File(sourceDirectory, packageParts[i]);
                }
            }
        }

        /**
         * Determine classes and interfaces names and what files to generate.
         */
        if (forRecomp) {
            generate = GENERATE_BOTH;
        }
        if (generate == GENERATE_CLASS) {
            unqualClassName = baseName;
        } else {
            unqualClassName = baseName + "Impl";
            unqualInterfaceName = baseName;
        }

        switch (generate) {
        case GENERATE_BOTH:
        case GENERATE_IMPLEMENTATION:
        case GENERATE_CLASS:
            javaClassSource = new File(sourceDirectory, unqualClassName + ".java");
        }
        switch (generate) {
        case GENERATE_BOTH:
        case GENERATE_INTERFACE:
            javaInterfaceSource = new File(sourceDirectory, unqualInterfaceName + ".java");
        }
    }

    /**
     * Default any options in the set that have not been specified.
     * Determination of document type, and hence validation and defaulting of
     * the parser base on document type, is handled outside of this class.
     */
    public void complete() throws XMLCException {
        if (sourceDocument == null) {
            if (printVersion) {
                return;  // Only print version, no other checking.
            }
            throw new XMLCException("Source documment must be specified");
        }

        // If source is being created, we need a package and class name.
        if (documentOutput == null) {
            setupClassName();
        }

        // Default relativeSourceFile 
        if ((relativeSourceFile == null) && (sourceDocument != null)) {
            if (packageName != null) {
                relativeSourceFile = packageName.replace('.', '/') + "/";
            }
            relativeSourceFile += sourceDocument.getName();
        }


        // Check for a valid parser.
        if (parser != null) {
            if (!containsIgnoreCase(parser, validParsers)) {
                throw new XMLCException("Invalid parser: "
                                        + expectedOneOf(parser, validParsers));
            }
            parser = parser.toLowerCase();
        }

        // Set parser options.
        if (validateDocument != null) {
            parserOptions |= (validateDocument.booleanValue() ?
                              XMLCParser.OPT_VALIDATE_DOCUMENT :
                              XMLCParser.OPT_NO_VALIDATE_DOCUMENT);
        }
    }

    /**
     * Save the contants of this file as a XMLC options file (*.xmlc).  This
     * is used in recompiling the JAVA class.  This is a short-term hack, the
     * long-term solution is to create an XML file.  Only options that make
     * sense for recompilation are saved.
     * <P>
     * Warning: this method will be removed in a future release.
     */
    public void save(File optionsFile) throws IOException {
        PrintWriter out = new PrintWriter(new FileOutputStream(optionsFile));
        try {
            XMLCOptions.saveRecompileParams(out, this);
        } finally {
            out.close();
        }
    }

    /**
     * Save the contants of this file as a XMLC options file (*.xmlc) in
     * the same directory and base name as the class (with a .xmlc extension).
     * <P>
     * Warning: this method will be removed in a future release.
     */
    public void save() throws IOException {
        String fileName = ((unqualInterfaceName != null) ? unqualInterfaceName : unqualClassName)
            + ".xmlc";
        File optionsFile = new File(getPackageDestDir(), fileName);
        save(optionsFile);
    }

    /**
     * Load compilation parms from a file.
     * <P>
     * Warning: this method will be removed in a future release.
     */
    public static CompileParams loadRecompileParams(String optFile,
                                                    InputStream optStream)
        throws IOException, XMLCException {
        return XMLCOptions.loadRecompileParams(optFile, optStream);
    }


    /**
     * Get values of object as a printable string.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("sourceDocument = " + sourceDocument + "\n");
        buf.append("parser = " + parser + "\n");
        buf.append("keep = " + keep + "\n");
        buf.append("noCompile = " + noCompile + "\n");
        buf.append("destDir = " + destDir + "\n");
        buf.append("printInfo = " + printInfo + "\n");
        buf.append("dump = " + dump + "\n");
        buf.append("className = " + className + "\n");
        buf.append("sourceOut = " + sourceOut + "\n");
        buf.append("packageName = " + packageName + "\n");
        buf.append("unqualClassName = " + unqualClassName + "\n");
        buf.append("verbose = " + verbose + "\n");
        buf.append("parseInfo = " + parseInfo + "\n");
        buf.append("printAccessMethods = " + printAccessMethods + "\n");
        buf.append("documentOutput = " + documentOutput + "\n");
        buf.append("javac = " + javac + "\n");
        buf.append("printVersion = " + printVersion + "\n");
        buf.append("validateDocument = " + validateDocument + "\n");
        buf.append("extendClassName = " + extendClassName + "\n");
        buf.append("domFactoryClassName = " + domFactoryClassName + "\n");
        buf.append("javaClassSource = " + javaClassSource + "\n");
        buf.append("javaInterfaceSource = " + javaInterfaceSource + "\n");
        buf.append("xCatalog = " + xCatalog + "\n");
        buf.append("generate = ");
        switch (generate) {
        case GENERATE_BOTH:
            buf.append("both");
            break;
        case GENERATE_INTERFACE:
            buf.append("interface");
            break;
        case GENERATE_IMPLEMENTATION:
            buf.append("implementation");
            break;
        case GENERATE_CLASS:
            buf.append("class");
            break;
        default:
            buf.append("unknown: " + generate);
            break;
        }

        buf.append("willImplement =\n");
        if (willImplement != null) {
            Enumeration impl = willImplement.elements();
            while (impl.hasMoreElements()) {
                buf.append("    " + (String)impl.nextElement());
            }
        }

        buf.append("urlEdits =\n");
        if (urlEdits != null) {
            Enumeration edits = urlEdits.elements();
            while (edits.hasMoreElements()) {
                buf.append("    " + ((URLEdit)(edits.nextElement())).toString() + "\n");
            }
        }
        buf.append("deleteElementClasses =\n");
        if (deleteElementClasses != null) {
            Enumeration deleteClasses = deleteElementClasses.elements();
            while (deleteClasses.hasMoreElements()) {
                buf.append("    " + (String)deleteClasses.nextElement()
                           + "\n");
            }
        }
        buf.append("javacArgs =\n");
        if (javacFlags != null) {
            for (int idx = 0; idx < javacFlags.size(); idx++) {
                buf.append("    " + (String)javacFlags.elementAt(idx) + "\n");
            }
        }
        if (javacOptions != null) {
            for (int idx = 0; idx < javacOptions.size(); idx++) {
                buf.append("    " + (String)javacOptions.elementAt(idx) + "\n");
            }
        }

        return buf.toString();
    }
}
