/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: EditDOM.java,v 1.3 2000/02/21 01:27:32 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.w3c.dom.*;
import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.dom.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

/**
 * Class to perform edits on a DOM.
 */
public class EditDOM {
    /**
     * Table of URL mappings, indexed by original URL,
     * referencing edits.
     */
    private Hashtable mappings = new Hashtable();
    
    /**
     * Vector of regexp replacements to make.
     */
    private Vector regexpMappings = new Vector();

    /**
     * Table of URL settings, indexed by id, referencing edits.
     */
    private Hashtable settings = new Hashtable();

    /**
     * Table of element classes to delete.
     */
    private Hashtable deleteElementClasses = new Hashtable();

    /**
     * Add a URL edit to the URLs to edit.
     * @param urlEdit The URLEdit object to add.
     */
    public void addURLEdit(URLEdit urlEdit) {
        if (urlEdit instanceof URLMapping) {
            URLMapping urlMapping = (URLMapping)urlEdit;
            if (urlMapping.getOriginalURL() == null) {
                // URL setting
                settings.put(urlMapping.getElementId(), urlMapping);
            } else {
                // URL mapping (maybe an id restriction)
                mappings.put(urlMapping.getOriginalURL(), urlMapping);
            }
        } else if (urlEdit instanceof URLRegexpMapping) {
            regexpMappings.addElement(urlEdit);
        } else {
            throw new XMLCError("BUG: Unimplemented URLEdit");
        }
    }

    /**
     * Add a table of URL edits.
     * @param urlMappings Mappings to add.
     */
    public void addURLEdits(URLEdit[] urlEdits) {
        for (int i = 0; i < urlEdits.length; i++) {
            addURLEdit(urlEdits[i]);
        }
    }

    /**
     * Add an element class name to the list of element classes to delete.
     * @param className The element class name.
     */
    public void addDeleteClass(String className) {
        if (deleteElementClasses == null) {
            deleteElementClasses = new Hashtable();
        }
        deleteElementClasses.put(className, className);
    }

    /**
     * Add a table of element class name to the list of element classes to delete.
     * @param classNames Table of element classes that are to be deleted.
     */
    public void addDeleteClasses(String[] classNames) {
        for (int i = 0; i < classNames.length; i++) {
            addDeleteClass(classNames[i]);
        }
    }

    /**
     * Check if an element is in one of the delete classes.
     */
    private boolean isInDeleteClass(XMLCDocument xmlcDoc,
                                    Element element) {
        // Check for the  element class
        String[] classNames = xmlcDoc.getElementClassNames(element);
        if (classNames != null) {
            // Check delete table for all classes this element is in.
            for (int idx = 0; idx < classNames.length; idx++) {
                if (deleteElementClasses.containsKey(classNames[idx])) {
                    return true;
                }
            }
        }
        return false;
    }

    /*
     * Map an element URL.
     */
    private void mapElementURL(Element element,
                               String id,
                               String urlAttr,
                               String oldURL) throws XMLCException {
        String newURL = null;

        // Try mappings in presidence order
        if (id != null) {
            URLMapping urlMapping = (URLMapping)settings.get(id);
            if (urlMapping != null) {
                newURL = urlMapping.getNewURL();
            }
        }
        if (newURL == null) {
            URLMapping urlMapping = (URLMapping)mappings.get(oldURL);
            if (urlMapping != null) {
                String mappingId = urlMapping.getElementId();
                if ((mappingId == null) || (mappingId.equals(id))) {
                    newURL = urlMapping.getNewURL();
                }
            }
        }
        if (newURL == null) {
            Enumeration regEnum = regexpMappings.elements();
            while (regEnum.hasMoreElements() && (newURL == null)) {
                URLRegexpMapping regExpMapping = (URLRegexpMapping)regEnum.nextElement();
                String mappingId = regExpMapping.getElementId();
                if ((mappingId == null) || (mappingId.equals(id))) {
                    newURL = regExpMapping.mapURL(oldURL);
                }
            }
        }
        if (newURL != null) {
            element.setAttribute(urlAttr, newURL);
        }
    }

    /**
     * Perform URL mappings on an element.
     */
    private void mapLinks(XMLCDocument xmlcDoc,
                          Element element) throws XMLCException {
        String[] attrs = xmlcDoc.getElementURLAttrs(element);
        if (attrs == null) {
            return;  // No URLs in this node.
        }
        String id = xmlcDoc.getElementId(element);

        for (int idx = 0; idx < attrs.length; idx++) {
            String oldURL = element.getAttribute(attrs[idx]);
            if ((oldURL != null) && (oldURL.length() > 0)) {
                mapElementURL(element, id, attrs[idx], oldURL);
            }
        }
    }

    /**
     * Recursively edit nodes.
     */
    private void editNodes(XMLCDocument xmlcDoc,
                           Node node,
                           Node parent)
            throws XMLCException {
        // Check for deletion first, as no other work will be done if deleted.
        if ((node instanceof Element)
            && isInDeleteClass(xmlcDoc, (Element)node)) {
            parent.removeChild(node);
            return;
        }

        // Mapping URL references.
        if (node instanceof Element) {
            mapLinks(xmlcDoc, (Element)node);
        }

        // Traverse children
        Node child = node.getFirstChild();
        while (child != null) {
            Node nextChild = child.getNextSibling(); // child might be deleted
            editNodes(xmlcDoc, child, node);
            child = nextChild;
        }
    }

    /**
     * Edit the nodes of the document.
     */
    public void edit(XMLCDocument xmlcDoc)
            throws XMLCException {
        editNodes(xmlcDoc, xmlcDoc.getDocument(), null);
    }
}
