/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: ElementIdTable.java,v 1.3 2000/02/21 01:27:32 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.dom.*;
import java.util.Hashtable;
import java.util.Enumeration;
import org.w3c.dom.*;

/**
 * Table of information about element ids.
 */
class ElementIdTable {
    /**
     * Container for document.
     */
    private XMLCDocument xmlcDoc;

    /**
     * Is this global id attribute mode.
     */
    private boolean globalIdAttrMode;

    /**
     * Table of tag name to id attribute name.  Null in global
     * id attribute mode.
     */
    private Hashtable tagName2IdAttrName;

    /**
     * Table of tag name to a Hashtable of ids.  Null in global
     * id attribute mode.
     */
    private Hashtable tagName2Ids;

    /**
     * Table of id to element class name.
     */
    public Hashtable id2ElementClassName = new Hashtable();

    /**
     * Hash table of element ids after Java name adjustment.  Avoids
     * collision.
     */
    private Hashtable adjustedIdNames = new Hashtable();

    /**
     * Table of elements ids that should have setTextNodes created.
     */
    private Hashtable getTextIds = new Hashtable();

    /**
     * Table of element ids that are not legal Java identifiers.
     */
    private Hashtable invalidElementIds = null;

    /**
     * Constructor.
     */
    public ElementIdTable(XMLCDocument xmlcDocObj) {
        xmlcDoc = xmlcDocObj;

        globalIdAttrMode = (xmlcDoc.getGlobalIdAttribute() != null);
        if (!globalIdAttrMode) {
            tagName2IdAttrName = new Hashtable();
            tagName2Ids = new Hashtable();
        }
    }

    /**
     * Adjust an element id to conform to the Java naming convention.
     * Upshifts the first character.
     */
    public String adjustElementId(String id) {
        return id.substring(0, 1).toUpperCase() + id.substring(1);
    }

    /**
     * Analyze an element and record information about it.
     * @return The element id, or null if the id will not result in
     *  an access method.
     */
    public String recordElementId(Element element) throws XMLCException {
        // Get attribute
        String idAttrName = xmlcDoc.getIdAttrName(element);
        if (idAttrName == null) {
            return null;  // element doesn't have an id attribute
        }
        String id = element.getAttribute(idAttrName);
        if ((id == null) || (id.length() == 0)) {
            return null;  // No id.
        }

        // Check if its a valid identifier?
        if (!JavaLang.legalJavaIdentifier(id)) {
            if (invalidElementIds == null) {
                invalidElementIds = new Hashtable();
            }
            invalidElementIds.put(id, id);
            return null;  // a Not legal Java identifier.
        }

        // Check that adjust name is unique.
        String adjustedId = adjustElementId(id);
        if (adjustedIdNames.containsKey(adjustedId)) {
            throw new XMLCException("Duplicate element id \"" + id
                                    + "\", id must be unique after first character is capitalized");
        }
        adjustedIdNames.put(adjustedId, adjustedId);

        // Record information
        String elementClass = xmlcDoc.nodeClassToInterface(element);
        id2ElementClassName.put(id, elementClass);

        if (!globalIdAttrMode) {
            String tagName = element.getTagName();

            tagName2IdAttrName.put(tagName, idAttrName);

            Hashtable idTable = (Hashtable)tagName2Ids.get(tagName);
            if (idTable == null) {
                idTable = new Hashtable();
                tagName2Ids.put(tagName, idTable);
            }
            idTable.put(id, elementClass);
        }

        // Remember ids that will have text accessors generated.
        if ((xmlcDoc.needsCreateTextSetMethod(element)
             && (XMLCUtil.findFirstText(element) != null))) {
            getTextIds.put(id, id);
        }
        return id;
    }

    /**
     * Get an enumeration of all ids.
     */
    public Enumeration getIds() {
        return id2ElementClassName.keys();
    }

    /**
     * Get an enumeration of all for a given tag name.
     */
    public Enumeration getIds(String tagName) {
        return ((Hashtable)tagName2Ids.get(tagName)).keys();
    }

    /**
     * Get the element class name for a given id.
     */
    public String getElementClassName(String id) {
        return (String)id2ElementClassName.get(id);
    }

    /**
     * Should a setText method be created for this element?
     */
    public boolean hasSetText(String id) {
        return getTextIds.containsKey(id);
    }

    /**
     * Get an enumeration over all tag names.
     */
    public Enumeration getTagNames() {
        if (globalIdAttrMode) {
            return null;  // Not kept for global.
        } else {
            return tagName2IdAttrName.keys();
        }
    }

    /**
     * Get the id attribute name for a tag.
     */
    public String getIdAttributeName(String tagName) {
        return (String)tagName2IdAttrName.get(tagName);
    }

    /**
     * Get an enumeration of element ids that are not legal
     * Java identifiers, or null if all ids were legal.
     */
    public Enumeration getInvalidElementIds() {
        if (invalidElementIds != null) {
            return invalidElementIds.keys();
        } else {
            return null;
        }
    }
}
