/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: ErrorReporter.java,v 1.3 2000/02/07 00:44:09 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import java.io.*;
import org.xml.sax.*;

/**
 * Object used to report error message for the user.  Writes to a file and
 * records the number of errors.  This is also a SAX compliant ErrorHandler.
 */
public class ErrorReporter implements ErrorHandler {
    /*
     * Error message are written to this file.
     */
    private PrintWriter out;

    /*
     * String buffered writer used when an external output stream is not
     * specified.  Is null if going to a stream.
     */
    private StringWriter stringOut;

    /*
     * Count of errors that have occured.
     */
    private int errorCnt = 0;

    /*
     * Count of warning that have occured.
     */
    private int warningCnt = 0;

    /**
     * Construct a new error report object.
     *
     * @param output File to recieve error output.
     */
    public ErrorReporter(PrintWriter output) {
        out = output;
    }

    /**
     * Construct a new error report object with output
     * saved in a string buffer.
     */
    public ErrorReporter() {
        stringOut = new StringWriter();
        out = new PrintWriter(stringOut, true);
    }

    /**
     * Get the count of errors that have occured.
     */
    public int getErrorCnt() {
        return errorCnt;
    }

    /**
     * Get the count of warnings that have occured.
     */
    public int getWarningCnt() {
        return warningCnt;
    }

    /**
     * Get the buffer containing the output if internal buffering
     * has been used.
     *
     * @return buffer containing output or null if output was written to
     *  an external stream.
     */
    public String getOutput() {
        if (stringOut == null) {
            return null;
        } else {
            return stringOut.toString();
        } 
    }

    /**
     * Generate a message with file name and line number.
     * This is in a format parsable by emacs and other
     * editors.
     */
    private void printFileLineMsg(String msg,
                                  boolean isError,
                                  String fileName,
                                  int lineNum) {
        out.println(fileName + ":" + lineNum + ": "
                    + (isError ? "Error: " : "Warning: ")
                    + msg);
    }

    /**
     * Report an error.
     */
    public void error(String msg) {
        out.print("Error: ");
        out.println(msg);
        errorCnt++;
    }

    /**
     * Report an error with file name and line number.
     */
    public void error(String msg,
                      String fileName,
                      int lineNum) {
        printFileLineMsg(msg, true, fileName, lineNum);
        errorCnt++;
    }

    /**
     * Report an error from an exception
     */
    public void error(String msg, Throwable except) {
        out.println("Error: " + msg + ": " + except.getClass().getName()
                    + ": " + except.getMessage());
        except.printStackTrace(out);
        errorCnt++;
    }

    /**
     * Report a warning.
     */
    public void warning(String msg) {
        out.print("Warning: ");
        out.println(msg);
        warningCnt++;
    }

    /**
     * Report a warning with file name and line number.
     */
    public void warning(String msg,
                        String fileName,
                        int lineNum) {
        printFileLineMsg(msg, false, fileName, lineNum);
        warningCnt++;
    }

    /**
     * Print a SAX parser message in a consistent style, consistent
     * with emacs compile mode (same as grep, cc, javac, etc).
     */
    private void printSAXMsg(String typeStr,
                             SAXParseException exception) {
        out.print(exception.getSystemId()
                  + ":" + exception.getLineNumber()
                  + ":");
        if (typeStr != null) {
            out.print(typeStr + ": ");
        }
        out.println(exception.getMessage());
        Exception cause = exception.getException();
        if (cause != null) {
            out.println("Cause: " + cause.getMessage());
            cause.printStackTrace(out);
        }
    }
    
    /**
     * Receive notification of a SAX warning.
     */
    public void warning(SAXParseException exception)
        throws SAXException {
        printSAXMsg("Warning", exception);
        warningCnt++;
    }


    /**
     * Receive notification of a SAX recoverable error.
     */
    public void error(SAXParseException exception)
        throws SAXException {
        printSAXMsg(null, exception);
        errorCnt++;
    }
    

    /**
     * Receive notification of a SAX non-recoverable error.
     */
    public void fatalError(SAXParseException exception)
        throws SAXException {
        printSAXMsg("Fatal error", exception);
        errorCnt++;
    }
}
