/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: IndentPrinter.java,v 1.3 2000/02/20 04:29:07 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.XMLCError;
import java.io.PrintWriter;
import java.io.StringWriter;

/**
 * Class to print lines with indentation.
 */
class IndentPrinter {
    /*
     * Output stream.
     */
    private PrintWriter out;

    /*
     * String buffering writer if no other writer specified.
     */
    private StringWriter stringWriter;

    /*
     * String used in indentation.
     */
    private static final String INDENT_STR = "    ";

    /*
     * Current level.
     */
    private int level = 0;

    /*
     * Are we at the beginning of the line.
     */
    boolean atBOL = true;

    /**
     * Constructor with specified writer.
     */
    public IndentPrinter(PrintWriter outWriter) {
        out = outWriter;
    }

    /**
     * Constructor to with internal string buffer.
     */
    public IndentPrinter() {
        stringWriter = new StringWriter();
        out = new PrintWriter(stringWriter);
    }

    /**
     * Get the string if internally buffered, otherwise null.
     */
    public String getStringBuffer() {
        if (stringWriter == null) {
            return null;
        } else {
            done();
            return stringWriter.toString();
        }
    }

    /**
     * Called when through generating code.  Flush the buffer
     * and make sure that indentation is zero.  If not, we might
     * have screawed something up.
     */
    public void done() {
        out.flush();
        if (level != 0) {
            // This could indicate something more serious.
            throw new XMLCError("BUG: code indentation not zero at end");
        }
    }

    /**
     * Increase indentation level.
     */
    public void enter() {
        level++;
    }

    /**
     * Decrease indentation level.
     */
    public void leave() {
        level--;
        if (level < 0) {
            // This could indicate something more serious.
            throw new XMLCError("BUG: code indentation underflow");
        }
    }

    /**
     * Write the current indentation and clear BOL flag.
     */
    private void writeIndent() {
        for (int cnt = 0; cnt < level; cnt++) {
            out.print(INDENT_STR);
        }
        atBOL = false;
    }

    /**
     * Write lines from a string, indenting at the beginning of the line.
     * The string is split at embedded newlines.
     */
    public void print(String str) {
        int idx = 0;
        int strlen = str.length();
        int nlIdx;

        // Write string upto last newline.
        while ((idx < strlen) &&
               ((nlIdx = str.indexOf('\n', idx)) >= 0)) {
            if (atBOL) {
                writeIndent();
            }
            for (;idx <= nlIdx; idx++) {
                out.print(str.charAt(idx));
            }
            atBOL = true;  // Last character was a newline.
        }
        
        // Write characters following the last newline.  If there
        // were no newlines in the string, this writes the entire string.
        if (idx < strlen) {
            if (atBOL) {
                writeIndent();
            }
            if (idx == 0) {
                // fast path
                out.print(str);
            } else {
                for (;idx < strlen; idx++) {
                    out.print(str.charAt(idx));
                }
            }
        }
    }

    /**
     * Write lines from a string, indenting at the beginning of the line.
     * The string is split at embedded newlines and a newlines is written
     * at the end.
     */
    public void println(String str) {
        print(str);
        out.println();
        atBOL = true;
    }

    /**
     * Print a newline.
     */
    public void println() {
        out.println();
        atBOL = true;
    }

    /**
     * Write an array of strings, one line per string. Indenting at the
     * beginning of the line.  The string is split at embedded newlines.
     */
    public void printlns(String[] strs) {
        for (int i = 0; i < strs.length; i++) {
            println(strs[i]);
        }
    }

}
