/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: JavaLang.java,v 1.4 2000/02/04 10:06:59 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;

/**
 * Collection of  methods dealing with the java language.
 */
public class JavaLang {

    /**
     * Prevent instanciation.
     */
    private JavaLang() {
    }

    /*
     * Parse a class or package name.
     */
    public static String[] parseClassName(String name) {
        StringTokenizer tokens = new StringTokenizer(name, ".");
        String[] names = new String[tokens.countTokens()];

        for (int i = 0; i < names.length; i++) {
            names[i] = tokens.nextToken();
        }
        return names;
    }

    /*
     * Get the class name without the package.
     * @param obj Any object or class object.
     */
    public static String simpleClassName(Object obj) {
        if (obj instanceof Class) {
            return simpleClassName(((Class)obj).getName());
        } else {
            return simpleClassName(obj.getClass().getName());
        }
    }

    /*
     * Get the class name without the package.
     */
    public static String simpleClassName(String className) {
        int lastDot = className.lastIndexOf('.');
        if (lastDot < 0) {
            return className;
        } else {
            return className.substring(lastDot+1);
        }
    }

    /*
     * Get the package name of a class.
     */
    public static String getPackageName(String className) {
        int lastDot = className.lastIndexOf('.');
        if (lastDot < 0) {
            return "";
        } else {
            return className.substring(0, lastDot);
        }
    }

    /**
     * Convert a unicode character to its \\uXXXX representation
     * and append to a buffer.
     */
    private static void toUnicodeEncoding(char ch, StringBuffer buf) {
        String numStr = Integer.toHexString(ch);
        buf.append("\\u");
        for (int cnt = 4-numStr.length(); cnt > 0; cnt--) {
            buf.append('0');
        }
        buf.append(numStr);
    }

    /**
     * Generate a Java string constant, quoting `\' and `"' and converting
     * newlines.
     *
     * @param text Text to quote.
     * @return The quoted string.
     */
    public static String createStringConst(String text) {
        if (text == null) {
            return "null";
        }

        StringBuffer str = new StringBuffer();
        int len = text.length ();

        str.append ("\"");
        for (int idx = 0; idx < len; idx++) {
            char textChar = text.charAt (idx);
            switch (textChar) {
            case '\\':
            case '"':
                str.append ('\\');
                str.append (textChar);
                break;
            case '\t':
                str.append ("\\t");
                break;
            case '\n':
                str.append ("\\n");
                break;
            case '\r':
                str.append ("\\r");
                break;
            default:
                if ((textChar > 127) || (Character.isISOControl(textChar))) {
                    toUnicodeEncoding(textChar, str);
                } else {
                    str.append (textChar);
                }
                break;
            }
        }
        str.append ("\"");
        return str.toString();
    }

    /*
     * Determine if a string is a legal Java identifier
     */
    public static boolean legalJavaIdentifier(String str) {
        if (str.length() == 0) {
            return false;
        }
        if (!Character.isJavaIdentifierStart(str.charAt(0))) {
            return false;
        }
        for (int j = 1; j < str.length(); j++) {
            if (!Character.isJavaIdentifierPart(str.charAt(j))) {
                return false;
            }
        }
        return true;
    }

    /**
     * Lame insertion sort for things that don't have to be
     * fast.  This is a hacked substitute for a real sort.
     * It will be repleaced with java 1.2 util classes one day.
     * FIXME: Replace with java.util.SortedMap class
     */
    public static void insertSorted(String str, Vector table) {
        int idx;
        for (idx = 0; idx < table.size(); idx++) {
            if (str.compareTo((String)table.elementAt(idx)) > 0) {
                break;
            }
        }
        table.insertElementAt(str, idx);
    }

    /**
     * Generate a sorted vector of strings from an enumeration.
     * FIXME: Replace with java.util.SortedMap class
     */
    public static Vector getSortedTable(Enumeration enum) {
        Vector table = new Vector();
        while (enum.hasMoreElements()) {
            JavaLang.insertSorted((String)enum.nextElement(), table);
        }
        return table;
    }
}
