/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: Javac.java,v 1.3 2000/02/14 18:48:36 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import java.io.*;
import java.util.Vector;

/**
 * Run a Java compiler as a child process.
 */
public class Javac extends ProcessRunner {
    /**
     * Default options.
     */
    private static final int DEFAULT_OPTIONS = PASS_STDOUT|PASS_STDERR;

    /**
     * Java program to run.
     */
    private String javacProg = "javac";

    /**
     * Construct a javac runner, specifying a java compiler.
     */
    public Javac(String javacCommand) {
        super(DEFAULT_OPTIONS);
        if (javacCommand != null) {
            javacProg = javacCommand;
        }
    }

    /**
     * Construct with default compiler.
     */
    public Javac() {
        this(null);
    }
    
    /**
     * Run a javac process.  Stdout/stderr are written to the standard
     * descriptors.
     *
     * @param javaSources List of java source files to compile.
     * @param verboseOut Write verbose message to this file if not null.
     * @param javacArgs Other arguments to pass to compiler.
     */
    public void compile(File[] javaSources,
                        ErrorReporter errorReporter,
                        PrintWriter verboseOut,
                        String[] javacArgs)
            throws XMLCException {
        // Assemble command arguments.
        Vector cmdVect = new Vector();
        cmdVect.addElement(javacProg);
        for (int idx = 0; idx < javacArgs.length; idx++) {
            cmdVect.addElement(javacArgs[idx]);
        }
        for (int idx = 0; idx < javaSources.length; idx++) {
            cmdVect.addElement(javaSources[idx].getPath());
        }

        String[] cmd = new String[cmdVect.size()];
        cmdVect.copyInto(cmd);

        // Run process
        run(cmd, errorReporter, verboseOut,
            "compile of generated java code failed");
    }

    /**
     * Run a javac process.  Stdout/stderr are written to the standard
     * descriptors.
     *
     * @param params Compilation parameters.
     */
    public void compile(CompileParams params,
                        ErrorReporter errorReporter,
                        PrintWriter verboseOut)
            throws XMLCException {
        int numSources = 0;
        if (params.getJavaClassSource() != null) {
            numSources++;
        }
        if (params.getJavaInterfaceSource() != null) {
            numSources++;
        }
        File[] sources = new File[numSources];
        int idx = 0;
        if (params.getJavaClassSource() != null) {
            sources[idx++] = params.getJavaClassSource();
        }
        if (params.getJavaInterfaceSource() != null) {
            sources[idx++] = params.getJavaInterfaceSource();
        }
        compile(sources,
                errorReporter,
                params.getVerbose() ? verboseOut : null,
                params.getJavacArgs());
    }
}
