/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: LineNumberMapReader.java,v 1.3.2.1 2000/04/04 05:29:52 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import java.io.Reader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.LineNumberReader;
import java.util.Vector;
import java.util.Enumeration;
import org.xml.sax.InputSource;

/**
 * A file reader that keeps a map of byte offsets to line numbers.
 */
public class LineNumberMapReader extends Reader {
    /**
     * Input file.
     */
    InputSource inputSource;

    /**
     * Actual reading class that tracks line numbers.
     */
    private LineNumberReader reader;

    /**
     * Structure to map byte offset to line numnber.
     */
    private class LineNumberOffset {
        public int lineNum;     //zero-based line number.
        public int byteOffset;  //zero-based byte offset.
    }

    /**
     * Map of line number to offset.
     */
    private Vector map = new Vector();
    
    /**
     * Current offset in the file.
     */
    private int curOffset = 0;

    /**
     * Add a line number to the map.
     */
    private void addLine(int line, int off) {
        LineNumberOffset lineEntry = new LineNumberOffset();
        lineEntry.lineNum  = line;
        lineEntry.byteOffset = off;
        map.addElement(lineEntry);
    }
    /**
     * Construct a new reader for the specified file.
     */
    public LineNumberMapReader(InputSource source)
            throws IOException {
        inputSource = source;
        reader = new LineNumberReader(InputHandling.obtainReader(source));
        addLine(1, 0);
    }

    /**
     * Get the file name or URL of the input.
     */
    public String getFileName() {
        return inputSource.getSystemId();
    }
    
    /**
     * Get the line number for a byte offset.  A one-based line number is
     * returned.
     */
    public int getLineNumber(int off) {
        // Start out with the first entry to remove special cases.
        LineNumberOffset lineEntry = null;
        LineNumberOffset prevLineEntry = (LineNumberOffset)map.elementAt(0);;

        // Search for an entry.
        Enumeration entries = map.elements();
        while (entries.hasMoreElements()) {
            lineEntry = (LineNumberOffset)entries.nextElement();
            if (off < lineEntry.byteOffset) {
                return prevLineEntry.lineNum + 1;
            }
            prevLineEntry = lineEntry;
        }
        // Offset is greater than any use, return the last.
        return prevLineEntry.lineNum + 1;
    }

    /**
     * Read characters into a portion of an array.
     *
     * @see Reader#read
     * @exception  IOException  If an I/O error occurs
     */
    public int read(char cbuf[], int off, int len) throws IOException {
        int lineNum = reader.getLineNumber();
        int num = reader.read(cbuf, off, len);
        if (num < 0) {
            return num;
        }
        for (int cnt = 0; cnt < num; cnt++) {
            curOffset++;
	    switch (cbuf[off+cnt]) {
	    case '\r':
	    case '\n':
                lineNum++;   //
                addLine(lineNum, curOffset);
	    }
        }
        return num;
    }

    /**
     * Close the stream.
     *
     * @see Reader#close
     * @exception  IOException  If an I/O error occurs
     */
    public void close() throws IOException {
        reader.close();
    }
}
