/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: Parse.java,v 1.6.2.1 2000/04/04 05:29:52 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.parsers.*;
import org.enhydra.xml.xmlc.html.HTMLParser;
import org.enhydra.xml.xmlc.html.HTMLProprietaryTags;
import org.enhydra.xml.xmlc.html.parsers.swing.SwingHTMLParser;
import org.enhydra.xml.xmlc.html.parsers.tidy.TidyHTMLParser;
import org.enhydra.xml.xmlc.parsers.xerces.XercesParser;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import java.io.*;


/**
 * Parse a XML or HTML document into a DOM.
 */
public class Parse {
    /**
     * Symbol used to specify the Xerces parser.
     */
    public static final String XERCES_PARSER = "xerces";

    /**
     * Symbol used to specify the HTML Swing parser.
     */
    public static final String HTML_SWING_PARSER = "swing";

    /**
     * Symbol used to specify the HTML Tidy parser.
     */
    public static final String HTML_TIDY_PARSER = "tidy";

    /**
     * XML parser object.
     */
    private XMLCParser xmlcParser;

    /**
     * Error output.
     */
    private ErrorReporter errorReporter;

    /**
     * Verbose tracing.
     */
    private ParseTracer tracer;

    /**
     * Construct a new file parser.
     *
     * @param errorReporter Object used to handle errors.
     * @param traceFile File to write parse trace information or null for no
     *  tracing.
     */
    public Parse(ErrorReporter errorReporter,
                 PrintWriter traceFile) {
        this.errorReporter = errorReporter;
        tracer = new ParseTracer(traceFile);
    }

    /*
     * Is this an the name of an HTML parser?
     */
    private boolean isHtmlParser(String name) {
        return name.equalsIgnoreCase(HTML_SWING_PARSER)
            || name.equalsIgnoreCase(HTML_TIDY_PARSER);
    }

    /*
     * Determine if a file is an XML document.
     * FIXME: This doesn't really work for, since one might not be able
     * to open the file twice.
     */
    private boolean isXMLDocument(InputSource input) 
        throws XMLCException, IOException {
        
        // Check for the beginning of the XMLDecl (<?xml
        Reader reader = InputHandling.obtainReader(input);
        try {
            char[] buf = new char[5];
            if (reader.read(buf, 0, buf.length) != buf.length) {
                return false;
            }
            return new String(buf).equals("<?xml");
        } finally {
            reader.close();
        }
    }

    /**
     * Determine the parser to use.
     */
    private void setupParser(String parserName,
                             int options,
                             boolean xmlDocument,
                             HTMLProprietaryTags htmlProprietaryTags)
        throws XMLCException, IOException {

        if (parserName == null) {
            // Use default parser based on document type.
            if (xmlDocument) {
                parserName = XERCES_PARSER;
            } else {
                parserName = HTML_TIDY_PARSER;
            }
        }

        // Check for conflicts with parser and document type.
        if (xmlDocument) {
            if (isHtmlParser(parserName)) {
                throw new XMLCException("Document appears to be an XML document; the "
                                        + parserName + " parser only supports HTML");
            }
            if (htmlProprietaryTags != null) {
                throw new XMLCException("proprietary HTML tags may not be specified for a XML document");
            }
        } else {
            if (!isHtmlParser(parserName)) {
                throw new XMLCException("Document appears to be an HTML document; the "
                                        + parserName + " parser only supports XML"
                                        + " (does the document start with `<?xml ... ?>'?)");
            }
        }

        // Load and initialize parser.
        if (parserName.equals(HTML_SWING_PARSER)) {
            xmlcParser = new SwingHTMLParser();
        } else if (parserName.equals(HTML_TIDY_PARSER)) {
            xmlcParser = new TidyHTMLParser(); 
        } else if (parserName.equals(XERCES_PARSER)) {
            xmlcParser = new XercesParser();
        } else {
            throw new XMLCError("Unknown parser \"" + parserName + "\"");
        }

        // Determine if options are legal.
        int validOptions = xmlcParser.getSupportedOptions();
        if (((options & XMLCParser.OPT_VALIDATE_DOCUMENT) != 0)
            && ((validOptions & XMLCParser.OPT_VALIDATE_DOCUMENT) == 0)) {
            throw new XMLCError(parserName + " parser does not support document validation");
        }
        if (((options & XMLCParser.OPT_NO_VALIDATE_DOCUMENT) != 0)
            && ((validOptions & XMLCParser.OPT_NO_VALIDATE_DOCUMENT) == 0)) {
            throw new XMLCError(parserName + " parser can't disable document validation");
        }

        if (htmlProprietaryTags != null) {
            ((HTMLParser)xmlcParser).setProprietaryTags(htmlProprietaryTags);
        }
    }

    /**
     * Load an optional parser.  Only available if XMLC was compiled
     * with support for the and it is in the classpath.
     */
    private XMLCParser loadOptionalParser(String parserClassName,
                                          String parserDesc)
        throws XMLCException {
        try {
            Class parserClass = Class.forName(parserClassName);
            return (XMLCParser)parserClass.newInstance();
        } catch (ClassNotFoundException except)  {
            throw new XMLCException("can't load " + parserDesc
                                    + " parser; support not compiled into xmlc",
                                    except);
        } catch (NoClassDefFoundError except) {
            throw new XMLCError("can't load " + parserDesc
                                + " parser; does not appear to be on CLASSPATH",
                                except);
        } catch (Exception except) {
            throw new XMLCError("can't load " + parserDesc + " parser", except);
        }
    }

    /*
     * Parse the page into the DOM and perform various checks and edits.
     *
     * @param params Compilation parameters.
     */
    public XMLCDocument parse(CompileParams params)
        throws XMLCException, IOException {

        setupParser(params.getParser(),
                    params.getParserOptions(),
                    isXMLDocument(params.getInputSource()),
                    params.getHTMLProprietaryTags());
        XMLCDocument xmlcDoc;
        try {
            xmlcDoc = xmlcParser.parse(params.getInputSource(), 
                                       params.getParserOptions(),
                                       params.getDOMFactoryClassName(),
                                       params.getExtendClassName(),
                                       params.getXCatalog(),
                                       errorReporter,
                                       tracer);
        } catch (SAXException except) {
            Exception useExcept = except.getException();
            if (useExcept == null) {
                useExcept = except;
            }
            throw new XMLCException("Parse of \"" + params.getSourceDocumentDesc()
                                    + "\" failed", useExcept);
        }
        int cnt = errorReporter.getErrorCnt();
        if (cnt > 0) {
            throw new XMLCException(cnt + " error" + ((cnt == 1) ? "" : "s")
                                    + " parsing document");
        }

        // Normalize the text nodes.
        Element root = xmlcDoc.getDocument().getDocumentElement();
        if (root != null) {
            root.normalize();
        }

        return xmlcDoc;
    }    
}
