/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: ProcessRunner.java,v 1.3 2000/02/14 18:48:37 markd Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import java.io.*;
import java.util.Vector;

/**
 * Run a process, optionally collecting stdout/stderr.
 */
public class ProcessRunner {
    /**
     * Options to enable passing or collecting of output.
     */
    public static final int PASS_STDOUT = 0x01;
    public static final int PASS_STDERR = 0x02;
    public static final int COLLECT_STDOUT = 0x04;
    public static final int COLLECT_STDERR = 0x08;
    public static final int STDERR_TO_REPORTER = 0x10;
    public static final int DUMP_STDERR_ON_FAIL = 0x20;
    public static final int NO_EXCEPTION_ON_ERR_EXIT = 0x40;

    /*
     * Output streams; maybe null.
     */
    private OutputStream stdoutStream;
    private OutputStream stderrStream;

    /*
     * Collection buffers; maybe null.
     */
    private StringBuffer stdoutBuf;
    private StringBuffer stderrBuf;

    /**
     * Options.
     */
    private int options;
    
    /**
     * Exit code of last process.
     */
    int exitCode;

    /**
     * Constructor.
     */
    public ProcessRunner(int opts) {
        setOptions(opts);
    }

    /**
     * Set the options
     */
    public void setOptions(int opts) {
        options = opts;
        if ((options & PASS_STDOUT) != 0) {
            stdoutStream = System.out;
        } else {
            stdoutStream = null;
        }
        if ((options & PASS_STDERR) != 0) {
            stderrStream = System.err;
        } else {
            stderrStream = null;
        }
        if ((options & DUMP_STDERR_ON_FAIL) != 0) {
            options |= COLLECT_STDERR;
        }
    }

    /**
     * Get the options
     */
    public int getOptions() {
        return options;
    }

    /*
     * Get exit code of process.
     */
    public int getExitCode() {
        return exitCode;
    }

    /*
     * Dump stderr if requested on a failure.
     */
    private void dumpStderr(ErrorReporter errorReporter) {
        if ((options & DUMP_STDERR_ON_FAIL) != 0) {
            errorReporter.error(stderrBuf.toString());
        }
    }

    /**
     * Monitor the running process.
     */
    private void monProc(Process proc) 
        throws IOException, InterruptedException {

        // Don't wipe out old buffer in case it was saved.
        if ((options & COLLECT_STDOUT) != 0) {
            stdoutBuf = new StringBuffer();
        } else {
            stdoutBuf = null;
        }
        if ((options & COLLECT_STDERR) != 0) {
            stderrBuf = new StringBuffer();
        } else {
            stderrBuf = null;
        }

        OutputMonitorThread outMon =
            new OutputMonitorThread(proc.getInputStream(),
                                    stdoutStream,
                                    stdoutBuf);
        OutputMonitorThread errMon =
            new OutputMonitorThread(proc.getErrorStream(),
                                    stderrStream,
                                    stderrBuf);
        try {
            outMon.start();
            errMon.start();
            proc.waitFor();
        } catch (InterruptedException except) {
            try {
                // Last ditch effort to clean up proc
                proc.destroy();
                proc.waitFor();
            } catch (InterruptedException drop) {
                // Already have an error...
            }
            outMon.stop();
            errMon.stop();
        } finally {
            outMon.join();
            errMon.join();
        }
        if (outMon.getException() != null) {
            throw outMon.getException();
        }
        if (errMon.getException() != null) {
            throw errMon.getException();
        }
    }

    /**
     * Run a process.  Stdout/stderr are optionally copied to the standard
     * descriptors and/or collected.
     *
     * @param cmd Program to run and arguments.
     * @param verboseOut Write verbose message to this file if not null.
     * @param failMsg Message to use in exception if process fails.
     */
    public void run(String[] cmd,
                    ErrorReporter errorReporter,
                    PrintWriter verboseOut,
                    String failMsg) throws XMLCException {
        if (verboseOut != null) {
            for (int i = 0; i < cmd.length; i++) {
                if (i > 0) {
                    verboseOut.print(" ");
                }
                verboseOut.print(cmd[i]);
            }
            verboseOut.println();
        }
        
        Process proc;
        exitCode = 0;
        try {
            proc = Runtime.getRuntime().exec(cmd);
            monProc(proc);
        } catch (IOException except) {
            dumpStderr(errorReporter);
            throw new XMLCException(except);
        } catch (InterruptedException except) {
            dumpStderr(errorReporter);
            throw new XMLCException(except);
        }
        exitCode = proc.exitValue();
        if (exitCode != 0) {
            dumpStderr(errorReporter);
            if ((options & NO_EXCEPTION_ON_ERR_EXIT) == 0) {
                throw new XMLCException(failMsg);
            }
        }
    }

    /**
     * Get accumulated stdout, or null if it was not collected.
     */
    public String getStdout() {
        return stdoutBuf.toString();
    }

    /**
     * Get accumulated stderr, or null if it was not collected.
     */
    public String getStderr() {
        return stderrBuf.toString();
    }
}
