/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: DocBuilder.java,v 1.5 2000/02/21 01:27:33 markd Exp $
 */

package org.enhydra.xml.xmlc.dom;

import org.enhydra.xml.xmlc.*;
import org.w3c.dom.*;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Class for building a Document DOM.
 * <P>
 * The document builder functions assume they are being called in the order the
 * document is parsed.  They keep a current node where new child nodes are
 * appended.
 */
public class DocBuilder {
    /**
     * Factory for creating the document.
     */
    XMLCDomFactory domFactory;

    /**
     * Object being used to build the DocumentType object. If an HTML
     * file is being parsed, this is null.
     */
    DocTypeBuilder docTypeBuilder;
    
    /**
     * Information from XML header.
     */
    private String docXMLVersion;
    private String docEncoding;
    private boolean docStandalone;

    /**
     * The document.
     */
    private Document document;

    /**
     * The current node that is being constructed.  This functions
     * as a stack during document construction.
     */
    protected Node currentNode;

    /**
     * Holds comments before the first element until the first element is
     * encountered.
     */
    private Vector pendingComments;

    /**
     * Constructor.
     * 
     * @param domFactory Factory class for Documents.
     * @param docTypeBuilder Object being used to construct the DocumentType.
     *  It will be called to get the document type when first needed.
     *  Should be null to indicate an HTML document.
     */
    public DocBuilder(XMLCDomFactory domFactory,
                      DocTypeBuilder docTypeBuilder) {
        this.domFactory = domFactory;
        this.docTypeBuilder = docTypeBuilder;
    }

    /**
     * Generate error about a method being called that should
     * be called before the document is created.
     */
    private void docNotCreatedError() {
        throw new XMLCError("Bug: parser event on document contents occured before document is created");
    }

    /**
     * Create the document.  Called only when needed to allow DocumentType
     * information to be collected.
     */
    protected Document createDocument(String rootTagName) {
        return domFactory.createDocument(docTypeBuilder.getDocumentTypeName(),
                                         docTypeBuilder.getCreateDocType());
    }

    /**
     * Hold a comment until the first element is encountered.
     */
    private void holdComment(String data) {
        if (pendingComments == null) {
            pendingComments = new Vector();
        }
        pendingComments.addElement(data);
    }

    /**
     * Add in comments that occured before the first element.
     */
    private void addPendingComments() {
        // Add comments before document element.
        Node first = document.getFirstChild();
        for (int idx = pendingComments.size()-1; idx >= 0; idx--) {
            Comment com = document.createComment((String)pendingComments.elementAt(idx));
            document.insertBefore(com, first);
        }
    }

    /**
     * Create the Document object when the first element of the document is
     * found.  This is delayed until the the first element, since all
     * information needed to build the document is not available until that
     * time.
     */
    private void makeDocument(String rootTagName) {
        // Set document name from root tag if necessary.
        if ((docTypeBuilder != null)
            && (docTypeBuilder.getDocumentTypeName() == null)) {
            docTypeBuilder.setDocumentTypeName(rootTagName);
        }

        // Create the document, which also creates the DocumentElement.
        document = createDocument(rootTagName);
        currentNode = document.getDocumentElement();
        if (currentNode == null) {
            throw new XMLCError("newly created document did not have a document (root) element");
        }

        if (pendingComments != null) {
            addPendingComments();
        }
    }

    /**
     * Get the document associated with this object.
     */
    public Document getDocument() {
        if (document == null) {
            throw new XMLCError("XMLC bug: Attempt to get document that hasn;t been created");
        }
        return document;
    }

    /**
     * Get the DOM factory associated with this object.
     */
    public XMLCDomFactory getDomFactory() {
        return domFactory;
    }

    /**
     * Get the node on the top of the stack during parsing.
     * FIXME: To support the broken swing parser.
     */
    public Node getCurrentNode() {
        return currentNode;
    }

    /**
     * Pop the current node off of the stack.  This is *only* used
     * during error recover from a broken parser.
     * FIXME: This and getCurrentNode() should go away when the
     * with the broken swing parser.
     */
    public void popCurrentNode() {
        currentNode = currentNode.getParentNode();
    }

    /**
     * Set the XML version.
     *
     * @param xmlVersion XML version string.
     */
    public void setXMLVersion(String xmlVersion) {
        docXMLVersion = xmlVersion;
    }

    /**
     * Get the XML version.
     */
    public String getXMLVersion() {
        return docXMLVersion;
    }

    /**
     * Set the standalone flag.
     *
     * @param standalone Is standalone specified?
     */
    public void setStandalone(boolean standalone) {
        docStandalone = standalone;
    }

    /**
     * Get the standalone flag.
     */
    public boolean getStandalone() {
        return docStandalone;
    }

    /**
     * Get the publicId.
     */
    public String getPublicId() {
        return (docTypeBuilder != null) ? docTypeBuilder.getPublicId() : null;
    }

    /**
     * Get the systemId.
     */
    public String getSystemId() {
        return (docTypeBuilder != null) ? docTypeBuilder.getSystemId() : null;
    }

    /**
     * Set the encoding for the document.
     *
     * @param encoding The encoding for the document
     */
    public void setEncoding(String encoding) {
        docEncoding = encoding;
    }

    /**
     * Set the encoding specified in the document.
     */
    public String getEncoding() {
        return docEncoding;
    }

    /**
     * Start a new Element.
     */
    public void startElement(String tagName) {
        if (document == null) {
            // root (first) element
            makeDocument(tagName);
        } else {
            // non-root element
            Element element = document.createElement(tagName);
            currentNode.appendChild(element);
            currentNode = element;
        }
    }
    
    /**
     * Add an attribute to the element on the top of the
     * stack.
     */
    public void addAttribute(String name, String value) {
        ((Element)currentNode).setAttribute(name, value);
    }

    /**
     * Finish the element being constructed. 
     */
    public void finishElement() {
        if (currentNode == null) {
            throw new XMLCError("node stack underflow; malformed document");
        }
        if (!(currentNode instanceof Element)) {
            throw new XMLCError("DOM node top of stack not a element for end tag");
        }
        currentNode = currentNode.getParentNode();
    }

    /**
     * Start an entity reference in the document (not DTD).
     */
    public void startEntityReference(String entityName) {
        if (document == null) {
            docNotCreatedError();
        }

        EntityReference entityRef = document.createEntityReference(entityName);
        currentNode.appendChild(entityRef);
        currentNode = entityRef;
    }

    /**
     * End an entity reference.
     */
    public void endEntityReference() {
        if (currentNode == null) {
            throw new XMLCError("node stack underflow; malformed document");
        }
        if (!(currentNode instanceof EntityReference)) {
            throw new XMLCError("DOM node top of stack not a EntityReference for end tag");
        }
        currentNode = currentNode.getParentNode();
    }

    
    /**
     * Add a <code>Text</code> node.
     */
    public void addTextNode(String data) {
        if (document == null) {
            docNotCreatedError();
        }
        currentNode.appendChild(document.createTextNode(data));
    }

    /**
     * Add a <code>Comment</code> node.
     */
    public void addComment(String data) {
        if (document == null) {
            holdComment(data);  // before first element..
        } else {
            currentNode.appendChild(document.createComment(data));
        }
    }

    /**
     * Add a <code>CDATASection</code> node.
     */
    public void addCDATASection(String data) {
        if (document == null) {
            docNotCreatedError();
        }
        currentNode.appendChild(document.createCDATASection(data));
    }

    /**
     * Add a <code>ProcessingInstruction</code> node.
     */
    public void addProcessingInstruction(String target, 
                                         String data) {
        if (document == null) {
            docNotCreatedError();
        }
        currentNode.appendChild(document.createProcessingInstruction(target, data));
    }

    /**
     * Add an EntityReference object.
     */
    public void addEntityReference(String name) {
        if (document == null) {
            docNotCreatedError();
        }
        currentNode.appendChild(document.createEntityReference(name));
    }
}
