/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: HTMLDocBuilder.java,v 1.4 2000/02/21 01:27:33 markd Exp $
 */

package org.enhydra.xml.xmlc.dom;

import org.enhydra.xml.xmlc.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;

/**
 * Class for building a HTMLDocument DOM.  Overrides some of the methods
 * to do some special handling for HTML.
 */
public class HTMLDocBuilder extends DocBuilder {
    /**
     * Factory for creating the document.
     */
    private XMLCDomFactory domFactory;

    /**
     * Is this a frame set?
     */
    private boolean isFrameSet;

    /**
     * Constructor.
     * 
     * @param domFactory Factory class for Documents.
     */
    public HTMLDocBuilder(XMLCDomFactory domFactory) {
        super(domFactory, null);
        this.domFactory = domFactory;
    }

    /**
     * Get the publicId.
     */
    public String getPublicId() {
        return "HTML";
    }

    /**
     * Mark the document as a frameset or standard HTML document.
     */
    public void setFrameSet(boolean frameSet) {
        isFrameSet = frameSet;
    }

    /**
     * Is the document as a frameset?
     */
    public boolean getIsFrameSet() {
        return isFrameSet;
    }

    /**
     * Create the document.
     */
    protected Document createDocument(String rootTagName) {
        return domFactory.createDocument(null, null);
    }

    /**
     * Test if a DOM node is a script or style element, which have
     * special content handling.
     */
    private boolean isScriptStyleElement(Node node) {
        return ((node instanceof HTMLScriptElement)
                || (node instanceof HTMLStyleElement));
    }

    /**
     * Special handling for children of script/style node.  Add nodes
     * as CDATA sections.
     */
    private void addScriptChild(String data,
                                boolean isComment) {
        String newData;
        if (isComment) {
            newData = "<!--" + data + "-->";
        } else {
            newData = data;
        }
        
        // Merge CDATASection children of SCRIPT tags, as not all normalize
        // functions do this
        Node child = currentNode.getLastChild();
        if ((child != null) && (child instanceof CDATASection)) {
            CDATASection cdataChild = (CDATASection)child;
            cdataChild.setData(cdataChild.getData() + newData);
        } else {
            addCDATASection(newData);
        }
    }

    /**
     * Add a <code>Text</code> node with handling for script/style;
     */
    public void addTextNode(String data) {
        if (isScriptStyleElement(currentNode)) {
            addScriptChild(data, false);
        } else {
            super.addTextNode(data);
        }
    }

    /**
     * Add a <code>Comment</code> node with handling for script/style;
     */
    public void addComment(String data) {
        if (isScriptStyleElement(currentNode)) {
            addScriptChild(data, true);
        } else {
            super.addComment(data);
        }
    }
}
