/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLCDocument.java,v 1.3 2000/03/07 19:36:17 markd Exp $
 */

package org.enhydra.xml.xmlc.dom;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.OutputOptions;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;

// Note: This has some HTML specific checks in it.  At one time, it was split
// into two classes, however it proved simpler to move most of the logic to
// the DocBuilder and XMLDomFactory and just create a special-case checks
// here.

/**
 * Class that is a container for the DOM Document being constructed. It
 * also provides various operations on that document.
 */
public class XMLCDocument {
    /**
     * Class name of default XML DOM factory.
     */
    protected static final String DEFAULT_XML_DOM_FACTORY_CLASS
        = "org.enhydra.xml.xmlc.dom.DefaultXMLDomFactory";
    
    /**
     * Class name of default HTML DOM factory.
     */
    protected static final String DEFAULT_HTML_DOM_FACTORY_CLASS
        = "org.enhydra.xml.xmlc.dom.DefaultHTMLDomFactory";
    
    /**
     * Is this an HTML document we are dealing with?
     */
    private boolean isHtmlDocument;
    
    /**
     * The DTD-specific DOM Factory.
     */
    private XMLCDomFactory domFactory;

    /**
     * DocumentType builder
     */
    private DocTypeBuilder docTypeBuilder;

    /**
     * Document builder.
     */
    private DocBuilder docBuilder;

    /**
     * Name for the base class of the generate classes.
     */
    private String baseClass;

    /**
     * Name for the base interface of the generate classes.
     */
    private String baseInterface;

    /**
     * Load the DOM factory given its class name.
     *
     * @param domFactoryClassName Name of the DOM factory to use.
     */
    static protected XMLCDomFactory loadDomFactory(String domFactoryClassName)
        throws XMLCException {
        
        try {
            Class domFactoryClass = Class.forName(domFactoryClassName);
            return (XMLCDomFactory)domFactoryClass.newInstance();
        } catch (Exception except) {
            throw new XMLCException("Load of XMLC DOM factory class " 
                                    + "\"" + domFactoryClassName
                                    + "\" failed", except);
        }
    }

    /**
     * Public constructor.
     *
     * @param domFactoryClassName Name of the DOM factory to use. If
     *  null, the default is used.
     * @param baseClassName Name of the base class for the generated
     *  object.  Must extend XMLObjectImpl.  Use null for default.
     * @param baseInterfaceName Name of the base interface for the generated
     *  object.  Must extend XMLObject.  Use null for default.
     * @param htmlDocument true if this is going to be an HTML, false
     *  if its XML
     */
    public XMLCDocument(String domFactoryClassName,
                        String baseClassName,
                        String baseInterfaceName,
                        boolean htmlDocument) throws XMLCException {
        isHtmlDocument = htmlDocument;

        // Create the DOM factory.
        if (isHtmlDocument) {
            if (domFactoryClassName == null) {
                domFactoryClassName = DEFAULT_HTML_DOM_FACTORY_CLASS;
            }
        } else {
            if (domFactoryClassName == null) {
                domFactoryClassName = DEFAULT_XML_DOM_FACTORY_CLASS;
            }
        }
        domFactory = loadDomFactory(domFactoryClassName);

        // Default other arguments.
        baseClass = (baseClassName != null)
            ? baseClassName : domFactory.getBaseClassName();
        baseInterface = (baseInterfaceName != null)
            ? baseInterfaceName : domFactory.getBaseInterfaceName();

        // Initialize builders
        docTypeBuilder = new DocTypeBuilder(domFactory);
        if (isHtmlDocument) {
            docBuilder = new HTMLDocBuilder(domFactory);
        } else {
            docBuilder = new DocBuilder(domFactory,
                                        docTypeBuilder);
        }
    }

    /**
     * Get the DOM factory associated with this object.
     */
    public XMLCDomFactory getDomFactory() {
        return domFactory;
    }

    /**
     * Get the document associated with this object.
     */
    public Document getDocument() {
        return docBuilder.getDocument();
    }

    /*
     * Get the DocumentType builder.
     */
    public DocTypeBuilder getDocTypeBuilder() {
        return docTypeBuilder;
    }

    /*
     * Get the Document builder.
     */
    public DocBuilder getDocBuilder() {
        return docBuilder;
    }

    /**
     * Get the fully qualified class name that the generated class
     * will extend.  This class must implement
     * <CODE>org.enhydra.xml.xmlc.XMLObjectImpl</CODE>.
     */
    public String getObjectBaseClass() {
        return baseClass;
    }

    /**
     * Get the fully qualified interface name that the generated class
     * will extend.  This class must implement
     * <CODE>org.enhydra.xml.xmlc.XMLObject</CODE>.
     */
    public String getObjectBaseInterface() {
        return baseInterface;
    }

    /**
     * Convert an implementation-specific DOM node class name to the
     * external interface or class name.
     *
     * @see XMLCDomFactory
     */
    public String nodeClassToInterface(Node node) {
        return domFactory.nodeClassToInterface(node);
    }


    /**
     * Get the global id attribute name.  A DTD supports global id mode if a
     * single attribute name is used for all elements that support ids and
     * never used as a non-id attribute name.  This is intended for the HTML
     * `id' attribute.  However, if a DTD conforms to this requirement, more
     * efficient code for cloning the object can be generated.   It has been
     * suggested that `id' be a defacto standard for XML as well.
     *
     * This method will not be called until after the DTD has been parsed, so
     * this can be determined.
     *
     * See the <A href="http://www.xml.com/axml/axml.html">
     * XML Annotated Specification</A>.
     *
     * @return The global id attribute name or null if there is not a global
     * ID attribute.
     */
    public String getGlobalIdAttribute() {
        if (isHtmlDocument) {
            return "id";
        } else {
            //FIXME Implement from docTypeBuilder, built its an optimization.
            return null;
        }
    }

    /**
     * Get the id attribute name for an element.  The attribute
     * doesn't need to actually exist in this element, this just gets
     * the name for the attribute.
     *
     * @return The attribute name or null if the element does not have
     *  an id attribute.
     */
    public String getIdAttrName(Element element) {
        if (isHtmlDocument) {
            return "id";
        } else {
            return docTypeBuilder.getIdAttribute(element.getTagName());
        }
    }

    /**
     * Extract the unique id for an element object.  This is specified with
     * an attribute of type <CODE>ID</CODE>.
     *
     * @return The id or null if the element does not have an id.
     */
    public String getElementId(Element element) {
        String idAttrName = getIdAttrName(element);
        if (idAttrName == null) {
            return null;  // element doesn't have an id attribute
        }
        String id = element.getAttribute(idAttrName);
        if ((id == null) || (id.length() == 0)) {
            return null;  // No id.
        }
        return id;
    }

    /**
     * Extract the class names for an element.
     *
     * @return An array of class names or null if the node has no classes.
     *  XML returns null.
     * @see XMLCDomFactory#getElementClassNames
     */
    public String[] getElementClassNames(Element element) {
        return domFactory.getElementClassNames(element);
    }
    
    /**
     * Extract the name for an element.  This is a non-unique name for an
     * element.  In HTML, the name is specified with the <CODE>name</CODE>
     * attribute.  It's not supported for XML.
     *
     * @return The element name or null if the node has no name.
     *  XML returns null.
     */
    public String getElementName(Element element) {
        if (isHtmlDocument) {
            String name = element.getAttribute("name");
            if (name.length() == 0) {
                return null;
            } else {
                return name;
            }
        } else {
            return null;
        }
    }

    /**
     * Get a list of attribute names for an element who's value can be a URL.
     * This is use when mapping URL values.  The return list does not need to
     * be restricted to attributes that are legal for the element.  The list
     * can be all attributes in the DTD that can have URLs if element 
     * context doesn't matter.  The attributes return should have values of
     * type %URL, %URI or %Script.
     *
     * @return A list of attribute names or null if none of the
     *  attributes.  The returned array will not be modified.
     */
    public String[] getElementURLAttrs(Element element) {
        return domFactory.getElementURLAttrs(element);
    }

    /**
     * Determine if this element should have an set method created
     * for it's first text child.
     */
    public boolean needsCreateTextSetMethod(Element element) {
        return domFactory.needsCreateTextSetMethod(element);
    }

    /**
     * Convert the document to a string representation of the document.  The
     * results are parsable into the same DOM hierarchy
     */
    public String toDocument() {
        OutputOptions options = new OutputOptions();
        if (isHtmlDocument) {
            options.setDropHtmlSpanIds(true);
        }            
        return new DOMFormatter(options).toString(getDocument());
    }
}
