/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLCDomFactory.java,v 1.9.2.2 2000/03/28 23:49:26 markd Exp $
 */

package org.enhydra.xml.xmlc.dom;

import org.w3c.dom.DocumentType;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.html.HTMLDocument;


/**
 * Interface for a factory class that can create DocumentType and Document
 * objects.  It also provides document-type specific functions. 
 * <P>
 * Classes implementing this must:
 * <UL>
 * <LI> Provide a constructor that takes no arguments.
 * <LI> Be thread-safe (reentrant is preferred).
 * </UL>
 * When a DTD-specific DOM is created, an class implementing this interface
 * must be supplied to XMLC.  This would normally be created by extending the
 * XMLCDomDefaultFactory class.
 */
public interface XMLCDomFactory {
    /**
     * Creates an empty <code>DocumentType</code> node.
     *
     * @param qualifiedName The document type name (same as the root element).
     * @param publicID The document type public identifier.
     * @param systemID The document type system identifier.
     * @param internalSubset The internal subset as a string.
     * @return A new <code>DocumentType</code> node with 
     *   <code>Node.ownerDocument</code> set to <code>null</code>.
     *   For HTML DOM factories, this returns null.
     */
    public DocumentType createDocumentType(String qualifiedName,
                                           String publicID,
                                           String systemID,
                                           String internalSubset);
    /**
     * Creates an XML <code>Document</code> object of the specified type with 
     * its document element.
     *
     * @param qualifiedName The document type name (same as the root element).
     *  Maybe null for HTML documents.
     * @param doctype The type of document to be created or <code>null</code>.
     *   When <code>doctype</code> is not <code>null</code>, its 
     *   <code>Node.ownerDocument</code> attribute is set to the document being 
     *   created. Maybe null for for documents without DTDs or HTML documents.
     * @return A new <code>Document</code> object.
     */
    public Document createDocument(String qualifiedName,
                                   DocumentType doctype);

    /**
     * Enable or disable optional DOM error checking.  DOM implementations
     * (like Xerces) may allow some error checking to be disabled to improve
     * performance.  If possible, this enables or disables error checking.
     * If the DOM does not support optional error checking, this does nothing.
     * <P>Note: as a short-term performance enhancement, this also
     * enables/disables DOM level 2 mutation events.  Don't count on this
     * functionality in future releases.
     *
     * @param doc the document to operate on.
     * @param check true to enable error checking, false to disable it.
     */
    public void setErrorChecking(Document doc,
                                 boolean check);

    /**
     * Get the base class name for generated classes.  It must extend
     * XMLObjectImpl.  This class maybe overridden for individual documents
     * that are compiled.
     */
    public String getBaseClassName();

    /**
     * Get the base interface name for generated interfaces.  It must extend
     * XMLObject.  This class maybe overridden for individual documents that
     * are compiled.
     */
    public String getBaseInterfaceName();

    /**
     * Convert an implementation-specific DOM node class name to the
     * external interface or class name that should be used to 
     * reference it.  This could be a <CODE>org.w3c.dom</CODE> interface or
     * other interface or class.
     *
     * @see org.enhydra.xml.xmlc#XMLObject
     */
    public String nodeClassToInterface(Node node);

    /**
     * Extract the class names for an element.  This is a class for grouping
     * elements, not the Java class.  In HTML, the class is specified with the
     * <CODE>class</CODE> attribute and with a value of a white-space
     * separated list of class names.  Its not specified for XML, however this
     * method can be implemented in DTD-specifc XMLDomFactories to enable this
     * functionality.
     *
     * @return An array of class names or null if the node has no classes.
     *  XML returns null.
     */
    public String[] getElementClassNames(Element element);

    /**
     * Get a list of attribute names for an element who's value can be a URL.
     * This is use when mapping URL values.  The return list does not need to
     * be restricted to attributes that are legal for the element.  The list
     * can be all attributes in the DTD that can have URLs if element 
     * context doesn't matter.  The attributes return should have values of
     * type %URL, %URI or %Script.
     *
     * @return A list of attribute names or null if none of the
     *  attributes.  The returned array will not be modified.
     */
    public String[] getElementURLAttrs(Element element);

    /**
     * Determine if this element should have an set method created
     * for it's first text child.
     */
    public boolean needsCreateTextSetMethod(Element element);
}
