/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: SwingParser.java,v 1.3 2000/02/21 01:27:34 markd Exp $
 */

package org.enhydra.xml.xmlc.html.parsers.swing;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.html.*;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.dom.*;
import org.w3c.dom.*;
import org.w3c.dom.Element;  // Work around compiler bug in above import.
import org.w3c.dom.html.*;
import java.util.*;
import java.net.*;
import java.io.*;
import javax.swing.text.BadLocationException;
import javax.swing.text.MutableAttributeSet;

import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.HTML;

/**
 * Interface to the Swing HTML parser.
 * This attempts to correct problem with using the SWING HTML
 * 3.2 parser to parse HTML 4.0 constructs and well as general
 * bugginess in the Swing parser.
 */
class SwingParser extends HTMLEditorKit.ParserCallback {
    /**
     * Reader of file being parsed, which keeps a map of byte offset to
     * line number.
     */
    private LineNumberMapReader reader;

    /**
     * Since the Swing parser saves an exception and continues to parse, 
     * we set this flag to stop the parsing on an error.
     */
    private boolean gotError = false;

    /**
     * The document builder.
     */
    private DocBuilder docBuilder;

    /**
     * Error thrown to stop parser.  This will be converted to an exception.
     */
    class ParserError extends Error {
        /**
         * Constructor.
         */
        public ParserError(String msg) {
            super(msg);
            gotError = true;
        }
    }

    /**
     * Table of prefixes of SWING error messages to ignore.
     */
    private static final String[] swingIgnoredPrefixes = {
        "invalid.tagatt",          // Allow any attribute name.
        "javascript.unsupported",  // JavaScript should always be ignored :-)
        "req.att",
        "tag.unrecognized",
        "end.unrecognized",
        "unmatched.endtag",
        "multi.tagatt",
        "tag.ignoreimg"
    };
    
    /**
     * Ignore body tag errors for frame sets.
     */
    private static final String ignoreBodyPrefixe = "tag.ignorebody";

    /**
     * Table of SWING unrecognized tags that we want to 
     * force into nesting as complex tags.
     */
    private static final Hashtable unrecognizedTagsToNest = new Hashtable();

    /**
     * Swing parser.  Uses special, modified version of some Swing parser
     * classes to work around some bugs.  import
     */
    private org.enhydra.swing.text.html.parser.ParserDelegator parser
        = new org.enhydra.swing.text.html.parser.ParserDelegator();

    /**
     * Error reporter.
     */
    private ErrorReporter reporter;

    /**
     * Verbose tracing.
     */
    private ParseTracer tracer;

    /**
     * Is this a frame set?
     */
    private boolean docIsFrameSet;

    /**
     * Static constructor.
     */
    static {
        Boolean someValue = new Boolean(true);
        unrecognizedTagsToNest.put("noscript", someValue);
        unrecognizedTagsToNest.put("frameset", someValue);
        unrecognizedTagsToNest.put("noframes", someValue);
        unrecognizedTagsToNest.put("span", someValue);
    }

    /**
     * Parse a file into a DOM.
     */
    //FIXME: xmlcDoc is not used...
    public void parse(XMLCDocument xmlcDoc,
                      ErrorReporter errorReporter,
                      ParseTracer tracer,
                      LineNumberMapReader input)
        throws XMLCException, IOException {
        try {
            reader = input;
            gotError = false;
            docBuilder = xmlcDoc.getDocBuilder();
            reporter = errorReporter;
            this.tracer = tracer;

            parser.parse(reader, this, true);
        } catch (ParserError err) {
            throw new XMLCException(err);
        }
        if (docIsFrameSet) {
            cleanupFrameSet(docBuilder.getDocument());
        }
        checkNodes(docBuilder.getDocument());
    }

    /**
     * Determine if an error should be ignored based on some crude
     * assumptions.
     */
    private boolean ignore(String errorMsg) {
        // Swing can't handle <body> in <noframes> section
        if (docIsFrameSet && errorMsg.startsWith(ignoreBodyPrefixe)) {
            return true;
        }
        for (int i = 0; i < swingIgnoredPrefixes.length; i++) {
            if (errorMsg.startsWith(swingIgnoredPrefixes[i])) {
                return true;
            }
        }
        return false;
    }

    /**
     * Handle a simple tag call on an unrecognized tag.
     *
     * @return true if the tag was handle, false if it should be entered
     *  as a simple tag.
     */
    private boolean handleUnrecognizedTag(HTML.Tag tag, MutableAttributeSet attrSet, int pos) {
        if (tracer.enabled()) {
            tracer.trace("unrecognized tag: " + tag.toString());
        }

        if (unrecognizedTagsToNest.containsKey(tag.toString().toLowerCase())) {
            // SWING sets an attribute endtag for unrecognized end tags.
            if (attrSet.isDefined(HTML.Attribute.ENDTAG)) {
                handleEndTag(tag, pos);
            } else {
                handleStartTag(tag, attrSet, pos);
            }
            return true;
        }

        // Don't include undefined end tags.
        if (attrSet.isDefined(HTML.Attribute.ENDTAG)) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Create a new element.  
     */
    private void newElement(HTML.Tag tag, MutableAttributeSet attrSet)
            throws ParserError {
        String tagName = tag.toString();

        if (HTMLParserBase.isFrameSetElement(tagName)) {
            docIsFrameSet = true;
        }
        
        docBuilder.startElement(tagName);
        if ((attrSet != null) && (attrSet.getAttributeCount() > 0)) {
            Enumeration keyes = attrSet.getAttributeNames();
            while (keyes.hasMoreElements()) {
                Object attrKey = keyes.nextElement();
                Object attrValue = attrSet.getAttribute(attrKey);
                if (attrValue != null) {
                    docBuilder.addAttribute(attrKey.toString(),
                                         attrValue.toString());
                }
            }
        }
    }

    /**
     * Flush method for interface; does nothing.
     */
    public void	flush() throws BadLocationException {
    }

    /**
     * Swing text handler.
     */
    public void handleText(char[] data, int pos) {
        if (gotError) {
            return;
        }
        if (tracer.enabled()) {
            tracer.trace("TEXT: " + JavaLang.createStringConst(new String(data)));
        }
        docBuilder.addTextNode(new String(data));
    }

    /**
     * Swing comment handler.
     */
    public void handleComment(char[] data, int pos) {
        if (tracer.enabled()) {
            tracer.trace("COMMENT: " + JavaLang.createStringConst(new String(data)));
        }
        docBuilder.addComment(new String(data));
    }
    /*
     * Output start tag trace info.
     */
    private void traceStartTag(HTML.Tag tag, MutableAttributeSet attrSet) {
        StringBuffer msg = new StringBuffer("BEGIN TAG: ");

        msg.append(tag.toString());
        
        if (attrSet != null) {
            msg.append(":");
            Enumeration keyes = attrSet.getAttributeNames();
            while (keyes.hasMoreElements()) {
                Object attrKey = keyes.nextElement();
                Object attrValue = attrSet.getAttribute(attrKey);
                msg.append(" " + attrKey + "='" + attrValue + "'"); 
            }
        }
        tracer.trace(msg.toString());
        tracer.enter();
    }

    /**
     * Swing tag start handler.
     */
    public void handleStartTag(HTML.Tag tag, MutableAttributeSet attrSet, int pos) {
        if (gotError) {
            return;
        }
        if (tracer.enabled()) {
            traceStartTag(tag, attrSet);
        }
        newElement(tag, attrSet);
    }

    /**
     * Generate basic close-tag mismatch message.
     */
    private String tagMisMatchMsg(HTML.Tag tag) {
        return "Close tag mismatch, got </"
            + tag.toString().toUpperCase() + ">, expected </"
            + docBuilder.getCurrentNode().getNodeName() + ">";
    }

    /**
     * Handle end tag mismatch, trying to correct the problem.
     * Swing seems to get very confused by </FORM>,  generating
     * a end-tag event before all other tags are closed.
     */
    private void handleEndTagMisMatch(HTML.Tag tag, int pos) {
        if (tag.toString().equalsIgnoreCase("form")
             || tag.toString().equalsIgnoreCase("span")) {
            // </FORM> when not expected.
            if (tracer.enabled()) {
                tracer.trace("IGNORED: " + tagMisMatchMsg(tag)
                      + ", not modifying node stack");
            }
            return;
        }
        Node currentNode = docBuilder.getCurrentNode();
        if (currentNode.getNodeName().equalsIgnoreCase("form")
             || currentNode.getNodeName().equalsIgnoreCase("span")) {
            // <FORM> still on the stack, lets drop.
            if (tracer.enabled()) {
                tracer.trace("IGNORED: " + tagMisMatchMsg(tag)
                      + ", appears to be left on stack from previous error; removing");
            }
            docBuilder.popCurrentNode();

            // Try again with next node on stack.
            handleEndTag(tag, pos);
            return;
        }

        // Don't have a clue whats going on, maybe a bad page.
        handleError(tagMisMatchMsg(tag), pos);
    }

    /**
     * Swing end tag start handler.
     */
    public void handleEndTag(HTML.Tag tag, int pos) {
        if (gotError) {
            return;
        }
        if (tracer.enabled()) {
            tracer.leave();
            tracer.trace("END TAG: " + tag.toString());
        }
        Node currentNode = docBuilder.getCurrentNode();
        if ((currentNode != null)
            && !currentNode.getNodeName().equalsIgnoreCase(tag.toString())) {
            handleEndTagMisMatch(tag, pos);
        } else {
            docBuilder.finishElement();
        }
    }

    /**
     * Swing simple tag start handler.
     */
    public void handleSimpleTag(HTML.Tag tag, MutableAttributeSet attrSet, int pos) {
        if (gotError) {
            return;
        }
        if (tag instanceof HTML.UnknownTag) {
            if (handleUnrecognizedTag(tag, attrSet, pos)) {
                return;
            }
        }
        if (tracer.enabled()) {
            tracer.trace("TAG: " + tag.toString());
        }
        newElement(tag, attrSet);
        docBuilder.finishElement();
    }

    /**
     * Swing parser error handler that tries to manage the message
     * generated by parser that should really be ignored.
     */
    public void handleError(String errorMsg, int pos) {
        if (gotError) {
            return;
        }
        boolean ignored = ignore(errorMsg);
        if (tracer.enabled()) {
            if (ignored) {
                tracer.trace("IGNORE: " + errorMsg + ": " + pos);
            } else {
                tracer.trace("ERROR: " + errorMsg + ": " + pos);
            }
        }
        if (!ignored) {
            reporter.error(errorMsg, reader.getFileName(), reader.getLineNumber(pos));
        }
    }

    /*
     * Make sure a script element contains a comment.  This is 
     * necessary because the SWING parser drops the contents
     * if its not in comments.
     */
    private void checkScriptElement(HTMLScriptElement scriptElement) {
        if (scriptElement.getFirstChild() == null) {
            reporter.error("Empty <SCRIPT> element found. The swing HTML parser discards javascript\n"
                           + " that is not in a comment");
        }
    }

    /**
     * Recursively perform somes checks on a node once complete.  This checks
     * for a script elements without comments, as text in such elements is
     * dropped.
     */
    private void checkNodes(Node node) {
        if (node instanceof HTMLScriptElement) {
            checkScriptElement((HTMLScriptElement)node);
        }
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            checkNodes(child);
        }
    }

    /**
     * Cleanup inserted top-level body in framesets that was added by Swing.
     */
    private void cleanupFrameSet(Document doc) {
        // Find <HTML> node.
        Node htmlElem = null;
        for (Node child = doc.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            if (child instanceof HTMLHtmlElement) {
                htmlElem = child;
                break;
            }
        }
        if (htmlElem == null) {
            return;
        }
        
        // Find <BODY> node
        Node bodyElem = null;
        for (Node child = htmlElem.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            if (child instanceof HTMLBodyElement) {
                bodyElem = child;
                break;
            }
        }
        if (bodyElem == null) {
            return;
        }

        // Move body children up one level.
        for (Node child = bodyElem.getFirstChild(); child != null;) {
            Node move = child;
            child = child.getNextSibling();

            bodyElem.removeChild(move);
            htmlElem.insertBefore(move, bodyElem);
        }
        
        // Finally, drop the <BODY>
        htmlElem.removeChild(bodyElem);
    }

    /**
     * Was the document parsed a frame set.
     */
    public boolean isFrameSet() {
        return docIsFrameSet;
    }
}
