/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: TidyHTMLParser.java,v 1.6.2.1 2000/04/04 05:29:54 markd Exp $
 */

package org.enhydra.xml.xmlc.html.parsers.tidy;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.html.*;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.html.HTMLParser;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.io.Encoding;
import org.w3c.dom.*;
import org.w3c.dom.html.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Enumeration;
import org.w3c.tidy.*;
import org.xml.sax.InputSource;

/**
 * XMLCParser object for HTML and HTML framesets that uses the Java version of
 * the W3C HTML tidy program.  It uses Tidy to convert HTML to XHTML and then
 * parses it with an XML parser.
 */
public class TidyHTMLParser extends HTMLParserBase implements XMLCParser {
    /**
     * Get the set of options supported by this parser.
     */
    public int getSupportedOptions() {
        return OPT_VALIDATE_DOCUMENT;
    }

    /**
     * Is this a frame set?
     */
    private boolean docIsFrameSet;

    /**
     * Verbose tracing.
     */
    private ParseTracer tracer;

    /**
     * The document builder.
     */
    private DocBuilder docBuilder;

    /**
     * Tidy parser.
     */
    private Tidy tidy = new Tidy();
    
    /**
     * Constructor.
     */
    public TidyHTMLParser() throws XMLCException {
    }

    /**
     * Get the type of a Tidy node as a string.
     */
    private String nodeTypeStr(short typeId) {
        switch (typeId) {
        case org.w3c.tidy.Node.RootNode: return "RootNode";
        case org.w3c.tidy.Node.DocTypeTag: return "DocTypeTag";
        case org.w3c.tidy.Node.CommentTag: return "CommentTag";
        case org.w3c.tidy.Node.ProcInsTag: return "ProcInsTag";
        case org.w3c.tidy.Node.TextNode: return "TextNode";
        case org.w3c.tidy.Node.StartTag: return "StartTag";
        case org.w3c.tidy.Node.EndTag: return "EndTag";
        case org.w3c.tidy.Node.StartEndTag: return "StartEndTag";
        case org.w3c.tidy.Node.AspTag: return "AspTag";
        default: return "**Unknown Node**";
        }
    }

    /**
     * Print information about a Tidy node.
     */
    private void printNodeInfo(org.w3c.tidy.Node tNode) {
        StringBuffer buf = new StringBuffer();
        buf.append(">>>" + nodeTypeStr(tNode.type) + ":");
        if (tNode.element != null) {
            buf.append(" " + tNode.element);
        }
        // Print attributes
        AttVal attVal = tNode.attributes;
        while (attVal != null) {
            buf.append("  " + attVal.attribute + "='" + attVal.value + "'");
            attVal = attVal.next;
        }

        // Print associated text
        if (tNode.type != org.w3c.tidy.Node.StartTag) {
            String value = tNode.getNodeValue();
            if (value != null) {
                buf.append(" '" + value + "'");
            }
        }
        tracer.trace(buf.toString());
    }

    /**
     * Recursively print information about a Tidy node.
     */
    private void printNode(org.w3c.tidy.Node tNode) {
        printNodeInfo(tNode);
        org.w3c.tidy.Node next = tNode.content;
        tracer.enter();
        while (next != null) {
            printNode(next);
            next = next.next;
        }
        tracer.leave();
    }

    /**
     * Create an Element node from a Tidy node.
     */
    private void createElement(org.w3c.tidy.Node tNode) {
        if (tNode.element == null) {
            return;
        }
        docBuilder.startElement(tNode.element);
        
        AttVal attVal = tNode.attributes;
        while (attVal != null) {
            String value = (attVal.value == null) ? "" : attVal.value;
            docBuilder.addAttribute(attVal.attribute, value);
            attVal = attVal.next;
        }
        
        // See if we have a frameset tag.
        if (isFrameSetElement(tNode.element)) {
            docIsFrameSet = true;
        }
    }

    /**
     * Create a DOM node from a Tidy node.
     * 
     * @param tNode Tidy node.
     */
    private void makeDomNode(org.w3c.tidy.Node tNode) {
        switch (tNode.type) {
        case org.w3c.tidy.Node.CommentTag:
            if (tNode.getNodeValue() != null) {
                docBuilder.addComment(tNode.getNodeValue());
            }
            break;
        case org.w3c.tidy.Node.TextNode:
            if (tNode.getNodeValue() != null) {
                docBuilder.addTextNode(tNode.getNodeValue());
            }
            break;
        case org.w3c.tidy.Node.StartTag:
            createElement(tNode);
            break;
        case org.w3c.tidy.Node.EndTag:
            new XMLCError("Internal error: Unexpected Tidy EndTag node");
            break;
        case org.w3c.tidy.Node.StartEndTag:;
            throw new XMLCError("Internal error: Unexpected Tidy StartEndTag node");
        case org.w3c.tidy.Node.RootNode:
        case org.w3c.tidy.Node.DocTypeTag:
        case org.w3c.tidy.Node.ProcInsTag:
        case org.w3c.tidy.Node.AspTag:
            break;
        default: 
            throw new XMLCError("Internal error: Unknown node");
        }
    }

    /**
     * Recursively convert a node of the Tidy tree to a DOM tree.
     */
    private void buildNode(org.w3c.tidy.Node tNode) {

        makeDomNode(tNode);

        // Recurse to process children.
        org.w3c.tidy.Node tNext = tNode.content;
        while (tNext != null) {
            buildNode(tNext);
            tNext = tNext.next;
        }
        if (tNode.type == org.w3c.tidy.Node.StartTag) {
            docBuilder.finishElement();
        }
    }

    /**
     * Check for a Tidy-supported encoding.  If one is found,
     * set it as the encoding.
     * @param htmlEncoding Encoding to check, null is assumed to
     *  be ASCII.
     * @return true if encoding is Tidy-supported, false if its
     *  not.
     */
    private boolean checkForTidyEncoding(String htmlEncoding) {
        if ((htmlEncoding == null)
            || (htmlEncoding.equalsIgnoreCase("US-ASCII"))
            || (htmlEncoding.equalsIgnoreCase("ASCII"))) {
            tidy.setCharEncoding(Configuration.ASCII);
            return true;
        } else if (htmlEncoding.equalsIgnoreCase("ISO-8859-1")) {
            tidy.setCharEncoding(Configuration.LATIN1);
            return true;
        } else if (htmlEncoding.equalsIgnoreCase("UTF-8")) {
            tidy.setCharEncoding(Configuration.UTF8);
            return true;
        } else if (htmlEncoding.equalsIgnoreCase("ISO-2022-JP")) {
            tidy.setCharEncoding(Configuration.ISO2022);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Convert the input to a UTF-8 byte array and make it into
     * an input stream.
     */
    private InputStream makeUTF8InputStream(InputSource input)
        throws IOException {
        
        // Read into a UTF byte array.
        Reader reader = InputHandling.obtainReader(input);
        try {
            ByteArrayOutputStream utf8Bytes = new ByteArrayOutputStream();
            OutputStreamWriter writer = new OutputStreamWriter(utf8Bytes,
                                                               "UTF8");
            char buffer[] = new char[4096];
            int readSize;
            while ((readSize = reader.read(buffer)) >= 0) {
                writer.write(buffer, 0, readSize);
            }
            writer.flush();

            // Now we can read it with Tidy
            tidy.setCharEncoding(Configuration.UTF8);
            return new ByteArrayInputStream(utf8Bytes.toByteArray());
            
        } finally {
            reader.close();
        }
    }

    /**
     * Get the input stream.  If the specified encoding is not one 
     * supported by Jtidy, the file is read into a byte array
     * of UTF characters and an input stream wrapped around that
     * array.
     */
    private InputStream getInputStream(InputSource input)
            throws IOException {
        String htmlEncoding = input.getEncoding();
        if (checkForTidyEncoding(htmlEncoding)) {
            return new FileInputStream(input.getSystemId());
        } else {
            return makeUTF8InputStream(input);
        }
    }

    /**
     * Build Document from Tidy parse tree.
     */
    private XMLCDocument buildDocument(org.w3c.tidy.Node tRoot,
                                       String baseClass) throws XMLCException {
        XMLCDocument xmlcDoc = createXMLCDocument(baseClass);
        docBuilder = xmlcDoc.getDocBuilder();

        buildNode(tRoot);
        setFrameSet(xmlcDoc, docIsFrameSet);

        return xmlcDoc;
    }

    /**
     * Parse an HTML file.
     * @see XMLCParser#parse
     */
    public XMLCDocument parse(InputSource input,
                              int options,
                              String documentClass,
                              String baseClass,
                              String[] xCatalog,
                              ErrorReporter errorReporter,
                              ParseTracer tracer)
        throws IOException, XMLCException {

        if (documentClass != null) {
            // FIXME: Put this check in some place common..
            throw new XMLCException("Document class may not be specified for HTML documents");
        }
        if (xCatalog.length != 0) {
            throw new XMLCException("XCatalog not support for HTML");
        }

        // FIXME: This doesn't handle full input source.
        if (input.getSystemId() == null) {
            throw new XMLCException("only InputSource with systemId currently supported");
        }
        this.tracer = tracer;
        tidy.setInputStreamName(input.getSystemId());
        tidy.setErrout(new TidyErrorHandler(errorReporter,
                                            input.getSystemId()));

        InputStream srcFileStream = getInputStream(input);
        org.w3c.tidy.Node tRoot;
        try {
            tRoot = tidy.parse(srcFileStream, null);
        } finally {
            srcFileStream.close();
        }
        //FIXME: need to get error count from parser..
        if (errorReporter.getErrorCnt() != 0) {
            handlePrarseErrors(errorReporter);
        }
        
        if (tracer.enabled()) {
            printNode(tRoot);
        }

        XMLCDocument xmlcDoc = buildDocument(tRoot, baseClass);
        setEncoding(xmlcDoc, input);
        
        return xmlcDoc;
    }

    /**
     * Add a proprietary tag to the set of allowed tags.
     */
    private void addTag(TagTable tagTable,
                        String tagName,
                        int contentModel) throws XMLCException {
        Parser tagParser = null;
        int model = 0;

        // Get basic content model, alow multiple to be combine,
        // although this only makes sense with empty and the others.
        if ((contentModel & HTMLProprietaryTags.TAG_CM_EMPTY) != 0) {
            model |= Dict.CM_EMPTY|Dict.CM_OPT;
            tagParser =  ParserImpl.getParseInline();
        }
        if ((contentModel & HTMLProprietaryTags.TAG_CM_INLINE) != 0) {
            model |= Dict.CM_INLINE;
            tagParser =  ParserImpl.getParseInline();
        }
        if ((contentModel & HTMLProprietaryTags.TAG_CM_BLOCK) != 0) {
            model |= Dict.CM_BLOCK;
            tagParser = ParserImpl.getParseBlock();
        }
        if (model == 0) {
            throw new XMLCException("must specify at least one on TAG_CM_EMPTY, TAG_CM_INLINE, or TAG_CM_BLOCK");
        }
        
        // Is close optional?
        if ((contentModel & HTMLProprietaryTags.TAG_CM_OPT) != 0) {
            model |= Dict.CM_OPT;
        }

        // Let them occur anywhere and contain anything
        model |= Dict.CM_HEAD|Dict.CM_HTML|Dict.CM_MIXED;

        tagTable.install(new Dict(tagName, Dict.VERS_PROPRIETARY,
                                  model, tagParser, null));
    }
    
    /**
     * Specify the list of proprietary tags and attributes.
     */
    public void setProprietaryTags(HTMLProprietaryTags tagsAttrs) throws XMLCException {
        TagTable tagTable = TagTable.getDefaultTagTable();
        Enumeration tags = tagsAttrs.getTags();
        while (tags.hasMoreElements()) {
            HTMLProprietaryTags.Tag tag = (HTMLProprietaryTags.Tag)tags.nextElement();
            addTag(tagTable, tag.tagName.toLowerCase(), tag.contentModel);
        }

        AttributeTable attributeTable = AttributeTable.getDefaultAttributeTable();
        Enumeration attrs = tagsAttrs.getAttributes();
        while (attrs.hasMoreElements()) {
            String name = ((String)attrs.nextElement()).toLowerCase();
            attributeTable.install(new Attribute(name, Dict.VERS_PROPRIETARY,
                                                 null));
        }
    }
}
