/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XercesParser.java,v 1.7.2.1 2000/04/04 05:29:55 markd Exp $
 */

package org.enhydra.xml.xmlc.parsers.xerces;

import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.compiler.ErrorReporter;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.dom.DocBuilder;
import org.enhydra.xml.xmlc.dom.DocTypeBuilder;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import org.xml.sax.EntityResolver;
import org.apache.xerces.framework.XMLParser;
import org.apache.xerces.framework.XMLAttrList;
import org.apache.xerces.framework.XMLValidator;
import org.apache.xerces.readers.XCatalog;
import java.io.IOException;

/**
 * Parse using the Xerces parser.
 */
public class XercesParser extends XMLParser implements XMLCParser, EntityResolver {
    /**
     * Verbose tracing.
     */
    private XercesTracer tracer;

    /**
     * The document builder.
     */
    private DocBuilder docBuilder;

    /**
     * The document type builder.
     */
    private DocTypeBuilder docTypeBuilder;

    /**
     * Is a CDATASection being processed?
     */
    private boolean processingCDATASection = false;

    /**
     * Currently processing a document.
     */
    private static final int PROCESSING_DOCUMENT = 0;

    /**
     * Currently processing the external subset.
     */
    private static final int PROCESSING_EXTERNAL_SUBSET = 1;

    /**
     * Currently processing the internal subset.
     */
    private static final int PROCESSING_INTERNAL_SUBSET = 2;

    /**
     * Part of the document being processed.
     */
    private int processingState;

    /**
     * XCatalog entity resolver.
     */
    private XCatalog xCatalogResolver;

    /**
     * Get the set of options supported by this parser.
     */
    public int getSupportedOptions() {
        return OPT_VALIDATE_DOCUMENT | OPT_NO_VALIDATE_DOCUMENT;
    }

    /**
     * @see XMLCParser
     */
    public XMLCDocument parse(InputSource input,
                              int options,
                              String domFactoryClassName,
                              String baseClass,
                              String[] xCatalog,
                              ErrorReporter errorReporter,
                              ParseTracer tracer)
        throws IOException, XMLCException, SAXException {

        //FIXME: what about baseInterface argument??
        XMLCDocument xmlcDoc = new XMLCDocument(domFactoryClassName,
                                                baseClass, null, false);
        this.tracer = new XercesTracer(fStringPool, tracer);
        docBuilder = xmlcDoc.getDocBuilder();
        docTypeBuilder = xmlcDoc.getDocTypeBuilder();
        processingState = PROCESSING_DOCUMENT;

        // Configure parser.
        setErrorHandler(errorReporter);
        setAllowJavaEncodings(true);
        setValidation((options & OPT_NO_VALIDATE_DOCUMENT) == 0);
        setSendCharDataAsCharArray(true);
        setEntityResolver(this);

        // Setup entity resolver
        if (xCatalog.length != 0) {
            xCatalogResolver = new XCatalog();
            for (int idx = 0; idx < xCatalog.length; idx++) {
                xCatalogResolver.loadCatalog(new InputSource(xCatalog[idx]));
            }
        }

        super.parse(input);

        return xmlcDoc;
    }

    /**
     * Get a string from the string pool.
     */
    private String getString(int index) {
        return fStringPool.toString(index);
    }

    /**
     * Resolve an entity.
     * @see EntityResolver#resolveEntity
     */
    public InputSource resolveEntity(String publicId,
                                     String systemId)
        throws SAXException, IOException {
        tracer.resolveEntity(publicId, systemId);

        InputSource source = null;
        if (xCatalogResolver != null) {
            source = xCatalogResolver.resolveEntity(publicId, systemId);
        }
        return source;
    }

    /** 
     * Handle start of document.
     * @see XMLParser#startDocument
     */
    public void startDocument(int version,
                              int encoding,
                              int standAlone) throws Exception {
        tracer.startDocument(version,
                             encoding,
                             standAlone);
        docBuilder.setXMLVersion(getString(version));
        docBuilder.setEncoding(getString(encoding));
        docBuilder.setStandalone(new Boolean(getString(standAlone)).booleanValue());
    }

    /**
     * Handle end document.
     * @see XMLParser#endDocument
     */
    public void endDocument() throws Exception {
        tracer.endDocument();
    }

    /** 
     * Handle start of a namespace declaration scope. 
     *
     * @see XMLParser#startNamespaceDeclScope
     */
    public void startNamespaceDeclScope(int prefix,
                                        int uri) throws Exception {
        tracer.startNamespaceDeclScope(prefix, uri);
        //FIXME: do some thing!!
        throw new XMLCError("need to implement namespace handling");
    }

    /** 
     * Handle end of a namespace declaration scope. 
     *
     * @see XMLParser#startNamespaceDeclScope
     */
    public void endNamespaceDeclScope(int prefix) throws Exception {
        tracer.endNamespaceDeclScope(prefix);
    }

    /**
     * Handle start of element.
     * @see XMLParser#startElement
     */
    public void startElement(int elementType,
                             XMLAttrList attrList,
                             int attrListHandle) throws Exception {
        tracer.startElement(elementType, attrList, attrListHandle);
        docBuilder.startElement(getString(elementType));

        int attrIndex = attrListHandle;
        while (attrIndex >= 0) {
            if (attrList.isSpecified(attrIndex)) {
                docBuilder.addAttribute(getString(attrList.getAttrName(attrIndex)),
                                        getString(attrList.getAttValue(attrIndex)));
            }
            attrIndex = attrList.getNextAttr(attrIndex);
        }
    }

    /**
     * Handle end element.
     * @see XMLParser#endElement
     */
    public void endElement(int elementNameIndex) throws Exception {
        tracer.endElement(elementNameIndex);
        docBuilder.finishElement();
    }

    /**
     * Handle the start of an entity reference.
     * @see XMLParser#startEntityReference
     */
    public void startEntityReference(int entityName,
                                     int entityType,
                                     int entityContext) throws Exception {
        tracer.startEntityReference(entityName, entityType, entityContext);

        // This called at the start and external DTD with a null name,
        // ignore these.
        if ((entityName >= 0) && (processingState == PROCESSING_DOCUMENT)) {
            docBuilder.startEntityReference(getString(entityName));
        }
    }

    /**
     * Handle the end of an entity reference.
     * @see XMLParser#endEntityReference
     */
    public void endEntityReference(int entityName,
                                   int entityType,
                                   int entityContext) throws Exception {
        tracer.endEntityReference(entityName, entityType, entityContext);

        if ((entityName >= 0) && (processingState == PROCESSING_DOCUMENT)) {
            docBuilder.endEntityReference();
        }
    }

    /**
     * Not used.
     * @see XMLParser#characters
     */
    public void characters(int data) throws Exception {
        throw new XMLCError("fatal error: method that should not be invoked called");
    }

    /**
     * Not used.
     * @see XMLParser#ignorableWhitespace
     */
    public void ignorableWhitespace(int data) throws Exception {
        throw new XMLCError("fatal error: method that should not be invoked called");
    }

    /** 
     * Handle start of CDATA section. 
     * @see XMLParser#startCDATA
     */
    public void startCDATA() {
        tracer.startCDATA();
        processingCDATASection = true;
    }

    /** 
     * Handle for end of CDATA section.
     * @see XMLParser#endCDATA
     */
    public void endCDATA() {
        tracer.endCDATA();
        processingCDATASection = false;
    }

    /** 
     * Handle processing instruction.
     * @see XMLParser#processingInstruction
     */
    public void processingInstruction(int target,
                                      int data) throws Exception {
        tracer.processingInstruction(target, data);
        docBuilder.addProcessingInstruction(getString(target),
                                            getString(data));
    }

    /**
     * Handle a comment.
     * @see XMLParser#comment
     */
    public void comment(int comment) throws Exception {
        tracer.comment(comment);
        docBuilder.addComment(getString(comment));
    }

    /**
     * Handle characters.
     * @see XMLParser#characters
     */
    public void characters(char ch[],
                           int start,
                           int length) throws Exception {
        tracer.characters(ch, start, length);
        if (processingCDATASection) {
            docBuilder.addCDATASection(new String(ch, start, length));
        } else {
            docBuilder.addTextNode(new String(ch, start, length));
        }
    }

    /**
     * Handle ignorable whitespace.
     * @see XMLParser#ignorableWhitespace
     */
    public void ignorableWhitespace(char ch[],
                                    int start,
                                    int length) throws Exception {
        tracer.ignorableWhitespace(ch, start, length);
        //FIXME: is this right, we ignore them.
    }

    /**
     * @see XMLParser#startDTD
     */
    public void startDTD(int rootElementType,
                         int publicId,
                         int systemId) {
        tracer.startDTD(rootElementType, publicId, systemId);
        if ((publicId < 0) && (systemId < 0)) {
            processingState = PROCESSING_INTERNAL_SUBSET;
        } else {
            processingState = PROCESSING_EXTERNAL_SUBSET;
        }
        docTypeBuilder.setDocumentTypeName(getString(rootElementType));
        docTypeBuilder.setPublicId(getString(publicId));
        docTypeBuilder.setSystemId(getString(systemId));
    }

    /**
     * @see XMLParser#internalSubset
     */
    public void internalSubset(int internalSubset) {
        tracer.internalSubset(internalSubset);
        docTypeBuilder.setInternalSubset(getString(internalSubset));
    }

    /** 
     * @see XMLParser#endDTD
     */
    public void endDTD() {
        tracer.endDTD();
        processingState = PROCESSING_DOCUMENT;
    }

    /**
     * <!ELEMENT Name contentspec>
     *
     * @see XMLParser#elementDecl
     */
    public void elementDecl(int elementType,
                            XMLValidator.ContentSpec contentSpec) throws Exception {
        tracer.elementDecl(elementType, contentSpec);
        docTypeBuilder.addElementDecl(getString(elementType),
                                      contentSpec.toString(),
                                      (processingState == PROCESSING_INTERNAL_SUBSET));
    }

    /**
     * <!ATTLIST Name AttDef>
     *
     * @see XMLParser#attlistDecl
     */
    public void attlistDecl(int elementType,
                            int attrName,
                            int attType,
                            String enumString,
                            int attDefaultType,
                            int attDefaultValue) throws Exception {
        tracer.attlistDecl(elementType, attrName, attType,
                           enumString, attDefaultType, attDefaultValue);

        if (getString(attType).equals("ID")) {
            docTypeBuilder.addIdAttribute(getString(elementType),
                                          getString(attrName),
                                          (processingState == PROCESSING_INTERNAL_SUBSET));

        }
        //FIXME: need to pass getString(attDefaultType) when full DTD handling
        // is implmented.
        docTypeBuilder.addAttributeDecl(getString(elementType),
                                        getString(attrName),
                                        getString(attType),
                                        enumString,
                                        getString(attDefaultValue),
                                        (processingState == PROCESSING_INTERNAL_SUBSET));
    }

    /**
     * <!ENTITY % Name EntityValue>
     *
     * @see XMLParser#internalPEDecl
     */
    public void internalPEDecl(int entityName,
                               int entityValue) {
        tracer.internalPEDecl(entityName, entityValue);
    }

    /**
     * <!ENTITY % Name ExternalID>
     *
     * @see XMLParser#externalPEDecl
     */
    public void externalPEDecl(int entityName,
                               int publicId,
                               int systemId) {
        tracer.externalPEDecl(entityName, publicId, systemId);
    }

    /**
     * <!ENTITY Name EntityValue>
     *
     * @see XMLParser#internalEntityDecl
     */
    public void internalEntityDecl(int entityName,
                                   int entityValue) {
        tracer.internalEntityDecl(entityName, entityValue);
    }

    /**
     * <!ENTITY Name ExternalID>
     *
     * @see XMLParser#externalEntityDecl
     */
    public void externalEntityDecl(int entityName,
                                   int publicId,
                                   int systemId) {
        tracer.externalEntityDecl(entityName, publicId, systemId);
    }

    /**
     * <!ENTITY Name ExternalID NDataDecl>
     *
     * @see XMLParser#unparsedEntityDecl
     */
    public void unparsedEntityDecl(int entityName,
                                   int publicId,
                                   int systemId,
                                   int notationName) {
        tracer.unparsedEntityDecl(entityName, publicId,
                                  systemId, notationName);
    }

    /**
     * <!NOTATION Name ExternalId>
     *
     * @see XMLParser#notationDecl
     */
    public void notationDecl(int notationName,
                             int publicId,
                             int systemId) {
        tracer.notationDecl(notationName, publicId, systemId);
    }
}
