/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XercesTracer.java,v 1.6 2000/03/07 19:36:18 markd Exp $
 */

package org.enhydra.xml.xmlc.parsers.xerces;

import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.parsers.*;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.parsers.XMLCDocumentLocator;
import org.enhydra.xml.xmlc.dom.DocBuilder;
import org.enhydra.xml.xmlc.dom.DocTypeBuilder;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.apache.xerces.utils.StringPool;
import org.apache.xerces.framework.XMLParser;
import org.apache.xerces.framework.XMLAttrList;
import org.apache.xerces.framework.XMLValidator;

/**
 * Tracing for parser callbacks.  Isolated in a class because it
 * was getting to complex and crowding the actuall handlers.
 */
class XercesTracer {
    /**
     * Verbose tracing.
     */
    private ParseTracer tracer;

    /**
     * String pool used to lookup strings by index.
     */
    private StringPool stringPool;

    /**
     * Is the tracer enabled?
     */
    private boolean isEnabled;

    /**
     * Constructor.
     */
    protected XercesTracer(StringPool stringPool,
                           ParseTracer tracer) {
        this.stringPool = stringPool;
        this.tracer = tracer;
        isEnabled = (tracer != null) && tracer.enabled();
    }

    /*
     * Determine if tracing is enabled.
     */
    public boolean enabled() {
        return isEnabled;
    }

    /**
     * Get a string from the string pool.
     */
    private String getString(int index) {
        return stringPool.toString(index);
    }

    /**
     * @see EntityHandler#resolveEntity
     */
    public void resolveEntity(String publicId,
                              String systemId) {
        if (isEnabled) {
            tracer.trace("resolveEntity: \""
                         + publicId + "\" \""
                         + systemId + "\"");
        }
        
    }

    /** 
     * @see XMLParser#startDocument
     */
    public void startDocument(int version,
                              int encoding,
                              int standAlone) {
        if (isEnabled) {
            tracer.trace("startDocument:" + getString(version)
                         + " " + getString(encoding)
                         + " " + getString(standAlone));
        }
    }

    /**
     * Handle end document.
     * @see XMLParser#endDocument
     */
    public void endDocument() {
        if (isEnabled) {
            tracer.trace("endDocument");
        }
    }

    /**
     * Handle start of element.
     * @see XMLParser#startElement
     */
    public void startElement(int elementType,
                             XMLAttrList attrList,
                             int attrListHandle) {
        if (isEnabled) {
            StringBuffer buf = new StringBuffer("startElement: " + 
                                                getString(elementType));
            int attrIndex = attrListHandle;
            while (attrIndex >= 0) {
                if (attrList.isSpecified(attrIndex)) {
                    buf.append(" ");
                    buf.append(getString(attrList.getAttrName(attrIndex)));
                    buf.append("='");
                    buf.append(getString(attrList.getAttValue(attrIndex)));
                    buf.append("'");
                }
                attrIndex = attrList.getNextAttr(attrIndex);
            }
            tracer.trace(buf.toString());
            tracer.enter();
        }
    }

    /**
     * Handle end element.
     * @see XMLParser#endElement
     */
    public void endElement(int elementType) {
        if (isEnabled) {
            tracer.leave();
            tracer.trace("endElement: " + getString(elementType));
        }
    }

    /** 
     * Handle start of a namespace declaration scope. 
     *
     * @see XMLParser#startNamespaceDeclScope
     */
    public void startNamespaceDeclScope(int prefix, int uri) {
        if (isEnabled) {
            tracer.trace("startNamespaceDeclScope: " + getString(prefix)
                         + " " + getString(uri));
            tracer.enter();
        }
    }

    /** 
     * Handle end of a namespace declaration scope. 
     *
     * @see XMLParser#startNamespaceDeclScope
     */
    public void endNamespaceDeclScope(int prefix) {
        if (isEnabled) {
            tracer.leave();
            tracer.trace("endNamespaceDeclScope: " + getString(prefix));
        }
    }

    /**
     * Handle the start of an entity reference.
     * @see XMLParser#startEntityReference
     */
    public void startEntityReference(int entityName,
                                     int entityType,
                                     int entityContext) {
        if (isEnabled) {
            tracer.trace("startEntityReference: "
                         + getString(entityName)
                         + " " +  XercesUtil.getEntityTypeString(entityType)
                         + " " +  XercesUtil.getEntityContextString(entityContext));
            tracer.enter();
        }
    }

    /**
     * Handle the end of an entity reference.
     * @see XMLParser#endEntityReference
     */
    public void endEntityReference(int entityName,
                                   int entityType,
                                   int entityContext) {
        if (isEnabled) {
            tracer.leave();
            tracer.trace("endEntityReference: " + getString(entityName));
        }
    }

    /** 
     * Handle start of CDATA section. 
     * @see XMLParser#startCDATA
     */
    public void startCDATA() {
        if (isEnabled) {
            tracer.trace("startCDATA");
            tracer.enter();
        }
    }

    /** 
     * Handle for end of CDATA section.
     * @see XMLParser#endCDATA
     */
    public void endCDATA() {
        if (isEnabled) {
            tracer.leave();
            tracer.trace("endCDATA");
        }
    }

    /** 
     * Handle processing instruction.
     * @see XMLParser#processingInstruction
     */
    public void processingInstruction(int target,
                                      int data) {
        if (isEnabled) {
            tracer.trace("processingInstruction: " + getString(target));
        }
    }

    /**
     * Handle a comment.
     * @see XMLParser#comment
     */
    public void comment(int comment) {
        if (isEnabled) {
            tracer.trace("comment");
        }
    }

    /**
     * Handle characters.
     * @see XMLParser#characters
     */
    public void characters(char ch[],
                           int start,
                           int length) {
        if (isEnabled) {
            tracer.trace("characters: \"" + new String(ch, start, length) + "\"");
        }
    }

    /**
     * Handle ignorable whitespace.
     * @see XMLParser#ignorableWhitespace
     */
    public void ignorableWhitespace(char ch[],
                                    int start,
                                    int length) {
        if (isEnabled) {
            tracer.trace("ignorableWhitespace: " + length);
        }
    }

    /**
     * @see XMLParser#startDTD
     */
    public void startDTD(int rootElementType,
                         int publicId,
                         int systemId) {
        if (isEnabled) {
            tracer.trace("startDTD: " + getString(rootElementType)
                         + " " + getString(publicId)
                         + " " + getString(systemId));
            tracer.enter();
        }
    }

    /**
     * @see XMLParser#internalSubset
     */
    public void internalSubset(int internalSubset) {
        if (isEnabled) {
            tracer.trace("internalSubset: " + getString(internalSubset));
        }
    }

    /** 
     * @see XMLParser#endDTD
     */
    public void endDTD() {
        if (isEnabled) {
            tracer.leave();
        }
    }

    /**
     * <!ELEMENT Name contentspec>
     *
     * @see XMLParser#elementDecl
     */
    public void elementDecl(int elementType,
                            XMLValidator.ContentSpec contentSpec) {
        if (isEnabled) {
            tracer.trace(getString(elementType)
                         + " " + contentSpec.toString());
        }
    }

    /**
     * <!ATTLIST Name AttDef>
     *
     * @see XMLParser#attlistDecl
     */
    public void attlistDecl(int elementType,
                            int attrName,
                            int attType,
                            String enumString,
                            int attDefaultType,
                            int attDefaultValue) {
        if (isEnabled) {
            String defaultValue = (attDefaultValue < 0) ? 
                "<null>" : getString(attDefaultValue);
            tracer.trace("attlistDecl: " + getString(elementType)
                         + " " + getString(attrName)
                         + " " + getString(attType)
                         + " " + enumString
                         + " " + getString(attDefaultType)
                         + " " + defaultValue);
        }
    }

    /**
     * <!ENTITY % Name EntityValue>
     *
     * @see XMLParser#internalPEDecl
     */
    public void internalPEDecl(int entityName,
                               int entityValue) {
        if (isEnabled) {
            tracer.trace("internalPEDecl: " + getString(entityName)
                         + " \"" + getString(entityValue) + "\"");
        }
    }

    /**
     * <!ENTITY % Name ExternalID>
     *
     * @see XMLParser#externalPEDecl
     */
    public void externalPEDecl(int entityName,
                               int publicId,
                               int systemId) {
        if (isEnabled) {
            tracer.trace("externalPEDecl: " + getString(entityName)
                         + " " + getString(publicId)
                         + " " + getString(systemId));
        }
    }

    /**
     * <!ENTITY Name EntityValue>
     *
     * @see XMLParser#internalEntityDecl
     */
    public void internalEntityDecl(int entityName,
                                   int entityValue) {
        if (isEnabled) {
            tracer.trace("internalEntityDecl: " + getString(entityName)
                         + " \"" + getString(entityValue) + "\"");
        }
    }

    /**
     * <!ENTITY Name ExternalID>
     *
     * @see XMLParser#externalEntityDecl
     */
    public void externalEntityDecl(int entityName,
                                   int publicId,
                                   int systemId) {
        if (isEnabled) {
            tracer.trace("externalEntityDecl: " + getString(entityName)
                         + " " + getString(publicId)
                         + " " + getString(systemId));
        }
    }

    /**
     * <!ENTITY Name ExternalID NDataDecl>
     *
     * @see XMLParser#unparsedEntityDecl
     */
    public void unparsedEntityDecl(int entityName,
                                   int publicId,
                                   int systemId,
                                   int notationName) {
        if (isEnabled) {
            tracer.trace("unparsedEntityDecl: " + getString(entityName)
                         + " " + getString(publicId)
                         + " " + getString(systemId)
                         + " " + getString(notationName));
        }
    }

    /**
     * <!NOTATION Name ExternalId>
     *
     * @see XMLParser#notationDecl
     */
    public void notationDecl(int notationName,
                             int publicId,
                             int systemId) {
        if (isEnabled) {
            tracer.trace("notationDecl: " + getString(notationName)
                         + " " + getString(publicId)
                         + " " + getString(systemId));
        }
    }
}
