/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: RunObject.java,v 1.17.2.1 2000/04/04 05:29:56 markd Exp $
 */

package org.enhydra.xml.xmlc.tests;
import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.io.*;
import java.io.*;
import java.util.Vector;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.Encoding;

/**
 * Simplistic runner for XMLObjects. Generates toDocument() output.
 *
 * The command line is: <BR>
 * <CODE>
 *  RunObject [options] objclass [output]
 * </CODE>
 *
 * The following options are supported:
 * <UL>
 * <LI> -dump file - Dump the DOM to the specified file.
 * <LI> -instanceof classname - Verify that the object is an instanceof this class.
 *       This maybe repeated.
 * <LI> -use-factory - Use a simplistic factor to get an instance of this object.
 * <LI> -reversetext - Reverse the text in all the text nodes.  Use to generate
 *      files to use in recompilation testing.  SPAN IDs are also preserved.
 * <LI> -encoding encoding - XML/HTML encoding.  This will be used for output
 *      and will be checked as the value of getEncoding().
 * <LI> -omit-encoding-header - Omit the encoding header (as used for
 *      WML).
 * <LI> -check-encoding encoding - Check that the encoding value matches.
 *      A blank argument checks for encoding being null.
 * </UL>
 */
public class RunObject {
    /**
     * Usage message.
     */
    private static final String USAGE = "wrong # args: RunObject [options] objclass [output]";

    // Parsed from command line.
    private static String dumpFile;
    private static String docClassName;
    private static boolean useFactory;
    private static boolean reverseText;
    private static Vector instanceOfList = new Vector();
    private static String xmlcObjClass;
    private static String outFile;
    private static String xmlEncoding;
    private static String javaEncoding;
    private static boolean omitEncodingHeader;
    private static String checkEncoding;

    /*
     * Create the test object.
     */
    private static XMLObject createTestObject()
        throws ClassNotFoundException, InstantiationException, IllegalAccessException {
        if (useFactory) {
            XMLCFactory factory = new XMLCStdFactory(null, null);
            // Append Impl to interface name to get class.
            Class xmlcInterface = Class.forName(xmlcObjClass);
            return factory.create(xmlcInterface);
        } else {
            Class objClass = Class.forName(xmlcObjClass);
            return (XMLObject)objClass.newInstance();
        }
    }

    /*
     * Validate instanceof.
     */
    private static void checkInstanceOf(XMLObject xmlObj)
        throws ClassNotFoundException {
        for (int idx = 0; idx < instanceOfList.size(); idx++) {
            String name = (String)instanceOfList.elementAt(idx);
            Class checkClass = Class.forName(name);
            if (!checkClass.isInstance(xmlObj)) {
                TestUtils.fatal("object is not an instance of " + name);
            }
        }
    }

    /**
     * Reverse all text in a DOM.
     */
    private static void reverseDOMText(Node node) {
        if ((node instanceof Text)&& !(node instanceof CDATASection)) {
            Text text = (Text)node;
            String data = text.getData();
            StringBuffer rdata = new StringBuffer(data.length());
            for (int idx = data.length()-1; idx >= 0; idx--) {
                rdata.append(data.charAt(idx));
            }
            text.setData(rdata.toString());
        } else {
            for (Node child = node.getFirstChild(); child != null;
                 child = child.getNextSibling()) {
                reverseDOMText(child);
            }
        }
    }
 

    /*
     * Validate document class.
     */
    private static void checkDocumentClass(XMLObject xmlObj)
        throws ClassNotFoundException {
        if (docClassName != null) {
            Class docClass = Class.forName(docClassName);
            if (!docClass.isInstance(((XMLObjectImpl)xmlObj).getDocument())) {
                TestUtils.fatal("document is not an instance of " + docClassName);
            }
        }
    }

    /*
     * Validate encoding.
     */
    private static void checkEncoding(XMLObject xmlObj) {
        if (checkEncoding != null) {
            if (checkEncoding.length() == 0) {
                if (xmlObj.getEncoding() != null) {
                    TestUtils.fatal("document encoding is \""
                                    + xmlObj.getEncoding()
                                    + "\", should be null");
                }
            } else {
                if (!checkEncoding.equalsIgnoreCase(xmlObj.getEncoding())) {
                    TestUtils.fatal("document encoding is \""
                                    + xmlObj.getEncoding()
                                    + "\", should be \""
                                    + checkEncoding + "\"");
                }
            }
        }
    }

    /**
     * Output an XML document. 
     */
    private static void outputDocument(XMLObject xmlObj)
        throws IOException {

        // Use toDocument() behavior for HTML files with nothing special.
        // This drops span tag ids which can't be done if reverseText
        // is set.
        if ((xmlObj instanceof HTMLDocument) && (!reverseText)
            && (javaEncoding == null) && (checkEncoding == null)) {
            TestUtils.outputDocument(xmlObj, outFile);
        } else {
            String docEncoding = xmlObj.getEncoding();
            OutputOptions options = new OutputOptions();
            if (javaEncoding != null) {
                options.setJavaEncoding(javaEncoding);
            } else {
                options.setXmlEncoding(docEncoding);
            }
            options.setOmitEncoding(omitEncodingHeader);
            TestUtils.outputDocument(xmlObj, outFile, options);
        }
    }

    /**
     * Dump DOM information to the specified output file.
     */
    private static void dumpInfo(XMLObject xmlObj) throws IOException {
        OutputStream outStream = new FileOutputStream(dumpFile);
        //FIXME: need to be able to specify the encoding:
        Writer out = new OutputStreamWriter(outStream, "UTF8");
        PrintWriter writer = new PrintWriter(out);
        if (xmlObj.getEncoding() != null) {
            writer.println("Document encoding: " + xmlObj.getEncoding());
        }
        XMLCUtil.printNode("DOM hierarchy", xmlObj, 
                           XMLCUtil.PRINT_TEXT,
                           writer);
        writer.flush();
        writer.close();
    }

    /*
     * Test entry point.
     */
    public static void main(String[] args) throws Throwable {
        int argIdx = 0;

        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-dump")) {
                TestUtils.checkOptArg(1, argIdx, args);
                dumpFile = args[++argIdx];
            } else if (args[argIdx].equals("-reversetext")) {
                reverseText = true;
            } else if (args[argIdx].equals("-instanceof")) {
                TestUtils.checkOptArg(1, argIdx, args);
                instanceOfList.addElement(args[++argIdx]);
            } else if (args[argIdx].equals("-docclass")) {
                TestUtils.checkOptArg(1, argIdx, args);
                docClassName = args[++argIdx];
            } else if (args[argIdx].equals("-use-factory")) {
                useFactory = true;
            } else if (args[argIdx].equals("-encoding")) {
                TestUtils.checkOptArg(1, argIdx, args);
                xmlEncoding = args[++argIdx];
            } else if (args[argIdx].equals("-omit-encoding-header")) {
                omitEncodingHeader = true;
            } else if (args[argIdx].equals("-check-encoding")) {
                TestUtils.checkOptArg(1, argIdx, args);
                checkEncoding = args[++argIdx];
            } else {
                TestUtils.fatal("Error: invalid option \"" + args[argIdx]
                                + "\", expected one of "
                                + "-dump, -reversetext, -instanceof, "
                                + "-docclass, -use-factory, "
                                + "-encoding, -omit-encoding-header, "
                                + "or -check-encoding");
            }
            argIdx++;
        }

        int argsLeft = args.length-argIdx;
        if ((argsLeft < 1) || (argsLeft > 2)) {
            TestUtils.fatal(USAGE);
        }
        if (xmlEncoding != null) {
            javaEncoding = Encoding.xml2JavaEncoding(xmlEncoding);
            if (javaEncoding == null) {
                TestUtils.fatal("invalid XML/HTML encoding: \""
                                + xmlEncoding + "\"");
            }
        }
        
        xmlcObjClass = args[argIdx];
        outFile = (argsLeft > 1) ? args[argIdx+1] : null;
        
        try {
            XMLObject xmlObj = createTestObject();

            checkInstanceOf(xmlObj);
            checkDocumentClass(xmlObj);
            checkEncoding(xmlObj);
            if (reverseText) {
                reverseDOMText(xmlObj);
            }

            if (dumpFile != null) {
                dumpInfo(xmlObj);
            }
            if (outFile != null) {
                outputDocument(xmlObj);
            }
        } catch (Throwable error) {
            TestUtils.handleError(error);
        }
        
    }
}
