/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: TestAutoComp.java,v 1.8 2000/02/21 09:57:21 markd Exp $
 */

package org.enhydra.xml.xmlc.tests;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.JavaLang;
import org.enhydra.xml.xmlc.autocomp.XMLCAutoCompFactory;
import org.enhydra.xml.xmlc.autocomp.StandardRecompilationHandler;
import com.lutris.classloader.*;
import com.lutris.testutil.*;
import com.lutris.logging.LogChannel;
import com.lutris.util.FatalExceptionError;
import java.io.*;
import java.util.Vector;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import org.w3c.dom.*;

/**
 * Tests for auto-compilation and class reloading.
 * 
 * The command line is: <BR>
 * <CODE>
 *  TestAutoComp [options] interfaceName [outfile]
 * </CODE>
 *
 * The following options are supported:
 * <UL>
 * <LI> -dump - Dump the DOM to stdout.
 * <LI> -sourcefile file - Path to the source file.  The root directory
 *      of the tree containing the file must be on the classpath.  This will
 *      will be overwritten.   
 * <LI> -copyfile file - File to copy over source file to generate an
 *      error or change the results.
 * </UL>
 * The directory containing 
 */
public class TestAutoComp {
    /*
     * Parsed from command line.
     */
    private boolean dump;
    private String interfaceName;
    private String outFile;
    private File sourceFile;
    private File copyFile;

    /*
     * Objects used during the tests.
     */
    private Class origClass;
    private XMLObject origObject;

    /**
     * Usage message.
     */
    private static final String USAGE = "wrong # args: TestAutoComp [options] interfaceName [output]";

    /**
     * Our package
     */
    private static final String TEST_PACKAGE
        = JavaLang.getPackageName(TestAutoComp.class.getName());

    /**
     * Test recording object.
     */
    private static Tester tester = new Tester();

    /**
     * Generate a failure error message and log.
     */
    private static void failure(String testId,
                                String msg) {
        TestUtils.log("Failure: " + testId + ": " + msg);
        tester.failure(testId, msg);
    }

    /**
     * Add a note to the log file.
     */
    private static void log(String msg) {
        TestUtils.log(msg);
    }

    /**
     * Add a note and exception to the log file.
     */
    private static void log(String msg, Throwable except) {
        TestUtils.log(msg, except);
    }

    /**
     * Create a class loader that includes the system class path.
     * This filters to only load test classes.
     */
    private static MultiClassLoader mkTestClassLoader() {
        MultiClassLoader loader
            = new MultiClassLoader(MultiClassLoader.getSystemClassPath(),

                                   TestUtils.getLogChannel());
        // Only load test classes.
        loader.addClassFilter(new MultiClassLoader.ClassFilter() {
                public int loadCheck(String className) {
                    if (className.startsWith(TEST_PACKAGE)) {
                        return MUST_LOAD;
                    } 
                    return DONT_LOAD;
                }
            });
        return loader;
    }
    
    /**
     * Find the source file for a class on the CLASSPATH.
     */
    private File findClassFile(MultiClassLoader loader,
                               String className)
            throws ClassNotFoundException {
        Resource res = loader.getClassResource(className);
        return ((LocalDirResource)res).getFile();
    }

    /**
     * Touch a class file.
     */
    private void touchClassFile(MultiClassLoader loader,
                                String className)
        throws ClassNotFoundException, XMLCException {
        TestUtils.touchFile(findClassFile(loader, className));
    }

    /**
     * Verify a delegate that must have changed.
     */
    private void verifyChangedDelegate(String testId,
                                       XMLObject newObject,
                                       XMLObject prevObject) {
        XMLObject delegate = newObject.getDelegate();
        if (delegate == null) {
            failure(testId + ".1", "delegate is null");
        } else {
            if (delegate.getClass() == newObject.getClass()) {
                failure(testId + ".2", "delegate class same as loaded class");
            }
            if (delegate.getClass().getClassLoader() == newObject.getClass().getClassLoader()) {
                failure(testId + ".3", "delegate class loader same as loaded class");
            }
        }
        if ((prevObject != null) && (prevObject.getDelegate() == delegate)) {
            failure(testId + ".4", "delegate same as previous delegate");
        }
    }

    /**
     * Create initial test object.
     */
    private void createTestObject(String testId,
                                  String className,
                                  MultiClassLoader loader) throws Exception {
        log("Test-" + testId + ".1");
        origClass = loader.loadClass(className);

        origObject = (XMLObject)origClass.newInstance();
        if (origObject.getDelegate() != null) {
            failure(testId + ".2", "delegate not null");
        }
        
    }

    /*
     * Test loading of a class given a class name.
     */
    private XMLObject doTest(String testId,
                             String interfaceName) throws Exception {
        XMLObject finalObj = null;
        String className = interfaceName + "Impl";

        MultiClassLoader loader = mkTestClassLoader();
        XMLCAutoCompFactory factory
            = new XMLCAutoCompFactory(new StandardRecompilationHandler(),
                                      loader, TestUtils.getLogChannel());

        // Load class for the first time, setting object variables.
        createTestObject(testId, className, loader);

        // Create another instance through the factory.
        log("Test-" + testId + ".2.1");
        XMLObject obj1 = factory.create(className);
        if (obj1.getClass() != origClass) {
            failure(testId + ".2.2", "class not the same as original");
        }
        if (origObject.getDelegate() != null) {
            failure(testId + ".2.3", "delegate not null");
        }
        if (origObject.getDocument() == null) {
            failure(testId + ".2.4", "document is null");
        }
        if (factory.getReloadCount() != 0) {
            failure(testId + ".2.5", "reload count wrong; got "
                    + factory.getReloadCount());
        }
        if (factory.getRecompileCount() != 0) {
            failure(testId + ".2.6", "recompile count wrong; got "
                    + factory.getRecompileCount());
        }

        // Modify the class file and try again.
        log("Test-" + testId + ".3.1");
        touchClassFile(loader, className);
        XMLObject obj2 = factory.create(className);
        if (obj2.getClass() != origClass) {
            failure(testId + ".3.2", "class not the same as original");
        }
        if (obj2.getDocument() == null) {
            failure(testId + ".3.4", "document is null");
        }
        if (factory.getReloadCount() != 1) {
            failure(testId + ".3.4", "reload count wrong; got " 
                    + factory.getReloadCount());
        }
        if (factory.getRecompileCount() != 0) {
            failure(testId + ".3.5", "recompile count wrong; got "
                    + factory.getRecompileCount());
        }
        verifyChangedDelegate(testId + ".3.6", obj2, null);
        finalObj = obj2;

        if (sourceFile != null) {
            if (copyFile != null) {
                // Overwrite it.
                log("Test: overwritting " + sourceFile);
                TestUtils.copyFile(copyFile, sourceFile);
            } else {
                // Just change mod date
                log("Test: touching " + sourceFile);
                TestUtils.touchFile(sourceFile);
            }
            log("Test-" + testId + ".4.1");
            XMLObject obj3 = factory.create(className);
            if (obj3.getClass() != origClass) {
                failure(testId + ".4.2", "class not the same as original");
            }
            if (origObject.getDocument() == null) {
                failure(testId + ".4.3", "document is null");
            }
            if (factory.getReloadCount() != 2) {
                failure(testId + ".4.4", "reload count wrong; got "
                        + factory.getReloadCount());
            }
            if (factory.getRecompileCount() != 1) {
                failure(testId + ".4.5", "recompile count wrong; got "
                        + factory.getRecompileCount());
            }
            verifyChangedDelegate(testId + ".4.6", obj3, obj2);
            finalObj = obj3;
            
        }
        if (finalObj == null) {
            throw new XMLCError("final object is null");
        }
        return finalObj;
    }

    /*
     * Exit based on the error count.
     */
    private static void finish() {
        if (tester.errorCnt != 0) {
            System.err.println("**** " + tester.errorCnt + " tests failed");
            System.exit(1);
        }
    }

    /**
     * Run tests.
     */
    private void runTests(String[] args) {
        int argIdx = 0;
        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-dump")) {
                dump = true;
            } else if (args[argIdx].equals("-sourcefile")) {
                TestUtils.checkOptArg(1, argIdx, args);
                sourceFile = new File(args[++argIdx]);
            } else if (args[argIdx].equals("-copyfile")) {
                TestUtils.checkOptArg(1, argIdx, args);
                copyFile = new File(args[++argIdx]);
            } else {
                TestUtils.fatal("Error: invalid option \"" + args[argIdx]
                                + "\", expected one of ");
            }
            argIdx++;
        }
        int argsLeft = args.length-argIdx;
        if ((argsLeft < 1) || (argsLeft > 2)) {
            TestUtils.fatal(USAGE);
        }
        interfaceName = args[argIdx];
        outFile = (argsLeft > 1) ? args[argIdx+1] : null;
        
        try {
            XMLObject xmlObj = doTest("1", interfaceName);
            if (dump) {
                PrintWriter out = new PrintWriter(System.out);
                XMLCUtil.printNode("DOM hierarchy", xmlObj, 
                                   XMLCUtil.PRINT_ALL,
                                   out);
                out.flush();
            }
            if (outFile != null) {
                TestUtils.outputDocument(xmlObj, outFile, null);
            }
            finish();
        } catch (Throwable error) {
            TestUtils.handleError(error);
        }
    }

    /*
     * Test entry point.
     */
    public static void main(String[] args) {
        new TestAutoComp().runTests(args);
    }
}
