/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: XMLParse.java,v 1.3.2.1 2000/04/04 05:29:56 markd Exp $
 */

package org.enhydra.xml.xmlc.tests;
import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.JavaLang;
import org.enhydra.xml.xmlc.compiler.ErrorReporter;
import java.io.*;
import org.xml.sax.*;
import org.w3c.dom.*;

/**
 * Class for experimenting with XML parser.
 */
public class XMLParse {
    /*
     * What to parse.
     */
    private static final int PARSE_XML = 1;
    private static final int PARSE_DTD = 2;
    private static final int PARSE_XHTML = 3;
    private static final int PARSE_XHTML_DTD = 4;

    /*
     * DTD URIs
     */
    private static final String XHTML_TRANSITIONAL_PATH = "org/openxml/dtd-catalog/xhtml1-transitional.dtd";
    private static final String XHTML_STRICT_PATH = "org/openxml/dtd-catalog/xhtml1-strict.dtd";
    private static final String XHTML_FRAMESET_PATH = "org/openxml/dtd-catalog/xhtml1-frameset.dtd";

    /*
     * Parse XML
     */
    private Document parseXML(Reader xmlIn, Class docClass) throws ParseException {
        XMLParser parser = new XMLParser(xmlIn, null);
        return parser.parseDocument(null, docClass);
    }

    /*
     * Parse a DTD.
     */
    private Document parseDTD(Reader xmlIn) throws ParseException {
        DTDParser parser = new DTDParser(xmlIn, null);
        return parser.parseDocument();
    }

    /*
     * Parse XML with a DTD.
     */
    private Document parseXMLWithDTD(Reader xmlIn,
                                     String dtdPath,
                                     Class docClass)
        throws ParseException, IOException {
        InputStream dtdIn = ClassLoader.getSystemResourceAsStream(dtdPath);
        if (dtdIn == null) {
            throw new FileNotFoundException("Can't find DTD as resource: " + dtdPath);
        }
        try {
            DTDParser dtdParser = new DTDParser(new InputStreamReader(dtdIn), null);
            DTDDocument dtd = (DTDDocument)dtdParser.parseDocument();
            XMLParser parser = new XMLParser(xmlIn, null);
            return parser.parseDocument(dtd, DOMFactory.DOCUMENT_HTML);
        } finally {
            dtdIn.close();
        }
    }
    
    /*
     * Parse and XML object and dump information about it.
     */
    private void parse(String inFile, int what, OutputStream out)
        throws IOException, ParseException {
        Reader xmlIn = new InputStreamReader(new FileInputStream(inFile));
        try {
            Document doc;

            switch (what) {
            case PARSE_XML: {
                doc = parseXML(xmlIn, null);
                break;
            }
            case PARSE_DTD: {
                doc = parseDTD(xmlIn);
                break;
            }
            case PARSE_XHTML: {
                doc = parseXML(xmlIn, DOMFactory.DOCUMENT_HTML);
                break;
            }
            case PARSE_XHTML_DTD: {
                doc = parseXMLWithDTD(xmlIn, XHTML_STRICT_PATH, DOMFactory.DOCUMENT_HTML);
                break;
            }
            default:
                throw new IllegalArgumentException("bad what parameter");
            }

            XMLCUtil.printNode("DOM hierarchy", doc, XMLCUtil.PRINT_ALL, 
                               new PrintWriter(out, true));
        } finally {
            xmlIn.close();
        }
    }

    /*
     * Error on more than one parse type option.
     */
    private static void multParseTypeError() {
        System.err.println("can only specify one of -dtd or -xhtml");
        System.exit(1);
    }

    /*
     * Entry point.
     */
    public static void main(String[] args) throws Throwable {
        int what = 0;
        int argIdx = 0;

        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-xml")) {
                if (what != 0) {
                    multParseTypeError();
                }
                what = PARSE_XML;
            } else if (args[argIdx].equals("-dtd")) {
                if (what != 0) {
                    multParseTypeError();
                }
                what = PARSE_DTD;
            } else if (args[argIdx].equals("-xhtml")) {
                if (what != 0) {
                    multParseTypeError();
                }
                what = PARSE_XHTML;
            } else if (args[argIdx].equals("-xhtml-dtd")) {
                if (what != 0) {
                    multParseTypeError();
                }
                what = PARSE_XHTML;
            } else {
                System.err.println("Error: invalid argument \"" + args[argIdx]
                                   + "\", expected one of -xml, -dtd, -xhtml, -xhtml-dtd");
                System.exit(1);
            }
            argIdx++;
        }

        if (what == 0) {
            what = PARSE_XML;
        }

        int argsLeft = args.length-argIdx;
        if ((argsLeft < 1) || (argsLeft > 2)) {
            System.err.println("wrong # args: test [-dtd] xml.file [output]");
            System.exit(1);
        }
        try {
            OutputStream out;
            if (argsLeft == 1) {
                out = System.out;
            } else {
                out = new FileOutputStream(args[argIdx+1]);
            }
            XMLParse xmlParse = new XMLParse();
            xmlParse.parse(args[argIdx], what, out);
            out.flush();
        } catch (Throwable error) {
            if (error instanceof ExceptionInInitializerError) {
                error = ((ExceptionInInitializerError)error).getException();
            }
            System.err.println(error.getClass().getName() + ": " + error.getMessage());
            error.printStackTrace();
            System.exit(2);
        }
        
    }
}
