/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: OasisTests.java,v 1.3 2000/02/21 01:27:34 markd Exp $
 */

package org.enhydra.xml.xmlc.tests.xml;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.*;
import java.io.*;
import org.w3c.dom.*;
import org.apache.xerces.parsers.DOMParser;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;

/**
 * Test XMLC using the OASIS XML parser conformance tests.
 */
public class OasisTests {
    /*
     * Location of file describing the tests.
     */
    private static final String XML_CONF = "conf/xmlconf.xml";

    /*
     * Elements and attributes in the testcases DTD.
     */
    private static final String TC_ELEM_TEST = "TEST";
    private static final String TC_ATTR_ID = "ID";
    private static final String TC_ATTR_TYPE = "TYPE";
    private static final String TC_ATTR_URI = "URI";

    /*
     * Separator between errors for tests.
     */
    private static final String sepLine1 =
        "==========================================================================";
    private static final String sepLine2 =
        "--------------------------------------------------------------------------";

    /*
     * Maximum number of failed tests before aborting.
     */
    private static final int MAX_FAILED_TESTS = 25;

    /*
     * Count of tests that have failed.
     */
    private int numFailedTest = 0;

    /*
     * Package for objects.
     */
    private static final String OASIS_PACKAGE =
        "org.enhydra.xml.xmlc.tests.xml.oasis";

    /*
     * Write error output to this stream.
     */
    PrintWriter errOut = new PrintWriter(System.err);


    /*
     * Handle a test error.
     */
    private static void handleTestError(String id,
                                        String uri,
                                        ByteArrayOutputStream errorOut,
                                        Exception except) {
        errOut.print(sepLine1);
        errOut.print("****Test Failed: " + id + ": " + uri);
        errOut.print(errorOut.toString());
        numFailedTest++;
        if (except != null) {
            except.printStackTrace(errOut);
        }
        errOut.print(sepLine2);
        if (numFailedTest > MAX_FAILED_TESTS) {
            errOut.println("*** " + numFailedTest + " tests failed");
            errOut.println("*** Maximum failed ");
            System.exit(1);
        }
    }

    /*
     * Parse one of the XML files containing test information.
     */
    private Document parseTestInfo(String xmlFile)
        throws SAXException, IOException {

        DOMParser parser = new DOMParser();
        parser.parse(new InputSource(xmlFile));
        return parser.getDocument();
    }

    /**
     * Compile a XMLC file using the XMLC API.
     

    /**
     * Compiler and run a object using the XMLC API.
     */
    private void runTestOnXML(String id, String uri,
                              boolean vaild) {
        ByteArrayOutputStream errorOut = new ByteArrayOutputStream();
        boolean gotError = false;
        int parseOptions =  valid ? XMLCParser.OPT_VALIDATE_DOCUMENT
            : XMLCParser.OPT_NO_VALIDATE_DOCUMENT;


        try {

            Parser parser = new Parser(new ErrorReporter(new PrintWriter(errorOut, true)));

            XMLCDocument xmlcDoc = parse(new File(uri),
                                         null, // default parser,
                                         XMLCDomFactory.class.getClass().getName(),
                                         null, null);
        CodeGenerator codeGen =
            new CodeGenerator(xmlcDoc,
                              (options.printAccessMethods ? traceOut : null));

        PrintWriter javaOut = new PrintWriter
            (new BufferedOutputStream(new FileOutputStream(options.javaSource)));
        try {
            codeGen.writeClass(javaOut, options.packageName, options.unqualClassName,
                               options.getImplements());
        } finally {
            javaOut.close();
        }
        if (options.verbose) {
            traceOut.println(">>> compiling code");
        }
        Javac javac = new Javac(options.javac);
        javac.compile(options.javaSource,
                      errorReporter,
                      options.verbose ? traceOut : null,
                      options.getJavacArgs());
        } catch (Exception except) {
            handleTestError(id, uri, errorOut, except);
            gotError = true;
        }
        if (!gotError && (errorOut.size() > 0)) {
            // Output but no exception...
            handleTestError(id, uri, errorOut, null);
        }
    }

    /*
     * Recursively traverse the test description, running
     * the desired tests.
     */
    private void run(Node node) {
        for (Node child = node.getFirstChild();
             child != null; child = child.getNextSibling()) {
            if (node instanceof Element) {
                Element elem = (Element)node;
                if (elem.getTagName().equals(TC_ELEM_TEST)) {
                    //runTest(elem.getAttribute(
                }
            }
            run(child);
        }
    }

    /*
     * Parse the test description file (see conf/testcases.dtd)
     * and recurse down the tree performing the desired tests.
     */
    private void testXmlc()
        throws SAXException, IOException {

        run(parseTestInfo(XML_CONF));
    }
    
    /*
     * Test entry point.
     */
    public static void main(String[] args) {
        if (args.length != 2) {
            System.err.println("wrong # args: outDir outsuffix");
            System.exit(1);
        }
        String outDir = args[0];
        String outSuffix = args[1];

        try {
            OasisTests tests = new OasisTests();
            tests.testXmlc();
        } catch (Throwable err) {
            System.err.println(err.getClass().getName() + ": " + err.getMessage());
            err.printStackTrace();
            System.exit(2);
        }
    }
}

/*
 * Local Variables:
 * mode: java
 * End:
 */

